Package: stringmagic
Type: Package
Title: Character String Operations and Interpolation, Magic Edition
Version: 1.3.0
Authors@R: 
  c(person(given = "Laurent R", family = "Berge", role = c("aut", "cre"), email = "laurent.berge@u-bordeaux.fr",
         comment = c(ORCID = "0000-0003-2736-1917")),
    person(given = "Morgan", family = "Jacob", role = "ctb"))
Imports: Rcpp(>= 1.0.5), utils, stats
Suggests: knitr, rmarkdown, data.table
LinkingTo: Rcpp
Description: Performs complex string operations compactly and efficiently. Supports string interpolation jointly with over 50 string operations. Also enhances regular string functions (like grep() and co). See an introduction at <https://lrberge.github.io/stringmagic/>.
License: GPL (>= 2)
Encoding: UTF-8
BugReports: https://github.com/lrberge/stringmagic/issues
URL: https://lrberge.github.io/stringmagic/,
        https://github.com/lrberge/stringmagic
VignetteBuilder: knitr
RoxygenNote: 7.3.2
NeedsCompilation: yes
Packaged: 2026-08-03 13:42:56 UTC; berge028
Author: Laurent R Berge [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2736-1917>),
  Morgan Jacob [ctb]
Maintainer: Laurent R Berge <laurent.berge@u-bordeaux.fr>
Repository: CRAN
Date/Publication: 2026-08-04 10:30:09 UTC

useDynLib(stringmagic, .registration = TRUE)

#
# string_*
#

export(string_ops, string_is, string_any, string_all, string_which, string_get, 
       string_clean, string_split2df, string_split2dt, string_fill, string_vec, 
       string_replace, string_extract, string_split)

# ... aliases; the rule is underscore before vowels unless long word follows
export(stops, stis, stany, stall, stextract, stwhich, stget, stclean, 
       stvec, streplace, stsplit)
       
export(string_ops_alias, string_vec_alias, string_clean_alias)

export(paste_conditional)

export(parse_regex_pattern)

#
# string_magic
#

export(string_magic, .string_magic)
export(string_magic_register_fun, string_magic_register_ops)
export(string_magic_alias)

export(cat_magic, cat_magic_alias)
export(message_magic, message_magic_alias)

export(timer_magic)

# ... aliases
export(sma, catma, mema)

#
# utilities
#

export(get_interpolated_expr, get_interpolated_vars)

#
# imports
#

importFrom(Rcpp, evalCpp)

importFrom("stats", terms)

importFrom("utils", "as.roman", "capture.output", "head", "tail")

# stringmagic

The purpose of `stringmagic` is to facilitate the manipulation of character strings.

It introduces various functions to facilitate pattern detections via
[regular expression (regex) logic](https://lrberge.github.io/stringmagic/articles/ref_regex_logic.html), or to efficienty [clean character vectors](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_clean) in a 
readable way. Consistently across the package, regular expressions gain [optional flags](https://lrberge.github.io/stringmagic/articles/ref_regex_flags.html) 
to monitor how the patterns should behave (fixed search? ignore case? add word boundaries? etc).
For more information, see the vignette on [string tools](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html).

The main contribution of this package, and flagship function, is `string_magic` which introduces 
a new language tailored to create complex character strings. 

### Getting started with `string_magic`

`string_magic` behaves in a similar way to the well known function [glue](https://glue.tidyverse.org/).
Use curly brackets to interpolate variables: i.e to insert their value directly in the string:
```r
library(stringmagic)
x = "John" ; y = "Mary"
string_magic("Hi {x}! How's {y} doing?")
#> [1] "Hi John! How's Mary doing?"
```

Almost anything `glue` can do, `string_magic` can do, and if you're thinking about speed, they're about
as fast (see the section [Performance](#Performance)).

The difference, and originality, of `string_magic` is that you can apply any arbitrary operation to 
the interpolated variables. You want to create an enumeration? Sure. Add quotation marks? Check. 
Change the case? No problem. Sort on a substring? Of course! There are over 50 built-in operations and the
magic is that applying these operations is about as simple as saying them.

**Operations.** The syntax to add operations is as follows:

![operation](man/figures/operation-template.png)

The `operations` are a comma separated sequence of keywords, each keyword being bound to a specific function. Here is an example in which we apply two operations:
```r
lovers = c("romeo", "juliet")
string_magic("Famous lovers: {title, enum ? lovers}.")
#> [1] "Famous lovers: Romeo and Juliet."
```

![example-lovers](man/figures/example-simple_operation.png)

**Arguments.** Some operations, like `split`, require arguments. Arguments are passed using quoted text
just before the operation. The syntax is:

![argument](man/figures/argument.png)

Let's take the example of splitting an email address and keeping the text before the domain:
```r
email = "John@Doe.com"
string_magic("This message comes from {'@'split, first ? email}.")
#> [1] "This message comes from John."
```

![example-argument](man/figures/example-argument.png)

**Options.** Many operations accept options. These options are keywords working like flags (i.e. things
that can be turned on) and change the behavior of the current operation. 
Add options using a dot separated sequence of keywords attached to the operation:

![options](man/figures/options.png)


We have seen the `enum` operation in an earlier example, let's add a couple of options to it.
```r
fields = c("maths", "physics")
string_magic("This position requires a PhD in either: {enum.i.or ? fields}.")
#> [1] "This position requires a PhD in either: i) maths, or ii) physics."
```

![example-options](man/figures/example-options.png)

**Anything else?** So far we have just scratched the surface of `string_magic` possiblities.
Other features are: advanced support for pluralization, nesting, conditional operations,
grouped operations, compact if-else statements, and unlimited customization.
Here is a list of resources:

+ [an introduction to `string_magic`](https://lrberge.github.io/stringmagic/articles/guide_string_magic.html)
+ [`string_magic`'s regular operations](https://lrberge.github.io/stringmagic/articles/ref_operations.html)
+ [`string_magic`'s special operations](https://lrberge.github.io/stringmagic/articles/ref_string_magic_special_operations.html)
+ [create new operations](https://lrberge.github.io/stringmagic/articles/guide_customization.html#sec_creation)

### Friendly errors

`string_magic` tries to be friendly to the user by providing useful error messages:
```r
x = c("Zeus", "Hades", "Poseidon")
string_magic("The {len?x} brothers: {anum?x}.")
#> Error: in string_magic("The {len?x} brothers: {anum?x}."): 
#> CONTEXT: Problem found in "The {len?x} brothers: {anum?x}.",
#>          when dealing with the interpolation `{anum?x}`.
#> PROBLEM: `anum` is not a valid operator. Maybe you meant `enum`?
#> 
#> INFO: Type string_magic(.help = "regex") or string_magic(.help = TRUE) for help.
#> Or look at the vignette: https://lrberge.github.io/stringmagic/articles/guide_string_magic.html

string_magic("The iris species are: {unik, sort, enum ? iris[['Species']}.")
#> Error: in string_magic("The iris species are: {unik, sort, enum ?...: 
#> CONTEXT: Problem found in "The iris species are: {unik, sort, enum ? iris[['Species']}.",
#>          when dealing with the interpolation `{unik, sort, enum ? iris[['Species']}.`.   
#> PROBLEM: in the expression `iris[['Species']`, a bracket (`[`) open is not closed.
```

The cost of this feature in terms of computing time is of the order of magnitude of 50 micro seconds (of course it depends on context). 
If you're not interested in informative error messages, `.string_magic` (note the `"."` prefix) is identical to `string_magic` but avoids error handling and is then slightly faster.

### Performance

**Basic interpolation.** For regular string interpolations, the performance of `string_magic` is similar to the performance of `glue`. That is to say, the price to pay for user experience is in the ballpark of 100 micro seconds (on my -- slow -- computer). Let's have a simple benchmark:

```r
library(microbenchmark)
library(glue)

x = "Romeo" ; y = "Juliet"
microbenchmark(      base    =   paste0(x, " seems to love ", y, "."),
                     glue    =     glue("{x} seems to love {y}."),
               string_magic  =   string_magic("{x} seems to love {y}."),
               .string_magic =  .string_magic("{x} seems to love {y}."))
#> Unit: microseconds
#>           expr  min    lq   mean median    uq   max neval
#>           base  1.3  1.50  1.604   1.60  1.70   2.4   100
#>           glue 77.7 83.70 93.560  87.45 92.90 470.6   100
#>   string_magic 69.9 73.75 78.257  77.50 80.75 106.5   100
#>  .string_magic 43.2 45.70 50.205  48.20 52.15 137.9   100
```

The difference with the base function `base::paste0` looks impressive (it looks 50 times faster), but is in fact not really important. Both `glue` and `string_magic` processing time is due to overheads: a fixed cost that does not depend on the size of the vectors in input. Hence for large vectors or operations that run in one millisecond or more, this difference is negligible.

As you can notice, `.string_magic`, `string_magic` without error handling, is about twice faster than `glue`. But I'm not sure that sacrificing user expericence for a 50us overhead is really worth it!

**Complex operations** The function `string_magic` shines when performing complex string manipulation. The question we ask here is: how much does it cost in terms of perfomance? Let's look at the following operation:

```r
x = c("Zeus", "Hades", "Poseidon")
string_magic("The {len?x} brothers: {enum?x}.")
#> [1] "The 3 brothers: Zeus, Hades and Poseidon."
```

Although the interface is different, let's compare `string_magic` to `glue` and `paste0`:

```r
x = c("Zeus", "Hades", "Poseidon")
microbenchmark(base = paste0("The ", length(x), " brothers: ", 
                        paste0(paste0(x[-length(x)], collapse = ", "), 
                               " and ", x[length(x)]), "."),
                               
               glue = glue("The {length(x)} brothers: {x_enum}.", 
                        x_enum = paste0(paste0(x[-length(x)], collapse = ", "), 
                                        " and ", x[length(x)])),
                                        
               string_magic = string_magic("The {len?x} brothers: {enum?x}."),
               
               string_magic_bis = string_magic("The {length(x)} brothers: {x_enum}.", 
                              x_enum = paste0(paste0(x[-length(x)], collapse = ", "), 
                                        " and ", x[length(x)])))
#> Unit: microseconds
#>              expr     min       lq      mean   median      uq       max neval
#>              base   4.701   5.5510  54.37105   6.2010   6.802  4792.601   100
#>              glue  93.900 104.6015 121.23295 118.8510 129.901   219.100   100
#>      string_magic 316.701 331.3010 628.79896 351.2510 391.401 26576.501   100
#>  string_magic_bis  94.601 106.5010 118.10602 114.8015 123.301   250.301   100
```

As we can see, the processing overhead of `string_magic` specific syntax is a few hundred microseconds. 
Remember that since regular interpolation can be performed, you can always fall back to `glue`-like processing (and benefit from the same performance), as is illustrated by the last command of the benchmark.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

# we preload it to avoid ugly (was comiled with R.x.x) warnings in the doc
library(stringmagic)

# Option to allow caching in non interactive mode
options("string_magic_string_get_forced_caching" = TRUE)

## -----------------------------------------------------------------------------
cars = row.names(mtcars)
cat_magic("All cars from mtcars:\n{C, 60 swidth ? cars}")

# cars with an 'a', an 'e', an 'i', and an 'o', all in lower case
string_get(cars, "a & e & i & o")

# cars with no 'e' and at least one digit
string_get(cars, "!e & \\d")

# flags apply to all
# contains the 'words' 2, 9 or l
# alternative syntax for flags: "wi/2 | 9 | l"
string_get(cars, "word, ignore/2 | 9 | l")

## -----------------------------------------------------------------------------
# string_get(cars, "a & e & i & o")
# cars with an 'a', an 'e', an 'i', and an 'o', all in lower case
string_get(cars, "a", "e", "i", "o")

# string_get(cars, "!e & \\d")
# cars with no 'e' and at least one digit
string_get(cars, "!e", "\\d")

# string_get(cars, "!/e & \\d")
# This example cannot be replicated directly, we need to apply logical equivalence
string_get(cars, "!e", "!\\d", or = TRUE)

# string_get(cars, "wi/2 | 9 | l")
# contains the 'words' 2, 9 or l
string_get(cars, "2", "9", "l", or = TRUE, word = TRUE, ignore.case = TRUE)

## -----------------------------------------------------------------------------
# cars without digits, then cars with 2 'a's or 2 'e's and a digit
string_get(cars, "!\\d", "i/a.+a | e.+e & \\d", seq = TRUE)

# let's get the first word of each car name
car_first = string_ops(cars, "extract.first")
# we select car brands ending with 'a', then ending with 'i'
string_get(car_first, "a$", "i$", seq = TRUE)
# seq.unik is similar to seq but applies unique()
string_get(car_first, "a$", "i$", seq.unik = TRUE)

## -----------------------------------------------------------------------------
# Since we used `car_first` in the previous example, we don't need to provide
# it explicitly now
# => brands containing 'M' and ending with 'a' or 'i'; brands containing 'M'
string_get("M & [ai]$", "M", seq.unik = TRUE)

## -----------------------------------------------------------------------------
# parsing an input: extracting the numbers
input = "8.5in, 5.5, .5 cm"
string_ops(input, "','split, tws, '^\\. => 0.'replace, '^\\D+|\\D+$'replace, num")


# Explanation------------------------------------------------------------------|
# ','split: splitting w.r.t. ','                                               |
# tws: trimming the whitespaces                                                |
# '^\\. => 0.'replace: adds a 0 to strings starting with '.'                   |
# '^\\D+|\\D+$'replace: removes non-digits on both ends of the string          |
# num: converts to numeric                                                     |


# now extracting the units
string_ops(input, "','split, '^[ \\d.]+'replace, tws")


# Explanation------------------------------------------------------------------|
# ','split: splitting w.r.t. ','                                               |
# '^[ \\d.]+'replace: removes the ' ', digit                                   |
#                     and '.' at the beginning of the string                   |
# tws: trimming the whitespaces                                                |

## -----------------------------------------------------------------------------
# Now using the car data
cars = row.names(mtcars)

# let's get the brands starting with an "m"
string_ops(cars, "'i/^m'get, x, unik")


# Explanation------------------------------------------------------------------|
# 'i/^m'get: keeps only the elements starting with an m,                       |
#            i/ is the 'regex-flag' "ignore" to ignore the case                |
#            ^m means "starts with an m" in regex language                     |
# x: extracts the first pattern. The default pattern is "[[:alnum:]]+"         |
#    which means an alpha-numeric word                                         |
# unik: applies unique() to the vector                                         |


# let's get the 3 largest numbers appearing in the car models
string_ops(cars, "'\\d+'x, rm, unik, num, dsort, 3 first")


# Explanation------------------------------------------------------------------|
# '\\d+'x: extracts the first pattern, the pattern meaning "a succession"      |
#          of digits in regex language                                         |
# rm: removes elements equal to the empty string (default behavior)            |
# unik: applies unique() to the vector                                         |
# num: converts to numeric                                                     |
# dsort: sorts in decreasing order                                             |
# 3 first: keeps only the first three element                                  |

## -----------------------------------------------------------------------------
monologue = c("For who would bear the whips and scorns of time",
              "Th' oppressor's wrong, the proud man's contumely,",
              "The pangs of despis'd love, the law's delay,",
              "The insolence of office, and the spurns",
              "That patient merit of th' unworthy takes,",
              "When he himself might his quietus make",
              "With a bare bodkin? Who would these fardels bear,",
              "To grunt and sweat under a weary life,",
              "But that the dread of something after death-",
              "The undiscover'd country, from whose bourn",
              "No traveller returns- puzzles the will,",
              "And makes us rather bear those ills we have",
              "Than fly to others that we know not of?")

# Cleaning a text
string_clean(monologue, 
          # use string_magic to: lower the case and remove basic stopwords
          "@lower, stopword",
          # remove a few extra stopwords(we use the flag word 'w/')
          "w/th, 's",
          # manually stem some verbs
          "despis'd => despise", "undiscover'd => undiscover", "(m|t)akes => \\1ake",
          # still stemming: dropping the ending 's' for words of 4+ letters, except for quietus
          "(\\w{3,}[^u])s\\b => \\1",
          # normalizing the whitespaces + removing punctuation
          "@ws.punct")


## -----------------------------------------------------------------------------
fruits = string_vec("orange, apple, pineapple, strawberry")
fruits

## -----------------------------------------------------------------------------
more_fruits = string_vec("lemon, {fruits}, peach")
more_fruits

## -----------------------------------------------------------------------------
more_fruits = string_vec("lemon, {6 Shorten ? fruits}, peach")
more_fruits

## -----------------------------------------------------------------------------
pkgs = string_vec("pandas, os, time, re")
imports = string_vec("import numpy as np, import {pkgs}")
imports

## -----------------------------------------------------------------------------
string_vec("1, 5,
            3, 2,
            5, 12", .nmat = TRUE)

## -----------------------------------------------------------------------------
string_vec(1, 5,
           3, 2,
           5, 12, .nmat = 3)

## -----------------------------------------------------------------------------
# you can add the column names directly in the argument .df
df = string_vec("1, john,
                 3, marie,
                 5, harry", .df = "id, name")
df

# automatic conversion of numeric values
df$id * 5

## -----------------------------------------------------------------------------
x = c("Nor rain, wind, thunder, fire are my daughters.",
      "When my information changes, I alter my conclusions.")

# we split at each word
sentences_split = string_split2df(x, "[[:punct:] ]+")
sentences_split

# recovering the original vectors (we only lose the punctuation)
paste_conditional(sentences_split$x, sentences_split$obs)

## -----------------------------------------------------------------------------
id = c("ws", "jmk")
# we add the identifier
base_words = string_split2df(x, "[[:punct:] ]+", id = list(author = id))

# merging back using a formula
paste_conditional(x ~ author, base_words)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(stringmagic)

## -----------------------------------------------------------------------------
code = c("DT = as.data.table(iris)", 
         "DT[, .(pl_sl = string_magic('PL/SL = {Petal.Length / Sepal.Length}')]")

string_get(code, "if/dt[")

## -----------------------------------------------------------------------------
unhappy = "Rumble thy bellyful! Spit, fire! spout, rain!
Nor rain, wind, thunder, fire are my daughters.
I tax not you, you elements, with unkindness.
I never gave you kingdom, call'd you children,
You owe me no subscription. Then let fall
Your horrible pleasure. Here I stand your slave,
A poor, infirm, weak, and despis'd old man."

# the ignore flag allows to retain words starting with the
# upper cased letters
# ex: getting words starting with the letter 'r' to 'z'
cat_magic("{'ignore/\\b[r-z]\\w+'extract, c, 60 swidth ? unhappy}")

## -----------------------------------------------------------------------------
x = "50 + 5 * 5 = 40"
string_clean(x, "f/+", "f/*", replacement = "-")

# Without the fixed flag, we would have gotten an error since '+' or '*'
# have a special meaning in regular expressions (it is a quantifier)
# and expects something before

# Here's the error
try(string_clean(x, "+", "*", replacement = "-"))

## -----------------------------------------------------------------------------
le_mont_des_oliviers = "S'il est vrai qu'au Jardin sacré des Écritures,
Le Fils de l'homme ai dit ce qu'on voit rapporté ;
Muet, aveugle et sourd au cri des créatures,
Si le Ciel nous laissa comme un monde avorté,
Alors le Juste opposera le dédain à l'absence
Et ne répondra plus que par un froid silence
Au silence éternel de la Divinité."

# we hide a few words from this poem
string_magic("{'wi/et, le, il, au, des?, ce => _'replace ? le_mont_des_oliviers}")

## -----------------------------------------------------------------------------
vowels ="aeiouy"
# let's keep only the vowels
# we want the pattern: "[^aeiouy]"
lmb = "'Tis safer to be that which we destroy
Than by destruction dwell in doubtful joy."
string_replace(lmb, "magic/[^{vowels}]", "_")

#
# Illustration of `string_magic` operations before regex application
#

cars = row.names(mtcars)
# Which of these models contain a digit?
models = c("Toyota", "Hornet", "Porsche")
# we want the pattern "(Toyota|Hornet|"Porsche).+\\d"
# we collapse the models with a pipe using '|'c
string_get(cars, "m/({'|'c ? models}).+\\d")

# alternative: same as above but we first comma-split the vector
models_comma = "Toyota, Hornet, Porsche"
string_get(cars, "m/({S, '|'c ? models_comma}).+\\d")

#
# Interpolation does not apply to regex-specific curly brackets
#

# We delete only successions of 2+ vowels
# {2,} has a rexex meaning and is not interpolated:
string_replace(lmb, "magic/[{vowels}]{2,}", "_")

## -----------------------------------------------------------------------------
cars_small = head(row.names(mtcars))
print(cars_small)

string_replace(cars_small, "ti/mazda", "Mazda: sold out!")

## -----------------------------------------------------------------------------
cars_small = head(row.names(mtcars))
print(cars_small)

string_replace(cars_small, "total, ignore/\\d & !e", "I don't like that brand!")

## -----------------------------------------------------------------------------
encounter = string_vec("Hi Cyclops., Hi you. What's your name?, Odysseus is my name.")
# we only remove the first word
string_replace(encounter, "single/\\w+", "...")

## -----------------------------------------------------------------------------
eq = "5/x = 3/2"
# escaping with backslashes
string_replace(eq, "(\\w)\\/(\\w)", "\\2/\\1")

## -----------------------------------------------------------------------------
path = "my/path/to/the/file.tex"
# we keep the directory only

# first try: an error bc flags are expected before the first '/'
try(string_replace(path, "/[^/]+$"))

# after escaping: works (only the first slash requires escaping)
string_replace(path, "\\/[^/]+$")

# if we did add a flag, we would need to double the slash
# compare...
string_replace(path, "i//[^/]+$")
# to...
string_replace(path, "i/[^/]+$")

#----------------------------------------------#
# Author: Laurent Berge
# Date creation: Wed Jul 27 09:42:17 2022
# ~: Misc funs for string.ops, mostly internal
#----------------------------------------------#


####
#### use-level ####
####

# I am exposing a tool that can be useful for developpers (well... me)
# => finding the variables in a sma call
# x = "bon{jour} {upper!les} {firstchar?gens} {ok} {s, enum!ou bien {haha}} {&is_ok;TRUE;{.}{damn}}"

#' Lists the expressions used for interpolation in a `string_magic` call
#' 
#' Tool intended for development: use `get_interpolated_expr` to obtain the list of expressions
#' which will be interpolated in a [string_magic()] call. 
#' The function `get_interpolated_vars` provides the variables instead.
#' 
#' @param x A character scalar for which the variables will be recovered. 
#' For example `x = "Hi {person}"` will  return `"person"` 
#' (the variable that will be interpolated).
#' @param parse Logical scalar, default is `FALSE`. If `TRUE`, the result is a list of R
#' expressions. If `FALSE` (default), the result is a character vector of expressions.
#' @param delim Character vector of length 1 or 2. Default is `c("{", "}")`. Defines 
#' the opening and the closing delimiters for interpolation. 
#' 
#' If of length 1, it must be of the form: 1) the opening delimiter, 
#' 2) a single space, 3) the closing delimiter. Ex: `".[ ]"` is equivalent to `c(".[", "]")`.
#' The default value is equivalent to `"{ }"`.
#' 
#' @details 
#' Note that this function captures even deeply nested interpolations.
#' 
#' @return 
#' If the argument `parse = FALSE`, the default, then this function returns a 
#' character vector containing all the expressions that will be interpolated. 
#' This vector can be empty if there is no interpolation. 
#' 
#' If the argument `parse = TRUE`, then a list is returned, containing the R expressions.
#' 
#' The function `get_interpolated_vars` always return a character vector.
#' 
#' @inherit string_clean seealso
#' 
#' @examples 
#' 
#' # let's create a simple interpolation
#' x = c("Ken", "Barbie")
#' sma_expr = "{' loves 'c ? x}. But does {' love 'c ? rev(x)}?"
#' string_magic(sma_expr)
#' 
#' # We recover the two expressions
#' (char = get_interpolated_expr(sma_expr))
#' 
#' # same with parsing
#' (expr = get_interpolated_expr(sma_expr, parse = TRUE))
#' 
#' # see the difference
#' eval(char[[1]])
#' eval(expr[[1]])
#' 
#' # and only the variables:
#' get_interpolated_vars(sma_expr)
#' 
get_interpolated_expr = function(x, parse = FALSE, delim = c("{", "}")){
  
  check_character(x, scalar = TRUE, mbt = TRUE)
  delim = check_set_delimiters(delim)
  check_logical(parse, scalar = TRUE)
  
  res = get_expr_internal(x, delim)
  cpp_trimws_in_place(res)  
  
  if(parse){
    n_res = length(res)
    res_list = vector("list", n_res)
    for(i in seq_len(n_res)){
      res_list[[i]] = str2lang(res[i])
    }
    
    res = res_list
  }
  
  res
}

#' @describeIn get_interpolated_expr Obtain the variables used in `string_magic()` interpolations
get_interpolated_vars = function(x, delim = c("{", "}")){
  
  check_character(x, scalar = TRUE, mbt = TRUE)
  delim = check_set_delimiters(delim)
  
  expr_all = get_expr_internal(x, delim)
  
  expr_all = unique(expr_all)
  n_expr = length(expr_all)
  
  if(n_expr == 0){
    return(character(0))
  }
  
  vars_list = vector("list", n_expr)
  for(i in seq_len(n_expr)){
    vars_list[[i]] = all.vars(str2lang(expr_all[i]))
  }
  
  unique(unlist(vars_list))
}


get_expr_internal = function(x, delim){
  
  vars = character()
  
  # This is not so straigtforward because of the many cases
  x_parsed = cpp_string_magic_parser(x, delim)
  for(elem in x_parsed){
    if(length(elem) == 1){
      next
    }
    
    op_slot = elem[[1]]
    var_slot = elem[[2]]
    
    if(var_slot == ""){
      if(op_slot[1] %in% c("&", "&&")){
        for(arg in op_slot[-1]){
          vars = c(vars, get_expr_internal(arg, delim))
          vars = setdiff(vars, ".")
        }
      }
      next
    }
    
    if(identical(tail(op_slot, 1), "!")){
      vars = c(vars, get_expr_internal(var_slot, delim))
      next
    }
    
    vars = c(vars, var_slot)
  }
  
  vars
}


####
#### string manipulation ####
####


string_to_ascii = function(x, options){
  # force to ASCII using iconv. Keeps track of encoding error

  opts = c("silent", "utf8")
  options = check_set_options(options, opts, "ascii", free = TRUE)

  is_silent = "silent" %in% options

  if("utf8" %in% options){
    from = "UTF-8"
  } else {
    from = setdiff(options, opts)
    if(length(from) == 0){
      from = ""
    }
  }

  old_char = try(nchar(x), silent = TRUE)

  if("try-error" %in% class(old_char)){
    # super nasty encoding => nchar can't be used
    x_conv = iconv(x, to = "ASCII//TRANSLIT//IGNORE", from = from)
    nbfail = 0

    if(!is_silent){
      warn_hook("In operation `ascii`, the encoding was a bit nasty (nchar couldn't be used!): please check the conversion to ASCII.",
              "\n(Ignore this message with the `silent` option [ex: `ascii.silent`]. Pass the native encoding as an option?)")
    }

  } else {
    x_conv = iconv(x, to = "ASCII//TRANSLIT//IGNORE", from = from)
    which_failed = which(old_char != nchar(x_conv) | grepl("?", x_conv, fixed = TRUE))
    nbfail = length(which_failed)
  }

  if(nbfail > 0){
    
    if(!is_silent){
      warni("When transforming to ASCII, there {#was, n ? nbfail} failed encoding{#s}. ",
            "Attribute `isFailed` is created (note that you can only access it if the variable is not transformed).",
            "\n(Ignore this message with the `silent` option [ex: `ascii.silent`]. Pass the native encoding as an option?)")
    }
    
    # we add the failures
    attr(x_conv, "isFailed") = which_failed
  }

  x_conv
}


####
#### internal ####
####


is_numeric_in_char = function(x){
  res = tryCatch(as.numeric(x), warning = function(x) "not numeric")
  !identical(res, "not numeric")
}


string_x = function(x, n, from){
  # string extract

  if(is.character(n)){
    n = nchar(n)
  }

  if(n < 0){
    if(!missing(from)) stop("Internal error: you should not provide arg. `from` when `n` is negative.")
    from = nchar(x) + n + 1
    to = nchar(x)
  } else if(!missing(from)){
    to = from + n - 1
  } else {
    from = 1
    to = n
  }

  substr(x, from, to)
}

string_trim = function(x, n_first = 0, n_last = 0){
  # positive values: first
  # negative values: last

  if(is.character(n_first)){
    n_first = nchar(n_first)
  } else if(n_first < 0){
    n_last = -n_first
    n_first = 0
  }

  substr(x, 1 + n_first, nchar(x) - n_last)
}


n_th = function(x, letters = TRUE, compact = FALSE){
  # The main purpose of this function is for smallish 'x'
  # only to print info, not performance oriented.

  if(is.character(x)) return(x)

  is_compact = FALSE
  if(length(x) > 1 && all(diff(x) == 1)){
    is_compact = TRUE
    x = x[c(1, length(x))]
  }

  res = character(length(x))

  if(letters){
    # We don't go all the way, it makes no sense

    res[x == 0] = "zeroth"

    dict = c("first", "second", "third", "fourth", "fifth", "sixth", "seventh",
             "eighth", "nineth", "tenth", "eleventh", "twelfth", "thirteenth",
             "fourteenth", "fifteenth", "sixteenth", "seventeenth", "eighteenth",
             "nineteenth", "twentyth")

    qui = x >= 1 & x <= 20
    res[qui] = dict[x[qui]]

    if(!any(res == "")){
      if(is_compact){
        res = paste0(res[1], " to ", res[2])
      }
      return(res)
    }
  }

  qui = res == ""
  rest = x[qui] %% 10
  rest[rest == 0 | rest >= 4] = 4
  rest[x[qui] %in% 11:13] = 4
  postfix = c("st", "nd", "rd", "th")

  res[qui] = paste0(x[qui], postfix[rest])

  if(is_compact){
    res = paste0(res[1], " to ", res[2])
  }

  res
}

# Internal fun, n is ALWAYS len 1, positive and not missing
n_letter = function(x){

  res = character(length(x))

  for(i in seq_along(x)){

    xi = x[i]
    
    if(xi >= 100){
      val = format(xi, big.mark = ",")
    } else {

      is_minus = xi < 0
      if(is_minus) xi = abs(xi)

      num2char = c("zero", "one", "two", "three", "four", "five", "six", "seven",
                  "eight", "nine", "ten", "eleven", "twelve", "thirteen",
                  "fourteen", "fifteen", "sixteen", "seventeen", "eighteen",
                  "nineteen")
      if(xi < 20){
        val = num2char[xi + 1]
      } else if(xi < 100){
        tens = xi %/% 10
        digit = xi %% 10
        tens_letter = c("twenty", "thirty", "forty", "fifty",
                        "sixty", "seventy", "eighty", "ninety")

        num2char = paste0("-", num2char)
        num2char[1] = ""

        val = paste0(tens_letter[tens - 1], num2char[digit + 1])
      }

      if(is_minus){
        val = paste0("minus ", val)
      }
    }
    
    res[i] = val
  }

  res
}


enum_letter = function(n){
  # returns only lowercase
  # oddity: there is no powers of z
  # => because otherwise the algorithm would have been too complex
  # and we don't care tbh


  if(n < 27){
    return(letters[1:n])
  }

  n_26 = log(n, 26)
  n_26 = floor(n_26) + (n_26 %% 1 == 0)

  rest = (1:n) - 1
  res = vector("list", n_26 + 1)
  i = 1
  for(p in n_26:0){
    num = 26**p

    if(p == 0){
      res[[i]] = letters[(rest %/% num) + 1]
    } else {
      res[[i]] = c("", letters)[(rest %/% num) + 1]
    }

    rest = rest %% num
    i = i + 1
  }

  res = do.call(base::paste0, res)

  res
}



# Internal fun, n is ALWAYS len 1, positive and not missing
n_times = function(x, letters = TRUE){
  # 1 and 2 are always once and twice

  res = character(length(x))

  for(i in seq_along(x)){
    xi = x[i]

    if(xi == 0) {
      if(letters){
        res[i] = "zero times"
      } else {
        res[i] = "0 times"
      }
    } else if(xi < 4 && letters){
      dict = c("once", "twice", "three times", "four times")
      res[i] = dict[xi]
    } else if(!letters && xi <= 2){
      res[i] = if(xi == 1) "once" else "twice"
    } else {
      if(letters){
        xi = n_letter(xi)  
      }      
      res[i] = paste0(xi, " times")
    }    
  }

  res
}

enum_main = function(x, options){

  # we construct the call to enum
  # default values
  quote = NULL
  enum = FALSE
  or = FALSE
  nmax = 7

  # Quote
  q_map = c("q" = "'", "Q" = '"', "bq" = "`")
  qui = which(options %in% names(q_map))
  if(length(qui) > 0){
    if(length(qui) > 1){
      stop_hook("In `enum`, you cannot use more than one quote keyword ",
                "(q, Q, or bq) in the enumeration. ")
    }

    q = options[qui]
    options = options[-qui]
    quote = q_map[q]
  }

  # or
  if("or" %in% options || "nor" %in% options){
    or = intersect(c("or", "nor"), options)
    options = setdiff(options, c("or", "nor"))
  }

  # i, a, etc
  qui = which(options %in% c("i", "I", "a", "A", "1"))
  if(length(qui) > 0){
    enum = options[qui[1]]
    options = options[-qui]
  }

  # digit
  qui = which(grepl("^[[:digit:]]+$", options) & options != "1")
  if(length(qui) > 0){
    nmax = as.numeric(options[qui[1]])
  }
  
  oxford = opt_equal(options, "oxford")
  comma = opt_equal(options, "comma")
  
  enumerate_items(x, quote = quote, or = or, enum = enum, nmax = nmax, 
                  oxford = oxford, comma = comma)
}

format_difftime = function(x, options = character()){
  # x: number of seconds or difftime or time

  if(is.character(x)){
    if(is_numeric_in_char(x)){
      # x: number of seconds
      x = as.numeric(x)
    } else {
      # When the data is not conform:
      # - should there be an error?
      # - should I return NA?

      if(!opt_equal(options, "silent")){
        warn_hook("Operation `difftime` could not be applied since the data",
                  " was not numeric, nor a POSIX time, nor a time-difference.",
                  "\n (To avoid this warning, use the option `silent`: `difftime.silent`.)")
      }

      return(rep("(difftime: NA)", length(x)))
    }
  }

  res = character(length(x))

  for(i in seq_along(x)){
    xi = x[i]
    
    if(inherits(xi, "POSIXt")){
      xi = Sys.time() - xi
    }
    
    if(inherits(xi, "difftime")){
      xi = as.double(xi, units = "secs")
    }
    
    if(opt_equal(options, "full")){
      # we don't go at the micro second precision: don't makes sense
      s = trunc(xi)
      if(s > 0){
        s_fmt = sprintf("% 3ss", s)
      } else {
        s_fmt = "    "
      }
      
      ms = sprintf("%.1f", (xi - s) * 1e3)
      ms_fmt = sprintf("% 5s", ms)
      
      res[i] = paste0(s_fmt, " ", ms_fmt, "ms")
      
    } else {
      if(xi > 3600){
        n_hour = xi %/% 3600
        rest_s = floor(xi %% 3600)
        n_min = rest_s %/% 60
        res[i] = string_magic("{n ? n_hour} hour{#s} {%02i ? n_min} min")
      } else if(xi > 60){
        n_min = xi %/% 60
        n_sec = floor(xi %% 60)
        res[i] = string_magic("{n ? n_min} min {%02i ? n_sec} sec")
      } else if(xi > 0.9){
        res[i] = paste0(fsignif(xi, 2, 1), "s")
      } else if(xi > 1e-3){
        res[i] = paste0(fsignif(xi * 1000, 2, 0), "ms")
      } else {
        res[i] = "<1 ms"
      }
    }
  }
  
  res
}

fsignif = function (x, s = 2, r = 0, commas = TRUE){
  # This is not intended to be applied to large vectors (not efficient)
  # Only for the in-print formatting of some numbers

  # It would supa dupa simple in c++...

  if(is.character(x)){
    return(x)
  }

  if(!is.numeric(x)){
    stop("The argumnent 'x' must be numeric.")
  }

  commas_single = function(x, s, r){

    if(!is.finite(x) || abs(x) < 1) return(as.character(x))

    if((p <- ceiling(log10(abs(x)))) < s){
      r = max(r, s - p)
    }

    x_sign = sign(x)
    x = abs(x)

    x_str = deparse(x)
    dec_string = ""

    if(grepl(".", x_str, fixed = TRUE)){
      decimal = gsub(".*\\.", ".", x_str)
      dec_string = substr(decimal, 1, 1 + r)

      if(dec_string == ".") dec_string = ""
    }

    entier = sprintf("%.0f", floor(x))
    quoi = rev(strsplit(entier, "")[[1]])
    n = length(quoi)
    sol = c()
    for (i in 1:n) {
      sol = c(sol, quoi[i])
      if (i%%3 == 0 && i != n) sol = c(sol, ",")
    }

    res = paste0(ifelse(x_sign == -1, "-", ""), paste0(rev(sol), collapse = ""), dec_string)

    res
  }

  signif_single = function(x, s, r) {
    if(is.na(x)) {
      return(NA)
    }

    if(abs(x) >= 10^(s - 1)){
      return(round(x, r))
    } else {
      return(signif(x, s))
    }
  }

  res = sapply(x, signif_single, s = s, r = r)

  if(commas) res = sapply(res, commas_single, s = s, r = r)

  qui0 = grepl("^0.", res, perl = TRUE)
  if(any(qui0)){
    qui_short = nchar(res) < s + 2
    if(any(qui_short)){
      for(i in which(qui0)[qui_short]){
        res[i] = as.vector(sprintf("%s%.*s", res[i], s + 2 - nchar(res[i]), "0000000000000000"))
      }
    }
  }

  res
}



enumerate_items = function (x, or = FALSE, quote = NULL,
                            enum = FALSE, nmax = 7, oxford = FALSE, comma = FALSE){
  # function that enumerates items
  # in argument type, you can have a mix of the different arguments, all separated with a "."

  if(length(x) == 0) return("");

  is_quote = !is.null(quote)
  n = length(x)

  # Ensuring it's not too long
  is_other = n > nmax
  if(is_other){

    nmax = nmax - 1

    other = paste0(n - nmax, " others")

    if(is_quote){
      x = c(paste0(quote, x[1:nmax], quote), other)
    } else {
      x = c(x[1:nmax], other)
    }

    n = length(x)
  } else if(is_quote){
    x = paste0(quote, x, quote)
  }

  is_enum = enum %in% c("i", "I", "a", "A", "1")
  if(is_enum){
    enum_comma = ","

    if(enum == "i"){
      enum_all = tolower(as.roman(1:n))
    } else if(enum == "I"){
      enum_all = as.roman(1:n)
    } else if(enum == "a"){
      enum_all = enum_letter(n)
    } else if(enum == "A"){
      enum_all = toupper(enum_letter(n))
    } else if(enum == "1"){
      enum_all = 1:n
    }

    enum_all = paste0(enum_all, ") ")

    if(is_other){
      enum_all[length(enum_all)] = ""
    }

  } else {
    enum_comma = ""
    enum_all = rep("", n)
  }

  if (n == 1) {
    res = x
  } else {
    
    and_or = " and "
    if(identical(or, "or") || isTRUE(or)){
      and_or = " or "
    } else if(identical(or, "nor")){
      and_or = " nor "
    }
    
    if(comma){
      if(is_enum){
        x = paste0(enum_all, x)
      }
      
      res = paste0(x, collapse = ", ")
    } else if(is_enum){
      res = paste0(paste0(enum_all[-n], x[-n], collapse = ", "), 
                   enum_comma, and_or, enum_all[n], x[n])
    } else {
      if(oxford){
        and_or = paste0(",", and_or)
      }
      res = paste0(paste0(x[-n], collapse = ", "), and_or, x[n])
    }

  }

  res
}


conjugate = function(verb, plural){

  # note that the parser squashes the spaces
  info_case = gsub("[[:upper:]]", "_", verb)
  verb = tolower(verb)

  if(verb %in% c("is", "are")){
    verb = "is"
  } else if(verb %in% c("has", "have")){
    verb = "has"
  } else if(verb %in% c("has not", "have not")){
    verb = "has not"
  } else if(verb %in% c("hasn't", "haven't")){
    verb = "hasn't"
  } else if(verb %in% c("donot", "doesnot")){
    verb = "do not"
  } else if(verb %in% c("don't", "doesn't")){
    verb = "don't"
  } else if(verb %in% c("isnot", "arenot")){
    verb = "is not"
  } else if(verb %in% c("isn't", "aren't")){
    verb = "isn't"
  } else if(verb %in% c("wasnot", "werenot")){
    ver = "was not"
  } else if(verb %in% c("wasn't", "weren't")){
    ver = "wasn't"
  } else if(verb %in% c("was", "were")){
    verb = "was"
  } else if(grepl("ies$", verb)){
    verb = gsub("ies$", "y", verb)

  } else if(grepl("(o|h)es$", verb)){
    verb = gsub("es$", "", verb)

  } else if(grepl("s$", verb)){
    verb = gsub("s$", "", verb)
  }

  pick = function(a, b) if(plural) b else a

  if(verb %in% c("is", "is not", "isn't", "has", "has not", "hasn't", 
                 "do", "do not", "don't", "was", "was not", "wasn't")){

    res = switch(verb,
                 is = pick("is", "are"),
                 "is not" = pick("is not", "are not"),
                 "isn't" = pick("isn't", "aren't"),
                 has = pick("has", "have"),
                 "has not" = pick("has not", "have not"),
                 "hasn't" = pick("hasn't", "haven't"),
                 do = pick("does", "do"),
                 "do not" = pick("does not", "do not"),
                 "don't" = pick("doesn't", "don't"),
                 "was" = pick("was", "were"),
                 "was not" = pick("was not", "were not"),
                 "wasn't" = pick("wasn't", "weren't"))

  } else {
    # by default, the rest is in plural form

    if(plural){
      res = verb

    } else {
      if(grepl("(h|o)$", verb)){
        res = paste0(verb, "es")

      } else if(grepl("[^aeo]y$", verb)){
        res = paste0(string_trim(verb, -1), "ies")

      } else {
        res = paste0(verb, "s")
      }
    }
  }

  if(substr(info_case, 1, 2) == "__"){
    res = toupper(res)
  } else if(substr(info_case, 1, 1) == "_"){
    substr(res, 1, 1) = toupper(substr(res, 1, 1))
  }

  res
}


missnull = function(x){
  missing(x) || is.null(x)
}

names_xpd = function(x){
  nm = names(x)
  if(is.null(nm)){
    nm = character(length(x))
  }
  nm
}

get_function_above = function(n = 2){
  sc = sys.calls()
  n_sc = length(sc)
  
  if(n_sc <= 2) return("nothing_above")
  
  fun_name = deparse_short(sc[[n_sc - 2]][[1]])
  fun_name
}

get_namespace_above = function(n = -2){
    
  if(sys.nframe() >= n){
    res = environmentName(environment(sys.function(n)))
    if(res == ""){
      res = "R_GlobalEnv"
    }
  } else {
    res = "R_GlobalEnv"
  }
  
  res
}


is_operator = function(x, op){
  # operators in formulas

  if(length(x) <= 1){
    res = FALSE
  } else {
    res = as.character(x[[1]])[1] %in% op
  }

  res
}

true_character = function(x){
  is.character(x) && identical(class(x), "character")
}

# fml = x ~ a + b + c*d + I(e + f)
fml_extract_elements = function(fml){

  stopifnot(inherits(fml, "formula"))

  rhs = fml[[3]]

  main_elements = list()
  positive = logical(0)

  is_plus = is_operator(rhs, "+")
  while(is_plus || is_operator(rhs, "-")){
    main_elements = append(main_elements, rhs[[3]])
    positive = append(positive, is_plus)
    rhs = rhs[[2]]
    is_plus = is_operator(rhs, "+")
  }
  main_elements = append(main_elements, rhs)

  # we clean the I()
  for(i in seq_along(main_elements)){
    elem = main_elements[[i]]
    if(is_operator(elem, "I")){
      main_elements[[i]] = elem[[2]]
    }
  }

  res = rev(main_elements)

  res
}

is_any_variable = function(x){
  # x: must be an expression
  length(all.vars(x, max.names = 1)) == 1
}

extract_first_variable = function(expr){
  # this function returns an expression
  # what is considered a variable is either a name (x), either a name subsetted (data$value)
  
  if(!is_any_variable(expr)){
    return(NULL)
  }
  
  # Needed to fix ensure compatibility with old R versions
  if(is.expression(expr)){
    expr = expr[[1]]
  }
  
  if(length(expr) == 1){
    return(expr)
  }
  
  if(is_operator(expr, c("[", "[[", "$"))){
    return(expr)
  } else {
    for(i in 2:length(expr)){
      if(is_any_variable(expr[[i]])){
        return(extract_first_variable(expr[[i]]))
      }
    }
  }
  
  return(NULL)
}

insert = function(x, y, i, replace = FALSE){
  # x = list(a = 1, b = 2, c = 3) ; y = list(u = 33, v = 55) ; i = 2 ; insert(x, y, i)
  # we insert y into x in location i
  # we don't lose the names!
  # replace: whether to erase the value at the index

  mode = mode(x)
  if(!mode %in% c("list", "numeric", "logical", "character", "integer")){
      stop("Internal error: the current mode (", mode, ") is node supported in insert().")
  }

  n_x = length(x)

  if(n_x == 0){
      return(y)
  }

  n_y = length(y)
  res = vector(mode, n_x + n_y - replace)

  names_x = names(x)
  if(is.null(names_x)) names_x = character(n_x)
  names_y = names(y)
  if(is.null(names_y)) names_y = character(n_y)

  if(replace && i <= n_x && length(x) == 1){
    # edge case of replacement
    res = setNames(y, names_y)
    return(res)
  }

  if(i > n_x){
      res[1:n_x] = x
      res[n_x + 1:n_y] = y
      names(res) = c(names_x, names_y)

  } else if(i == 1){
      res[1:n_y] = y
      if(replace){
        res[n_y + 2:n_x] = x
        names(res) = c(names_y, names_x[-1])
      } else {
        res[n_y + 1:n_x] = x
        names(res) = c(names_y, names_x)
      }      

  } else {
      res[1:(i-1)] = x[1:(i-1)]
      res[(i-1) + 1:n_y] = y
      if(replace){
        if(i == n_x){
          names(res) = c(names_x[1:(i-1)], names_y)
        } else {
          res[(i+n_y):(n_x + n_y - 1)] = x[(i + 1):n_x]
          names(res) = c(names_x[1:(i-1)], names_y, names_x[(i + 1):n_x])
        }
      } else {
        res[(i+n_y):(n_x + n_y)] = x[i:n_x]
        names(res) = c(names_x[1:(i-1)], names_y, names_x[i:n_x])
      }
  }

  return(res)
}

insert_at = function(x, y, index){
  # x = list(a = 1, b = 2, c = 3) ; y = list(u = 33, v = 55) ; i = 2 ; insert(x, y, i)
  # we insert y into x in location i
  # we don't lose the names!

  if(length(y) != length(index)){
    stop("Internal error: the length of the object to be inserted must match the length of the indexes.")
  }

  if(max(index) > length(x)){
    stop("Internal error: the value of the index cannot be greater than the length of the object in which ot insert.")
  }

  mode = mode(x)
  if(!mode %in% c("list", "numeric", "logical", "character", "integer")){
      stop("Internal error: the current mode (", mode, ") is node supported in insert().")
  }

  n_x = length(x)
  n_y = length(y)
  res = vector(mode, n_x + n_y)
  names_res = character(n_x + n_y)

  names_x = names(x)
  if(is.null(names_x)) names_x = character(n_x)
  names_y = names(y)
  if(is.null(names_y)) names_y = character(n_y)

  x_index = 1:n_x
  is_insert = +(x_index %in% index)
  x_index = x_index + c(0, cumsum(is_insert)[seq_len(n_x - 1)])

  res[x_index] = x
  names_res[x_index] = names_x

  y_index = index + cumsum(rep(1, n_y))
  res[y_index] = y
  names_res[y_index] = names_y
  
  names(res) = names_res

  return(res)
}






eval_dt = function(call, data = list(), envir){
  
  is_data = TRUE
  if(missing(envir)){
    envir = data
    is_data = FALSE
  }

  dt_data = attr(envir, "dt_data")
  
  if(is_data){
    for(v in names(data)){
      dt_data[[v]] = data[[v]]
    }
  }

  if(".BY" %in% names(dt_data)){
    # we subset the data
    all_vars = all.vars(call)
    for(v in intersect(all_vars, names(dt_data))){
      value = dt_data[[v]]
      if(length(value) > 1){
        dt_data[[v]] = value[dt_data[[".I"]]]
      }
    }
  }

  eval(call, dt_data, envir)
}



# substitute to sprintf which does not handle char length properly
# slower but safer
simple_string_fill = function(x, n = NULL, symbol = " ", right = FALSE, 
                              center = FALSE, center.right = TRUE){
  
  x_nc = nchar(x)
  if(is.null(n)){
    n = max(x_nc)
  }

  pattern = sprintf("% *s", n, "  ")
  if(symbol != " "){
    pattern = gsub(" ", symbol, pattern, fixed = TRUE)
  }
  
  i_add = which(x_nc < n)
  if(length(i_add)){
    x_add = x[i_add]
    extra = substr(rep(pattern, length(x_add)), 1, n - x_nc[i_add])
    if(center){
      nc_extra = nchar(extra)
      mid = if(center.right) ceiling(nc_extra / 2) else floor(nc_extra / 2)
      extra_left = substr(extra, 1, mid)
      extra_right = substr(extra, mid + 1, nc_extra)
      x[i_add] = paste0(extra_left, x_add, extra_right)
      
    } else if(right){
      x[i_add] = paste0(extra, x_add)
    } else {
      x[i_add] = paste0(x_add, extra)
    }
    
  }
  
  x
}



renvir_get = function(key){
  # Get the values of envir variables
  # we also evaluate them

  value_raw = Sys.getenv(key)

  if(value_raw == ""){
      return(NULL)
  }

  # Any default value should be able to be evaluated "as such"
  value_clean = gsub("__%%;;", "\n", value_raw)
  value_clean = gsub("&quot;", '"', value_clean)
  value_clean = gsub("&apos;", "'", value_clean)

  value = eval(str2lang(value_clean))

  return(value)
}

is_string_magic_root = function(){
  isTRUE(renvir_get("string_magic_ROOT"))
}

is_box_open = function(x, .delim){
  substr(x, 1, nchar(.delim[1])) == .delim[1]
}

is_box_close = function(x, .delim){
  n = nchar(x)
  substr(x, n - nchar(.delim[2]) + 1, n) == .delim[2]
}

escape_newline = function(x){
  x = gsub("\n", "\\\\n", x)
  gsub("\t", "\\\\t", x)
}

extract_pipe = function(x, op, double = FALSE, numeric = FALSE, data = NULL, mbt = FALSE){
  # returns:
  # $value
  # $extra
  # $is_double
  
  if(!is.character(x)){
    x = as.character(x)
  }
  
  res = cpp_extract_pipe(x, double)
  
  if(x == "" || res$value == ""){
    
    if(identical(x, "|")){
      res$value = "|"
    }
    
    if(!mbt){
      return(res)
    } else {
      stop_hook("In `string_magic`, the operator {bq?op} must take an argument which has no default value.",
                "\nPROBLEM: no argument is given. Please provide one.")
    }
  }
  
  value_raw = res$value
  if(!is.null(data)){
    # we evaluate
    new_value = try(eval(str2lang(res$value), data), silent = TRUE)
    if(inherits(new_value, "try-error")){
      stop_hook("In `string_magic`, the operator {bq?op} must take a numeric argument.",
                "\nPROBLEM: {bq?res$value} could not be evaluated. See problem below:",
                "\n{new_value}")
    }
    res$value = new_value
  }
  
  if(numeric){
    if(!is.numeric(res$value) && !is_numeric_in_char(res$value)){
      stop_hook("In `string_magic`, the operator {bq?op} must take a numeric argument.",
                "\nPROBLEM: {bq?value_raw} is not numeric.")
    }
    
    res$value = as.numeric(res$value)
  }
  
  res
}


extract_first = function(x, pattern, fixed = FALSE){
  if(length(x) != 1){
    stop("Internal error: x must be of length 1.")
  }
  
  x_pat = regexpr(pattern, x, fixed = fixed, perl = !fixed)
  
  x1 = substr(x, min(1, x_pat - 1), x_pat - 1)
  x2 = substr(x, x_pat + attr(x_pat, "match.length"), 1e5)
  
  c(x1, x2)
}

tag_gen = function(){
  id = getOption("string_magic_tag")
  
  if(is.null(id)){
    id = 1
  }
    
  options(string_magic_tag = id + 1)
  
  items = letters
  n_shifts = id %% 26
  n_reshuffle = id %/% 26
  if(n_reshuffle > 0){
    for (i in 1:n_reshuffle) {
      items = sample(items)
    }
  }
  
  if(n_shifts > 0){
    items = c(items[-(1:n_shifts)], items[1:n_shifts])
  }
  
  tag = paste0(sample(items, 6, replace = TRUE), collapse = "")
  return(tag)
}

setNames = function(x, nm){
  names(x) = nm
  x
}

fix_pkgwdown_path = function(){
  # https://github.com/r-lib/pkgdown/issues/1218
  # just because I use google drive... it seems pkgdown cannot convert to relative path...

  # This is to ensure it only works for me
  if(!is_string_magic_root()) return(NULL)
  # we check we're in the right directory (otherwise there can be prblms with Rmakdown)
  if(!isTRUE(file.exists("R/string_magic_main.R"))) return(NULL)

  all_files = list.files("docs/articles/", full.names = TRUE, pattern = "html$")

  for(f in all_files){
    my_file = file(f, "r", encoding = "UTF-8")
    text = readLines(f)
    close(my_file)
    if(any(grepl("../../../", text, fixed = TRUE))){
      # We embed the images directly: safer

      # A) we get the path
      # B) we transform to URI
      # C) we replace the line

      pat = "<img.+\\.\\./.+/stringmagic/.+/images/"
      qui = which(grepl(pat, text))
      for(i in qui){
        # ex: line = "<img src = \"../../../Google drive/stringmagic/stringmagic/vignettes/images/etable/etable_tex_2021-12-02_1.05477838.png\">"
        line = text[i]
        line_split = strsplit(line, "src *= *\"")[[1]]
        path = gsub("\".*", "", line_split[2])
        # ROOT is always stringmagic
        path = gsub(".+stringmagic/", "", path)
        path = gsub("^articles", "vignettes", path)

        URI = knitr::image_uri(path)

        rest = gsub("^[^\"]+\"", "", line_split[2])
        new_line = .sma('{line_split[1]} src = "{URI}"{rest}')

        text[i] = new_line
      }

      my_file = file(f, "w", encoding = "UTF-8")
      writeLines(text, f)
      close(my_file)
    }
  }

}


####
#### timer ####
####


timer = function(type = "simple"){
  
  if(type == "total"){
    tm = getOption("stringmagic_timer_origin")
  } else {
    tm = getOption("stringmagic_timer")
  }
  
  if(is.null(tm)){
    return("(Error: Please set up the timer first with timer_magic().)")
  }  
  
  time_new = Sys.time()
  difftime = format_difftime(time_new - tm, "full")
  
  if(type == "simple"){
    options(stringmagic_timer = time_new)
  } 
  
  return(difftime)
}

#
# compatibility issues
#

# Nota:
# - gregexec, used only in format_help(), does not exist in R 3.5.0
#   => we can live without it
#   


if(!exists("str2lang", asNamespace("base"))){
  str2lang = function(x){
    parse(text = x, keep.source = FALSE)[[1]]
  }
}

change_defaults = function(fun_name, ...){
  # I don't rewrite the function by attaching the body of the fun_name
  # because I'm a bit wary of namespaces
  #
  # it would work for base R stuff but would be dangerous if extended to
  # functions from imported packages
  # it would also be a problem if some functions I use are conflicted with internal base R funs
  # (although I think they're all exposed)
  #
  # So here it's pretty innocuous, it's a simple rewrite of the call
  # but a bit less efficient
  # 
  
  defaults = list(...)
  for(i in seq_along(defaults)){
    arg_val = parse(text = deparse(defaults[[i]], width.cutoff = 500), keep.source = FALSE)
    if(is.expression(arg_val)){
      arg_val = arg_val[[1]]
    }
    defaults[[i]] = arg_val
  }
  
  fun = parse(text = fun_name, keep.source = FALSE)
  if(is.expression(fun)){
    fun = fun[[1]]
  }
  
  function(...){
    mc = match.call()
    mc_names = names(mc)
    for(arg_name in setdiff(names(defaults), mc_names)){
      mc[[arg_name]] = defaults[[i]]
    }
    
    mc[[1]] = fun
    eval(mc, parent.frame())
  }
  
}

if(is.factor(data.frame(x = "bonjour")$x)){
  data.frame = change_defaults("base::data.frame", stringsAsFactors = FALSE)
  
  as.data.frame = change_defaults("base::as.data.frame", stringsAsFactors = FALSE)
}


#
# startup
#


.onLoad = function(libname, pkgname){
  # setting some options
  
  # operations
  setup_operations()
  
  # generating functions (only if root)
  generate_help_extensive()

  # help
  setup_help_extensive()
  
  # To circumvent a peculiar behavior from pkgdown
  fix_pkgwdown_path()

  invisible()
}

# 
# Generated automatically with indexthis::indexthis_vendor
# this is indexthis version 2.2.0
# 


to_index = function(..., list = NULL, sorted = FALSE, items = FALSE,
                    items.simplify = TRUE){
  return_items = items
  IS_DOT = TRUE
  if(!missing(list) && !is.null(list)){
    if(!is.list(list)){
      stop("The argument `list` must be a list of vectors of the same length.",
           "\nPROBLEM: currently it is not a list.")
    } else if(length(list) == 0){
      stop("The argument `list` must be a list of vectors of the same length.",
           "\nPROBLEM: currently this list is empty.")
    }
    dots = list
    IS_DOT = FALSE
  } else {
    dots = list(...)
  }  
  Q = length(dots)
  n_all = lengths(dots)
  n = n_all[1]
  if(length(unique(n_all)) != 1){
    stop("All elements in `...` should be of the same length (current lenghts are ", 
         paste0(n_all, collapse = ", "), ").")
  }
  if(n == 0){
    res = integer(0)
    if(return_items){
      items = integer(0)
      if(items.simplify){
        items = data.frame()
      }
      res = list(index = res, items = items)
    }
    return(res)
  }
  info = cpp_to_index(dots)
  if(isTRUE(info$is_error)){
    stop(info$error_msg)
  }
  index = info$index
  if(sorted || return_items){
    items_unik = vector("list", Q)
    for (q in 1:Q) {
      items_unik[[q]] = dots[[q]][info$first_obs]
    }
    if(sorted){
      x_order = do.call(order, items_unik)
      index = order(x_order)[index]
      for (q in 1:Q) {
        items_unik[[q]] = items_unik[[q]][x_order]
      }
    }
    items = NULL
    if(items.simplify && Q == 1){
      items = items_unik[[1]]
    } else {
      user_names = names(dots)
      if(is.null(user_names)){
        user_names = character(Q)
      }
      if(IS_DOT){
        mc_dots = match.call(expand.dots = FALSE)[["..."]]
      }
      for(q in 1:Q){
        if(nchar(user_names[q]) == 0){
          is_done = FALSE
          if(IS_DOT){
            mcq = mc_dots[[q]]
            if(is.name(mcq)){
              user_names[q] = as.character(mcq)[1]
              is_done = TRUE
            } else if(is.call(mcq) && as.character(mcq[[1]])[1] == "$"){
              user_names[q] = as.character(mcq[[3]])[1]
              is_done = TRUE
            }
          }
          if(!is_done){
            user_names[q] = paste0("x", q)
          }          
        }
      }
      names(items_unik) = user_names
      items = as.data.frame(items_unik)
      row.names(items) = 1:nrow(items)
    }
    if(return_items){
      res = list(index = index, items = items)
    } else {
      res = index
    }
  } else {
    res = index
  }
  res
}

#----------------------------------------------#
# Author: Laurent Berge
# Date creation: Wed Jul 27 09:40:24 2022
# ~: string ops core functions
#----------------------------------------------#


####
#### User-level ####
####


#' Register custom operations to apply them in `string_magic`
#' 
#' Extends the capabilities of [string_magic()] by adding any custom operation
#' 
#' @param fun A function which must have at least the arguments 'x' and '...'. 
#' Additionnaly, it can have the arguments: 'argument', 'options', 'group', 'group_flag'.
#' This function must return a vector.
#' This function will be internally called by `string_magic` in the form 
#' `fun(x, argument, options, group, group_flag)`.`x`: the value to which the 
#' operation applies. `argument`: the quoted `string_magic` argument (always character). 
#' `options`: a character vector of `string_magic` options. The two last arguments are of use
#' only in group-wise operations if `fun` changes the lengths of vectors. `group`: an index of
#' the group to which belongs each observation (integer). `group_flag`: value between 0
#' and 2; 0: no grouping operation requested; 1: keep track of groups; 2: apply grouping.
#' @param ops Character scalar representing a valid chain of `string_magic` operations. It should
#' be of the form `"op1, 'arg'op2, etc"`. For example `"'80|-'fill"` fills the line
#' with dashes up to 80 characters. 
#' @param alias Character scalar, the name of the operation.
#' @param valid_options A character vector or NULL (default). Represents a list of 
#' valid options for the operation. This is used: a) to enable auto-completion,
#' b) for error-handling purposes.
#' @param namespace Character scalar or `NULL` (default). **Only useful for package developers.**
#' As a regular end-user you shouldn't care! If your package uses `string_magic`, you should care. 
#' If the function `string_magic_register_*` is located in the `onLoad` function (see `help("onLoad")`), 
#' there is nothing to do. Otherwise, pass the name of your package in this argument to 
#' make all the new operation definitions scoped (i.e. only your package can access it and 
#' it can't be messed up by end users).
#' 
#' 
#' @details 
#'  
#' We try to strongly check the new operations since it's always better to find out problems
#' sooner than later. This means that when the function is defined, it is also 
#' tested.
#' 
#' If you pass a function, note that it should work for non-character arguments in `x`.
#' 
#' @section Writing a package using `string_magic`:
#' 
#' If you want to use `string_magic` in your package and want to make use of custom operations:
#' 
#' - place any `string_magic_register_fun` and `string_magic_register_ops` in your `.onLoad` function
#' (see `help("onLoad")`). The .onLoad function is run whenever the package is loaded 
#' for the first time. It's a function that you can place anywhere in your `R/*` files 
#' and which looks like this:
#' ```{r}
#' .onLoad = function(libname, pkgname){
#'   # string_magic custom operations
#'   string_magic_register_ops("'80|-'fill", "h1")
#' 
#'   invisible()
#' }
#' ```
#' - if you don't want to place the `string_magic_register_*` functions in the .onLoad function, 
#' you can, but then you **must** provide the argument `namespace`:
#' ```{r}
#' string_magic_register_ops("'80|-'fill", "h1", namespace = "myPackageName")
#' ```
#' - you must create an [string_magic_alias()] to create an alias to [string_magic()] and use the 
#' argument `.namespace = "myPackageName"`. Use this opportunity to change the 
#' defaults if you wish. You can even override the `string_magic` function:
#'  ```{r}
#' # creating an alias with the same name + changing the delimiter
#' string_magic = stringmagic::string_magic_alias(.namespace = "myPackageName", .delim = "{{ }}")
#' ```
#' 
#' @author 
#' Laurent R. Berge
#' 
#' @return 
#' These function do not return anything. They register new operations to be used in the 
#' `string_magic` family of functions by placing them in the options (later fetched by 
#' `string_magic()` at run-time).
#' 
#' @family related to string_magic
#' 
#' @inherit string_clean seealso
#' 
#' @examples
#' 
#' # let's create an operation that adds markdown emphasis
#' # to elements of a vector
#' 
#' # A) define the function
#' fun_emph = function(x, ...) paste0("*", x, "*")
#' 
#' # B) register it
#' string_magic_register_fun(fun_emph, "emph")
#' 
#' # C) use it
#' x = string_vec("right, now")
#' string_magic("Take heed, {emph, c? x}.")
#' 
#' #
#' # now let's add the option "strong"
#' fun_emph = function(x, options, ...) {
#'   if("strong" %in% options){
#'     paste0("***", x, "***")
#'   } else {
#'     paste0("*", x, "*")
#'   }
#' }
#' 
#' string_magic_register_fun(fun_emph, "emph", "strong")
#' 
#' x = string_vec("right, now")
#' string_magic("Take heed, {emph.strong, c? x}.")
#' 
#' #
#' # now let's add an argument
#' fun_emph = function(x, argument, options, ...){
#'   arg = argument
#'   if(nchar(arg) == 0) arg = "*"
#'   
#'   if("strong" %in% options){
#'     arg = paste0(rep(arg, 3), collapse = "")
#'   }
#'   
#'   paste0(arg, x, arg)
#' }
#' 
#' string_magic_register_fun(fun_emph, "emph", "strong")
#' 
#' x = string_vec("right, now")
#' string_magic("Take heed, {'_'emph.s, c? x}.")
#' 
#' #
#' # using string_magic_register_ops
#' #
#' 
#' # create a 'header' maker
#' string_magic_register_ops("tws, '# 'paste, ' 'paste.right, '40|-'fill", "h1")
#' cat_magic("{h1 ! My title}\n my content")
#' 
#' 
#' 
#' 
string_magic_register_fun = function(fun, alias, valid_options = NULL, namespace = NULL){
  # fun: must be a function with x and ... as arguments
  # the argument names must be in:
  # x, argument, options, group, conditonnal_flag

  if(missing(fun)){
    stop("The argument `fun` must be provided. PROBLEM: it is missing.")
  }

  if(!is.function(fun)){
    stopi("The argument `fun` must be a function. ",
          "PROBLEM: it is not a function, instead it is of class {enum.bq?class(fun)}.")
  }

  check_character(alias, scalar = TRUE, mbt = TRUE)
  check_character(valid_options, no_na = TRUE, null = TRUE)

  fun_args = names(formals(fun))
  
  arg_must = c("...", "x")
  arg_missing = setdiff(arg_must, fun_args)
  if(length(arg_missing) > 0){
    stopi("The argument `fun` must be a function with {enum.bq?arg_missing} in its arguments.",
              "\nPROBLEM: it has no argument {enum.bq.or?arg_missing}.")
  }

  valid_args = string_vec("x, argument, options, group, group_flag")
  arg_pblm = setdiff(setdiff(fun_args, "..."), valid_args)
  if(length(arg_pblm) > 0){
    stopi("The argument `fun` must have specific argument names. Valid arguments are {enum.bq.or?valid_args}.",
              "\nPROBLEM: the argument{$s, enum.bq, are?arg_pblm} invalid.")
  }
  
  # we only test for simple cases
  if(!"argument" %in% fun_args){
    options = head(valid_options, 1)
    run = try(fun(x = 1:5, options = options, group_flag = 0, group = rep(1, 5)), silent = TRUE)
    if(isError(run)){
      mc = match.call()
      fun_dp = if(length(mc$fun) == 1) deparse_short(fun) else ""
      stopi("The function defined in argument `fun` failed to run for a simple case.",
            "\n{&nchar(fun_dp)>0 ; {.}(x = 1:5) ; Running it on x = 1:5} leads to an error,",
            " see below:",
            "\n{run}")
    }  
  }
  
  
  #
  # adding attributes
  #
  
  if(!is.null(valid_options)){
    attr(fun, "valid_options") = valid_options
  }
  
  if("group_flag" %in% fun_args){
    attr(fun, "group_flag") = TRUE
  }
  
  #
  # saving within a namespace
  #
  
  if(get_function_above() == ".onLoad"){
    namespace = get_namespace_above()
  }  
  
  save_user_fun(fun, alias, namespace)
}

#' @describeIn string_magic_register_fun Create new combinations of `string_magic` operations
string_magic_register_ops = function(ops, alias, namespace = NULL){
  
  #
  # checking
  #
  
  check_character(ops, scalar = TRUE, mbt = TRUE)
  check_character(alias, scalar = TRUE, mbt = TRUE)
  
  run = try(string_magic("x{1:5}", .last = ops), silent = TRUE)
  
  if(isError(run)){
    stopi("The argument `ops` must refer to valid `string_magic` operations.", 
          "\nPROBLEM: running `string_magic(\"x\\{1:5}\", .last = {Q?ops})` leads ",
          "to an error, see below:",
          "\n{run}")
  }
  
  #
  # adding attributes
  #
  
  if(!is.null(ops)){
    attr(ops, "simple_ops") = TRUE
  }
  
  #
  # saving within a namespace
  #
  
  if(get_function_above() == ".onLoad"){
    namespace = get_namespace_above()
  } 
  
  save_user_fun(ops, alias, namespace)  
}

save_user_fun = function(fun, alias, namespace){
  # We save the user function in the global options
  
  if(is.null(namespace)){
    namespace = "R_GlobalEnv"
  }
  
  # We forbid the redefinition of internal operations
  # => only for V1, so that later versions don't break code
  #    if user did defined operations with the same name of the new ones
  OPERATORS_v1.0.0 = getOption("string_magic_operations_v1.0.0")
  if(alias %in% OPERATORS_v1.0.0){
    stopi("The argument `alias` must not be equal to an existing internal argument.",
              "\nPROBLEM: the operation {bq?alias} is already an internal operation.")
  }
  
  user_ops_all = getOption("string_magic_user_ops")
  if(is.null(user_ops_all) || !is.list(user_ops_all)){
    user_ops_all = list()
  }
  
  user_info = user_ops_all[[namespace]]
  if(is.null(user_info)){
    funs = list()
    funs[[alias]] = fun
    operators_default = getOption("string_magic_operations_default")
    operators = c(operators_default, alias)
  } else {
    funs = user_info$funs
    funs[[alias]] = fun
    operators = user_info$operators
    operators = unique(c(operators, alias))
  }
  
  user_info = list(funs = funs, operators = operators)
  user_ops_all[[namespace]] = user_info

  options("string_magic_user_ops" = user_ops_all)
}

#' Display messages using interpolated strings
#' 
#' Utilities to display messages using `string_magic` interpolation and operations to generate the message.
#' 
#' @inheritParams string_magic
#' 
#' @param .end Character scalar, default is `""` (the empty string) for `cat_magic`, 
#' and `"\n"` (a newline) for `message_magic`. 
#' This string will be collated at the end of the message (a common alternative is `"\n"`).
#' @param .width Can be 1) a positive integer, 2) a number in (0;1), 3) `FALSE` (default 
#' for `cat_magic`), 4) `NULL` (default for `message_magic`), or 5) a one-sided formula. It represents the target 
#' width of the message on the user console. Newlines will be added *between words* to fit the
#' target width.
#' 
#' 1. positive integer: number of characters
#' 2. number (0;1): fraction of the screen
#' 3. `FALSE`: does not add newlines
#' 4. `NULL`: the min between 100 characters and 90% of the screen width
#' 5. one-sided formula: an expression in which you can use the special variable
#' `.sw` which represents the current screen width
#' 
#' Note that the value `NULL` is equivalent to using `~min(100, 0.9*.sw)`.
#' @param .leader Character scalar, default is `TRUE`. Only used if argument `.width` is not `FALSE`. 
#' Whether to add a leading character string right after the extra new lines.
#' 
#' @details 
#' These functions are [base::cat()]/[message()] wrappers aroung [string_magic()]. There is one notable difference
#' with respect to `cat`/`message`. It's the ability to add newlines after words for 
#' the message to fit a target width. This is controlled with the argument `.width`. This is 
#' active by default for `message_magic` (default is `.width = NULL` which leads to the 
#' minimum betwen 100 characters and 90% of the screen width).
#' 
#' You can very easily change the default values with the alias generators `cat_magic_alias` and 
#' `message_magic_alias`.
#' 
#' `[Advanced]` A note for package developers who would use these functions **and**
#' also use custom `string_magic` operations created with [string_magic_register_fun()] or
#' [string_magic_register_ops()]. To ensure forward compatibility the new operations created 
#' should be defined in the package namespace (see the *ad hoc* section in [string_magic_register_fun()] help).
#' To access these operators in their specific namespaces, you must use an alias with 
#' `cat_magic_alias` or `message_magic_alias` with the argument `.namespace = "myPackageName"` 
#' (to avoid having to provide the `.namespace` argument repeatedly).
#' 
#' @return 
#' The functions `cat_magic()` and `message_magic()` do not return anything, they simply print on the console. 
#' 
#' The function `cat_magic_alis()` returns a function behaving identically to [cat_magic()] but for which the
#' default values have been altered.
#' Same for `message_magic_alias()`.
#' 
#' @family tools with aliases
#' 
#' @inherit string_clean seealso
#' 
#' @examples 
#' 
#' start = Sys.time()
#' Sys.sleep(0.05)
#' message_magic("This example has run in {difftime ? start}.")
#' 
#' cat_magic("Let's write a very long message to illustrate how .width work.", 
#'           .width = 40)
#' 
#' # Let's add a leader
#' cat_magic("Let's write a very long message to illustrate how `.width` work.", 
#'           "And now we add `.leader`.", .width = 40, .leader = "#> ")
#' 
#' # newlines respect the introductory spaces
#' cat_magic("Here's a list:", 
#'           "    + short item", 
#'           "    + this is a very long item that likely overflows", 
#'          .width = 30, .sep = "\n")
#' 
#' 
cat_magic = function(..., .sep = "", .end = "", .width = FALSE, .leader = "", .data = list(),
                     .envir = parent.frame(), .delim = c("{", "}"), .last = NULL, 
                     .trigger = TRUE, .check = TRUE, .help = NULL, 
                     .namespace = NULL){
  
  if(!isTRUE(.trigger)) return(invisible(NULL))
  
  set_pblm_hook()
  
  if(.check){
    dots = check_set_dots(..., nofun = TRUE)
    check_character(.end, scalar = TRUE)
  } else {
    dots = list(...)
  }
  
  if(length(dots) == 0){
    return(invisible(NULL))
  }
  
  if(all(lengths(dots) == 1)){
    txt = string_magic(..., .envir = .envir, .data = .data, .delim = .delim, .last = .last,
                       .sep = .sep, .check = .check, .help = .help, .split = FALSE,
                       .namespace = .namespace)
  } else {
    txt = string_vec(..., .envir = .envir, .data = .data, .delim = .delim, .last = .last,
                     .check = .check, .help = .help, .split = FALSE,
                     .namespace = .namespace)
  }
  
  # all this is needed to implement lazy default values with yet to be evaluated expressions (.sw)
  if(is.character(.width)){
    .width = str2lang(.width)
  }
  is_call = isTRUE(try(is.call(.width), silent = TRUE))
  .width = if(is_call) .width else substitute(.width)
  .width = check_set_width(.width)
  if(is.finite(.width)){
    txt = fit_screen(txt, .width, leader = .leader)
  }
  
  if(length(txt) > 1){
    txt = paste0(txt, collapse = .sep)
  }
  
  if(nchar(.end) > 0){
    txt = paste0(txt, .end)
  }
  
  .width = check_set_width(.width)
  if(is.finite(.width)){
    txt = fit_screen(txt, .width, leader = .leader)
  }

  cat(txt)
}

#' @describeIn cat_magic Alias to `cat_magic`
catma = cat_magic

#' @describeIn cat_magic Display messages using interpolated strings
message_magic = function(..., .sep = "", .end = "\n", .width = NULL, 
                         .leader = "", .envir = parent.frame(), .data = list(), 
                         .delim = c("{", "}"), 
                         .last = NULL, .trigger = TRUE,
                         .check = TRUE, .help = NULL, 
                         .namespace = NULL){
  
  if(!isTRUE(.trigger)) return(invisible(NULL))
  
  set_pblm_hook()
  
  if(.check){
    dots = check_set_dots(..., nofun = TRUE)
    check_character(.end, scalar = TRUE)
  } else {
    dots = list(...)
  }
  
  if(length(dots) == 0){
    return(invisible(NULL))
  }
  
  if(all(lengths(dots) == 1)){
    txt = string_magic(..., .envir = .envir, .data = .data, .delim = .delim, .last = .last,
                       .sep = .sep, .check = .check, .help = .help, .split = FALSE,
                       .namespace = .namespace)
  } else {
    txt = string_vec(..., .envir = .envir, .data = .data, .delim = .delim, .last = .last,
                     .check = .check, .help = .help, .split = FALSE,
                     .namespace = .namespace)
  }
    
  if(length(txt) > 1){
    txt = paste0(txt, collapse = .sep)
  }
  
  if(nchar(.end) > 0){
    txt = paste0(txt, .end)
  }
  
  .width = check_set_width(.width)
  if(is.finite(.width)){
    txt = fit_screen(txt, .width, leader = .leader)
  }

  message(txt, appendLF = FALSE)
}

#' @describeIn cat_magic Alias to `message_magic`
mema = message_magic

#' Sets up a timer that can be used within `_magic` functions
#' 
#' Sets up a timer which can later be summoned by [string_magic()] functions via
#' the `.timer`, `.timer_lap` and `.timer_total` variables. Useful to report 
#' timings within functions with the function [cat_magic()] or [message_magic()].
#' 
#' @details
#' This functions sets up a timer with [base::Sys.time()]. This timer can then be tracked 
#' and modified with the `.timer`, `.timer_lap` and `.timer_total` variables within 
#' [cat_magic()] or [message_magic()].
#' 
#' Note that the timer is precise at +/- 1ms, hence it should **not** be used to time 
#' algorithms with very short execution times. 
#' 
#' It works by saving the current system time in R options (`stringmagic_timer` and `stringmagic_timer_origin`).
#' Hence, since it uses options, it should not be used in parallel processes.
#' 
#' @return 
#' This function does not return anything and is only intended to be used in
#' conjunction with future calls of [string_magic()]. 
#' 
#' 
#' @author 
#' Laurent Berge
#' 
#' @inherit cat_magic seealso
#' 
#' @examples 
#' 
#' # simple example where we time the execution of some elements in a function
#' # we trigger the message conditionally on the value of the argument `debug`.
#' rnorm_crossprod = function(n, mean = 0, sd = 1, debug = FALSE){
#'   # we set the timer
#'   timer_magic()
#'   # we compute some stuff
#'   x = rnorm(n, mean, sd)
#'   # we can report the time with .timer
#'   message_magic("{15 align ! Generation}: {.timer}", .trigger = debug)
#'   
#'   res = x %*% x
#'   message_magic("{15 align ! Product}: {.timer}",
#'                 "{15 align ! Total}: {.timer_total}", 
#'                 .sep = "\n", .trigger = debug)
#'   res
#' }
#' 
#' rnorm_crossprod(1e5, TRUE)
#' 
#' 
timer_magic = function(){
  tm = Sys.time()
  options(stringmagic_timer = tm)
  options(stringmagic_timer_origin = tm)
}

####
#### ... string_magic ####
####


#' @describeIn string_magic String interpolation with operation chaining
string_magic = function(..., .envir = parent.frame(), .data = list(), 
                        .sep = "", .vectorize = FALSE, 
                        .delim = c("{", "}"), .last = NULL, .post = NULL, .nest = FALSE,
                        .collapse = NULL, .invisible = FALSE, .default = NULL,
                        .trigger = TRUE, 
                        .check = TRUE, .class = NULL, .help = NULL, 
                        .namespace = NULL){
  
  if(!isTRUE(.trigger)) return(invisible(NULL))
  
  if(.check){
    set_pblm_hook()
    
    if(!missing(.envir)) check_envir(.envir)
    if(!missing(.data)){
      if(!is.list(.data)){
        stop("The argument `.data` must be a list or a data.frame.",
             "PROBLEM: instead it is of class ", class(.data)[1], ".")
      }
    }
    if(!missing(.sep)) check_character(.sep, scalar = TRUE)
    if(!missing(.vectorize)) check_logical(.vectorize, scalar = TRUE)
    if(!missing(.nest)) check_logical(.nest, scalar = TRUE)
    if(!missing(.delim)){
      .delim = check_set_delimiters(.delim)
    }
    if(!missing(.last)) check_last(.last)
    if(!missing(.collapse)) check_character(.collapse, null = TRUE, scalar = TRUE)
    if(!missing(.invisible)) check_logical(.invisible, scalar = TRUE)
    if(!missing(.default)) check_character(.default, null = TRUE, scalar = TRUE)
  }
    
  if(length(.delim) == 1){
    .delim = strsplit(.delim, " ", fixed = TRUE)[[1]]
    if(.check){ 
      check_delimiters(.delim)
    }
  }  

  if(...length() == 0){
    if(missnull(.help)){
      return("")
    }
  } 
  
  res = string_magic_internal(..., .delim = .delim, 
                              .envir = .envir, .data = .data, .sep = .sep,
                              .vectorize = .vectorize, .help = .help, .nest = .nest,
                              .collapse = .collapse, .is_root = TRUE, 
                              .namespace = .namespace, .default = .default,
                              .check = .check, .last = .last)
  
  if(!is.null(attr(res, "group_index"))){
    # cleaning artifacts
    attr(res, "group_index") = NULL
  }
  
  if(!is.null(.post)){
    # .post must be a function
    # => we **don't** catch the arguments (too error prone, and pretty useless TBH)
    check_function(.post, argname = ".post")
    res = .post(res)
  }
  
  if(!is.null(.class)){
    class(res) = c(.class, class(res))
  }
  
  if(.invisible){
    return(invisible(res))
  }
  
  res
}

#' @describeIn string_magic A simpler version of `string_magic` without any error handling to save a few micro seconds
.string_magic = function(..., .envir = parent.frame(), .data = list(), 
                         .sep = "", .vectorize = FALSE,
                         .delim = c("{", "}"), .collapse = NULL, .last = NULL, .nest = FALSE,
                         .trigger = TRUE, .namespace = NULL){
  
  if(!.trigger) return(invisible(NULL))
  
  if(length(.delim) == 1){
    .delim = strsplit(.delim, " ", fixed = TRUE)[[1]]
  }

  string_magic_internal(..., .delim = .delim, 
                        .envir = .envir, .data = .data,
                        .sep = .sep, .namespace = .namespace, .nest = .nest,
                        .vectorize = .vectorize, .is_root = TRUE, 
                        .collapse = .collapse,
                        .check = FALSE, .last = .last)
}

# This is an internal alias (not exported)
.sma = .string_magic

#' @describeIn string_magic Alias to `string_magic`
sma = string_magic

####
#### Internal ####
####


string_magic_internal = function(..., .delim = c("{", "}"), .envir = parent.frame(), 
                                 .data = list(),
                                 .sep = "", .vectorize = FALSE,
                                 .collapse = NULL, .last = NULL, .nest = FALSE,
                                 .help = NULL, .is_root = FALSE, 
                                 .namespace = NULL, .user_funs = NULL,
                                 .valid_operators = NULL, .default = NULL,
                                 .check = FALSE, .plural_value = NULL){
  
  # flag useful to easily identify this environment (used in error messages)
  is_string_magic_internal = TRUE

  if(!is.null(.help)){
    on.exit(string_magic_dynamic_help(.help))
    stop_up("string_magic: Help requested.")  
  }
  
  if(...length() == 0){
    return("")
  }
  
  if(.is_root){
    if(is.null(.namespace)){
      .namespace = "R_GlobalEnv"
    }
    user_ops_all = getOption("string_magic_user_ops")
    # beware the sneaky assignment!
    if(!is.null(user_ops_all) && !is.null(user_info <- user_ops_all[[.namespace]])){
      #                                             ^^ sneaky!
      .user_funs = user_info$funs
      .valid_operators = user_info$operators          
    } else {
      .valid_operators = getOption("string_magic_operations_default")
    }
    if(length(.data) > 0){
      .data = unclass(.data)
    }
  }
  
  if(...length() == 1){
    
    # Note: using ...names() conditionnally adds 110us!!!! (getRversion() >= "4.1.3" is slow)
    # otherwise the line below is about half a micro second
    dots_nm = names(list(...))
    is_no_names = !is.null(dots_nm)
    
    if(is_no_names){
      stop_hook("`string_magic` requires at least one character scalar to work.",
                "\nNamed arguments are only used as variables on which to apply interpolation.",
                "\nFIX: please provide at least one non-named argument.")
    }
    
    if(.check){
      x = as.character(..1)
    } else {
      x = check_expr(as.character(..1), 
                     "In `string_magic`, the argument could not be evaluated.")
    }
    
    
    if(length(x) > 1){
      stop_hook("`string_magic` can only be applied to character scalars. ",
                "\nPROBLEM: the argument is not of length 1, it is of length {len ? x}.")
    }

  } else {

    if(.check){
      dots = check_expr(list(...), "In `string_magic`, one element of ... could not be evaluated.")
    } else {
      dots = list(...)
    }
    
    # we check if some variables were passed in the arguments
    if(.is_root){
      # ONLY HAPPENS IN THE ROOT CALL
      dot_names = names(dots)
      if(!is.null(dot_names)){
        is_var = dot_names != ""
        for(i in which(is_var)){
          if(is.null(dots[[i]])){
            .data[dot_names[i]] = list(NULL)
            
          } else {
            .data[[dot_names[i]]] = dots[[i]]
          }
        }
        dots[is_var] = NULL      
      }
    }
    
    if(length(dots) == 0){
      stop_hook("`string_magic` requires at least one character scalar to work.",
                "\nNamed arguments are only used as variables on which to apply interpolation.",
                "\nPROBLEM: all arguments are named.",
                "\nFIX: please provide at least one non-named argument.")
    }
    
    if(any(lengths(dots) > 1)){
      qui = which(lengths(dots) > 1)[1]
      stop_hook("`string_magic` can only be applied to character scalars.",
                "\nPROBLEM: The ", n_th(qui),
                " elment in ... is of length ", length(dots[[qui]]), ".")
    }

    if(!.vectorize){
      # Note: using paste(..1, ..2, sep = .sep) explicitly only saves 2us vav do.call
      # not worth it.

      dots$sep = .sep
      x = do.call(paste, dots)
    } else {
      
      if(.nest){
        stop_hook("The argument `.nest` is not compatible with the argument `.vectorize`.",
                  "One of the two must be set to `FALSE`.")
      }
      
      # vectorize
      n = length(dots)
      res = vector("list", n)
      
      now_done = FALSE
      date_done = FALSE
      
      for(i in 1:n){
        
        di = dots[[i]]
        if(.is_root && length(di) == 1 && is.character(di)){
          if(!now_done && grepl(".now", di, fixed = TRUE)){
            if(grepl(".now(", di, fixed = TRUE)){
              .data[[".now"]] = function(x) format(Sys.time(), x)
            } else {
              .data[[".now"]] = Sys.time()
            }
            
            now_done = TRUE
          }
          
          if(grepl(".timer", di, fixed = TRUE)){
            if(grepl(".timer([^_]|$)", di)){
              .data[[".timer"]] = timer("simple")
            }
            
            if(grepl(".timer_lap", di, fixed = TRUE)){
              .data[[".timer_lap"]] = timer("lap")
            }
            
            if(grepl(".timer_total", di, fixed = TRUE)){
              .data[[".timer_total"]] = timer("total")
            }
          }
          
          if(!date_done && grepl(".date", di, fixed = TRUE)){
            .data[[".date"]] = Sys.Date()
            date_done = TRUE
          }
        }
        
        res[[i]] = string_magic_internal(dots[[i]], .delim = .delim, .envir = .envir, 
                                         .data = .data, .check = .check,
                                         .user_funs = .user_funs, 
                                         .valid_operators = .valid_operators)
      }
      
      res = unlist(res)
      
      if(!true_character(res)){
        res = as.character(res)
      }
      
      if(!is.null(.last)){
        if(is.function(.last)){
          res = .last(res)
        } else {
          res = apply_simple_operations(res, ".last", .last, .check, .envir, .data,
                                        group_flag = 1 * grepl("~", .last, fixed = TRUE), 
                                        .delim, .user_funs = .user_funs, 
                                        .valid_operators = .valid_operators)
        }
      }

      return(res)
    }
  }

  if(is.na(x) || length(x) == 0){
    return(x)
  }
  
  if(.nest){
    x = paste0(.delim[1], x, .delim[2])
  }

  BOX_OPEN = .delim[1]
  
  if(!grepl(BOX_OPEN, x, fixed = TRUE)){
    if(is.null(.last)){
      return(x)
    } else {
      x_parsed = list(x)
    }    
  } else {
    x_parsed = cpp_string_magic_parser(x, .delim)
    if(length(x_parsed) == 1 && isTRUE(attr(x_parsed, "error"))){
      report_string_magic_parsing_error(x, x_parsed, .delim)
    }
    
    # we add extra variables
    if(.is_root){
      if(grepl(".now", x, fixed = TRUE)){
        if(grepl(".now(", x, fixed = TRUE)){
          .data[[".now"]] = function(x) format(Sys.time(), x)
        } else {
          .data[[".now"]] = Sys.time()
        }
      }
      
      if(grepl(".timer", x, fixed = TRUE)){
        if(grepl(".timer([^_]|$)", x)){
          .data[[".timer"]] = timer("simple")
        }
        
        if(grepl(".timer_lap", x, fixed = TRUE)){
          .data[[".timer_lap"]] = timer("lap")
        }
        
        if(grepl(".timer_total", x, fixed = TRUE)){
          .data[[".timer_total"]] = timer("total")
        }
      }
      
      if(grepl(".date", x, fixed = TRUE)){
        .data[[".date"]] = Sys.Date()
      }
    }
  }
  

  # NOTE on the parsing:
  # - parsing returns a list.
  #   If, eg: x = "str1{stuff1}str2{stuff2}"
  #   we will obtain a list of 4 elements
  #   note that the empty string between two boxes is an element
  # - an element can be either of length 1 or of length 2
  #   * l=1: means this is a regular string
  #   * l=2: this is a list of two elements:
  #     + the first element is a vector of operators, with ? or ! always at the end (when available)
  #     + the second element is the string to be interpolated (when available, some operators don't
  #       require interpolation, like pluralization)
  #   ex: "hello: {u, C ! {S!luc, domi}}"
  #     [[1]]: "hello: "
  #     [[2]]:
  #           [[1]]: vector c("u", "C")
  #           [[2]]: scalar "{S!luc, domi}"
  # - note that nestedness is dealt with in here and NOT in the parser.
  #

  ANY_PLURAL = isTRUE(attr(x_parsed, "plural"))
  if(ANY_PLURAL){
    # we need to book keep the evaluations to 'know' to which the pluralization applies
    x_values_all = vector("list", length(x_parsed))
  }

  n_x_all = lengths(x_parsed)
  
  n = length(x_parsed)
  res = character(0)
  i_done = FALSE
  for(i in 1:n){

    if(i_done){
      i_done = FALSE
      next
    }

    xi = x_parsed[[i]]

    if(length(xi) == 1){
      # This is a regular string (no operations involved)

      if(i == 1){
        res = xi
      } else if(nchar(xi) > 0){
        res = paste0(res, xi)
      }

    } else {
      #
      # operations
      #

      operators = xi[[1]]
      xi = xi[[2]]

      if(length(operators) == 0){

        # we need to evaluate xi
        xi_call = check_set_string_magic_parsing(xi, .check, .delim)

        if(is.character(xi_call)){
          xi = xi_call
        } else {
          xi = check_set_string_magic_eval(xi_call, .data, .envir, .check)
        }

        if(ANY_PLURAL){
          # we save
          x_values_all[[i]] = xi
        }
        
        if(!is.null(.default)){
          # We apply the default operations
          xi = apply_simple_operations(xi, ".default", .default, .check, .envir, .data,
                                       group_flag = 1 * grepl("~", .default, fixed = TRUE), 
                                       .delim, .user_funs = .user_funs, 
                                       .valid_operators = .valid_operators)
        }

      } else {

        n_op = length(operators)
        verbatim = operators[n_op] == "!"
        
        if(operators[1] %in% c("&", "&&")){
          .data[["len"]] = function(x) length(x)
        }

        is_ifelse = FALSE
        concat_nested = FALSE
        is_xi_done = FALSE
        is_plural = ANY_PLURAL && operators[1] %in% c("$", "#")
        i_candidate = i # => used when evaluating later xi by anticipation
        if(ANY_PLURAL){

          if(!is_plural){
            # if the xi has already been evaluated by anticipation, we catch it
            if(lengths(x_values_all)[i] > 0){
              xi = x_values_all[[i]]
              is_xi_done = TRUE
            }
          }

          if(is_plural){
            #
            # Finding out candidates for pluralization
            #

            if(identical(xi, "")){

              if(!is.null(.plural_value)){
                # we are in a nested pluralization call:
                # ex: "{$(;; Maybe you meant: {enum.bq.4}?) ? sugg)}"
                # here 'sugg' would be in `plural_value`
                # we would be in the evaluation of 
                # " Maybe you meant: {enum.bq.4}?"
                # if we did not pass it along, there would have been an error

                xi = .plural_value
                is_xi_done = TRUE

              } else {
                if(length(operators) == 1){
                  example = 'Example: x = c("Mark", "Sarah"); string_magic("{$enum, is ? x} away.")'
                  .stop_hook("In `string_magic`, the pluralization operator (`", operators[1], 
                             "`) must contain operations. ", '\n', example)
                }

                # We need to find it out!!!!!!
                xl = lengths(x_values_all)
                if(all(xl == 0)){
                  # We need to evaluate the next ones
                  ok = FALSE
                  if(i < n){
                    for(j in (i + 1):n){
                      if(length(x_parsed[[j]]) == 2 && !identical(x_parsed[[j]][[2]], "")){
                        ok = TRUE
                        break
                      }
                    }
                  }

                  if(!ok){
                    pblm = paste0(operators[1], paste0(operators[-1], collapse = ", "))
                    example = 'Example: x = c("Mark", "Sarah"); string_magic("{$enum, is ? x} away.")'

                    .stop_hook("In `string_magic`, the pluralization (`", pblm, "`) did not find any value to pluralize on. ",
                            "Please provide one by adding it after a question mark.\n", example)
                  }

                  i_candidate = j
                  xi = x_parsed[[j]][[2]]

                } else {
                  # Means that stg has been evaluated before
                  qui = tail(which(xl > 0), 1)
                  xi = x_values_all[[qui]]
                  is_xi_done = TRUE
                }
              }

            } else {
              # xi != ""

              # last item is ? or !, we trim it
              operators = operators[-n_op]
            }
          }
        }

        if(!is_plural) {
          # regular operators

          if(operators[1] %in% c("&", "&&")){
            # if else operator
            is_ifelse = TRUE
            xi_raw = xi = operators[2]
            verbatim = FALSE
          } else {
            operators = operators[-n_op]
            # The two separators ? and ! have no default operation
          }

          # If split operator, we need to concatenate
          concat_nested = length(operators) > 0 && grepl("(?:^|[^[:alpha:]])(?:s|S)$", operators[[1]])
        }


        #
        # evaluation of xi
        #


        if(!is_xi_done){
          if(verbatim){
            if(grepl(BOX_OPEN, xi, fixed = TRUE)){
              xi = string_magic_internal(xi, .delim = .delim, .envir = .envir, .data = .data, 
                                         .vectorize = concat_nested, .check = .check, 
                                         .user_funs = .user_funs, 
                                         .valid_operators = .valid_operators)
            }
          } else if(!verbatim){
            # evaluation
            xi_call = check_set_string_magic_parsing(xi, .check, .delim)
            is_deparse = length(operators) > 0 && grepl("(dp|dep)[^ ]*$", operators[1])
            if(is_deparse){
              op_parsed = sma_char2operator(operators[1], .valid_operators)
              is_deparse = op_parsed$operator %in% c("dp", "deparse")
            }
                           
            xi = check_set_string_magic_eval(xi_call, .data, .envir, .check, is_deparse)
          }
        }

        if(ANY_PLURAL){
          # we save the results
          x_values_all[[i]] = xi
          if(i != i_candidate){
            x_values_all[[i_candidate]] = xi
          }
        }


        #
        # Applying the operators
        #

        if(is_plural){
          xi = sma_pluralize(operators, xi, .delim, .envir, .data, .check, 
                             .user_funs, .valid_operators)

        } else if(is_ifelse){

          xi_val = NULL
          xi_expr = str2lang(xi_raw)
          vars = all.vars(xi_expr)

          if(length(vars) == 0){
            if(operators[1] == "&&"){
              # ERROR
              form = "{&& cond ; true}"
              example = '\nEXAMPLE: x = c(5, 700); string_magic("Value: {&&x > 20 ; > 20}")'

              .stop_hook("The if-else operator `&&`, of the form ", form,
                      ", requires at least one variable to be evaluated in the condition.",
                      "\nPROBLEM: no variable could be found.", example)
            }
            
          } else {
            do_eval = TRUE
            if(operators[1] == "&"){
              do_eval = grepl(BOX_OPEN, operators[3], fixed = TRUE)
              if(!do_eval && length(operators) == 4){
                do_eval = grepl(BOX_OPEN, operators[4], fixed = TRUE)
              }
            }
            
            if(do_eval){
              xi_call = extract_first_variable(xi_expr)
              xi_val = check_set_string_magic_eval(xi_call, .data, .envir, .check)
            }
            
            if(operators[1] == "&&" && length(xi) != length(xi_val)){
              form = "{&& cond ; true}"
              example = 'EXAMPLE: x = c(5, 700); string_magic("Value: {&&x > 20 ; > 20}")'

              stop_hook("The if-else operator `&&`, of the form {form} ",
                      "requires that the first variable used in the condition ",
                      "(here {bq?vars[1]}) is of the same length as the condition.",
                      "\nPROBLEM: the condition is of length {len?xi} while the ",
                      "variable is of length {len?xi_val}.",
                      "\n{example}")
            }
          }
          
          if(ANY_PLURAL){
            # we save the variable over which the ifelse is done
            x_values_all[[i]] = xi_val
            if(i != i_candidate){
              x_values_all[[i_candidate]] = xi_val
            }
          }

          xi = sma_ifelse(operators, xi, xi_val, .envir = .envir, .data = .data,
                          .check = .check, .delim = .delim, 
                          .user_funs = .user_funs, .valid_operators = .valid_operators)
        } else {
          #
          # REGULAR OPERATORS
          #

          group_flag = any(string_x(operators, 1) == "~")
          
          for(j in seq_along(operators)){
            opi = operators[[j]]
            op_parsed = sma_char2operator(opi, .valid_operators)

            if(op_parsed$eval){
              argument_call = check_set_oparg_parse(op_parsed$argument, opi, .check)
              argument = check_set_oparg_eval(argument_call, .data, .envir, opi, .check)
            } else {
              argument = op_parsed$argument
            }

            if(.check){
              xi = check_expr(sma_operators(xi, op_parsed$operator, op_parsed$options, argument,
                                            .check = .check, .envir = .envir, 
                                            group_flag = group_flag,
                                            .delim = .delim, .user_funs = .user_funs, 
                                            .valid_operators = .valid_operators),
                              get_string_magic_context(), " See error below:",
                              verbatim = TRUE, up = 1)
            } else {
              xi = sma_operators(xi, op_parsed$operator, op_parsed$options, argument, 
                                 .check = .check,
                                 .envir = .envir, group_flag = group_flag,
                                 .delim = .delim, .user_funs = .user_funs, 
                                 .valid_operators = .valid_operators)
            }
          }
        }
      }
      
      if(i == 1){
        res = xi
      } else{
        if(i < n && n_x_all[i + 1] == 1){
          if(.vectorize){
            res = c(res, xi, x_parsed[[i + 1]])
          } else {
            res = paste0(res, xi, x_parsed[[i + 1]])
          }

          i_done = TRUE
        } else {
          if(.vectorize){
            res = c(res, xi)
          } else {
            res = paste0(res, xi)
          }
        }
      }
    }
  }
  
  if(!true_character(res)){
    # we always return something in character form
    res = as.character(res)
  }  
  
  if(!is.null(.last)){
    if(is.function(.last)){
      res = .last(res)
    } else {
      res = apply_simple_operations(res, ".last", .last, .check, .envir, .data,
                                    group_flag = 1 * grepl("~", .last, fixed = TRUE), 
                                    .delim, .user_funs = .user_funs, 
                                    .valid_operators = .valid_operators)
    }
  }

  if(!is.null(.collapse) && length(res) > 1){
    res = paste0(res, collapse = .collapse)
  }

  return(res)
}

ROUND_SIGNIF_OPS = c(paste0("r", 0:6), paste0("s", 0:6))

sma_char2operator = function(x, .valid_operators){

  op_parsed = cpp_parse_operator(x)
  
  ok = FALSE
  op = op_parsed$operator

  if(nchar(op) == 0){
    .stop_hook("In `string_magic`, if a quoted value is present, the operators must ",
              "be of the form 'value'op, ",
              "with 'op' an operator. ",
              "\nPROBLEM: In `", escape_newline(x), "` the operator is missing.")
  }
  
  argument = op_parsed$argument
  do_eval = op_parsed$eval
  
  # we partially match operators if needed
  if(!op %in% .valid_operators){
    op = check_set_options(op, .valid_operators, case = TRUE, free = TRUE)
    op_parsed$operator = op
  } else if(op %in% ROUND_SIGNIF_OPS){
    digits = as.numeric(substr(op, 2, 2))
    op = if(substr(op, 1, 1) == "r") "round" else "signif"
    op_parsed$options[[length(op_parsed$options) + 1]] = digits
    op_parsed$operator = op
  }

  if(nchar(argument) == 0 && !substr(x, 1, 1) %in% c("'", "\"", "`")){
    # no argument provided: we need to:
    # - set default values if available
    # - send error if operator requires argument
    
    # default values
    argument = switch(op,
                      s = " ", S = ",[ \t\n]*", split = " ", Split = ",[ \t\n]*",
                      x = "[[:alnum:]]+", X = "[[:alnum:]]+", extract = "[[:alnum:]]+",
                      c = " ", C = ", | and ", collapse = " ", Collapse = ", | and ",
                      head = "6",
                      times = 1, each = 1,
                      first = 1, last = 1,
                      firstchar = 1, lastchar = 1,
                      trim = 1, 
                      "")
    
    op_parsed$argument = argument

    if(op %in% c("R", "r", "%", "k", "K", "paste", "get", "is")){
      ex = c("R" = 'x = "She loves me."; string_magic("{\'s\\b => d\'R ? x}")',
            "r" = 'x = "Amour"; string_magic("{\'ou => e\'r ? x}...")',
            "replace" = 'x = "Amour"; string_magic("{\'ou => e\'r ? x}...")',
            "%" = 'string_magic("pi is: {%.03f ? pi}")',
            "k" = 'string_magic("The first 8 letters of the longuest word are: {8k, q ! the longuest word}.")',
            "shorten" = 'string_magic("The first 8 letters of the longuest word are: {8 shorten, q ! the longuest word}.")',
            "Shorten" = 'string_magic("The first 8 letters of the longuest word are: {8 Shorten, q ! the longuest word}.")',
            "K" = 'x = 5:9; string_magic("The first 2 elements of `x` are: {2K, C ? x}.")',
            "get" = 'x = row.names(mtcars) ; string_magic("My fav. cars are {\'toy\'get.ignore, \'the \'app, enum ? x}.")',
            "is" = 'x = c("Bob", "Pam") ; string_magic("{\'am\'is ? x}")',
            "paste" = 'x = "those, words"; string_magic("Let\'s emphasize {S, \'**\'paste.both, c ? x}.")')

      .stop_hook("The operator `", op, "` has no default value, you must provide values explicitly.", 
                 " EXAMPLE: ", ex)
    }
  }

  if(op %in% c("Ko", "KO")){
    # special case
    text = if(op == "Ko") "||:rest: others" else "||:REST: others"
    argument = paste0(argument, text)
    op_parsed$operator = op = "K"
    op_parsed$argument = argument
  }
    
  if(!ok && !op %in% .valid_operators){
    
    op = check_set_options(op, .valid_operators, case = TRUE, free = TRUE)
    op_parsed$operator = op
    
    if(!op %in% .valid_operators){
      
      context = get_string_magic_context()

      sugg_txt = suggest_item(op, .valid_operators, newline = FALSE, info = "operator")
      
      op = escape_newline(op)
      op = gsub("\t", "\\\\t", op)

      msg = .sma("{context}",
              "\nPROBLEM: {bq?op} is not a valid operator. ", sugg_txt,
              "\n\nINFO: Type string_magic(.help = \"regex\") or string_magic(.help = TRUE) for help.",
              "\nOr look at the vignette: https://lrberge.github.io/stringmagic/articles/guide_string_magic.html")

      .stop_hook(msg)
    }

  }

  op_parsed
}


sma_operators = function(x, op, options, argument, .check = FALSE, .envir = NULL, 
                         .data = list(), group_flag = 0, .delim = c("{", "}"),  
                         .user_funs = NULL, .valid_operators = NULL){
  
  
  #
  # step 1: retro compatibility ####
  #
  
  
  # now: >=1.2.0 operator width becomes similar to fill
  # we need take care of not breaking existing code!!!!
  # special case: width is like swidth (old behavior, < 1.2.0)
  if(op == "width" && grepl("sw", argument)){
    op = "swidth"
  }
  
  #
  # step 2: user functions ####
  #
  

  # group_flag:  0 nothing
  #                    1 keep track of conditional things
  #                    2 apply conditional

  group_index = attr(x, "group_index")
  
  # beware the sneaky assignment!!!!!
  if(!is.null(.user_funs) && !is.null(fun <- .user_funs[[op]])){
    #                                     ^ sneaky (but saves half a us)
    # user-defined operations ####
    
    # EXPLANATION: why do we check user operations first?
    # - I want this package to last and be stable
    # - I also want forward stability for packages that use stringmagic as a dependency
    # - example of use case:
    #   * pkg matchmaker uses stringmagic and defines the operation 'h1'
    #   * all works fine with SM 1.0.0
    #   * now SM 1.1.0 introduces the operator h1!!!!
    #   * what happens?
    #     + matchmaker should work for both SM 1.0.0 and SM 1.1.0
    #     + it should work even if h1 has a new definition in SM 1.1.0 => the h1 used
    #       should be the one of the user, defined in the package.
    # - hence I need to:
    #   * allow users to redefine existing operations
    #   * check user defined version first
    #
    
    #
    # step 3: built in functions ####
    #
    
    
    if(isTRUE(attr(fun, "simple_ops"))){
      res = apply_simple_operations(x, op, fun, .check, .envir,  .data,
                                    group_flag = group_flag, .delim, 
                                    .user_funs = .user_funs, 
                                    .valid_operators = .valid_operators)
                                
      # res already has a group_index attribute
      
    } else {
      # This is a regular function
      valid_options = attr(fun, "valid_options")
      if(!is.null(valid_options)){
        options = check_set_options(options, valid_options)
      }

      res = fun(x = x, argument = argument, options = options, group = group_index, 
                group_flag = group_flag)

      # NOTA: for the users I use 'group', internally I use 'group_index'
      group_index = attr(res, "group")
      if(group_flag != 0 && is.null(group_index) && !isTRUE(attr(fun, "group_flag")) 
          && length(x) == length(res)){
        # This is the case in which there is grouping but we don"t impose
        # to the user to define the grouping behavior
        # if the lengths are different: it's messed up and so be it
        attr(res, "group_index") = attr(x, "group_index")
      }
    }
    
    return(res)
  } 
  
  # We carry on
  
  if(op %in% c("c", "C", "collapse", "Collapse")){
    # C, collapse ####
    # collapse

    # argument of the form: 'main_sep|last_sep'
    sep_last = ""
    is_last = FALSE
    if(length(x) > 1 && grepl("|", argument, fixed = TRUE) && nchar(argument) > 1){
      info = extract_pipe(argument, op)
      is_last = info$is_pipe
      argument = info$value
      sep_last = info$extra
    }
    sep = argument

    if(!is.null(group_index) && group_flag == 2){
      x = as.character(x)
      res = cpp_paste_conditional(x, group_index, sep, sep_last)
      group_index = seq_along(res)
      
    } else {
      n_x = length(x)
      if(is_last){
        # This is the "last" operator
        if(n_x == 2){
          res = paste(x, collapse = sep_last)
        } else {
          res = paste(x[-n_x], collapse = sep)
          res = paste0(res, sep_last, x[n_x])
        }
      } else {
        res = paste(x, collapse = sep)
      }

      if(group_flag == 1){
        group_index = NULL
      }

    }
    
  } else if(op %in% c("s", "S", "split", "Split", "r", "R", "replace", "clean", "get", 
                      "is", "which", "x", "X", "extract")){
    # split, replace, clean, extract, get, is, which ####
    
    valid_options = c("word", "ignore", "fixed")
    if(op == "extract"){
      valid_options = c(valid_options, "first")
    } else if(op %in% c("r", "R", "clean", "replace")){
      valid_options = c(valid_options, c("total", "single"))
    } else if(op %in% c("get", "is", "which")){
      valid_options = c(valid_options, "equal", "in")
    }

    options = check_set_options(options, valid_options)
    is_word = "word" %in% options
    is_ignore = "ignore" %in% options
    is_fixed = "fixed" %in% options

    if(op == "extract"){
      if("first" %in% options){
        op = "x"
      } else {
        op = "X"
      }
    }

    if(op %in% c("s", "S", "split", "X", "x", "extract")){
      # otherwise => dealt with in string_is or string_clean
      # we don't repeat the processing (otherwise => bugs)
      
      pat_parsed = format_simple_regex_flags(argument, fixed = is_fixed, word = is_word, 
                                             ignore = is_ignore, magic = TRUE, 
                                             envir = .envir)
      argument = pat_parsed$pattern
      is_fixed = pat_parsed$fixed
    } 

    #
    # now the operations
    #
    
    if(!true_character(x) && op != "which"){
      x = as.character(x)
    }
    
    if(op %in% c("s", "S", "split", "Split")){
      # strsplit applied to "" returns character(0)
      # if occurring: need to fix
      
      x_split = strsplit(x, argument, fixed = is_fixed, perl = !is_fixed)
      
      fix_empty = any(x == "")

      if(group_flag != 0){
        # we keep track of the group index
        x_len_all = lengths(x_split)
        
        if(fix_empty){
          x_len_all[x_len_all == 0] = 1
        }
        
        group_index = rep(seq_along(x_len_all), x_len_all)
      }
      
      # Note that unlist removes empty strings
      if(fix_empty){
        x_split[x == ""] = ""
      }

      res = unlist(x_split)
      
    } else if(op %in% c("r", "R", "replace", "clean")){
      is_total = "total" %in% options
      is_single = "single" %in% options

      pipe = "=>"
      if(grepl(" => ", argument, fixed = TRUE)){
        pipe = " => "
      }
      
      split = if(op == "clean") ",[ \t\n]+" else ""
      
      res = string_clean(x, argument, pipe = pipe, split = split, 
                      ignore.case = is_ignore, fixed = is_fixed, word = is_word, 
                      total = is_total, single = is_single, envir = .envir)

    } else if(op == 'x'){
      x_pat = regexpr(argument, x, fixed = is_fixed, perl = !is_fixed)

      res = substr(x, x_pat, x_pat - 1 + attr(x_pat, "match.length"))
    } else if(op == "X"){
      # extract all patterns

      x_list = regmatches(x, gregexpr(argument, x, fixed = is_fixed, perl = !is_fixed))

      if(group_flag != 0){
        # we keep track of the group index
        x_len_all = lengths(x_list)
        # I could avoid the last line... later
        group_index = rep(seq_along(x_len_all), x_len_all)
        group_index = cpp_recreate_index(group_index)
      }

      res = unlist(x_list)
    } else if(length(argument) == 1 && argument == "" && op == "which"){
      
      if(!is.logical(x)){
        stop_hook("The operation `which` must apply only to logical values.",
                 "\nPROBLEM: the current value to which it is applied is not logical ",
                 "(instead is is of class {enum.bq?class(x)}).")
      }

      res = which(x)

      if(group_flag == 1){
        group_index = group_index[res]
        group_index = cpp_recreate_index(group_index)
      } else if(group_flag == 2){
        stop_hook("The operation `which` cannot be applied in ",
                  "conditional statements (inside `~()`).",
                  "\nFIX: simply put this operation after the conditional statement.")
      }
      
    } else {
      # default is string_is
      if("equal" %in% options || "in" %in% options){
        if("equal" %in% options){
          res = x == argument
        } else {
          arg_split = strsplit(argument, ",[ \t\n]+")
          res = x %in% unlist(arg_split)
        }
        
      } else {
        res = string_is(x, pattern = argument, fixed = is_fixed, 
                     ignore.case = is_ignore, word = is_word, envir = .envir)
      }

      if(op %in% c("get", "which")){

        if(group_flag == 1){
          group_index = group_index[res]
          group_index = cpp_recreate_index(group_index)
        } else if(group_flag == 2){
          stop_hook("The operation `{q?argument}{op}` cannot be applied in ",
                    "conditional statements (inside `~()`).",
                    "\nFIX: simply put this operation after or before the conditional statement.")
        }

        if(op == "get"){
          res = x[res]
        } else if(op == "which"){
          res = which(res)
        }

      }
    } 
    

  } else if(op %in% c("each", "times")){
    # times/each ####

    if(!is_numeric_in_char(argument)){
      msg = paste0("In `string_magic`: the operator `", op, "` must have numeric arguments, `",
                   argument, "` is not numeric.")
      .stop_hook(msg)
    }
    
    if(length(argument) != 1){
      msg = paste0("In `string_magic`: the operator `", op, "` must have an argument of length 1.",
                   "\nPROBLEM: the argument is of length ", length(argument), ".")
      .stop_hook(msg)
    }
    
    if(op == "each"){
      res = rep(x, each = as.numeric(argument))
    } else {
      res = rep(x, as.numeric(argument))
    }

    if("c" %in% options){
      res = paste(res, collapse = "")
    }

    group_index = NULL

  } else if(op == "upper"){
    # upper, lower, title, Q, ####
    
    options = check_set_options(options, c("first", "sentence"))
        
    if("first" %in% options){
      # First letter only, if relevant
      res = as.character(x)
      substr(res, 1, 1) = toupper(substr(x, 1, 1))  
    } else if("sentence" %in% options){
      # we add upper case like in sentences
      # does not work for spanish for instance (with inverted !? coming first)
      res = gsub("(^|[.?!])(\\s*)(\\p{Ll})", "\\1\\2\\U\\3", x, perl = TRUE)
    } else {
      res = toupper(x)  
    }
    
  } else if(op == "lower"){
    res = tolower(x)

  } else if(op == "title"){
    # title case

    options = check_set_options(options, c("force", "ignore"))

    if("force" %in% options){
      # puts in lowercase first
      x = tolower(x)
    }

    is_ignore = "ignore" %in% options
    if(is_ignore){
      # we add __IGNORE in front of the items so that upper casing won't proc
      IGNORE = c("a", "the", "in", "on", "at", "of", 
                 "for", "to")
      pattern = paste0("(?<=[^:!?.]) (", paste0(IGNORE, collapse = "|"), ")(?=(\\s|$))")
      x = gsub(pattern, " __IGNORE\\1", x, perl = TRUE)
    }
    
    # should I add unicode awareness with (*UCP)? Option?
    # I would say yes by default bc you only apply this function
    # to small strings

    res = gsub("(*UCP)(^|[^\\p{L}])(\\p{Ll})", "\\1\\U\\2", x, perl = TRUE)
    
    if(is_ignore){
      res = gsub("__IGNORE", "", res, fixed = TRUE)
    }

  } else if(op == "q"){
    res = paste0("'", x, "'")

  } else if(op == "Q"){
    res = paste0("\"", x, "\"")

  } else if(op == "bq"){
    res = paste0("`", x, "`")

  } else if(op %in% c("format", "Format")){
    # Format, % ####

    options = check_set_options(options, c("0", "zero", "left", "right", "center"))

    is_zero = any(options %in% c("0", "zero"))
    is_right = "right" %in% options || op == "Format" || is_zero
    is_center = any(options == "center")
    is_left = !is_right && !is_center
    
    pos = "left"
    if(is_right) pos = "right"
    if(is_center) pos = "center"
    
    if(nchar(argument) > 0){
      res = format(x, argument, justify = "right", big.mark = ",")
    } else {
      res = format(x, justify = "right", big.mark = ",")
    }

    if(is.numeric(x)){
      if(is_zero){
        res = gsub(" ", "0", res, fixed = TRUE)
      } else if(!is_right){
        # we have to do that.... not sure the use cases need to be optimized
        pos = if(is_center) "centre" else "left"
        res = format(cpp_normalize_ws(res), justify = pos)
      }
    } else if(is_zero && is_numeric_in_char(x)){
      res = gsub(" ", "0", res, fixed = TRUE)
    }

  } else if(op == "%"){
    res = sprintf(paste0("%", argument), x)

  } else if(op == "ws"){
    # White spaces, tws, trim ####

    # w: only whitespaces
    # p: punct
    # d: digits
    # i: isolated
    
    options = check_set_options(options, c("punct", "digit", "isolated"))

    clean_punct = "punct" %in% options
    clean_digit = "digit" %in% options
    clean_isolated = "isolated" %in% options

    res = cpp_normalize_string(x, clean_punct, clean_digit, clean_isolated)
    
  } else if(op == "tws"){
    # trimming WS

    # note that the changes are 'in place' => it does not matter here
    # but I should not use cpp_trimws in other functions
    res = cpp_trimws_in_place(x)

  } else if(op == "trim"){
    # we trim

    nb = argument
    if(!is_numeric_in_char(nb)){
      msg = paste0("In `string_magic`: the operator `", op, "` must first contain a numeric argument. PROBLEM: `",
                   argument, "` is not numeric.")
      .stop_hook(msg)
    }
    nb = as.numeric(nb)

    is_both = opt_equal(options, "both")
    is_right = any(c("r", "right") %in% options) || nb < 0
    
    if(!true_character(x)){
      x = as.character(x)
    }

    nx = nchar(x)
    if(is_both){
      # on both sides
      nb = abs(nb)

      res = substr(x, 1, nx - nb)
      res = substr(res, 1 + nb, nchar(res))

    } else if(is_right){
      # we revert: start from the end
      nb = abs(nb)
      res = substr(x, 1, nx - nb)
    } else {
      res = substr(x, 1 + nb, nx)
    }

  } else if(op %in% c("k", "shorten", "Shorten", "K", "head")){
    # keep: either the nber of characters (k) or the number of elements (K)

    #
    # Keep, shorten, head ####
    #
    
    if(op %in% c("k", "shorten", "Shorten")){
      options = check_set_options(options, c("include", "dots"))
      is_dots = "dots" %in% options || substr(op, 1, 1) == "S"
      op = "k"
    } else {
      options = check_set_options(options, "include")
    }
    
    info = extract_pipe(argument, op, double = TRUE, numeric = TRUE, mbt = TRUE)
    nb = info$value
    add = info$extra
    is_included = info$is_double

    if(op == "k"){
      if(!true_character(x)){
        x = as.character(x)
      }
      
      n_longer = n_rm = 0
      if(is_dots){
        n_longer = 1
        n_rm = 1
        add = ".."
      } else if(is_included || "include" %in% options){
        n_rm = nchar(add)
      }
      
      qui = nchar(x) > (nb + n_longer)
      res = x
      
      res[qui] = substr(res[qui], 1, nb - n_rm)

      if(nchar(add) > 0){
        res[qui] = paste0(res[qui], add)
      }

    } else {
      # K or head

      if(nb == 0){
        res = character(0)
      } else {
        res = if(nb < length(x)) x[1:nb] else x
        
        if(group_flag == 2){
          stop_hook("The operation `{q?argument}{op}` cannot be applied in ",
                    "conditional statements (inside `~()`).",
                    "\nFIX: simply put this operation after or before the conditional statement.")
        }

        if(any(grepl(":(?:n|N|rest|REST):", add))){
          n = length(x)
          N = ""
          if(grepl(":N:", add, fixed = TRUE)){
            N = n_letter(n)
            add = gsub(":N:", N, add, fixed = TRUE)
          }

          add = gsub(":n:", n, add, fixed = TRUE)
          n_rest = n - nb + is_included
          if(n_rest > 0){
            add = gsub(":rest:", n_rest, add, fixed = TRUE)
            add = gsub(":REST:", n_letter(n_rest), add, fixed = TRUE)
          }
        }

        if(length(x) > nb){
          if(is_included){
            res = res[-nb]
          }

          if(nchar(add) > 0){
            res = c(res, add)
          }
        }
      }
    }
  } else if(op %in% c("dp", "deparse")){
    
    options = check_set_options(options, c("long", "short", "vec"))
    
    if(nchar(argument) > 0){
      argument = as.numeric(argument)
    } else {
      argument = 50
    }
    
    if("vec" %in% options){
      x_dp = deparse(x, nlines = argument)
      
    } else if("short" %in% options){
      x_dp = paste0(deparse(x, nlines = 5), collapse = "\\n")
      
      if(nchar(x_dp) > argument){
        x_dp = substr(x_dp, 1, argument - 2)
        x_dp = gsub(" *$", "...", x_dp)
      }
    } else {
      # long is the new default
      x_dp = paste0(deparse(x, nlines = argument), collapse = "\n")
      
    }
    
    res = x_dp
    
  } else if(op == "enum"){
    # enum ####
    # NOTE:
    # for conditional operations, the code is slow.
    # but I don't see a use case for large vectors....
    # we use enum for smallish things so it should be OK.

    if(!is.null(group_index) && group_flag == 2){

      res = unname(tapply(x, group_index, enum_main, options = options))
      group_index = seq_along(group_index)

    } else {
      res = enum_main(x, options = options)

      if(group_flag == 1){
        group_index = NULL
      }

    }

  } else if(op %in% c("first", "last", "firstchar", "lastchar")){
    # first, last, firstchar, lastchar ####

    nb = argument
    
    nb_extra = NULL
    if(op == "first" && grepl("|", argument, fixed = TRUE)){
      arg_split = strsplit(argument, "|", fixed = TRUE)[[1]]
      nb = arg_split[1]
      nb_extra = arg_split[2]
      
      if(!is.null(nb_extra) && !is_numeric_in_char(nb_extra)){
        msg = paste0("In `string_magic`: in the operator `", op, "` the argument can be of the form ",
                     "'n1|n2' with `n1` and `n2` numbers. ",
                     "\nPROBLEM: the second element `", nb_extra, "` is not numeric.")
        .stop_hook(msg)
      }
      
      nb_extra = as.numeric(nb_extra)
    }
    
    if(!is_numeric_in_char(nb)){
      msg = paste0("In `string_magic`: the operator `", op, "` must have a numeric argument. ",
                   "\nPROBLEM: `", argument, "` is not numeric.")
      .stop_hook(msg)
    }    

    # the numeric argument can also be passed in options
    qui_num = which(grepl("^\\d+$", options))
    if(length(qui_num)){
      nb = options[qui_num[1]]
    }

    nb = as.numeric(nb)

    if(string_x(op, -4) == "char"){
      # we select the first/last characters
      
      if(!true_character(x)){
        x = as.character(x)
      }
      
      if(nb < 0){
        nb = abs(nb)
        nx = nchar(x)
        if(op == "firstchar"){
          res = substr(x, nb + 1, nx)
        } else {
          res = substr(x, 1, nx - nb)
        }
      } else {
        if(op == "firstchar"){
          res = substr(x, 1, nb)
        } else {
          nx = nchar(x)
          res = substr(x, nx - nb + 1, nx)
        }
      }
      
    } else {
      # we select the first/last elements

      if(!is.null(group_index) && group_flag == 2){
        is_last = op == "last"
        qui = cpp_find_first_index(group_index, nb, is_last)
        
        res = x[qui]
        group_index = group_index[qui]
        
        if(!is.null(nb_extra)){
          qui = cpp_find_first_index(group_index, nb, TRUE)
          res = c(res, x[qui])
          group_index = c(group_index, group_index[qui])
          
          my_order = order(group_index)
          res = res[my_order]
          group_index = group_index[my_order]
        }
        
      } else {
        if(op == "first"){
          
          if(nb < 0){
            
            if(!is.null(nb_extra)){
              msg = paste0("In `string_magic`: in the operator `", op, "` the argument can be of the form ",
                          "'n1|n2' with `n1` and `n2` positive numbers. ",
                          "\nPROBLEM: the first element `", nb, "` is negative.")
              .stop_hook(msg)
            }
            
            nb = abs(nb)
            if(nb > length(x)){
              res = character(0)
            } else {
              res = tail(x, length(x) - nb)
            }            
          } else {
            res = head(x, nb)
            
            if(!is.null(nb_extra)){
              if(nb_extra < 0){
                msg = paste0("In `string_magic`: in the operator `", op, "` the argument can be of the form ",
                            "'n1|n2' with `n1` and `n2` positive numbers. ",
                            "\nPROBLEM: the second element `", nb_extra, "` is negative.")
                .stop_hook(msg)
              }
              
              res = c(res, tail(x, nb_extra))
            }
            
          }    
        } else {
          if(nb < 0){
            nb = abs(nb)
            if(nb > length(x)){
              res = character(0)
            } else {
              res = head(x, length(x) - nb)
            }            
          } else {
            res = tail(x, nb)
          }
        }

        if(group_flag == 1){
          # this operation destroys the groups
          group_index = NULL
        }
      }
    }

  } else if(op == "rev"){
    # rev, sort, dsort ####
    
    if(!is.null(group_index) && group_flag == 2){
      qui = cpp_group_rev_index(group_index)
      res = x[qui]
    } else {
      res = rev(x)
    }

  } else if(op %in% c("sort", "dsort")){
    
    is_decreasing = op == "dsort"
    x2sort = x
    is_modified = FALSE
    if(nchar(argument) > 0){
      is_modified = TRUE
      x2sort = string_clean(x, argument, envir = .envir)
    }
    
    options = check_set_options(options, "num")
    if("num" %in% options){
      is_modified = TRUE
      x2sort = suppressWarnings(as.numeric(x2sort))
    }
    
    if(group_flag == 1){
      # it makes no sense to keep track of the indexes when sorting the full string
      # ex: a1 b2 => split => a 1 b 2 => sort => a b 1 2
      # after that, the conditional operations (on a1 and b2) don't make much sense
      group_index = NULL
      group_flag = 0
    }

    if(!is.null(group_index) && group_flag == 2){
      
      if(is_decreasing){
        new_order = order(-group_index, x2sort, decreasing = TRUE)
      } else {
        new_order = order(group_index, x2sort)
      }

      res = x[new_order]
      group_index = cpp_recreate_index(group_index[new_order])
    } else {
      if(is_modified){
        new_order = order(x2sort, decreasing = is_decreasing)
        res = x[new_order]
      } else {
        res = sort(x, decreasing = is_decreasing)
      }
      
    }

  } else if(op %in% c("paste", "append", "insert")){
    # paste, append, insert ####
    # paste: at the beginning/end of all strings
    # insert: element at the beginning/end of the vector 
    
    if(op == "append"){
      op = "paste"
    }
    
    valid_options = c("left", "both", "right")
    if(op == "paste"){
      valid_options = c(valid_options, "front", "back", "delete")
    }
    
    options = check_set_options(options, valid_options)
    
    left = right = ""
    if(!any(c("both", "right") %in% options) && grepl("|", argument, fixed = TRUE)){
      args = extract_pipe(argument, op)
      left = args$value
      right = args$extra
    } else {
      if("both" %in% options){
        left = right = argument
      } else if("right" %in% options){
        right = argument
      } else {
        left = argument
      }
    }
    
    if("delete" %in% options){
      x = character(length(x))
    }

    if(nchar(argument) == 0){
      res = x

    } else if(op == "paste"){

      if(length(x) == 0){
        res = x

      } else if(any(options %in% c("front", "back"))) {
        res = x
        if("front" %in% options){
          res[[1]] = paste0(left, x[1], right)
        }
        
        if("back" %in% options){
          n = length(x)
          res[[n]] = paste0(left, x[n], right)
        }        
        
      } else {

        # We replace the special values
        # :1:, :i:, :a:
        # we allow only one special value
        n_x = length(x)

        pat = c(":1:", ":01:", ":i:", ":a:", ":I:", ":A:")

        for(i in 1:2){
          tmp = if(i == 1) left else right
          any_done = FALSE

          for(p in pat){
            if(grepl(p, tmp, fixed = TRUE)){
              any_done = TRUE
              tmp_split = strsplit(tmp, p, perl = TRUE)[[1]]
              if(length(tmp_split) > 1){
                txt = switch(p,
                             ":1:" = 1:n_x,
                             ":01:" = string_fill(1:n_x, right = TRUE, symbol = "0"),
                             ":i:" = tolower(as.roman(1:n_x)),
                             ":I:" = as.character(as.roman(1:n_x)),
                             ":a:" = enum_letter(n_x),
                             ":A:" = toupper(enum_letter(n_x)))

                if(length(tmp_split) == 1){
                  tmp = paste0(tmp_split[[1]], txt)
                } else  {
                  tmp_new = paste0(tmp_split[[1]], txt, tmp_split[[2]])
                  k = 2
                  while(k + 1 <= length(tmp_split)){
                    k = k + 1
                    tmp_new = paste0(tmp_new, txt, tmp_split[[k]])
                  }
                  tmp = tmp_new
                }
              }

              break
            }
          }

          if(any_done){
            if(i == 1){
              left = tmp
            } else {
              right = tmp
            }
          }
        }

        if(any(nchar(left) > 0)){
          if(any(nchar(right) > 0)){
            res = paste0(left, x, right)
          } else {
            res = paste0(left, x)
          }
        } else {
          res = paste0(x, right)
        }
      }
    } else if(op == "insert"){
      # inserts an ELEMENT at the beginning/end
      
      if(group_flag == 2){
        stop_hook("The operation `{q?argument}{op}` cannot be applied in ",
                  "conditional statements (inside `~()`).",
                  "\nFIX: simply put this operation after or before the conditional statement.")
      }

      if(nchar(left) > 0){
        if(nchar(right) > 0){
          res = c(left, x, right)
        } else {
          res = c(left, x)
        }
      } else {
        res = c(x, right)
      }

    }

    # END: paste/insert

  } else if(op %in% c("fill", "align", "width")){
    # fill, align ####
    
    #
    # regular case
    #
    
    # all are equivalent
    valid_options = c("left", "right", "center")
    options = check_set_options(options, valid_options, op = op)
    
    info = extract_pipe(argument, op, double = FALSE, numeric = TRUE, mbt = FALSE)
    argument = info$value
    symbol = info$extra
    
    center = "center" %in% options
    right = "right" %in% options
    
    if(symbol == ""){
      symbol = " "
    }
    
    if(nchar(symbol) != 1){
      stop_hook("In the operator `fill`, the symbol used to fill must be of length 1.",
                "\nPROBLEM: the symbol, equal to {bq?symbol}, is of length {n?nchar(symbol)}.",
                "\nEXAMPLE: to fill with 0s: `fill.0`; of length 10 with underscores and right-aligned text: `'10|_'fill.right`.")
    }
    
    if(argument == ""){
      argument = NULL
    }
    
    if(!is.null(argument)){
      if(is_numeric_in_char(argument)){
        argument = as.numeric(argument)
      } else {
        stop_hook("In the operator `fill`, the argument (giving the length of the fill) must be numeric.",
                "\nPROBLEM: the argument, equal to {bq?argument}, is not numeric ",
                "(instead it is of class {enum ? class(argument)}.",
                "\nEXAMPLE: a fill of size 10: `10 fill`; of size 10 (using quotes) on the right: `'10'fill.right`.")
      }
    }
    
    res = string_fill(x, argument, symbol = symbol, right = right, center = center)
    
  } else if(op == "join"){
    # join, escape ####
    res = gsub(" *\\\\ *\n *", " ", x, perl = TRUE)    
  } else if(op == "escape"){
    # later, I can inclue latex, markdown, etc
    options = check_set_options(options, c("nl", "tab"))
    if(length(options) == 0){
      options = c("nl", "tab")
    }
    
    res = x
    if("nl" %in% options){
      res = gsub("\n", "\\\\n", res, perl = TRUE)
    }
    
    if("tab" %in% options){
      res = gsub("\t", "\\\\t", res, perl = TRUE)
    }
    
  } else if(op == "unik"){
    # unik, table ####

    if(!is.null(group_index) && group_flag == 2){

      # not really fast 
      # => to be improved
      df = data.frame(g = group_index, x = x, stringsAsFactors = FALSE)
      df_unik = unique(df)

      group_index = df_unik$g
      res = df_unik$x

    } else {
      
      res = unique(x)

      if(group_flag == 1){
        group_index = NULL
      }

    }

  } else if(op == "table") {
    
    if(!is.null(group_index) && group_flag == 2){
      stop_hook("The `table` operation is not implemented for groups.")
    }
    
    # The default is decreasing frequency sort
    options = check_set_options(options, c("sort", "dsort", "fsort", "dfsort", "nosort"))
    
    if(argument == ""){
      argument = "{x} ({n ? n})"
      
    } else {
      # we check the validity of the argument
      vars = get_interpolated_vars(argument)
      if(length(vars) == 0){
        stop_hook("The operation `table` requires an argument with at least one of the variables `x` (value), `n` (frequency), or `s` (share) in curly braces.",
                  "\nPROBLEM: no interpolated variable found in the argument {q?argument}",
                  "\nEXAMPLE: sma(\"Freq. of months: \\{'\\{x} (\\{n})'table.nosort, enum ? month.name[airquality$Month]}.\")")
      }
      
      pblm = setdiff(vars, c("x", "n", "s"))
      if(length(pblm) > 0){
        stop_hook("The operation `table` requires an argument with the variables `x` (value), `n` (frequency), and/or `s` (share) in curly braces (and only those variables).",
                  "\nPROBLEM: the following invalid variable{$s, is} used {enum.bq ? pblm}.",
                  "\nEXAMPLE: sma(\"Freq. of months: \\{'\\{x} (\\{n})'table.nosort, enum ? month.name[airquality$Month]}.\")")
      }
    }
    
    # We apply a SMA interpolation with the following variables:
    # - x: the character value
    # - n: the frequency
    # - s: the share
    
    info = to_index(x, items = TRUE, items.simplify = TRUE)
    
    n_items = length(info$items)
    x_tab = cpp_table(info$index, n_items)
    
    table_data = list(x = info$items, n = x_tab, s = x_tab / length(x))
    
    # NOTA: there can be no ties in x (by definition)
    #       but ties can occur in frequencies
    
    my_order = NULL
    if("sort" %in% options){
      if("fsort" %in% options){
        my_order = order(table_data$n, table_data$x)
        
      } else if("dfsort" %in% options){
        my_order = order(-table_data$n, table_data$x)
        
      } else {
        my_order = order(table_data$x)
      }
      
    } else if("dsort" %in% options){
      if("fsort" %in% options){
        my_order = order(-table_data$n, table_data$x, decreasing = TRUE)
        
      } else if("dfsort" %in% options){
        my_order = order(table_data$n, table_data$x, decreasing = TRUE)
        
      } else {
        my_order = order(table_data$x, decreasing = TRUE)
      }
      
    } else if("fsort" %in% options){
      my_order = order(table_data$n)
    } else if("nosort" %in% options){
      # nothing
    } else {
      # the default
      my_order = order(table_data$n, decreasing = TRUE)
    }
    
    if(!is.null(my_order)){
      table_data$x = table_data$x[my_order]
      table_data$n = table_data$n[my_order]
      table_data$s = table_data$s[my_order]
    }
    
    res = string_magic_internal(argument, .delim = .delim, .envir = .envir,
                                .data = table_data, .check = .check, 
                                .user_funs = .user_funs, .valid_operators = .valid_operators)
    
    
  } else if(op %in% c("round", "signif")){
    #
    # round, signif ####
    #
    
    is_round = op == "round"
    
    digits_opt = grep("^\\d$", options, value = TRUE)
    if(length(digits_opt) > 0){
      digits = as.numeric(digits_opt)
      options = setdiff(options, digits_opt)
    } else {
      digits = if(is_round) 0 else 1
    }
    
    if(is_round){
      n_round = digits
      n_signif = 0
      signif_info = grep("^s\\d", options, value = TRUE)
      if(length(signif_info) > 0){
        options = setdiff(options, signif_info)
        n_signif = as.numeric(substr(signif_info[1], 2, 2))
      }
    } else {
      n_signif = digits
      n_round = 1
      round_info = grep("^r\\d", options, value = TRUE)
      if(length(round_info) > 0){
        options = setdiff(options, round_info)
        n_round = as.numeric(substr(round_info[1], 2, 2))
      }
    }
    
    options = check_set_options(options, c("int", "nocomma"))
    
    int_as_double = !"int" %in% options
    
    info_pipe = extract_pipe(argument)
    if(info_pipe$is_pipe){
      prefix = info_pipe$value
      suffix = info_pipe$extra
    } else {
      prefix = NULL
      suffix = info_pipe$value
    }
    
    if("nocomma" %in% options){
      big_mark = ""
    } else {
      big_mark = ","
    }
    
    if(!is.numeric(x)){
      if(!is.atomic(x)){
        stop_hook("The operation {bq?op} can only be applied on atomic vectors. ",
                  "\nPROBLEM: the value is of class {enum ? class(x)}.")
      }
      
      x_num = suppressWarnings(as.numeric(x))
      
      if(anyNA(x_num)){
        id_ok = which(!is.na(x_num))
        res = as.character(x)
        res[id_ok] = cpp_format_numeric(x_num[id_ok], R_digits = n_round, R_signif = n_signif, 
                                        R_int_as_double = int_as_double, 
                                        minus_sign = "-", decimal = ".", big_mark = big_mark, 
                                        small_mark = "", prefix = prefix, suffix = suffix)
      } else {
        res = cpp_format_numeric(x_num, R_digits = n_round, R_signif = n_signif, 
                                 R_int_as_double = int_as_double, 
                                 minus_sign = "-", decimal = ".", big_mark = big_mark, 
                                 small_mark = "", prefix = prefix, suffix = suffix)
      }
      
    } else {
      res = cpp_format_numeric(x, R_digits = n_round, R_signif = n_signif, 
                               R_int_as_double = int_as_double, 
                               minus_sign = "-", decimal = ".", big_mark = big_mark, 
                               small_mark = "", prefix = prefix, suffix = suffix)
    }
    
    
    
    
  } else if(op %in% c("nth", "Nth", "ntimes", "Ntimes", "n", "N", "len", "Len")){
    #
    # nth, ntimes, n, len ####
    #
    
    # parsing the default values of the options and if we display in letters
    is_letter = substr(op, 1, 1) %in% c("N", "L")
    if(is_letter){
      op = tolower(op)
    }
    
    if(op == "nth"){
      opt_default = c("letter", "upper", "compact")
    } else if(op == "n"){
      opt_default = c("letter", "upper", "0", "zero", "roman", "Roman")
    } else if(op == "len"){
      opt_default = c("letter", "upper", "num")
    } else {
      opt_default = c("letter", "upper")
    }

    options = check_set_options(options, opt_default)
    
    if(!is_letter){
      is_letter = any(c("letter", "upper") %in% options)
    }

    # operations
    if(op == "nth"){
      res = n_th(x, letters = is_letter, compact = "compact" %in% options)

    } else if(op == "ntimes"){
      res = n_times(x, letters = is_letter)

    } else if(op == "n"){
      is_zero = any(options %in% c("0", "zero"))
      # we force the conversion to numeric
      mark = if(is_zero) "" else ","
      if(is.numeric(x)){
        if(any(options %in% c("roman", "Roman"))){
          res = as.roman(x)
          if("roman" %in% options){
            res = tolower(res)
          }
        } else if(is_letter){
          res = n_letter(x)
        } else {
          res = format(x, big.mark = mark)
          
          if(is_zero){
            res = gsub(" ", "0", res, fixed = TRUE)
          } else {            
            res = cpp_trimws_in_place(res)
          }
        }
      } else {
        x_num = suppressWarnings(as.numeric(x))
        is_x_num = which(!is.na(x_num))
        num_val = x_num[is_x_num]

        if(any(options %in% c("roman", "Roman"))){
          res_num = as.roman(num_val)
          if("roman" %in% options){
            res_num = tolower(res_num)
          }
        } else if(is_letter){
          res_num = n_letter(num_val)
        } else {
          res_num = format(num_val, big.mark = mark)
          if(is_zero){
            res_num = gsub(" ", "0", res_num, fixed = TRUE)
          } else {            
            res_num = cpp_trimws_in_place(res_num)
          }
        }

        res = x
        res[is_x_num] = res_num
      }
      
    } else if(op == "len"){

      # if conditional: lengths of all the groups
      if(!is.null(group_index) && group_flag == 2){
        res = tabulate(group_index)
        group_index = seq_along(res)
      } else {
        if(group_flag == 1){
          group_index = 1
        }

        res = length(x)
      }

      if(is_letter){
        res = n_letter(res)
      } else if(opt_equal(options, "num")){
        # we do nothing => we keep it as numeric
      } else {
        res = format(res, big.mark = ",")
      }
    }

    if("upper" %in% options){
      res = gsub("^(.)", "\\U\\1", res, perl = TRUE)
    }
  } else if(op == "swidth"){
    # width, difftime ####
    
    sw = getOption("width")
    data = list(.sw = sw)

    info = extract_pipe(argument, op, double = FALSE, numeric = TRUE, data = data, mbt = FALSE)
    nb = info$value
    comment = info$extra
    
    if(nb == ""){
      nb = NULL
    }

    if(string_x(comment, -1) == "_"){
      # the underscore means that we don't add a space
      comment = string_trim(comment, -1)
    } else if(comment != ""){
      comment = paste0(comment, " ")
    }

    # code is slow but I don't see why would one want it to be super fast
    # (only for user-content, not performance/data-analysis content)

    res = character(length(x))
    for(i in seq_along(x)){
      res[i] = fit_screen(x[i], width = nb, leader = comment)
    }
    
  } else if(op == "difftime"){
    res = format_difftime(x, options)

  } else if(op == "erase"){
    # erase, rm, nuke, num ####
    #

    options = check_set_options(options, c("fixed", "ignore", "word"))

    if(argument == ""){
      res = character(length(x))
    } else {
      is_fixed = "fixed" %in% options
      is_word = "word" %in% options
      is_ignore = "ignore" %in% options

      qui = string_is(x, pattern = argument, fixed = is_fixed, ignore.case = is_ignore, 
                   word = is_word, envir = .envir)

      res[qui] = ""
    }

  } else if(op %in% c("rm", "nuke")){
    options = check_set_options(options, c("empty", "blank", "noalpha", "noalnum", "all", 
                                           "fixed", "ignore", "word"))
                                           
    if(op == "nuke"){
      res = character(0)
      return(res)
    }
    
    res = NULL
    if(argument != ""){
      is_fixed = "fixed" %in% options
      is_word = "word" %in% options
      is_ignore = "ignore" %in% options

      qui = !string_is(x, pattern = argument, fixed = is_fixed, ignore.case = is_ignore, 
                    word = is_word, envir = .envir)
      
      options = setdiff(options, c("fixed", "ignore", "word"))
      if("blank" %in% options){
        # 10 times faster than grepl("\\S", x, perl = TRUE)
        qui_blank = cpp_which_empty(x)
        if(length(qui_blank) > 0){
          qui[qui_blank] = FALSE
        } 
        
      } else if("empty" %in% options){
        qui = qui & x != ""
      }

    } else {
      options = setdiff(options, c("fixed", "ignore", "word"))
      if(length(options) == 0){
        opt = "empty"
      } else {
        opt = options[1]
      }

      # beware I define here the ones we should keep, **not* the ones we should drop

      if(opt == "empty"){
        qui = x != ""
      } else if(opt == "blank"){
        # 10 times faster than grepl("\\S", x, perl = TRUE)
        qui = cpp_which_empty(x)
        if(length(qui) == 0){
          qui = seq_along(x)
        } else {
          qui = -qui
        }
      } else if(opt == "noalpha"){
        qui = grepl("[\\p{L}]", x, perl = TRUE)
      } else if(opt == "noalnum"){
        qui = grepl("[\\p{L}[:digit:]]", x, perl = TRUE)
      } else if(opt == "all"){
        qui = integer(0) 
      }
    }

    if(is.logical(qui)){
      qui = which(qui)
    }

    res = x[qui]
    if(!is.null(group_index) && group_flag != 0){
      group_index = group_index[qui]
      group_index = cpp_recreate_index(group_index)
    }

  } else if(op == "num"){
    options = check_set_options(options, c("warn", "soft", "rm", "clear"))

    # warn: warning if failed conversions
    is_warn = "warn" %in% options
    # soft: if some numbers couldn't be parsed to numeric: no conversion is done
    is_soft = "soft" %in% options
    # rm: the on numeric are removed
    is_rm = "rm" %in% options
    # clear: the non numeric are turned into empty strings
    is_clear = "clear" %in% options

    is_valid_num = TRUE
    if(is_warn || is_soft){
      is_valid_num = is_numeric_in_char(x)
      if(!is_valid_num && is_warn){
        msg = if(!is_soft) "" else " No conversion is performed."
        warning("In operation `num`: when trying to convert to numeric, NAs were created.", 
                 msg, call. = FALSE)
      }
    }

    if(is_valid_num || !is_soft){
      res = suppressWarnings(as.numeric(x))
      if(is_rm || is_clear){
        if(anyNA(res)){
          qui_na = which(is.na(res))
          if(is_rm){
            res = res[-qui_na]
            if(!is.null(group_index) && group_flag != 0){
              group_index = group_index[-qui_na]
              group_index = cpp_recreate_index(group_index)
            }
          } else { 
            # is_clear
            res[qui_na] = ""
          }
        }
      }
    } else {
      res = x
    }

  } else if(op == "ascii"){
    # ascii, stop ####
    res = string_to_ascii(x, options)

  } else if(op == "stopwords"){

    # current limitation: does not work for quoted words
    #                     but quoted words are not stopwords usually
    # + the algo is quite inefficient

    # Snowball stopwords
    # These come from http://snowballstem.org/algorithms/english/stop.txt
    stopwords = c("i", "me", "my", "myself", "we", "our", "ours", "ourselves",
                  "you", "your", "yours", "yourself", "yourselves", "he", "him",
                  "his", "himself", "she", "her", "hers", "herself", "it", "its",
                  "itself", "they", "them", "their", "theirs", "themselves", "what",
                  "which", "who", "whom", "this", "that", "these", "those", "am", "is",
                  "are", "was", "were", "be", "been", "being", "have", "has", "had",
                  "having", "do", "does", "did", "doing", "would", "should",
                  "could", "ought", "i'm", "you're", "he's", "she's", "it's",
                  "we're", "they're", "i've", "you've", "we've", "they've",
                  "i'd", "you'd", "he'd", "she'd", "we'd", "they'd", "i'll",
                  "you'll", "he'll", "she'll", "we'll", "they'll", "isn't",
                  "aren't", "wasn't", "weren't", "hasn't", "haven't", "hadn't",
                  "doesn't", "don't", "didn't", "won't", "wouldn't", "shan't",
                  "shouldn't", "can't", "cannot", "couldn't", "mustn't", "let's",
                  "that's", "who's", "what's", "here's", "there's", "when's", "where's",
                  "why's", "how's", "a", "an", "the", "and", "but", "if", "or",
                  "because", "as", "until", "while", "of", "at", "by", "for",
                  "with", "about", "against", "between", "into", "through",
                  "during", "before", "after", "above", "below", "to", "from",
                  "up", "down", "in", "out", "on", "off", "over", "under",
                  "again", "further", "then", "once", "here", "there", "when",
                  "where", "why", "how", "all", "any", "both", "each", "few",
                  "more", "most", "other", "some", "such", "no", "nor", "not",
                  "only", "own", "same", "so", "than", "too", "very")

    n = length(x)
    x_split = strsplit(x, "(?<=[[:alnum:]])(?=[^[:alnum:]'])|(?<=[^[:alnum:]'])(?=[[:alnum:]])", perl = TRUE)
    x_len = lengths(x_split)
    x_vec = unlist(x_split)

    id = rep(1:n, x_len)

    # Lowering is costly, checking costs a bit less
    if(any(grepl("[[:upper:]]", x))){
      qui_drop = which(tolower(x_vec) %in% stopwords)
    } else {
      qui_drop = which(x_vec %in% stopwords)
    }

    x_vec = x_vec[-qui_drop]
    id = id[-qui_drop]
    
    res = cpp_paste_conditional(x_vec, id)

  } else if(op %in% c("if", "vif")){
    #
    # Conditions: if, vif ####
    #
    
    info = strsplit(argument, "_;;;_", fixed = TRUE)[[1]]
    
    cond_raw = info[[1]]
    true_section = info[[2]]
    is_false_section = length(info) == 3
    if(is_false_section){
      false_section = info[[3]]  
    } else {
      false_section = NULL
    }

    # computing the condition
    # a) parsing
    if(.check){
      cond_parsed = check_expr(str2lang(cond_raw), 
            .sma("The string operation {bq?op} should be of the form ",
                 "{op}(condition ; true ; false). ",
                 "\nPROBLEM: the condition {bq?cond_raw} could not be parsed, see error below:"), 
            verbatim = TRUE)
    } else {
      cond_parsed = str2lang(cond_raw)
    }
    
    # b) evaluating
    if(is.character(cond_parsed)){
      # REGEX
      
      if(.check){
        cond = check_expr(string_is(x, cond_parsed, envir = .envir), 
                  .sma("The operation is of the form {bq?op}(cond ; true ; false). ",
                  "When `cond` is a pure character string, the function `string_is()` ",
                  "is applied with `cond` being the searched pattern.",
                  "\nPROBLEM: in {op}({'_;;;_ => ;'R ? argument}), the condition ",
                  "{bq?cond_raw} led to an error, see below:"),
                  verbatim = TRUE)
      } else {
        cond = string_is(x, cond_parsed, envir = .envir)
      }      
      
    } else {
      vars = all.vars(cond_parsed)
      my_data = list()
      for(v in vars){
        if(v %in% c(".N", ".len")){
          my_data[[v]] = length(x)
        } else if(v %in% c(".nchar", ".C")){
          if(!true_character(x)){
            x = as.character(x)
          }
          my_data[[v]] = nchar(x)
        } else if(v == "."){
          my_data[[v]] = x
        }
      }
      
      if(.check){
        cond = check_expr(eval(cond_parsed, my_data, .envir), 
                  .sma("The string operation {bq?op} should be of the form ",
                       "{op}(condition ; true ; false). ",
                       "\nPROBLEM: the condition {bq?cond_raw} could not be ",
                       "evaluated, see error below:"), 
                   verbatim = TRUE)  
      } else {
        cond = eval(cond_parsed, my_data, .envir)
      }
    }
    
    # further error checking
    if(!length(cond) %in% c(1, length(x))){
      stop_hook("In {bq?op} operations, the condition (here {bq?cond_raw}) must ",
                "be either of length 1 (applying to the full string), ",
                "either of the length of the interpolated value.", 
                "\nPROBLEM: the condition is of length {len?cond} while the ",
                "interpolated value is of length {len?x}.")
    }
    
    if(is.numeric(cond)){
      # numeric values are converted to logical
      cond = as.logical(cond)
    }
    
    if(!is.logical(cond)){
      stop_hook("In {bq?op} operations, the condition (here {bq?cond_raw}) must be logical.",
                "\nPROBLEM: the condition is not logical, instead it is of class ",
                "{enum.bq ? class(cond)}.")
    }
    
    # NA = FALSE
    if(anyNA(cond)){
      cond[is.na(cond)] = FALSE
    }
    
    #
    # Computing the operations
    #    
    
    is_elementwise = length(cond) > 1
    
    if(is_elementwise){
      x_true = x[cond]
      x_false = x[!cond]
    } else {
      if(cond){
        x_true = x
        x_false = character(0)
      } else {
        x_true = character(0)
        x_false = x
      }
    }
    
    # All operations that change the length of the element should be forbidden
    # we check that on the fly

    for(state in c("true", "false")){

      if(state == "true"){
        xi = x_true
      } else {
        xi = x_false
      }
      
      if(length(xi) == 0){
        next
      }
      
      if(state == "false" && !is_false_section){
        break
      }
      
      instruction = if(state == "true") true_section else false_section
      
      if(op == "if"){
        n_old = length(xi)
        cond_flag = +grepl("~(", instruction, fixed = TRUE)
        xi = apply_simple_operations(xi, op, instruction, .check, .envir, .data,
                                     group_flag = cond_flag, .delim,
                                     .user_funs = .user_funs, 
                                     .valid_operators = .valid_operators)
        
        if(is_elementwise && !length(xi) %in% c(0, n_old)){
          stop_hook("In {bq?op} operations, when conditions are of length > 1, the ",
                   "operations either a) can remove the string completely ",
                   "(e.g. with `nuke`), or b) must not change the length of the element.",
                   "\nPROBLEM: the original vector is of length {n?n_old} while after ",
                   "the operations it becomes of length {len ? xi}")
        }
      } else {
        # verbatim if
        if(grepl(".", instruction, fixed = TRUE)){
          .data[["."]] = xi
        }
        
        xi = string_magic_internal(instruction, .delim = .delim, .envir = .envir, 
                             .data = .data, .check = .check, 
                             .user_funs = .user_funs, .valid_operators = .valid_operators)
      }
      
      if(state == "true"){
        x_true = xi
      } else {
        x_false = xi
      }
    }
    
    if(is_elementwise){
      
      if(length(x_true) == 0 && length(x_false) == 0){
        res = character(0)
        if(group_flag != 0){
          group_index = NULL
        }
      } else if(length(x_true) == 0){
        if(op == "vif" && length(x_false) == 1){
          res = x
          res[!cond] = x_false
        } else {
          res = x_false  
        }
        
        if(group_flag != 0){
          group_index = group_index[!cond]
        }
      } else if(length(x_false) == 0){
        if(op == "vif" && length(x_true) == 1){
          res = x
          res[cond] = x_true
        } else {
          res = x_true
        }
        
        if(group_flag != 0){
          group_index = group_index[cond]
        }
      } else {
        res = x
        res[cond] = x_true
        res[!cond] = x_false
      }
    } else {
      res = if(cond) x_true else x_false
    }

  } else if(op == "~"){
    # Group wise: ~ ####
    res = apply_simple_operations(x, op, argument, .check, .envir, .data,
                                  group_flag = 2, .delim, .user_funs = .user_funs, 
                                  .valid_operators = .valid_operators)
                                  
    group_index = attr(res, "group_index")

  } else {
    msg = paste0("In `string_magic`: the operator `", op, "` is not recognized. ",
                "Internal error: this problem should have been spotted beforehand.")
    .stop_hook(msg)
  }

  if(length(group_index) > 0){
    attr(res, "group_index") = group_index
  }

  return(res)
}



sma_pluralize = function(operators, xi, .delim, .envir, .data, .check, 
                         .user_funs, .valid_operators){

  plural_len = operators[1] == "$"

  if(!plural_len){
    # xi should be a number
    is_pblm = length(xi) != 1 || !is.atomic(xi) || any(is.na(xi))

    if(!is_pblm && !is.numeric(xi)){
      xi = suppressWarnings(as.numeric(xi))
      is_pblm = any(is.na(xi))
    }

    if(!is_pblm){
      is_pblm = xi < 0
    }

    if(is_pblm){
      example = 'Example: x = 5; string_magic("There {#is, N ? x} cat{#s} in the room.")'

      extra = ""
      reason = NULL
      if(length(xi) != 1){
        extra = " single"
        reason = "PROBLEM: it is not of length 1."
      } else if(!is.atomic(xi)){
        reason = paste0("PROBLEM: it is not atomic (it is of class", class(xi)[1], ").")
      } else if(is.na(xi)){
        reason = "PROBLEM: It is NA."
      } else if(xi < 0){
        extra = " positive"
        reason = "PROBLEM: It is negative."
      }

      .stop_hook("The pluralization operator `#` applies to a", extra, " number. ", reason, ".\n",
              example)
    }
  }

  n = if(plural_len) length(xi) else xi
  IS_ZERO = n == 0
  IS_PLURAL = n > 1

  operators = operators[-1]
  n_op = length(operators)
  res = rep(NA_character_, n_op)
  for(i in 1:n_op){
    op_parsed = cpp_parse_operator(operators[i])

    op = op_parsed$operator
    options = op_parsed$options
    argument = op_parsed$argument
    if(op_parsed$eval){
      argument_call = check_set_oparg_parse(argument, op, .check)
      argument = check_set_oparg_eval(argument_call, .data, .envir, op, .check)
    }

    if(op %in% c("es", "s", "y", "ies")){
      zero_case = opt_equal(options, c("zero", "0"))

      is_really_plural = IS_PLURAL || (zero_case && IS_ZERO)

      if(op %in% c("s", "es")){
        if(is_really_plural){
          res[i] = op
        } else {
          res[i] = ""
        }
      } else {
        if(is_really_plural){
          res[i] = "ies"
        } else {
          res[i] = "y"
        }
      }

    } else if(op %in% c("n", "N", "len", "Len")){
      
      options = check_set_options(options, c("letter", "upper", "no", "No"))

      is_letter = any(options %in% c("letter", "upper")) || substr(op, 1, 1) %in% c("N", "L")
      is_upper = "upper" %in% options
      
      if(IS_ZERO && nchar(argument) > 0){
        res[i] = argument
      } else if(IS_ZERO && any(options %in% c("no", "No"))){
        if(is_upper || "No" %in% options){
          res[i] = "No"
        } else {
          res[i] = "no"
        }
      } else if(is_letter){
        val = n_letter(n)

        if(is_upper){
          val = gsub("^(.)", "\\U\\1", val, perl = TRUE)
        }

        res[i] = val
      } else {
        res[i] = format(n, big.mark = ",")
      }
      
    } else if(op %in% c("nth", "Nth")){
      is_compact = opt_equal(options, "compact") 
      is_letter = opt_equal(options, c("letter", "upper")) || substr(op, 1, 1) == "N"
      val = n_th(n, letters = is_letter, compact = is_compact)
      if(opt_equal(options, "upper")){
        val = gsub("^(.)", "\\U\\1", val, perl = TRUE)
      }
      res[i] = val

    } else if(op %in% c("ntimes", "Ntimes")){
      is_letter = opt_equal(options, "letter") || substr(op, 1, 1) == "N"
      res[i] = n_times(n, letters = is_letter)

    } else if(op %in% c("zero", "singular", "plural")){
      if((op == "zero" && IS_ZERO) || 
         (op == "singular" && !IS_PLURAL && !(any(grepl("zero$", operators)) && IS_ZERO)) || 
         (op == "plural" && IS_PLURAL)){
          
          value = string_magic_internal(argument, .delim = .delim, .envir = .envir, .data = .data,
                                        .check = .check, .plural_value = xi, 
                                        .user_funs = .user_funs, 
                                        .valid_operators = .valid_operators)
        if(value != ""){
          res[i] = value
        }        
      }
    } else if(op == "enum"){

      res[i] = enum_main(xi, options = options)

    } else {
      # The verb is always last

      if(nchar(op) < 2){
        example = 'Example: x = c("Charles", "Alice"); string_magic("{$Is, enum ? x} crazy? Yes {$(he:they), are}.")'
        .stop_hook("In pluralization, `", op, "` is expected to be a verb ",
                   "and verbs must always be composed of at least two letters.\n", example)
      }

      zero_case = opt_equal(options, c("zero", "0"))
      is_really_plural = IS_PLURAL || (zero_case && IS_ZERO)

      res[i] = conjugate(op, is_really_plural)
    }

  }

  res = res[!is.na(res)]
  paste(res, collapse = " ")
}

sma_ifelse = function(operators, xi, xi_val, .envir, .data, .delim, .check, 
                      .user_funs, .valid_operators){

  if(is.numeric(xi)){
    xi = xi != 0
  }
  
  amp = operators[1]
  is_double_amp = amp == "&&"
  if(!is.logical(xi)){
    form = "{&cond ; true ; false}"
    if(is_double_amp) form = "{&&cond ; true}"
    example = '\nEXAMPLE: x = Sys.time(); string_magic("Hello {&format(x, \'%H\') < 20 ; Sun ; Moon}!")'

    .stop_hook("The if-else operator `", amp, "`, of the form ", form,
            ", accepts only logical values in the condition. ",
            "PROBLEM: the value is not logical, but of class `",
            class(xi)[1], "`.", example)
  }

  if(anyNA(xi)){
    form = "{&cond ; true ; false}"
    if(is_double_amp) form = "{&&cond ; true}"
    example = '\nEXAMPLE: x = Sys.time(); string_magic("Hello {&format(x, \'%H\') < 20 ; Sun ; Moon}!")'

    .stop_hook("The if-else operator `", amp, "`, of the form ", form,
              ", accepts only non-NA logical values.\n",
              "PROBLEM: the condition contains NA values.", example)
  }

  BOX_OPEN = .delim[1]
  true = operators[3]
  false = operators[4]
  
  if(is_double_amp && !is.na(false)){
    .stop_hook("The if-else operator `&&`, of the form ", 
               "{&&cond ; true} does not accept a 'false' statement (since it will ",
               "be filled with the value of the variable in the condition). ",
              "\nFIX: remove the false statement or use the if-else operator `&` (single ampersand).",
              "\nEXAMPLE: x = 1:5",
              "           compare string_magic(\"{x} = {&x %% 2;odd;even}\")",
              "\n              to string_magic(\"{x} = {&x %% 2;odd}\")",
              "\n              to string_magic(\"{x} = {&&x %% 2;odd}\")")
  }
  
  if(is.na(false)) false = ""
  
  true = cpp_extract_quote_from_op(true)
  false = cpp_extract_quote_from_op(false)
  
  new_values_done = FALSE
  n_x = length(xi)
  is_vector = n_x > 1
  
  if(is_vector){
    # if vector: true/false can be vectors
    for(i in 1:2){
      log_op = if(i == 1) true else false
      if(grepl(BOX_OPEN, log_op, fixed = TRUE)){
        if(!new_values_done){
          new_values_done = TRUE
          .data[["."]] = xi_val
          .data[[".N"]] = length(xi_val)
          .data[[".len"]] = length(xi_val)
        }
        
        log_op_eval = string_magic_internal(log_op, .delim = .delim, .envir = .envir, .data = .data,
                                            .check = .check, .plural_value = xi_val, 
                                            .user_funs = .user_funs, 
                                            .valid_operators = .valid_operators)
        
        if(!length(log_op) %in% c(1, n_x)){
          form = "{&cond ; true ; false}"
          if(is_double_amp) form = "{&&cond ; true}"
          .stop_hook("The if-else operator of the form {form} accepts values in ",
                      "{&i==1;true;false} which can be interpolated, as is the case here ",
                      "for {bq?log_op}.",
                      "\nThe length of the interpolated value must be equal to 1 or the length ",
                      "of the condition.",
                      "PROBLEM: the length of the condition is {n?n_x} while the length of ",
                      "`{&i==1;true;false}` is {len?log_op_eval}.")
        }
        
        if(i == 1){
          true = log_op_eval
        } else {
          false = log_op_eval
        }        
      }
    }
  }

  if(operators[1] == "&"){
    # we replace TRUE and FALSE with the strings, even empty
    
    true_long = length(true) > 1
    false_long = length(false) > 1
    
    if(true_long || false_long){
       if(true_long){
        res = true
        if(false_long){
          index_false = which(!xi)
          res[index_false] = false[index_false]
        } else {
          res[!xi] = false
        }
       } else {
        res = false
        res[xi] = true
       }
       
       if(anyNA(xi)){
        res[is.na(xi)] = NA
       }
    } else {
      res = c(false, true)[xi + 1]
    }
    
  } else {
    # we change only the ones that need to be changed
    res = xi_val
    
    if(length(true) > 1){
      res[xi] = true[xi]
    } else {
      res[xi] = true
    }
  }

  # we allow nestedness only for single values
  if(length(res) == 1 && grepl(BOX_OPEN, res, fixed = TRUE)){
    if(!new_values_done){
        new_values_done = TRUE
        .data[["."]] = xi_val
        .data[[".N"]] = length(xi_val)
        .data[[".len"]] = length(xi_val)
      }
    res = string_magic_internal(res, .delim = .delim, .envir = .envir, .data = .data,
                                .check = .check, .plural_value = xi_val, 
                                .user_funs = .user_funs, .valid_operators = .valid_operators)
  }

  res
}


apply_simple_operations = function(x, op, operations_string, .check = FALSE, .envir = NULL, 
                                   .data = list(), group_flag = 0, .delim = c("{", "}"), 
                                   .valid_operators = NULL, .user_funs = NULL){
                                  
  op_all = cpp_parse_simple_operations(operations_string, .delim)
    
  if(length(op_all) == 0){
    return(x)
  }
  
  if(identical(op_all[1], "_ERROR_")){
    op_msg = "'arg'op1, op2"
    if(op =='~') op_msg = "~(op1, op2)" 
    if(op == "if") op_msg = "if(cond ; op1, op2 ; op3, op4)"
    
    last = gsub("_;;;_", ";", tail(op_all, 1))
    extra = ""
    if(last %in% c("!", "?")){
      extra = string_magic("The character {bq?last} is forbidden in operations.")
    } else if(last != "_ERROR_"){
      extra = string_magic("Operations must be of the form `'arg'operator.option` but",
                     " the value {bq?last} is ill formed.")
    } else {
      extra = string_magic("The value {bq?operations_string} could not be parsed.")
    }
      
    msg = string_magic("The operator {bq?op} expects a suite of valid operations (format: {bq?op_msg}). ",
               "\nPROBLEM: the operations were not formatted correctly. ", extra)
    
    .stop_hook(msg)
  }
  
  #
  # applying the operations
  #
  
  xi = x

  for(i in seq_along(op_all)){
    opi = op_all[i]

    op_parsed = sma_char2operator(opi, .valid_operators)

    if(op_parsed$eval){
      if(.check){
        argument_call = check_expr(str2lang(op_parsed$argument), 
                                   .sma("In the operation `{op}()`, the ",
                                        "{&length(op_all) == 1 ; operation ; chain of operations} ",
                                        " {bq?operations_string} led to a problem.", 
                                        "In operation {bq?opi}, the argument in backticks ", 
                                        "(equal to {bq?op_parsed$argument}) is evaluated from the calling environment.",
                                        "\nPROBLEM: the argument could not be parsed."),
                                    verbatim = TRUE)

        argument = check_expr(eval(argument_call, .envir),
                              "In the operation `{op}()`, the ",
                              "{&length(op_all) == 1 ; operation ; chain of operations} ",
                              " {bq?operations_string} led to a problem.", 
                              "In operation {bq?opi}, the argument in backticks ", 
                              "(equal to {bq?op_parsed$argument}) is evaluated from the calling environment.",
                              "\nPROBLEM: the argument could not be evaluated.")
      } else {
        argument = eval(str2lang(op_parsed$argument), .envir)
      }

    } else {
      argument = op_parsed$argument
    }

    if(.check){
      xi = check_expr(sma_operators(xi, op_parsed$operator, op_parsed$options, argument, 
                                    group_flag = group_flag, .delim = .delim, .envir = .envir,
                                    .user_funs = .user_funs, .valid_operators = .valid_operators),
                      "In the operation `{op}()`, the ",
                      "{&length(op_all) == 1 ; operation ; chain of operations} ",
                      " {bq?operations_string} led to a problem.", 
                      "\nPROBLEM: the operation {opi} failed. Look up the doc?")
    } else {
      xi = sma_operators(xi, op_parsed$operator, op_parsed$options, argument, 
                         group_flag = group_flag, .delim = .delim, .envir = .envir,
                         .user_funs = .user_funs, .valid_operators = .valid_operators)
    }
  }

  res = xi

}

setup_operations = function(){
  OPERATORS = c("s", "S", "split", "Split", "x", "X", "extract", 
                "c", "C", "collapse", "Collapse", "r", "R", "replace", "clean",
                "times", "each", "align", "fill", "join", "escape",
                "~", "if", "vif",
                "upper", "lower", "q", "Q", "bq", 
                "format", "Format", "%",
                "erase", "rm", "nuke", "append", "paste", "insert", 
                "k", "shorten", "Shorten", 
                "K", "head", "last", "first",
                "firstchar", "lastchar", "unik", "num", "enum",
                "rev", "sort", "dsort", "ascii", "title",
                "ws", "tws", "trim", "get", "is", "which",
                "n", "N", "len", "Len", "width", "difftime",
                "stopwords", "nth", "Nth", "ntimes", "Ntimes")
                
  options("string_magic_operations_v1.0.0" = sort(OPERATORS))
  
  # dp: v1.1.0
  # swidth: v1.1.3
  # table, round, r0-r5, signif, s0-s5: v1.1.3
  OPERATORS = c(OPERATORS, 
                # v1.1.0
                "dp", "deparse", 
                # v1.1.3
                "swidth", "table", 
                "round", paste0("r", 0:6), 
                "signif", paste0("s", 0:6))
  
  options("string_magic_operations_default" = sort(OPERATORS))
  
  options("string_magic_user_ops" = NULL)
}

#' String operations and interpolation: magic edition
#' 
#' The stringmagic package introduces a new language tailored to create complex character strings. 
#' Use the function string_magic to interpolate a variable inside a character string and to apply, 
#' on-the-fly, any arbitrary string operation.
#' 
#' It supports over 50 basic operations, seamless customization, nesting, 
#' pluralization, and much more. Writing complex character strings has never been that easy!
#' 
#' To get an overview of the package: see the [Readme](https://lrberge.github.io/stringmagic/index.html). 
#' To get started with string_magic,
#'  see [the vignette](https://lrberge.github.io/stringmagic/articles/guide_string_magic.html).
#' 
#' This package also enhances regular string functions (like grep and co) with
#'  specialized functions to:
#' 
#' - detect combinations of regular expressions
#' - chain basic string operations
#' - clean character string vectors
#' 
#' See more in the [dedicated vignette](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html).
#' 
#' @author 
#' Laurent Berge
#' 
#' For a few functions (e.g. string_split2df) I use internally an algorithm to turn vectors 
#' into indexes (i.e. integer vectors of the same length ranging to 1 to the number of groups).
#' This algorithm was inspired by Sebastian Krantz's [collapse](https://cran.r-project.org/package=collapse)
#' (if you don't know about it, you should definitely check it out!) who uses an algorithm
#' from Morgan Jacob's [kit](https://cran.r-project.org/package=kit) package.
#' 
#' @keywords internal
#' 
"_PACKAGE"

#------------------------------------------------------------------------------#
# Author: Laurent R. Bergé
# Created: 2023-06-02
# ~: string_magic help setup
#------------------------------------------------------------------------------#
  



generate_help_extensive = function(){
  
  if(!is_string_magic_root()) return(NULL)
  
  if(!identical(parent.frame(), .GlobalEnv) && 
         !identical(deparse(sys.calls()[[1]]), "devtools::load_all()")){
    return(NULL)
  }
  
  mtime_origin = floor(as.numeric(file.info("R/string_magic_doc.R")["mtime"]))
  mtime_destination = readLines("R/AUTO_help.R", n = 1)
  mtime_destination = as.numeric(string_trim(mtime_destination, 2))
    
  if(mtime_origin <= mtime_destination){
    return(NULL)
  }
  
  # we check if browser is used
  for(f in list.files("R", full.names = TRUE)){
    if(grepl("help.R", f, fixed = TRUE)){
      next
    }
    txt = readLines(f)
    if(any(grepl("browser()", txt, fixed = TRUE))){
      return(NULL)
    }
  }
  
  message("Help rewritten.")
  
  string_magic_txt = readLines("R/string_magic_doc.R")
  
  i_string_magic = string_which(string_magic_txt, "^NULL")
  doc = string_magic_txt[1:(i_string_magic - 1)]
  i_start_doc = max(string_which(doc, "!^#'")) + 1
  
  doc = doc[i_start_doc:length(doc)]
  
  # We select only the sections
  i_at = string_which(doc, "#' @")
  is_sec = string_is(doc[i_at], "#' @section")
  i_end = c(i_at[-1], length(doc)) - 1
  
  # we extract + format the section
  text = c()
  for(id_sec in which(is_sec)){
    my_sec = doc[i_at[id_sec]:i_end[id_sec]]
        
    title = string_ops(my_sec[1], "'.+@section'r, tws, ':$'r, '# 'paste, ' ----|'paste.right")
    content = string_ops(my_sec[-1], "'^#.'r, tws, '_ENDLINE_'paste.right, ''c")
    
    content = string_clean(content, 
                        "^(_ENDLINE_)+|(_ENDLINE_)+$",
                        "(_ENDLINE_){2,} => _NEWLINE__NEWLINE_",
                        "_ENDLINE_([+-]) => _NEWLINE_  \\1",
                        " *_ENDLINE_ * =>  ")
                        
    content = string_ops(content, "'_NEWLINE_'s")
    
    text = c(text, "", title, "", content)
  }
  
  # modifying the text so that it can be written and be interpreted as code
  text_dp = capture.output(dput(as.character(text)))
  text_dp[1] = paste0("txt = ", text_dp[1])
  
  fun_txt = c(paste0("# ", mtime_origin), 
              "# DO NOT EDIT BY HAND: generated with generate_help_extensive() in help.R",
              "",
              "setup_help_extensive = function(){",
              text_dp,
              "",
              "  options(string_magic_help_extensive = txt)",
              "}")
  
  writeLines(fun_txt, "./R/AUTO_help.R")
}


format_help = function(pattern = NULL, x = NULL){
  # pattern = "replace"
  
  if(is.null(x)){
    x = getOption("string_magic_help_extensive")
  }
  
  select = logical(length(x))

  for(i in seq_along(pattern)){
    p = pattern[i]

    qui = string_is(x, p, ignore.case = TRUE)

    if(!any(qui)){
      stop_up("In argument `.help`, the pattern `", p, "` did not match any element of the documentation.")
    }

    select = select | qui
  }

  # We select the section
  # of each selected line

  line_id = which(select)
  section_id = string_which(x, "^# ")
  empty_id = which(x == "")

  start = end = c()
  for(i in seq_along(line_id)){
    line = line_id[i]
    start[i] = max(section_id[section_id <= line])
    if(line == length(x)){
      end[i] = line
    } else if(line %in% section_id){
      if(line == max(section_id)){
        end[i] = length(x)
      } else {
        next_section = min(section_id[section_id > line])
        end[i] = next_section - 1
      }
    } else {
      end[i] = min(empty_id[empty_id > line])
    }
  }

  info = cbind(start, end)
  info = info[order(start, -end), , drop = FALSE]
  info = info[!duplicated(info[, 1]), , drop = FALSE]

  text = c()
  for(i in 1:nrow(info)){
    text = c(text, x[info[i, 1]:info[i, 2]])
  }
  
  # we format the text
  text = fit_screen(paste(text, collapse = "\n"))
  text = strsplit(text, "\n")[[1]]
  
  # we format the sections
  if(any(string_is(text, "fixed/---|"))){
    qui_sec = string_is(text, "fixed/---|")
    text[qui_sec] = .sma("{80k ! {'f/|'r ? text[qui_sec]}{80 times.c ! -}}|")
  }
  
  # we highlight the selections
  highlights = list(line = NULL, value = NULL)
  for(i in seq_along(pattern)){
    p = pattern[i]
    qui = string_is(text, p, ignore.case = TRUE)
    
    pat_parsed = format_simple_regex_flags(p, ignore = TRUE)
    p = pat_parsed$pattern
    is_fixed  = pat_parsed$fixed
    
    for(j in which(qui)){
      line = text[j]
      
      pos = which(highlights$line == j)
      line_exists = length(pos) > 0
      if(line_exists){
        line_new = highlights$value[pos]
      } else {
        line_new = gsub(".", " ", line)
      }
      
      info = gregexec(p, line, fixed = is_fixed, perl = !is_fixed)[[1]]
      
      start = as.numeric(info)
      len = attr(info, "match.length")
      
      replace = string_fill("", max(len), symbol = "^")
      
      for(k in seq_along(start)){
        substr(line_new, start[k], start[k] + len[k] - 1) = replace
      }
      
      if(line_exists){
        highlights$value[pos] = line_new
      } else {
        highlights$line = c(highlights$line, j)
        highlights$value = c(highlights$value, line_new)
      }      
    }
  }
  
  # now adding the highlights to the text
  hlines = highlights$line
  hvalues = highlights$value
  
  text[hlines] = paste0(text[hlines], "\n", hvalues)
  
  n = length(text)
  
  section_id = string_which(text, "^# ")
  empty_id = which(text == "")
  match_id = string_which(text, "(^| +)\\^+( +|$)")
  bullet_id = string_which(text, "^ *[+-]")
  
  msg = function(x, start = 1, len = 1){
    my_range = start:min(start + len - 1, length(text))
    cat(paste(text[my_range], collapse = "\n"))
  }
  
  # if small: we print the full text
  if(length(text) < 10){
    msg(text, 1, 500)
    return(invisible(NULL))
  }
  
  message(.sma("Welcome to string_magic dynamic help:\n", 
               "enter = continue ; p = next paragraph ; n = next match ; N = next section ; q or Q = quit"))
  
  msg(text, 1, 5)
  i = 6
  
  while(i <= n){
    choice = readline(text[i])
    i = i + 1
    if(i >= n) break
    
    if(choice == ""){
      msg(text, i, 1)
      i = i + 1
    } else if(choice == "n"){
      # next match
      
      i_next = match_id[match_id > i]
      
      if(length(i_next) == 0){
        message("---> no additional match found")
        return(invisible(NULL))
      }
      
      message("---> jumping to the next match\n")
      
      # we add context
      i_min = i
      i = min(i_next)
      
      while(i >= i_min){
        if(text[i] == ""){
          # end
          i = i + 1
          break
        } else if(i %in% bullet_id){
          # end 
          break
        } else {
          i = i - 1
        }
      }
            
      msg(text, i, 1)
      i = i + 1
    } else if(choice %in% c("p", "N")){
      
      if(choice == "p"){
        # next paragraph
        i_next = min(empty_id[empty_id > i] + 1, bullet_id[bullet_id > i])
      } else {
        # next section
        i_next = section_id[section_id > i]
      }
      
      what = if(choice == "p") "paragraph" else "section"
      
      if(length(i_next) == 0){
        message("---> this was the last ", what, "\n")
        return(invisible(NULL))
      }
      i_next = min(i_next)
      
      message("---> jumping to next ", what, "\n")
      i = i_next
      
      msg(text, i, 1)
      i = i + 1      
    } else if(choice %in% c("Q", "q")){
      return(invisible(NULL))
    } else {
      message(.sma("Command {bq?choice} is not recognized, please whoose among: {S, bq, C ! enter, p, n, N, Q}"))
    }
  }
}

general_help = function(){
  
  message("Welcome to string_magic help. Please choose which sections to read.")
  
  x = getOption("string_magic_help_extensive")
  
  #
  # showing the titles
  #
  
  section_titles = string_ops(x, "'^#'get, '^#|-+\\|'r, tws")
  
  max_w = getOption("width") / 2 - 5
  
  section_titles = string_ops(section_titles, "`max_w`width, ':01:: 'paste, '\n => \n    'r")
  n_sec = length(section_titles)
  
  if(length(section_titles) %% 2 == 1){
    section_titles = c(section_titles, " ")
  }
    
  if(any(string_is(section_titles, "\n"))){
    no_nl = string_which(section_titles, "!\n")
    section_titles[no_nl] = paste0(section_titles[no_nl], "\n ")
    
    section_titles_split = unlist(strsplit(section_titles, "\n"))
    index = seq_along(section_titles_split)
    main_part = section_titles_split[index %% 2 == 1]
    extra_part = section_titles_split[index %% 2 == 0]
    index_extra = seq_along(extra_part)
    extra_part_1 = extra_part[index_extra %% 2 == 1]
    extra_part_2 = extra_part[index_extra %% 2 == 0]
    
    insert_index = index_extra[index_extra %% 2 == 0]
    tmp = insert_at(main_part, extra_part_1, insert_index)
    section_titles = insert_at(tmp, extra_part_2, insert_index + seq_along(extra_part_1))
  } 
  
  index = seq_along(section_titles)
  titles_odd = section_titles[index %% 2 == 1] 
  titles_even = section_titles[index %% 2 == 0]
  
  mat_titles = cbind(string_fill(titles_odd), string_fill(titles_even))
  is_empty = string_is(titles_odd, "^\\s*$") & string_is(titles_even, "^\\s*$")
  
  mat_titles = mat_titles[!is_empty, , drop = FALSE]
  
  mat_titles = apply(mat_titles, 1, paste, collapse = "  ")
  cat(mat_titles, sep = "\n")
  
  while(TRUE){
    choice = readline("Which to select (Q = quit)? (ex: 5, 11): ")
    if(choice %in% c("q", "Q")){
      return(invisible(NULL))
    }  
    
    choice_fmt = try(eval(str2lang(.sma("c({choice})"))), silent = TRUE)
    if(inherits(choice_fmt, "try-error")){
      message(fit_screen(.sma("Sorry the parsing failed, please provide a comma separated ", 
                              "list of numbers, like '1, 2, 6' for example.")))
    } else if(any(!choice_fmt %in% 1:n_sec)){
      message(fit_screen(.sma("All numbers must be in between 1 and {n_sec}.")))
    } else {
      break
    }
  }
  
  #
  # The help
  #
  
  section_choice = choice_fmt
  
  section_id = string_which(x, "^#")
  text = c()
  for(id in section_choice){
    i_start = section_id[id]
    i_end = if(i_start == max(section_id)) length(x) else min(section_id[section_id > i_start]) - 1
    text = c(text, x[i_start:i_end])
  }

  msg = function(x, start = 1, len = 1){
    my_range = start:min(start + len - 1, length(text))
    cat(paste(text[my_range], collapse = "\n"))
  }
  
  # we format the sections
  if(any(string_is(text, "fixed/---|"))){
    qui_sec = string_is(text, "fixed/---|")
    text[qui_sec] = .sma("{80k ! {'f/|'r ? text[qui_sec]}{80 times.c ! -}}|")
  }
  
  text_origin = text
  text = strsplit(fit_screen(paste0(text, collapse = "\n")), "\n")[[1]]
  
  # if small: we print the full text
  if(length(text) < 10){
    msg(text, 1, 500)
    return(invisible(NULL))
  }
  
  message("Available controls:")
  message("enter = continue ; p = next paragraph ; N = next section ; s = search a regex ; Q = quit")
  
  empty_id = which(text == "")
  section_id = string_which(text, "^#")
  bullet_id = string_which(text, "^ *[+-]")
  
  msg(text, 1, 5)
  
  n = length(text)
  i = 6 
  while(i < n){
    choice = readline(text[i])
    i = i + 1
    if(i == n) break
    
    if(choice == ""){
      msg(text, i, 1)
      i = i + 1
    } else if(choice %in% c("p", "N")){
      
      if(choice == "p"){
        # next paragraph
        i_next = min(empty_id[empty_id > i] + 1, bullet_id[bullet_id > i])
      } else {
        # next section
        i_next = section_id[section_id > i]
      }
      
      if(length(i_next) == 0){
        message("---> this was the last ", if(choice == "p") "paragraph" else "section", "\n")
        return(invisible(NULL))
      }
      i_next = min(i_next)
      
      message("---> jumping to next ", if(choice == "p") "paragraph" else "section", "\n")
      i = i_next
      
      msg(text, i, 1)
      i = i + 1      
    } else if(choice == "s"){
      message(fit_screen(.sma("Please write a regex pattern (flags `fixed` and `word` accepted,",
                              " ignore case is always on; ex: 'w/width'):")))
                   
      pattern = readline("")
      if(pattern %in% c("q", "Q")){
        return(invisible(NULL))
      }
      
      while(!any(string_is(text, pattern))){
        message(.sma("The pattern {bq?pattern} does not match any item. Please provide a new one:"))
        pattern = readline("")
        if(pattern %in% c("q", "Q")){
          return(invisible(NULL))
        }
      }
      
      format_help(pattern, text)
      return(invisible(NULL))
      
    } else if(choice %in% c("Q", "q")){
      return(invisible(NULL))
    } else {
      message(.sma("Command {bq?choice} is not recognized, please whoose among: {S, bq, C ! enter, p, s, N, Q}"))
    }
    
  }
  
}

string_magic_dynamic_help = function(help){
  if(isTRUE(help)){
    general_help()
  } else {
    format_help(help)
  }
}

#----------------------------------------------#
# Author: Laurent Berge
# Date creation: Wed Jul 27 10:06:18 2022
# ~: checking functions
#----------------------------------------------#


# NOTE: it's unusual for me but there will be A LOT of code duplication
#       the objective is to limit the most possible the overheads, hence this decision
#       + the functions are simple and to the point (much less complex than what is possible with dreamerr), so that's manageable

####
#### checking ####
####


check_logical = function(x, null = FALSE, scalar = FALSE, l0 = FALSE, no_na = FALSE, mbt = FALSE, up = 0){
  set_up(up + 1)

  if(missing(x)){
    if(mbt){
      x_dp = deparse_short(substitute(x))
      stop_up("The argument `", x_dp, "` must be provided. PROBLEM: it is missing.")
    }

    return()
  }

  if(null && is.null(x)){
    return()
  }

  if(!is.logical(x)){
    x_dp = deparse_short(substitute(x))
    nullable = if(null) "(nullable) " else ""
    type = if(scalar) "scalar" else "vector"
    stop_up("The ", nullable, "argument `", x_dp, "` must be a logical ", type, ".",
            " PROBLEM: it is not logical, it is of class {enum?class(x)}.")
  }

  len_x = length(x)
  if(len_x == 0 &&  l0) return()

  if(scalar){
    if(missing(no_na)) no_na = TRUE

    if(len_x != 1){
      x_dp = deparse_short(substitute(x))
      nullable = if(null) "(nullable) " else ""
      stop_up("The ", nullable, "argument `", x_dp, "` must be a logical scalar.",
              " PROBLEM: it is not of length 1, it is of length ", length(x), ".")
    }
  }

  if(no_na && anyNA(x)){
    x_dp = deparse_short(substitute(x))
    nullable = if(null) "(nullable) " else ""
    type = if(scalar) "scalar" else "vector without NA"
    problem = if(scalar) "it is equal to NA" else "it contains NA values"
    stop_up("The ", nullable, "argument `", x_dp, "` must be a logical ", type, ". ",
            "PROBLEM: ", problem, ".")
  }

}

check_numeric = function(x, null = FALSE, scalar = FALSE, l0 = FALSE, no_na = FALSE, 
                         mbt = FALSE, integer = FALSE, complex = FALSE, up = 0){
  set_up(up + 1)
  
  if(missing(x)){
    if(mbt){
      x_dp = deparse_short(substitute(x))
      stop_up("The argument `", x_dp, "` must be provided. PROBLEM: it is missing.")
    }

    return()
  }

  if(null && is.null(x)){
    return()
  }

  if((!complex && !is.numeric(x)) || (complex && !is_num_complex(x))){
    x_dp = deparse_short(substitute(x))
    nullable = if(null) "(nullable) " else ""
    vector_type = if(scalar) "scalar" else "vector"
    type = if(integer) "an integer" else "a numeric"
    if(complex) paste0(type, ", or complex,")
    
    stop_up("The ", nullable, "argument `", x_dp, "` must be ", type, " ", vector_type, ".",
            " PROBLEM: it is not {'^.+ 'r ? type} it is of class {enum?class(x)}.")
  }

  len_x = length(x)
  if(len_x == 0 &&  l0) return()

  if(scalar){
    if(missing(no_na)) no_na = TRUE

    if(len_x != 1){
      x_dp = deparse_short(substitute(x))
      nullable = if(null) "(nullable) " else ""
      type = if(integer) "an integer" else "a numeric"
      
      stop_up("The ", nullable, "argument `", x_dp, "` must be ", type, " scalar.",
              " PROBLEM: it is not of length 1, it is of length ", length(x), ".")
    }
  }

  if(no_na && anyNA(x)){
    x_dp = deparse_short(substitute(x))
    nullable = if(null) "(nullable) " else ""
    vector_type = if(scalar) "scalar" else "vector without NA"
    type = if(integer) "an integer" else "a numeric"
    problem = if(scalar) "it is equal to NA" else "it contains NA values"
    
    stop_up("The ", nullable, "argument `", x_dp, "` must be ", type, " ", vector_type, ". ",
            "PROBLEM: ", problem, ".")
  }
  
  if(integer){
    if(complex){
      for(part in 1:2){
        z = if(part == 1) Re(x) else Im(x)
        if(any(z != round(z))){
          x_dp = deparse_short(substitute(x))
          i = which(z != round(z))[1]
          intro = ""
          if(part == 1 && all(Im(x) == 0)){
            msg = .sma("is equal to {z}")
          } else {
            msg = .sma("has {&i==1;a real;an imaginary} part equal to {z}")
          }
          stop_up("The argument {bq?x_dp} must be an integer {&scalar ; scalar ; vector}.",
                  "\nPROBLEM: {&len(x)==1 ; it ; the {nth ? i} value} ",
                  "{msg}, not an integer.")
        }
      }
    } else {
      if(any(x != round(x))){
        x_dp = deparse_short(substitute(x))
        i = which(x != round(x))[1]
        stop_up("The argument {bq?x_dp} must be an integer {&scalar ; scalar ; vector}.",
                "\nPROBLEM: {&length(x)==1 ; it is equal to {x} ; the {nth ? i} ",
                "value is equal to {x[i]}}, not an integer.")
      }
    }    
  }

}

check_character = function(x, null = FALSE, scalar = FALSE, l0 = FALSE, 
                          no_na = FALSE, mbt = TRUE, up = 0){
                            
  set_up(up + 1)

  if(missing(x)){
    if(mbt){
      x_dp = deparse_short(substitute(x))
      stop_up("The argument `", x_dp, "` must be provided. PROBLEM: it is missing.")
    }

    return()
  }

  if(null && is.null(x)){
    return()
  }

  if(!is.character(x)){
    x_dp = deparse_short(substitute(x))
    nullable = if(null) "(nullable) " else ""
    type = if(scalar) "scalar" else "vector"
    stop_up("The ", nullable, "argument `", x_dp, "` must be a character ", type, ".",
            " PROBLEM: it is not character, it is of class {enum?class(x)}.")
  }

  len_x = length(x)
  if(len_x == 0 && l0 && !scalar) return()

  if(scalar){
    if(missing(no_na)) no_na = TRUE

    if(len_x != 1){
      x_dp = deparse_short(substitute(x))
      nullable = if(null) "(nullable) " else ""
      stop_up("The ", nullable, "argument `", x_dp, "` must be a character scalar.",
              " PROBLEM: it is not of length 1, it is of length ", length(x), ".")
    }
  }

  if(no_na && anyNA(x)){
    x_dp = deparse_short(substitute(x))
    nullable = if(null) "(nullable) " else ""
    type = if(scalar) "scalar" else "vector without NA"
    problem = if(scalar) "it is equal to NA" else "it contains NA values"
    stop_up("The ", nullable, "argument `", x_dp, "` must be a character ", type, ". ",
            "PROBLEM: ", problem, ".")
  }

}

check_set_character = function(x, null = FALSE, scalar = FALSE, l0 = FALSE, 
                               no_na = FALSE, mbt = TRUE){

  if(missing(x)){
    if(mbt){
      x_dp = deparse_short(substitute(x))
      stop_up("The argument `", x_dp, "` must be provided. PROBLEM: it is missing.")
    }

    return(NULL)
  }

  if( (null && is.null(x)) || (l0 && length(x) == 0) ){
    return(NULL)
  }

  if(!is.atomic(x)){
    x_dp = deparse_short(substitute(x))
    stop_up("Argument {bq?x_dp} must be atomic. ",
            "\nPROBLEM: Currently it is of the non-atomic class {bq?class(x)[1]}.")
  }

  if(!is.character(x) || !identical(class(x), "character")){
    x = as.character(x)
  }

  len_x = length(x)
  if(len_x == 0 && l0 && !scalar){
    return(character(0))
  }

  if(scalar){
    if(missing(no_na)) no_na = TRUE

    if(len_x != 1){
      x_dp = deparse_short(substitute(x))
      nullable = if(null) "(nullable) " else ""
      stop_up("The ", nullable, "argument `", x_dp, "` must be a character scalar.\n",
              " PROBLEM: it is not of length 1, it is of length ", length(x), ".")
    }
  }

  if(no_na && anyNA(x)){
    x_dp = deparse_short(substitute(x))
    nullable = if(null) "(nullable) " else ""
    type = if(scalar) "scalar" else "vector without NA"
    problem = if(scalar) "it is equal to NA" else "it contains NA values"
    stop_up("The ", nullable, "argument `", x_dp, "` must be a character ", type, ". \n",
            " PROBLEM: ", problem, ".")
  }
  
  return(x)
}


check_list = function(x){
  if(!is.list(x)){
    x_dp = deparse_short(substitute(x))
    stop_up("The argument `", x_dp, "` must be a list (e.g. like a data.frame). \n",
              "PROBLEM: it is not a list, it is of class {enum?class(x)}.")
  }
}

check_envir = function(x){

  if(!inherits(x, "environment")){
    x_dp = deparse_short(substitute(x))
    if(is.list(x)){
      stop_up("The argument `", x_dp, "` must be an environment (ex: parent.frame()). ",
              "\nPROBLEM: it is not an environment, ",
              "it is a {&is.data.frame(x) ; data.frame ; list}.",
              "\nUse the argument `.data` instead.")
    } else {
      stop_up("The argument `", x_dp, "` must be an environment (ex: parent.frame()). ",
              "PROBLEM: it is not an environment, it is of class {enum?class(x)}.")
    }
  }

}

check_function = function(x, null = FALSE, up = 0, argname = NULL){
  set_up(up + 1)
  
  if(null && is.null(x)){
    return(NULL)
  }
  
  if(!is.function(x)){
    if(is.null(argname)){
      x_dp = deparse_short(substitute(x))
    }
    
    stop_up("The argument `", x_dp, "` must be a function. ",
            "PROBLEM: it is not a function, it is of class {enum?class(x)}.")
  }
}


check_set_dots = function(..., mc = NULL, mbt = FALSE, character = FALSE,
                          no_na = FALSE, scalar = FALSE, nofun = FALSE){
  # check the dots arguments

  n = ...length()
  if(n == 0){

    if(mbt){
      stop_up("At least one element in `...` must be provided. PROBLEM: `...` is empty.")
    } else {
      return(list())
    }

  }

  if(is.null(mc)){
    sysOrigin = sys.parent()
    mc = match.call(definition = sys.function(sysOrigin),
                    call = sys.call(sysOrigin), expand.dots = FALSE)
  }

  dots = vector("list", n)
  
  dots_nm = names(mc[["..."]])
  
  # We first catch evaluation problems
  for(i in 1:n){
    elem = tryCatch(...elt(i), silent = TRUE, 
                    error = function(e) structure(conditionMessage(e), 
                                                  class = "stringmagic-error"))

    if(inherits(elem, "stringmagic-error") || (nofun && is.function(elem))){
      nm = if(is.null(dots_nm)) "" else dots_nm[i]
      if(is.na(nm)) nm = ""

      mc_dots = mc[["..."]]
      value = deparse_short(mc_dots[[i]])

      nm = string_magic(" ({if(.C<4 ; erase) ! {nm} = }{value})")
      
      if(inherits(elem, "stringmagic-error")){
        if(grepl("try(...", elem, fixed = TRUE)){
          elem = gsub("^[^:]+:", "", elem)
        }

        stop_up("In the argument `...`, the {#nth ? i} element{nm} raises an error:\n",
                elem)
      } else {
        stop_up("In the argument `...`, the elements must not be functions.",
                "\nPROBLEM: the {#nth ? i} element{nm} is a function.")
      }
    }

    dots[[i]] = elem
  }
  
  # now we catch type problems
  if(scalar){
    if(any(lengths(dots) != 1)){
      len_dots = lengths(dots)
      i_pblm = which(len_dots != 1)
      len_pblm = len_dots[i_pblm]

      # I purposefully copy-paste in each block
      # otherwise the code becomes too complicated and especially more difficult to debug

      # We try to give as much info as possible
      n_pblm = length(i_pblm)
      nm_pblm = character(n_pblm)

      if(!is.null(dots_nm)){
        rep = dots_nm[i_pblm]
        rep[is.na(rep)] = ""
        nm_pblm = rep
      }

      # the value of the variables
      mc_dots = mc[["..."]]
      value_all = sapply(i_pblm, function(i) deparse_short(mc_dots[[i]]))

      info_call = .string_magic("`{if(.C<4 ; erase) ! {nm_pblm} = }{value_all}`")

      stop_up("In the argument `...`, all elements must be scalars (i.e. of length 1).\nPROBLEM: ",
              "{'\n'c ! The {nth ? i_pblm} element ({info_call}) is ",
              "of length {format, ws ? len_pblm}.}")
    }
  }

  if(character){
    # we convert to character => requirement is atomicity
    if(!all(sapply(dots, is.atomic))){      
      i_pblm = which(!sapply(dots, is.atomic))

      # We try to give as much info as possible
      n_pblm = length(i_pblm)
      nm_pblm = character(n_pblm)

      if(!is.null(dots_nm)){
        rep = dots_nm[i_pblm]
        rep[is.na(rep)] = ""
        nm_pblm = rep
      }

      # the value of the variables
      mc_dots = mc[["..."]]
      value_all = sapply(i_pblm, function(i) deparse_short(mc_dots[[i]]))

      info_call = string_magic("`{if(.C<4 ; erase) ! {nm_pblm} = }{value_all}`")

      cls_pblm = sapply(dots[i_pblm], function(x) string_magic("{bq ! {enum ? class(x)}}"))
      stop_up("In the argument `...`, all elements must be atomic (i.e. convertible to a character string).",
              "\nPROBLEM: {'\n         'c ! The {Nth ? i_pblm} element ({info_call}) is not character",
              " (instead it is of class: {cls_pblm}).}")
    }

    true_character = function(x) is.character(x) && identical(class(x), "character")
    i_no_char = which(!sapply(dots, true_character))
    for(i in i_no_char){
      dots[[i]] = as.character(dots[[i]])
    }
  }

  if(no_na){
    if(any(sapply(dots, anyNA))){
      i_pblm = which(sapply(dots, anyNA))

      # We try to give as much info as possible
      n_pblm = length(i_pblm)
      nm_pblm = character(n_pblm)

      if(!is.null(dots_nm)){
        rep = dots_nm[i_pblm]
        rep[is.na(rep)] = ""
        nm_pblm = rep
      }

      # the value of the variables
      mc_dots = mc[["..."]]
      value_all = sapply(i_pblm, function(i) deparse_short(mc_dots[[i]]))

      info_call = string_magic("`{if(.C<4 ; erase) ! {nm_pblm} = }{value_all}`")

      stop_up("In the argument `...`, all elements must be without NA.\nPROBLEM: ",
              "The {nth, enum ? i_pblm} element{$s} ({C ? info_call})",
              " {$contain} NA values.")
    }
  }
  
  names(dots) = dots_nm

  return(dots)
}


check_set_options = function(x, options, op = NULL, free = FALSE, case = FALSE){
  # x: always a character vector
  # options: the options to match
  if(length(x) == 0) return(x)

  n = length(x)
  res = x
  for(i in 1:n){
    v = x[i]

    pm = pmatch(v, options)
    if(is.na(pm) && !case){
      pm = pmatch(tolower(v), tolower(options))
    }
    
    if(is.na(pm) && !free){
      # absence of a match
      stop_hook("The option {bq?v} is not valid {&is.null(op) ; for the current operation ; for the operator {bq?op}}.\n",
                "FYI the option{$s ? options} available {$are, enum.bq}.")
    }

    if(!is.na(pm)){
      res[i] = options[pm]
    }
  }

  res
}

check_last = function(.last){
  if(is.null(.last) || is.function(.last)){
    # ok, no check
  } else if(!is.character(.last)){
    stop_hook("Argument `.last` mut be equal to an `string_magic` operations chain", 
              " or a function.",
              "\nPROBLEM: instead it is of class {enum.bq?class(.last)}.")
  } else {
    check_character(.last, null = TRUE, scalar = TRUE)
  }
}

check_set_delimiters = function(.delim){
  check_character(.delim, no_na = TRUE, up = 1)
  
  if(length(.delim) == 1){
    .delim = strsplit(.delim, " ", fixed = TRUE)[[1]]
  }
  
  check_delimiters(.delim)
  
  .delim
}

check_set_width = function(width_expr){
  sw = getOption("width") 
  data = list(.sw = sw)
  width = eval(width_expr, data, parent.frame(2))
  
  if(inherits(width, "formula")){
    
    if(length(width) != 2){
      stop_up("If the argument `.width` is a formula, it must be a one sided formula.",
              "\nPROBLEM: this is a two sided formula.")
    }
    
    width = eval(width[[2]], data, parent.frame(2))
    
  } else if(is.null(width)){
    width = min(100, 0.9 * sw)
    
  } else if(isFALSE(width)){
    width = Inf
    
  } else {
    
    if(length(width) != 1){
      stop_up("The argument `.width` must be either: 1) an integer scalar, 2) a number between 0 and 1, 3) FALSE, 4) NULL, 5) a one sided formula with the variable .sw (representing the current screen width).",
              "\nPROBLEM: it is not FALSE, nor NULL nor a formula, but is not of length 1. ",
              "Instead it is of length ", length(width), ".")
    } 
    
    if(!is.numeric(width)){
      stop_up("The argument `.width` must be either: 1) an integer scalar, 2) a number between 0 and 1, 3) FALSE, 4) NULL, 5) a one sided formula with the variable .sw (representing the current screen width).",
              "\nPROBLEM: it is not FALSE, nor NULL nor a formula, but is not numeric. ",
              "Instead it is of class ", class(width)[1], ".")
    }
    
    if(width < 2){
      if(width < 0){
        stop_up("The argument `.width` must be either: 1) an integer scalar, 2) a number between 0 and 1, 3) FALSE, 4) NULL, 5) a one sided formula with the variable .sw (representing the current screen width).",
              "\nPROBLEM: it is not FALSE, nor NULL nor a formula, but is not a positive number. ",
              "It is lower than 0 (", width, ").")
      }
      
      width = width * sw
    }
    
  }
  
  width
}

check_set_split = function(split){
  if(missing(split)){
    if(isTRUE(split)){
      split = ","
    }
    return(split)
  }
  
  if(is.logical(split)){
    check_logical(split, scalar = TRUE, up = 1)
    if(split){
      split = ","
    }
  } else if(is.character(split)){
    check_character(split, scalar = TRUE, up = 1)
    if(nchar(split) != 1){
      stop_hook("The argument `.split` must be a logical scalar or a single character symbol.",
              "\nPROBLEM: it is currently a character scalar ({10 Short, Q ? split}) but ",
              "contains {nchar(split)} characters.")
    }
  } else {
    stop_hook("The argument `.split` must be a logical scalar or a single character symbol.",
              "\nPROBLEM: instead it is of class {enum.bq ? class(split)}.")
  }
  
  split
}

is_num_complex = function(x){
  is.numeric(x) || is.complex(x)
}

check_set_mat = function(cmat, nmat, df){
  res = TRUE
  
  if(is.logical(cmat)){
    check_logical(cmat, scalar = TRUE)
    if(cmat){
      attr(res, "auto") = TRUE
      return(res)
    }    
  } else if(is_num_complex(cmat)){
    check_numeric(cmat, scalar = TRUE, complex = TRUE)
    attr(res, "row_col") = c(Re(cmat), Im(cmat))
    return(res)
  } else {
    stop_hook("The argument `.cmat` must be a logical scalar, an integer (giving the ",
              "nber of rows) or an imaginary (ex: `3i`, giving the number of columns).",
              "\nPROBLEM: instead it is of class {enum.bq ? class(split)}.")
  }
  
  if(is.logical(nmat)){
    check_logical(nmat, scalar = TRUE)
    if(nmat){
      attr(res, "auto") = TRUE
      attr(res, "num") = TRUE
      return(res)
    }    
  } else if(is_num_complex(nmat)){
    check_numeric(nmat, scalar = TRUE, complex = TRUE)
    attr(res, "num") = TRUE
    attr(res, "row_col") = c(Re(nmat), Im(nmat))
    return(res)
  } else {
    stop_hook("The argument `.cmat` must be a logical scalar, an integer (giving the ",
              "nber of rows) or an imaginary (ex: `3i`, giving the number of columns).",
              "\nPROBLEM: instead it is of class {enum.bq ? class(split)}.")
  }
  
  if(is.logical(df)){
    check_logical(df, scalar = TRUE)
    if(df){
      attr(res, "auto") = TRUE
      attr(res, "df") = TRUE
      return(res)
    }    
  } else if(is_num_complex(df)){
    check_numeric(df, scalar = TRUE, complex = TRUE)
    attr(res, "df") = TRUE
    attr(res, "row_col") = c(Re(df), Im(df))
    return(res)
  } else if(is.character(df)){
    if(length(df) == 1){
      colnames = strsplit(df, ",[ \t\n]*")[[1]]
    } else {
      colnames = df
    }
    
    attr(res, "df") = TRUE
    attr(res, "colnames") = colnames
    attr(res, "row_col") = c(0, length(colnames))
    return(res)
  } else {
    stop_hook("The argument `.df` must be a logical scalar, an integer (giving the ",
              "nber of rows), an imaginary (ex: `3i`, giving the number of columns), ",
              "or a character scalar giving the (comma separated) column names.",
              "\nPROBLEM: instead it is of class {enum.bq ? class(split)}.")
  }
  
  res = FALSE
  res
}

check_delimiters = function(.delim){
  
  if(!length(.delim) == 2){
    stop_hook("The argument .`delim` must lead to a character vector of length 2",
              " (after splitting the space if relevant).",
              "\nPROBLEM: it is of length {len?.delim}.")
  }
  
  if(any(nchar(.delim) == 0)){
    info = c("opening", "closing")[nchar(.delim) == 0]
    stop_hook("Argument `.delim` must be composed of non-empty delimiters.",
          "\nPROBLEM: the {C?info} delimiter{$s, are} equal to the empty string.")
  }
  
  forbid = "();?! "
  forbid_regex = paste0("[", forbid, "]")
  if(any(grepl(forbid_regex, .delim))){
    i = which(grepl(forbid_regex, .delim))[1]
    pblm = string_ops(.delim[i], paste0("'", forbid_regex, "'X"))
    info = string_magic("{&i == 1;opening;closing} delimiter (equal to {bq?.delim[i]})")
    stop_hook("Argument `.delim` cannot contain the following, reserved, ",
              "characters: {''s, bq, C?forbid}.",
              "\nPROBLEM: the {info} contains the forbidden character{$s ? pblm}: {$enum.bq}.")
  }
}

get_string_magic_context = function(){
  # this is specific to functions running within string_magic_internal
  
  up = 1
  while(!get0("is_string_magic_internal", parent.frame(up), inherits = FALSE, ifnotfound = FALSE)){
    up = up + 1
    if(identical(parent.frame(up), .GlobalEnv)){
      return("")
    }
  }
  
  x = get("x", parent.frame(up))
  .delim = get(".delim", parent.frame(up))
  
  i = get("i", parent.frame(up))
  n = get("n", parent.frame(up))
  
  # we reparse
  x_parsed = cpp_string_magic_parser(x, .delim, TRUE)
  context_specific = x_parsed[[i]]
  
  # normalizing newlines or very confusing
  x = escape_newline(x)
  context_specific = escape_newline(context_specific)
  
  # normalizing the if-else semi-colon
  x = gsub("_;;;_", ";", x)
  context_specific = gsub("_;;;_", ";", context_specific)
  
  same_context = gsub(" ", "", x) == gsub(" ", "", context_specific)
  
  if(same_context){
    res = .sma("CONTEXT: Problem found in {bq?context_specific}.")
  } else {
    res = .sma("CONTEXT: Problem found in {'80|..'k, Q?x},",
               "\n         when dealing with the interpolation {bq?context_specific}.")
  }
  
  res    
}

check_set_string_magic_parsing = function(x, check, .delim){
  # x is a character string to be turned into an R expression
  
  if(check){
    x_call = try(parse(text = x, keep.source = FALSE), silent = TRUE)
  } else {
    x_call = parse(text = x, keep.source = FALSE)
    return(x_call)
  }  
  
  if(inherits(x_call, "try-error")){
    
    open = .delim[1]
    close = .delim[2]
    
    context = get_string_magic_context()
    first_char = substr(x, 1, 1)
    
    if(first_char %in% c("#", "$")){
      msg = string_magic("PROBLEM: there is a syntax error in the pluralization (the ", 
                  "interpolation starting with {bq?first_char}).",
                  "\nFor more information on the syntax, type `string_magic(.help = TRUE)` and go to the section ",
                  "{Q!Interpolation and string operations: Principle}")
      
    } else if(first_char == "&"){
      msg = string_magic("PROBLEM: there is a syntax error in the if-else (the ", 
                  "interpolation starting with {'^&+'x, bq?x}).",
                  "\nFor more information on the syntax, type `string_magic(.help = TRUE)` and go to the section ",
                  "{Q!Special interpolation: if-else}")
    } else {
      if(grepl("[!?]", x)){
        type = paste0("2 (`", open, "ops ? expr", close, "`) or 3 (`", open, "ops ! verbatim", close, "`)")
        if(!grepl("!", x)){
          type = paste0("2 (`", open, "ops ? expr", close, "`)")
        } else if(!grepl("[?]", x)){
          type = paste0("3 (`", open, "ops ! verbatim", close, "`)")
        }
        
        extra = string_magic("\nCurrently the interpolation has been parsed as the first case: ",
                     "`{open}expr{close}`.",
                     "\nMaybe there is a syntax mistake preventing it to be interpreted as case {type}?")
      } else {
        extra = "\nCurrently it seems that the expression to be interpolated is not a valid R expression."
      }
      
      help_suggest = .sma("\nFor more information on the syntax, type `string_magic(.help = TRUE)` and go to the section ",
                          "{!Operations: General syntax}")
      
      x_call_clean = gsub("Error in parse[^:]+: ?", "Error when parsing: ", x_call)
      
      msg = .sma("PROBLEM: The expression {bq?x} could not be parsed, see error below:",
                 "\n{x_call_clean}",
                 "\nINFO: Interpolations can be of the form `{open}expr{close}`, ",
                 "`{open}op1, op2?expr{close}`, or", 
                 " `{open}op1, op2!verbatim{close}`. ",
                 "{extra}", 
                 "{help_suggest}")
    }
    
    .stop_hook(context, "\n", msg)
    
  }
  
  x_call
}

check_set_string_magic_eval = function(call, data, envir, check, is_deparse = FALSE){
  # we need to eval the call
  
  is_dt = "dt_data" %in% names(attributes(envir))
  
  if(!check){
    if(is_dt){
      x = eval_dt(call, data, envir)
    } else {
      x = eval(call, data, envir)
    }
    
  } else {
    # now with check = TRUE
    
    if(is_dt){
      x = tryCatch(eval_dt(call, data, envir), silent = TRUE, 
                   error = function(e) structure(conditionMessage(e), 
                                                 class = "stringmagic-error"))
    } else {
      x = tryCatch(eval(call, data, envir), silent = TRUE,
                   error = function(e) structure(conditionMessage(e), 
                                                 class = "stringmagic-error"))
    }
    
    if(inherits(x, "stringmagic-error")){
      context = get_string_magic_context()
      call_dp = deparse_short(call)
      if(grep("^expression\\(", call_dp)){
        call_dp = gsub("^expression\\(|\\)", "", call_dp)
      }
      x_clean = gsub("^Error in eva[^:]+: ?", "Error: ", x)
      
      msg = .sma("PROBLEM: The expression {bq ? call_dp} could not be evaluated, see error below:",
                "\n{x_clean}")
      
      stop_hook("{context}\n{msg}")
    }
  }  
  
  if(is.function(x) && !is_deparse){
    context = get_string_magic_context()
    call_dp = deparse_short(call)
    msg = .sma("EXPECTATION: interpolated variables should be coercible to character strings.",
               "\nPROBLEM: the expression {bq?call_dp} is a **function**. Please provide a character string.")
    stop_hook("{context}\n{msg}")
  }
  
  x
}


# oparg: operator's argument
check_set_oparg_parse = function(argument, operator, check){
  
  if(check){
    call = try(str2lang(argument), silent = TRUE)
  } else {
    call = str2lang(argument)
    return(call)
  }
  
  
  if(inherits(call, "try-error")){
    context = get_string_magic_context()
    msg = .sma("EXPECTATION: In operation \"{bq?argument}{operator}\" the argument ",
               "in backticks is evaluated from the calling environment.",
               "\nPROBLEM: {bq?argument} is not a valid R-expression, see parsing error below:",
               "\n{call}")
    stop_hook("{context}\n{msg}")
  }
  
  call
}

check_set_oparg_eval = function(call, data, envir, operator, check){
  
  if(check){
    x = try(eval(call, data, envir), silent = TRUE)
  } else {
    x = eval(call, data, envir)
    return(x)
  }
  
  if(inherits(x, "try-error")){
    context = get_string_magic_context()
    msg = .sma("EXPECTATION: In operation {bq?operator} the argument ",
               "in backticks is evaluated from the calling environment.",
               "\nPROBLEM: {bq?deparse_short(call)} could not be evaluated, see error below:",
               "\n{x}")
    stop_hook("{context}\n{msg}")
  }
  
  x
}

report_string_magic_parsing_error = function(x, x_parsed, .delim, error = TRUE){
  
  x_parsed = x_parsed[[1]]
  error_msg = x_parsed[1]
  xi = x_parsed[2]
  
  #
  # getting the context
  #
  
  open = .delim[1]
  close = .delim[2]
  
  x = escape_newline(x)
  xi = escape_newline(xi)
  
  same_context = gsub(" ", "", x) == gsub(" ", "", xi)
  if(same_context){
    context = .sma("CONTEXT: Problem found in {bq?xi}.")
  } else {
    context = .sma("CONTEXT: Problem found in {'80|..'k, Q?x},",
                   "\n         when dealing with the interpolation {bq?xi}.")
  }
  
  #
  # the error message
  #
  
  suggest = ""
  if(error_msg == "slash operator not ending correctly"){
    if(is_box_open(xi, .delim) && !is_box_close(xi, .delim)){
      msg = .sma("PROBLEM: the opening delimiter ({bq?.delim[1]}) is not ", 
                 "matched with a closing delimiter ({bq?.delim[2]}).",
                 "\nNOTE: to escape the meaning of the delimiter, use a ",
                 "double backslash: \\\\{.delim[1]}")
      
      suggest= "INFO: see string_magic(.help = TRUE) and go to the section 'Escaping and special cases'"
    } else {
      # we diagnose the substring
      if(is_box_open(xi, .delim)){
        xi_small = string_trim(xi, 1)
      } else {
        xi_small = string_trim(xi, nchar(open), nchar(close))
      }
      
      xi_small_parsed = cpp_string_magic_parser(xi_small, .delim)
      
      if(length(xi_small_parsed) == 1 && isTRUE(attr(xi_small_parsed, "error"))){
        msg = report_string_magic_parsing_error(xi_small_parsed, xi_small, .delim, FALSE)
        msg = gsub("^CONTEXT:", "         ", msg)
      } else {
        # here I don't knwo what the error can be
        # in theory, we should not be here
        stop("Internal error: uncaught parsing error, could you report your string_magic() call?")
      }
      
      # We don't add a suggestion since it should be in the 'msg'
    }
    
  } else if(error_msg == "no closing bracket"){
    
    # {unik, sort ? iris[['Species']}
    # we first try on iris[['Species'] only
    
    done = FALSE
    expr_msg = ""
    if(grepl("?", xi, fixed = TRUE)){
      xi_expr = gsub(".+[?]", "", xi)
      pblm = cpp_find_closing_problem(xi_expr, .delim)
      if(pblm %in% c("", "delim")){
        # no problem found or delim prblm (we check with the full xi)
      } else {
        done = TRUE
        close = .delim[2]
        xi_expr_clean = string_ops(xi_expr, "'mf/{close}'s, -1 last, `close`collapse, tws")
        expr_msg = .sma("in the expression {bq?xi_expr_clean}, ")
      }
    }
    
    if(!done){
      pblm = cpp_find_closing_problem(xi, .delim)
    }    
    
    if(pblm == "delim"){
      suggest = "INFO: see string_magic(.help = TRUE) and go to the section 'Escaping and special cases'"
    }
    
    delim_msg = if(all(.delim == c("{", "}"))) "bracket" else "delimiter"
    
    pblm = switch(pblm, 
                  "delim" = .sma("the opening {delim_msg} ({bq?.delim[1]}) is not ", 
                                  "matched with a closing {delim_msg} ({bq?.delim[2]}).",
                                  "\n\nNOTE: to escape the meaning of the {delim_msg}, use a ",
                                  "double backslash: \\\\{.delim[1]}."),
                  "'" = "a single quote (') open is not closed.",
                  "\"" = "a double quote (\") open is not closed.",
                  "`" = "a backtick (`) open is not closed.",
                  "(" = "a parenthesis (`(`) open is not closed.",
                  "{" = "a bracket (`{`) open is not closed.",
                  "[" = "a bracket (`[`) open is not closed.",
                  "it seems that something open is not closed."
                  )
    msg = .sma("PROBLEM: {expr_msg}{pblm}")  
    
  } else if(string_x(error_msg, 7) == "if-else"){
    type = if(string_x(xi, 2, 1 + nchar(open)) == "&&") "&&" else "&"
    syntax = if(type == "&&") "{&&cond ; verb_true}" else "{&cond ; verb_true ; verb_false}"
    intro = .sma("The syntax of the if-else operation is {syntax}, with ", 
                 "`verb_.` a verbatim expression (possibly containing interpolations).")
    
    if(error_msg == "if-else: error extracting condition"){
      
      msg = .sma("{intro}\nPROBLEM: no semi-colon (used to delimit the condition) ",
                 "was found so the condition could not be parsed.")
                 
    } else if(error_msg == "if-else: error extracting the first part"){
      msg = .sma("{intro}\nPROBLEM: no semi-colon {&type == '&' ; or closing bracket }was ",
                 "found to delimit `verb_true`.")
    } else if(error_msg == "if-else: error extracting the second part"){
      msg = .sma("{intro}\nPROBLEM: no closing bracket was found to delimit `verb_false`.")
    }
    
    suggest = "INFO: see string_magic(.help = TRUE) and go to the section 'Special interpolation: if-else'"
    
  } else {
    # pluralization
    
    suggest = "INFO: see string_magic(.help = TRUE) and go to the section 'Special interpolation: Pluralization'"
    type = string_x(xi, 1, 1 + nchar(open))
    syntax = .sma("{open}{type}op1, op2{close} or {open}{type}op1, op2 ? variable{close}")
    
    if(error_msg == "pluralization: operators cannot contain parentheses"){
      msg = .sma("In pluralization, the parenthesis is a forbidden character *except* ",
                 "for the special operation `(v1;v2)` (without any letter before the paren!).",
                 "\nPROBLEM: a parenthesis was found attached to an operator.")
      
    } else if(error_msg == "pluralization: the interpolation ends incorrectly"){
      msg = .sma("The syntax of the pluralization is {syntax}, with `op` pluralization operations.", 
                 "\nPROBLEM: The pluralization did not end correctly.")
    } else {
      # (v1;v2) case
      type = string_ops(error_msg, "(first|second|third) x")
      pblm = switch(type, 
                    "first" = "The semi-colon to delimit the first part is missing.",
                    "second" = "The semi-colon or closing parenthesis to delimit the second part is missing.",
                    "third" = "The closing parenthesis to delimit the third part is missing.")
      
      msg = .sma("In pluralizations, the special verbatim operation must be of the ",
                 "form (v1;v2) or (v1;v2;v3), with `v.` verbatim expressions ", 
                 "(possibly containing interpolations).",
                 "\nPROBLEM: the {type} verbatim expression could not be extracted. {pblm}")
    }
    
  }
  
  full_msg = .sma("{context}\n{msg}{if(.C>0;'\n'paste) ? suggest}")
  
  if(error){
    .stop_hook(full_msg)
  } else {
    return(full_msg)
  }
  
}

check_set_eval_fun = function(fun__, x__, ...){
  # we postfix weirdly to avoid common argument names
  # the .x is pretty common
  
  check_function(fun__, up = 1, argname = deparse(substitute(fun__)))
  
  # NOTA: formalArgs does not work for primitives, like sum
  arg_names = names(formals(args(fun__)))
  dots = list(...)
  args = list()
  args[[1]] = x__
  if(!is.null(arg_names)){
    arg_ok = intersect(names(dots), arg_names)
    args[arg_ok] = dots[arg_ok]
  }
  
  res = try(do.call(fun__, args), silent = TRUE)
  if(isError(res)){
    fun_arg = deparse(substitute(fun__))
    stop_hook("The evaluation of the function in argument {bq?fun_arg} ", 
              "led to an error, see below:",
              "\n{'^.+do\\.call[^\n]+\n'r?res}")
  }
  
  res
}

####
#### utilities ####
####


suggest_item = function(x, items, write_msg = TRUE, newline = TRUE, info = "variable"){
  # typical use: x is not in items
  #              we want to suggest possible values
  # returns vector of length 0 if no suggestion
  
  items_origin = items

  # 1: with case
  nx = nchar(x)
  items_nx = substr(items, 1, nx)

  qui = items_nx == x
  if(length(x) == 0){
    res = character(0)
    
  } else if(any(qui)){
    res = items[qui]

  } else {
    # 2: without case

    x = tolower(x)
    items_nx = tolower(items_nx)

    qui = items_nx == x
    if(any(qui)){
      res = items[qui]
    } else {
      # 3: with spelling mistakes, keeping the order
      # only if x > 3 characters

      if(nx < 3) return(character(0))

      # lazy algorithm
      score = numeric(length(items))
      for(i in 1:nx){
        score = score + (substr(items_nx, i, i) == substr(x, i, i))
      }

      if(any(score > (nx / 2))){
        s_order = order(score, decreasing = TRUE)
        score = score[s_order]
        items = items[s_order]
        
        res = items[score > nx / 2]
        
      } else {
        # 4: with spelling mistakes, ignoring the order
        x_letters = strsplit(x, "")[[1]]
        score = numeric(length(items))
        for(i in 1:nx){
          score = score + (substr(items_nx, i, i) %in% x_letters)
        }

        s_order = order(score, decreasing = TRUE)
        score = score[s_order]
        items = items[s_order]
        
        res = items[score > (nx * 0.65)]
      }
    }
  }

  if(write_msg){
    if(length(res) == 0){
      if(length(items_origin) <= 5){
        res = string_magic("FYI the {info}{$s, are, enum.bq ? items_origin}.")
      } else {
        res = string_magic("\nFYI the {info}s are: {sort, ', 'c ? items_origin}.")
      }
    } else {
      res = string_magic("Maybe you meant {enum.bq.or ? res}?")
    }

    if(newline){
      res = paste0("\n", res)
    }
  } else {
    if(length(res) == 0){
      res = head(items_origin, 5)
    }
  }

  res
}

opt_equal = function(x, option){
  any(!is.na(pmatch(x, option)))
}

deparse_short = function(x){
  x_dp = deparse(x)
  if(length(x_dp) > 1){
    x_dp = paste0(x_dp, "...")
  }

  x_dp
}

deparse_long = function(x){
  x_dp = deparse(x, width.cutoff = 500L)
  if(length(x_dp) > 1){
    x_dp = paste0(x_dp, collapse = "\n")
  }

  x_dp
}


####
#### dreamerr's copies ####
####


setDreamerr_show_stack = function(show_full_stack = FALSE){

  check_logical(show_full_stack, scalar = TRUE, mbt = TRUE)

  options("dreamerr_show_full_stack" = show_full_stack)

}

set_up = function(.up = 1){
  if(length(.up) == 1 && is.numeric(.up) && !is.na(.up) && .up == floor(.up) && .up >= 0){
    assign("DREAMERR__UP", .up, parent.frame())
  } else {
    stop("Argument '.up' must be an integer scalar greater or equal to 1. ",
         "This is currently not the case.")
  }
}

set_pblm_hook = function(){
  assign("string_magic_HOOK", 1, parent.frame())
}

get_up_hook = function(){
  # up with set_up
  f = parent.frame()
  up = 1
  while(!identical(f, .GlobalEnv)){
    if(exists("string_magic_HOOK", f)){
      break
    }
    up = up + 1
    f = parent.frame(up + 1)
  }
  
  # we accept direct nestedness
  f_up = parent.frame(up + 2)
  while(!identical(f_up, .GlobalEnv) && exists("string_magic_HOOK", f_up)){
    up = up + 1
    f = f_up
    f_up = parent.frame(up + 2)
  }

  if(identical(f, .GlobalEnv)){
    up = 1
  }

  up
}

.stop_hook = function(..., msg = NULL, envir = parent.frame()){
  # verbatim version
  up = get_up_hook()

  stop_up(..., up = up, msg = msg, envir = envir, verbatim = TRUE)
}

stopi = function(..., envir = parent.frame()){
  stop_up(..., up = 1, envir = envir, verbatim = FALSE)
}

stop_hook = function(..., msg = NULL, envir = parent.frame(), verbatim = FALSE){
  up = get_up_hook()

  stop_up(..., up = up, msg = msg, envir = envir, verbatim = verbatim)
}

stop_up = function(..., up = 1, msg = NULL, envir = parent.frame(), verbatim = FALSE){

  if(verbatim){
    main_msg = paste0(...)
  } else {
    main_msg = .string_magic(..., .envir = envir)
  }

  # up with set_up
  mc = match.call()
  if(!"up" %in% names(mc)){
    up_value = mget("DREAMERR__UP", parent.frame(), ifnotfound = 1)
    up = up_value[[1]]
  }
  
  up = up + 1

  show_full_stack = isTRUE(getOption("dreamerr_show_full_stack"))

  sc = sys.calls()
  
  if(show_full_stack){
    # The user requests the full stack
    my_call = sapply(sc, function(x) deparse(x, width.cutoff = 200L, nlines = 1))
    my_call = string_magic("{'\n'c ! [{format.0 ? 1:length(my_call)}] {'100|...'k ? my_call}}")

    intro = paste0("the full stack is shown (set this off with setDreamerr_show_stack(FALSE))\n", my_call)

  } else {
    # only the original call
    my_call = sys.call(sys.parent(up))
    my_call = deparse(my_call)[1]
    nmax = 50
    if(nchar(my_call) > nmax) my_call = paste0(substr(my_call, 1, nmax - 1), "...")

    intro = paste0("in ", my_call)
  }  

  main_msg = fit_screen(main_msg)

  if(!is.null(msg)){
    if(length(msg) > 1){
      msg = paste(msg, collapse = "")
    }
    msg = fit_screen(msg)
    on.exit(message(msg))
  }

  stop(intro, ": \n", main_msg, call. = FALSE)

}

warn_no_named_dots = function(dots, extra_args = NULL, extra_funName = ""){
  if(!is.null(names(dots))){
    is_extra = !is.null(extra_args)
    args_fun = names(formals(sys.function(sys.parent())))
    args_fun = unique(c(args_fun, extra_args))
    args_fun = setdiff(args_fun, "...")

    dot_names = names(dots)
    dot_names = dot_names[nchar(dot_names) > 0]
    sugg_txt = suggest_item(dot_names[1], args_fun, info = "argument")

    warn_up("The arguments in `...` shall not have names{&is_extra; or may refer to the ",
                      "arguments of the function in {bq?extra_funName}}. ",
            "The name{$s, enum.bq ? dot_names}",
            " may refer to {$(an;some)} argument{$s}?", sugg_txt)

  }
}

.warn_hook = function(..., envir = parent.frame(), immediate. = FALSE){
  up = get_up_hook()

  warn_up(..., up = up, envir = envir, verbatim = TRUE, immediate. = immediate.)
}

warn_hook = function(..., envir = parent.frame(), immediate. = FALSE, verbatim = FALSE){
  up = get_up_hook()

  warn_up(..., up = up, envir = envir, verbatim = verbatim, immediate. = immediate.)
}

warni = function(..., envir = parent.frame(), immediate. = FALSE){
  warn_up(..., up = 1, envir = envir, verbatim = FALSE, immediate. = immediate.)
}

warn_up = function (..., up = 1, immediate. = FALSE, envir = parent.frame(), verbatim = FALSE){

  if(verbatim){
    message = paste0(...)
  } else {
    message = .string_magic(..., .envir = envir)
  }
  
  mc = match.call()

  if (!"up" %in% names(mc)) {
      up_value = mget("DREAMERR__UP", parent.frame(), ifnotfound = 1)
      up = up_value[[1]]
  }
  
  up = up + 1
  
  my_call = sys.call(sys.parent(up))
  my_call = deparse(my_call)[1]
  
  nmax = 50
  if (nchar(my_call) > nmax){
    my_call = paste0(substr(my_call, 1, nmax - 1), "...")
  }
      
  warning("In ", my_call, ":\n ", fit_screen(message),
          call. = FALSE, immediate. = immediate.)
}


fit_screen = function(msg, width = NULL, leading_ws = TRUE, leader = ""){
  # makes a message fit the current screen, by cutting the text at the appropriate location
  # msg must be a character string of length 1
  
  if(length(msg) == 0) return(msg)

  # Note that \t are NOT handled
  
  # eval
  width = check_set_width(width)

  N_LEAD = nchar(leader)

  if(width > 1){
    MAX_WIDTH = width
  } else {
    MAX_WIDTH = getOption("width") * width
  }

  MAX_WIDTH = max(MAX_WIDTH, 15)

  res = c()

  msg_split = cpp_split_newlines(msg)

  for(m in msg_split){
    if(nchar(m) <= MAX_WIDTH){
      res = c(res, paste0(leader, m))
    } else {
      # we apply a splitting algorithm

      lead_ws = gsub("^([ \t]*).*", "\\1", m, perl = TRUE)
      m = trimws(m)
      N_LEAD_WS = nchar(lead_ws)
      add_lead = TRUE
      first = TRUE

      m_split = strsplit(m, "(?<=[^ ]) ", perl = TRUE)[[1]]

      while(TRUE){

        if(add_lead){
          width = MAX_WIDTH - N_LEAD_WS - N_LEAD
          prefix = paste0(leader, lead_ws)
        } else {
          width = MAX_WIDTH - N_LEAD
          prefix = leader
        }

        if(sum(nchar(m_split) + 1) - 1 <= width){
          res = c(res, paste0(prefix, paste(m_split, collapse = " ")))
          break
        }

        where2split = which.max(cumsum(nchar(m_split) + 1) - 1 > width) - 1
        res = c(res, paste0(prefix, paste(m_split[1:where2split], collapse = " ")))
        m_split = m_split[-(1:where2split)]

        if(!leading_ws && first){
          add_lead = FALSE
          first = FALSE
        }

      }
    }
  }

  paste(res, collapse = "\n")
}

isError = function(x){
  inherits(x, "try-error")
}

####
#### fixest copies ####
####


check_expr = function(expr, ..., clean, up = 0, arg_name, verbatim = FALSE){

  res = tryCatch(expr, error = function(e) structure(list(conditionCall(e),
                                                          conditionMessage(e)), 
                                                     class = "stringmagic-error"))
  if(inherits(res, "stringmagic-error")){

    if(missing(up)){
      # up with set_up
      f = parent.frame()
      up = 1
      while(!identical(f, .GlobalEnv)){
        if(exists("string_magic_HOOK", f)){
          break
        }
        up = up + 1
        f = parent.frame(up + 1)
      }

      if(identical(f, .GlobalEnv)){
        up = 1
      }
    }
    
    up = up + 1 
    
    if(verbatim){
      msg = paste0(..., collapse = "")
    } else {
      msg = .string_magic(..., .envir = parent.frame())
    }
    
    if(nchar(msg) == 0){
      if(missing(arg_name)){
        arg_name = deparse(substitute(expr))
      }
      msg = paste0("Argument '", arg_name, "' could not be evaluated: ")
      stop_up(msg, res[[2]], verbatim = verbatim, envir = parent.frame(), up = up)
      
    } else {
      call_non_informative = deparse(substitute(expr),100)[1]

      call_error = deparse(res[[1]], 100)[1]

      if(call_error == call_non_informative || call_error == "NULL" || grepl("^(doTry|eval)", call_error)){
        call_error = ""
      } else {
        call_error = paste0("In ", call_error, ": ")
      }

      err = res[[2]]
      if(grepl("^in eval\\(str[^:]+:\n", err)){
        err = sub("^in eval\\(str[^:]+:\n", "", err)
      }
            
      # cleaning ugly call repetition
      current_call = deparse(sys.call(sys.parent(up)))[1]
      first_err = gsub("\n.+", "", err)
      if(grepl(substr(current_call, 1, 20), first_err, fixed = TRUE)){
        err = gsub("^[^\n]+\n", "", err)
      }
      
      # cleaning extra artifacts
      i_clean = 0
      msg_split = strsplit(msg, "\n")[[1]]
      err_split = strsplit(err, "\n")[[1]]
      while(i_clean < min(length(msg_split), length(err_split))){
        if(string_x(msg_split[i_clean + 1], 15) == string_x(err_split[i_clean + 1], 15)){
          i_clean = i_clean + 1
        } else {
          break
        }
      }
      
      for(i in seq_len(i_clean)){
        err = gsub("^[^\n]+\n", "", err)
      }
      
      if(!missing(clean)){
        err = string_clean(err, clean)
      }
      stop_up(msg, "\n", call_error, err, verbatim = TRUE, up = up)
    }
  }
  res
}

#----------------------------------------------#
# Author: Laurent Berge
# Date creation: Mon Aug  8 16:26:42 2022
# ~: string tools
#----------------------------------------------#


#' Chains basic operations to character vectors
#'
#' Simple tool to perform multiple operations to character vectors.
#' 
#' @inheritParams string_clean
#'
#' @param x A character vector. If not a character vector but atomistic (i.e. not a list), 
#' it will be converted to a character vector.
#' @param ... Character **scalars**. Character scalar containing the comma separated values 
#' of operations to perform to the vector. The 50+ operations are detailed in the help
#' page of [string_magic()].
#' @param op Character **vector** or `NULL` (default). Character scalar containing the comma separated values 
#' of operations to perform to the vector. The 50+ operations are detailed in the help
#' page of [string_magic()]. Note that if this argument is provided, then the values in 
#' `...` are ignored.
#' @param pre_unik Logical scalar, default is `NULL`. Whether to first unique the vector 
#' before applying the possibly costly string operations, and merging back the result. 
#' For very large vectors with repeated values the time gained can be substantial. By 
#' default, this is `TRUE` for vector of length 1M or more.
#' 
#' @details 
#' This function is a simple wrapper around string_magic. Formally, `string_ops(x, "op1, op2")`
#' is equivalent to `string_magic("{op1, op2 ? x}")`.
#'
#' @return
#' In general it returns a character vector. It may be of a length different from the original
#'  one, depending on the operations performed. 
#'
#' @author
#' Laurent R. Berge
#' 
#' @inherit string_clean seealso
#' 
#' @family tools with aliases
#'
#' @examples
#' 
#' # data on car models
#' cars = row.names(mtcars)
#' 
#' # let's get the brands starting with an "m"
#' string_ops(cars, "'i/^m'get, x, unik")
#' 
#' # Explainer:
#' # 'i/^m'get: keeps only the elements starting with an m,
#' #            i/ is the 'regex-flag' "ignore" to ignore the case
#' #            ^m means "starts with an m" in regex language
#' # x: extracts the first pattern. The default pattern is "[[:alnum:]]+"
#' #    which means an alpha-numeric word
#' # unik: applies unique() to the vector
#' # => see help in ?string_magic for more details on the operations
#' 
#' 
#' # let's get the 3 largest numbers appearing in the car models
#' string_ops(cars, "'\\d+'x, rm, unik, num, dsort, 3 first")
#' 
#' # Explainer:
#' # '\\d+'x: extracts the first pattern, the pattern meaning "a succession"
#' #          of digits in regex language
#' # rm: removes elements equal to the empty string (default behavior)
#' # unik: applies unique() to the vector
#' # num: converts to numeric
#' # dsort: sorts in decreasing order
#' # 3 first: keeps only the first three elements
#' 
#' # You can use several character vectors as operations:
#' string_ops(cars, 
#'            "'\\d+'x, rm, unik",
#'            "num, dsort, 3 first")
#' 
string_ops = function(x, ..., op = NULL, pre_unik = NULL, namespace = NULL, envir = parent.frame()){

  if(missing(x)){
    stop("Argument `x` must be provided. PROBLEM: it is currently missing.")
  }
  
  if(is.null(x) || length(x) == 0){
    return(character(0))
  }

  if(!is.atomic(x)){
    stop("Argument `x` must be atomic. Currently it is of class `", class(x)[1], "`")
  }
  
  if(length(x) == 0){
    return(x)
  }

  if(!true_character(x)){
    x = as.character(x)
  }
  
  set_pblm_hook()
  
  check_character(op, null = TRUE, no_na = TRUE)
  if(!is.null(op)){
    all_ops = op
  } else {
    all_ops = check_set_dots(..., mbt = TRUE, scalar = TRUE, character = TRUE)
    all_ops = unlist(all_ops)
  }
  
  check_logical(pre_unik, null = TRUE, scalar = TRUE)

  # For very large vectors, we unique
  n = length(x)
  if(is.null(pre_unik)){
    pre_unik = n > 1e6
  }

  if(pre_unik){
    x_int = to_index(x)
    x_small = x[!duplicated(x_int)]

    res_small = string_ops(x_small, op = all_ops, pre_unik = FALSE, 
                           namespace = namespace, envir = envir)
    res = res_small[x_int]
  } else {
    
    if(is.null(namespace)){
      namespace = "R_GlobalEnv"
    }
    
    user_ops_all = getOption("string_magic_user_ops")
    # beware the sneaky assignment!
    if(!is.null(user_ops_all) && !is.null(user_info <- user_ops_all[[namespace]])){
      .user_funs = user_info$funs
      .valid_operators = user_info$operators          
    } else {
      .user_funs = NULL
      .valid_operators = getOption("string_magic_operations_default")
    }
    
    res = x
    for(op in all_ops){
      group_flag = 1 * grepl("~", op, fixed = TRUE)
      res = apply_simple_operations(res, "string_ops", op, .check = TRUE, .envir = envir,
                                    .data = list(), group_flag = group_flag, 
                                    .delim = c("{", "}"), .user_funs = .user_funs, 
                                    .valid_operators = .valid_operators)  
    }
    
  }

  if("group_index" %in% names(attributes(res))){
    attr(res, "group_index") = NULL
  }

  res
}





#' Detects whether a pattern is in a character string
#'
#' Function that detects if one or more patterns are in a string. The patterns can be 
#' chained, by default this is a regex search but special flags be triggered with a 
#' specific syntax, supports negation.
#'
#' @param x A character vector.
#' @param ... Character scalars representing the patterns to be found. By default they are (perl) regular-expressions.
#' Use ' & ' or ' | ' to chain patterns and combine their result logically (ex: `'[[:alpha:]] & \\d'` gets strings
#' containing both letters and numbers). You can negate by adding a `!` first (ex: `"!sepal$"` will 
#' return `TRUE` for strings that do not end with `"sepal"`).
#' Add flags with the syntax 'flag1, flag2/pattern'. Available flags are: 'fixed', 'ignore', 'word' and 'magic'.
#' Ex: "ignore/sepal" would get "Sepal.Length" (wouldn't be the case w/t 'ignore'). 
#' Shortcut: use the first letters of the flags. Ex: "if/dt[" would get `"DT[i = 5]"` (flags 'ignore' + 'fixed').
#' For 'word', it adds word boundaries to the pattern. The `magic` flag first interpolates
#' values directly into the pattern with "{}".
#' @param or Logical, default is `FALSE`. In the presence of two or more patterns, 
#' whether to combine them with a logical "or" (the default is to combine them with a logical "and").
#' @param pattern (If provided, elements of `...` are ignored.) A character vector representing the 
#' patterns to be found. By default a (perl) regular-expression search is triggered. 
#' Use ' & ' or ' | ' to chain patterns and combine their result logically (ex: `'[[:alpha:]] & \\d'` gets strings
#' containing both letters and numbers). You can negate by adding a `!` first (ex: `"!sepal$"` will 
#' return `TRUE` for strings that do not end with `"sepal"`).
#' Add flags with the syntax 'flag1, flag2/pattern'. Available flags are: 'fixed', 'ignore', 'word' and 'magic'.
#' Ex: "ignore/sepal" would get "Sepal.Length" (wouldn't be the case w/t 'ignore'). 
#' Shortcut: use the first letters of the flags. Ex: "if/dt[" would get `"DT[i = 5]"` (flags 'ignore' + 'fixed').
#' For 'word', it adds word boundaries to the pattern. The `magic` flag first interpolates
#' values directly into the pattern with "{}".
#' @param fixed Logical scalar, default is `FALSE`. Whether to trigger a fixed search instead of a 
#' regular expression search (default).
#' @param word Logical scalar, default is `FALSE`. If `TRUE` then a) word boundaries are added to the pattern, 
#' and b) patterns can be chained by separating them with a comma, they are combined with an OR logical operation.
#' Example: if `word = TRUE`, then pattern = "The, mountain" will select strings containing either the word
#' 'The' or the word 'mountain'.
#' @param ignore.case Logical scalar, default is `FALSE`. If `TRUE`, then case insensitive search is triggered. 
#' @param last A function or `NULL` (default). If a function, it will be applied to the vector 
#' just before returning it.
#' @param envir Environment in which to evaluate the interpolations if the flag `"magic"` is provided.
#' Default is `parent.frame()`.
#' 
#' @details
#' The internal function used to find the patterns is [base::grepl()] with `perl = TRUE`.
#' 
#' @section Generic regular expression flags:
#' 
#' All `stringmagic` functions support generic flags in regular-expression patterns. 
#' The flags are useful to quickly give extra instructions, similarly to *usual* 
#' [regular expression flags](https://javascript.info/regexp-introduction).
#' 
#' Here the syntax is "flag1, flag2/pattern". That is: flags are a comma separated list of flag-names 
#' separated from the pattern with a slash (`/`). Example: `string_which(c("hello...", "world"), "fixed/.")` returns `1`. 
#' Here the flag "fixed" removes the regular expression meaning of "." which would have otherwise meant *"any character"*.
#' The no-flag verion `string_which(c("hello...", "world"), ".")` returns `1:2`.
#' 
#' Alternatively, and this is recommended, you can collate the initials of the flags instead of using a
#' comma separated list. For example: "if/dt[" will apply the flags "ignore" and "fixed" to the pattern "dt[".
#' 
#' The four flags always available are: "ignore", "fixed", "word" and "magic". 
#' 
#' + "ignore" instructs to ignore the case. Technically, it adds the perl-flag "(?i)" 
#' at the beginning of the pattern.
#' 
#' + "fixed" removes the regular expression interpretation, so that the characters ".", "$", "^", "[" 
#' (among others) lose their special meaning and are treated for what they are: simple characters. 
#' 
#' + "word" adds word boundaries (`"\\b"` in regex language) to the pattern. Further, the comma (`","`) 
#' becomes a word separator. Technically, "word/one, two" is treated as "\\b(one|two)\\b". Example: 
#' `string_clean("Am I ambushed?", "wi/am")` leads to " I ambushed?" thanks to the flags "ignore" and "word".
#'
#' + "magic" allows to interpolate variables inside the pattern before regex interpretation. 
#' For example if `letters = "aiou"` then `string_clean("My great goose!", "magic/[{letters}] => e")` 
#' leads to `"My greet geese!"`
#'
#' @return
#' It returns a logical vector of the same length as `x`.
#' 
#' The function `string_which` returns a numeric vector. 
#' 
#' @author 
#' Laurent R. Berge
#' 
#' @inherit string_clean seealso
#'
#' @examples
#' 
#' # NOTA: using `string_get` instead of `string_is` may lead to a faster understanding 
#' #       of the examples 
#'
#' x = string_vec("One, two, one... two, microphone, check")
#'
#' # default is regular expression search
#' # => 3 character items
#' string_is(x, "^...$")
#'
#' # to trigger fixed search use the flag 'fixed'
#' string_is(x, "fixed/...")
#' # you can just use the first letter
#' string_is(x, "f/...")
#'
#' # to negate, use '!' as the first element of the pattern
#' string_is(x, "f/!...")
#'
#' # you can combine several patterns with " & " or " | "
#' string_is(x, "one & c")
#' string_is(x, "one | c")
#' 
#' #
#' # word: adds word boundaries
#' #
#' 
#' # compare
#' string_is(x, "one")
#' # with
#' string_is(x, "w/one")
#' 
#' # words can be chained with commas (it is like an OR logical operation)
#' string_is(x, "w/one, two")
#' # compare with
#' string_is(x, "w/one & two")
#' # remember that you can still negate
#' string_is(x, "w/one & !two")
#' 
#' # you can combine the flags
#' # compare
#' string_is(x, "w/one")
#' # with
#' string_is(x, "wi/one")
#' 
#' #
#' # the `magic` flag
#' #
#' 
#' p = "one"
#' string_is(x, "m/{p}")
#' # Explanation:
#' # - "p" is interpolated into "one"
#' # - we get the equivalent: string_is(x, "one")
#' 
#'
#' #
#' # string_which
#' #
#'
#' # it works exactly the same way as string_is
#' # Which are the items containing an 'e' and an 'o'?
#' string_which(x, "e", "o")
#' # equivalently
#' string_which(x, "e & o")
#'
#'
string_is = function(x, ..., fixed = FALSE, ignore.case = FALSE, word = FALSE, 
                     or = FALSE, pattern = NULL, envir = parent.frame(), last = NULL){

  x = check_set_character(x, mbt = TRUE, l0 = TRUE)
  if(length(x) == 0){
    return(logical(0))
  }
  
  check_logical(ignore.case, scalar = TRUE)
  check_logical(fixed, scalar = TRUE)
  check_logical(word, scalar = TRUE)
  check_logical(or, scalar = TRUE)
  check_character(pattern, null = TRUE, no_na = TRUE)
  check_function(last, null = TRUE)

  if(missnull(pattern)){
    dots = check_set_dots(..., mbt = TRUE, character = TRUE, scalar = TRUE, no_na = TRUE)
    if(!is.null(names(dots))){
      if(is.null(last)){
        warn_no_named_dots(dots)
      } else {
        # we remove the arguments that go to "last"
        dot_names = names(dots)
        args_last = names(formals(args(last)))
        if(!is.null(args_last)){
          args_ok = intersect(dot_names, args_last)
          dots[args_ok] = NULL
          if(sum(nchar(dot_names) > 0) != length(args_ok)){
            warn_no_named_dots(dots, extra_args = args_last, extra_funName = "last")
          }
        }        
      }
    }
    
    pattern = unlist(dots)
  }

  or_origin = or
  negate = FALSE
  logical_op = function(a, b, or) if(or) a | b else a & b

  res = NULL
  for(i in seq_along(pattern)){
    pat = pattern[i]
    first_char = substr(pat, 1, 1)
    
    pat_parsed = parse_regex_pattern(pat, c("fixed", "word", "ignore", "magic"), 
                                     envir = envir)
    is_or = pat_parsed$is_or
    flags = pat_parsed$flags
    all_patterns = pat_parsed$patterns
    negate_all = pat_parsed$is_not

    is_fixed_origin = fixed || "fixed" %in% flags
    is_word = word || "word" %in% flags
    is_ignore = ignore.case || "ignore" %in% flags
    
    res_current = NULL
    n_pat = length(all_patterns)
    for(j in 1:n_pat){
      is_fixed = is_fixed_origin
      p = all_patterns[j]
      sub_negate = negate_all[j]
      
      p = format_pattern(p, fixed = is_fixed, word = is_word, ignore = is_ignore) 
      is_fixed = attr(p, "fixed")
      
      res_tmp = tryCatch(grepl(p, x, perl = !is_fixed, fixed = is_fixed), 
                         error = function(e) structure(conditionMessage(e), class = "try-error"),
                         warning = function(w) structure(conditionMessage(w), class = "try-warning"))
        
      is_warn = inherits(res_tmp, "try-warning")
      warn_msg = ""
      if(is_warn){
        # is there an error?
        warn_msg = res_tmp
        res_tmp = tryCatch(suppressWarnings(grepl(p, x, perl = !is_fixed, fixed = is_fixed)), 
                           error = function(e) structure(conditionMessage(e), class = "try-error"))
      }
      
      if(inherits(res_tmp, "try-error")){
        pat_raw = pattern[i]
        stopi("CONTEXT: {&p != pat_raw;evaluation of {Q?pat_raw}\n         }",
              "pattern = {Q?p} ",
              "\nEXPECTATION: the pattern must be a valid regular expression",
              "\nPROBLEM: `grepl` led to an error, see below:",
              "\n{res_tmp}",
              "{&nchar(warn_msg) > 0;\n{.}}")
      } else if(is_warn){
        pat_raw = pattern[i]
        warni("CONTEXT: {&p != pat_raw;evaluation of {Q?pat_raw}\n         }",
              "pattern = {Q?p} ",
              "\nA warning was raised when evaluating the pattern:",
              "\n{warn_msg}", immediate. = TRUE)
      }
      
      if(sub_negate){
        res_tmp = !res_tmp
      }
      if(is.null(res_current)){
        res_current = res_tmp
      } else {
        res_current = logical_op(res_current, res_tmp, is_or[j])
      }
    }

    if(is.null(res)){
      res = res_current
    } else {
      res = logical_op(res, res_current, or)
    }
  }
  
  if(!is.null(last)){
    res = check_set_eval_fun(last, res, ...)
  }

  res
}

#' @describeIn string_is Detects if at least one element of a vector matches a regex pattern
string_any = function(x, ..., fixed = FALSE, ignore.case = FALSE, word = FALSE, 
                     or = FALSE, pattern = NULL, envir = parent.frame()){

  check_character(pattern, null = TRUE, no_na = TRUE)

  if(missnull(pattern)){
    dots = check_set_dots(..., mbt = TRUE, character = TRUE, scalar = TRUE, no_na = TRUE)
    warn_no_named_dots(dots)
    pattern = unlist(dots)
  }

  any(string_is(x, fixed = fixed, ignore.case = ignore.case, word = word, 
               or = or, pattern = pattern, envir = envir))
}

#' @describeIn string_is Detects if all elements of a vector match a regex pattern
string_all = function(x, ..., fixed = FALSE, ignore.case = FALSE, word = FALSE, 
                     or = FALSE, pattern = NULL, envir = parent.frame()){

  check_character(pattern, null = TRUE, no_na = TRUE)

  if(missnull(pattern)){
    dots = check_set_dots(..., mbt = TRUE, character = TRUE, scalar = TRUE, no_na = TRUE)
    warn_no_named_dots(dots)
    pattern = unlist(dots)
  }

  all(string_is(x, fixed = fixed, ignore.case = ignore.case, word = word, 
               or = or, pattern = pattern, envir = envir))
}

#' @describeIn string_is Returns the indexes of the values in which a pattern is detected
string_which = function(x, ..., fixed = FALSE, ignore.case = FALSE, word = FALSE, 
                     or = FALSE, pattern = NULL, envir = parent.frame()){

  check_character(pattern, null = TRUE, no_na = TRUE)

  if(missnull(pattern)){
    dots = check_set_dots(..., mbt = TRUE, character = TRUE, scalar = TRUE, no_na = TRUE)
    warn_no_named_dots(dots)
    pattern = unlist(dots)
  }

  which(string_is(x, fixed = fixed, ignore.case = ignore.case, word = word, 
               or = or, pattern = pattern, envir = envir))
}

#' Gets elements of a character vector
#'
#' Convenient way to get elements from a character vector.
#'
#' @inheritParams string_is
#'
#' @param x A character vector.
#' @param seq Logical, default is `FALSE`. The argument `pattern` accepts a vector of 
#' patterns which are combined with an `and` by default. If `seq = TRUE`, then it is like 
#' if `string_get` was called sequentially with its results stacked. See examples.
#' @param seq.unik Logical, default is `FALSE`. The argument `...` (or the argument `pattern`) accepts 
#' a vector of patterns which are combined with an `and` by default. If `seq.unik = TRUE`, then 
#' `string_get` is called sequentially with its results stacked, and `unique()` is 
#' applied in the end. See examples.
#' 
#' @details 
#' This function is a wrapper to [string_is()].
#' 
#' @inheritSection string_is Generic regular expression flags
#' 
#' @section Caching:
#' 
#' In an exploratory stage, it can be useful to quicky get values from a vector with the 
#' least hassle as possible. Hence `string_get` implements caching, so that users do not need
#' to repeat the value of the argument `x` in successive function calls, and can concentrate
#' only on the selection patterns.
#' 
#' Caching is a feature only available when the user calls `string_get` from the global environment.
#' If that feature were available in regular code, it would be too dangerous, likely leading to hard to debug bugs. 
#' Hence caching is disabled when used within code (i.e. inside a function or inside an 
#' automated script), and function calls without the main argument will lead to errors in such scripts.
#' 
#' @author 
#' Laurent R. Berge
#'
#' @return
#' It always return a character vector.
#' 
#' @inherit string_clean seealso
#' 
#'
#' @examples
#'
#' x = rownames(mtcars)
#'
#' # find all Mazda cars
#' string_get(x, "Mazda")
#' # same with ignore case flag
#' string_get(x, "i/mazda")
#' 
#' # all cars containing a single digit (we use the 'word' flag)
#' string_get(x, "w/\\d")
#'
#' # finds car names without numbers AND containing `u`
#' string_get(x, "!\\d", "u")
#' # equivalently
#' string_get(x, "!\\d & u")
#'
#' # Stacks all Mazda and Volvo cars. Mazda first
#' string_get(x, "Mazda", "Volvo", seq = TRUE)
#'
#' # Stacks all Mazda and Volvo cars. Volvo first
#' string_get(x, "Volvo", "Mazda", seq = TRUE)
#'
#' # let's get the first word of each car name
#' car_first = string_ops(x, "extract.first")
#' # we select car brands ending with 'a', then ending with 'i'
#' string_get(car_first, "a$", "i$", seq = TRUE)
#' # seq.unik is similar to seq but applies unique()
#' string_get(car_first, "a$", "i$", seq.unik = TRUE)
#' 
#' #
#' # flags
#' #
#' 
#' # you can combine the flags
#' x = string_magic("/One, two, one... Two!, Microphone, check")
#' # regular
#' string_get(x, "one")
#' # ignore case
#' string_get(x, "i/one")
#' # + word boundaries
#' string_get(x, "iw/one")
#' 
#' # you can escape the meaning of ! with backslashes
#' string_get(x, "\\!")
#' 
#' #
#' # Caching
#' #
#' 
#' # Caching is enabled when the function is used interactively
#' # so you don't need to repeat the argument 'x'
#' # Mostly useful at an exploratory stage
#' 
#' if(interactive() && is.null(sys.calls())){
#'    
#'    # first run, the data is cached
#'    string_get(row.names(mtcars), "i/vol")
#' 
#'    # now you don't need to specify the data
#'    string_get("i/^m & 4")
#' }
#'
#'
#'
#'
string_get = function(x, ..., fixed = FALSE, ignore.case = FALSE, word = FALSE, 
                   or = FALSE, seq = FALSE, seq.unik = FALSE, 
                   pattern = NULL, envir = parent.frame()){

  x = check_set_character(x, mbt = TRUE, l0 = TRUE)
  if(length(x) == 0){
    return(character(0))
  }
  
  # data caching in interactive mode
  is_caching = FALSE
  is_forced_caching = isTRUE(getOption("string_magic_string_get_forced_caching"))
  if(is_forced_caching || (interactive() && identical(parent.frame(), .GlobalEnv))){
    mc = match.call()
    if(is.character(mc$x) && !is.null(getOption("stringmagic_string_get_cache"))){
      is_caching = TRUE
      x_pattern = x
      x = getOption("stringmagic_string_get_cache")
    } else if(length(x) > 1){
      options(stringmagic_string_get_cache = x)
    }
  }

  check_character(pattern, null = TRUE, no_na = TRUE)
  check_logical(or, scalar = TRUE)
  check_logical(seq, scalar = TRUE)
  check_logical(seq.unik, scalar = TRUE)

  if(missnull(pattern)){
    if(is_caching){
      dots = check_set_dots(..., mbt = FALSE, character = TRUE, scalar = TRUE, no_na = TRUE)
      dots = append(dots, x_pattern, 0)
    } else {
      dots = check_set_dots(..., mbt = TRUE, character = TRUE, scalar = TRUE, no_na = TRUE)  
    }
    
    warn_no_named_dots(dots)
    pattern = unlist(dots)
  }

  if(seq.unik){
    seq = TRUE
  }

  if(seq){
    for(i in seq_along(pattern)){
      value = string_get(x, pattern = pattern[i], or = or, seq = FALSE, 
                      fixed = fixed, ignore.case = ignore.case, 
                      word = word, envir = envir)
      if(i == 1){
        res = value
      } else {
        res = c(res, value)
      }
    }

    if(seq.unik){
      res = unique(res)
    }

    return(res)
  }

  index = string_is(x, fixed = fixed, ignore.case = ignore.case, word = word, 
                 pattern = pattern, or = or, envir = envir)
  x[index]
}




#' Splits a character string wrt a pattern
#' 
#' Splits a character string with respect to pattern
#' 
#' @inheritParams string_split2df
#' 
#' @param x A character vector.
#' @param simplify Logical scalar, default is `TRUE`. If `TRUE`, then when the vector input `x`
#'  is of length 1, a character vector is returned instead of a list.
#' 
#' @inheritSection string_is Generic regular expression flags
#' 
#' @return 
#' If `simplify = TRUE` (default), the object returned is:
#' + a character vector if `x`, the vector in input, is of length 1: the character vector contains
#' the result of the split.
#' + a list of the same length as `x`. The ith element of the list is a character vector
#' containing the result of the split of the ith element of `x`.
#' 
#' If `simplify = FALSE`, the object returned is always a list.
#' 
#' @examples 
#' 
#' time = "This is the year 2024."
#' 
#' # we break the sentence
#' string_split(time, " ")
#' 
#' # simplify = FALSE leads to a list
#' string_split(time, " ", simplify = FALSE)
#' 
#' # let's break at "is"
#' string_split(time, "is")
#' 
#' # now breaking at the word "is"
#' # NOTE: we use the flag `word` (`w/`)
#' string_split(time, "w/is")
#' 
#' # same but using a pattern from a variable
#' # NOTE: we use the `magic` flag
#' pat = "is"
#' string_split(time, "mw/{pat}")
#' 
#' 
string_split = function(x, split, simplify = TRUE, fixed = FALSE,
                        ignore.case = FALSE, word = FALSE, 
                        envir = parent.frame()){
  
  x = check_set_character(x, mbt = TRUE, l0 = TRUE)

  check_logical(simplify, scalar = TRUE)
  check_logical(fixed, scalar = TRUE)
  check_logical(ignore.case, scalar = TRUE)
  check_logical(word, scalar = TRUE)
  
  pat_parsed = format_simple_regex_flags(split, ignore = ignore.case, 
                                         fixed = fixed, word = word, 
                                         magic = TRUE, envir = envir)
  split = pat_parsed$pattern
  is_fixed = pat_parsed$fixed

  x_split = strsplit(x, split, fixed = is_fixed, perl = !is_fixed)
  
  if(simplify && length(x) == 1){
    x_split = x_split[[1]]
  }
  
  x_split
}


#' Splits a character vector into a data frame
#'
#' Splits a character vector and formats the resulting substrings into a data.frame
#' 
#' @inheritParams string_is
#'
#' @param x A character vector or a two-sided formula. If a two-sided formula, then the 
#' argument `data` must be provided since the variables will be fetched in there. 
#' A formula is of the form `char_var ~ id1 + id2` where `char_var` on the left is a 
#' character variable and on the right `id1` and `id2` are identifiers which will be 
#' included in the resulting table. Alternatively, you can provide identifiers via 
#' the argument `id`.
#' @param data Optional, only used if the argument `x` is a formula. It should 
#' contain the variables of the formula.
#' @param split A character scalar. Used to split the character vectors. By default 
#' this is a regular expression. You can use flags in the pattern in the form `flag1, flag2/pattern`.
#' Available flags are `ignore` (case), `fixed` (no regex), word (add word boundaries), 
#' magic (add interpolation with `"{}"`). Example:
#' if "ignore/hello" and the text contains "Hello", it will be split at "Hello". 
#' Shortcut: use the first letters of the flags. Ex: "iw/one" will split at the word 
#' "one" (flags 'ignore' + 'word').
#' @param id Optional. A character vector or a list of vectors. If provided, the 
#' values of `id` are considered as identifiers that will be included in the resulting table.
#' @param add.pos Logical, default is `FALSE`. Whether to include the position of each split element.
#' @param id_unik Logical, default is `TRUE`. In the case identifiers are provided, 
#' whether to trigger a message if the identifiers are not unique. Indeed, if
#'  the identifiers are not unique, it is not possible to reconstruct the original texts.
#' @param fixed Logical, default is `FALSE`. Whether to consider the argument `split` 
#' as fixed (and not as a regular expression).
#' @param dt Logical, default is `FALSE`. Whether to return a `data.table`. See also the function `string_split2dt`.
#' @param ... Not currently used.
#'
#' @return
#' It returns a `data.frame` or a `data.table` which will contain: i) `obs`: the observation index, 
#' ii) `pos`: the position of the text element in the initial string (optional, via add.pos), 
#' iii) the text element, iv) the identifier(s) (optional, only if `id` was provided).
#' 
#' @inherit string_clean seealso
#' 
#' @examples
#'
#' x = c("Nor rain, wind, thunder, fire are my daughters.",
#'       "When my information changes, I alter my conclusions.")
#'
#' id = c("ws", "jmk")
#'
#' # we split at each word
#' string_split2df(x, "[[:punct:] ]+")
#'
#' # we add the 'id'
#' string_split2df(x, "[[:punct:] ]+", id = id)
#'
#' # TO NOTE:
#' # - the second argument is `data`
#' # - when it is missing, the argument `split` becomes implicitly the second
#' # - ex: above we did not use `split = "[[:punct:] ]+"`
#'
#' #
#' # using the formula
#'
#' base = data.frame(text = x, my_id = id)
#' string_split2df(text ~ my_id, base, "[[:punct:] ]+")
#'
#' #
#' # with 2+ identifiers
#'
#' base = within(mtcars, carname <- rownames(mtcars))
#'
#' # we have a message because the identifiers are not unique
#' string_split2df(carname ~ am + gear + carb, base, " +")
#'
#' # adding the position of the words & removing the message
#' string_split2df(carname ~ am + gear + carb, base, " +", id_unik = FALSE, add.pos = TRUE)
#'
#'
string_split2df = function(x, data = NULL, split = NULL, id = NULL, add.pos = FALSE,
                           id_unik = TRUE, fixed = FALSE, ignore.case = FALSE,
                           word = FALSE, envir = parent.frame(), dt = FALSE, ...){

  if(missing(x)){
    stop("Argument 'x' must be provied. PROBLEM: it is missing.")
  }

  if(missnull(data) && missnull(split)){
    stop("Argument `split` must be provided. PROBLEM: it is missing.")
  }

  if(!missnull(data) && is.character(data) && length(data) == 1 && missnull(split)){
    # argument sliding
    split = data
    data = NULL
  }

  #
  # Formatting the input data
  #

  # used for the names
  dots = list(...)
  internal = FALSE
  if("mc" %in% names(dots)){
    mc = dots$mc
    internal = TRUE
  } else {
    mc = match.call()
  }

  id_names = NULL
  x_name = NULL
  if(inherits(x, "formula")){

    fml = x

    if(!missnull(data) && !is.list(data)){
      stop("When `x` is a formula, the argument `data` must be a list containing ",
           "the variables in the formula.",
           "\nPROBLEM: `data` is not a list.")
    }

    if(length(fml) < 3){
      stop("When `x` is a formula, it must be two sided (on the left the variable,",
           " on the right the identifier).\nPROBLEM: it is currently only one sided.")
    }

    x = try(eval(fml[[2]], data, enclos = parent.frame()), silent = TRUE)

    # error handling
    if(inherits(x, "try-error") || !is.atomic(x)){
      vars = all.vars(fml[[2]])

      if(!is.null(names(data))){
        var_pblm = setdiff(vars, names(data))
        if(length(var_pblm) > 0){
          stop("The evaluation of the left side of `x` raised an error:\n",
               string_magic("PROBLEM: the variable{$s, enum.bq, is ? var_pblm} not in the data set (`"),
               deparse_short(mc$data), "`).")
        }
      }

      if(inherits(x, "try-error")){
        stop("The evaluation of the left side of `x` raised an error:\n", x)
      }

      stop("The evaluation of the left side of `x` raised an error:\n",
           "PROBLEM: `x` is not a character vector, instead it is of class ", class(x)[1])
    }

    #
    # elements from id
    # we don't apply terms: we only separate with sum

    terms_right = fml_extract_elements(fml)

    id = list()
    for(i in seq_along(terms_right)){
      term = terms_right[[i]]
      id_name = deparse_short(term)

      val = try(eval(term, data, enclos = parent.frame()), silent = TRUE)

      # error handling
      if(inherits(val, "try-error") || !is.atomic(val)){
        vars = all.vars(term)

        intro = string_magic("The evaluation of the right side of `x` raised an error.\n",
                    "VALUE TO EVAL: {bq ? id_name}\n")

        if(!is.null(names(data))){
          var_pblm = setdiff(vars, names(data))
          if(length(var_pblm) > 0){
            stop(intro,
                 string_magic("PROBLEM: the variable{$s, enum.bq, is ? var_pblm} not in the data set (`"),
                 deparse_short(mc$data), "`).")
          }
        }

        if(inherits(val, "try-error")){
          stop(intro, "ERROR: ", val)
        }

        stop(intro,
             "PROBLEM: this identifier is not atomic, instead it is of class ", class(x)[1])
      }

      id[[id_name]] = val
    }
    
    if(!is.character(x)){
      x = as.character(x)
    }

    x_name = deparse_short(fml[[2]])
    id_names = names(id)

  } else {
    check_set_character(x)
    x_name = "x"
  }

  n = length(x)
  if(!missnull(id)){
    # id: a) a vector, b) a data frame. Same length as x

    if(is.list(id)){
      n_id = unique(lengths(id))
      if(length(n_id) != 1 || max(n_id) != n){
        extra = ""
        if(max(n_id) != n) extra = string_magic("\nPROBELM: len x: {#n ? n} len id: {#n ? max(n_id)}.")
        stop("The argument `id` must be either a vector of identifiers or a data.frame ",
             "of identifiers of the same length as `x`.", extra)
      }

      id_names = names(id)
      if(is.null(id_names)){
        id_txt = deparse_short(mc$id)
        if(length(id_names) > 1){
          id_names = paste0(id_txt, "..", 1:length(id))
        } else {
          id_names = id_txt
        }
      }

    } else {
      n_id = length(id)
      if(n != n_id){
        stopi("The argument `id` must be either a vector of identifiers or a data.frame ",
             "of identifiers of the same length as `x`.",
             "\nPROBELM: len x: {n ? n}; len id: {n ? n_id}.")
      }
      if(!is.atomic(id)){
        stop("The argument `id` must be either a vector of identifiers or a data.frame ",
             "of identifiers of the same length as `x`.",
             "\nPROBLEM: `id` is not an atomic vector")
      }

      id_names = deparse_short(mc$id)
      id = list(id)
    }
  }

  add.id = FALSE
  obs = 1:n
  if(!missnull(id)){
    add.id = TRUE
    id_raw = id
    id = to_index(list = id)

    if(id_unik && max(id) != length(id)){
      message("The identifiers are not unique, you will not be able to reconstruct the data using only them.")
    }

  }
  
  # flags
  pat_parsed = format_simple_regex_flags(split, ignore = ignore.case, 
                                         fixed = fixed, word = word, 
                                         magic = TRUE, envir = parent.frame(1 + internal))
  split = pat_parsed$pattern
  is_fixed = pat_parsed$fixed

  x_split = strsplit(x, split, fixed = is_fixed, perl = !is_fixed)

  n_all = lengths(x_split)

  # the result table
  if(add.pos){
    obs_all = rep(obs, n_all)
    pos = cpp_create_pos(obs_all)
    res = data.frame(obs = obs_all, pos = pos, x = unlist(x_split),
                     row.names = NULL)
  } else {
    res = data.frame(obs = rep(obs, n_all), x = unlist(x_split),
                     row.names = NULL)
  }


  # we add the names
  names(res)[names(res) == "x"] = x_name

  # we add the ID
  if(add.id){
    # we add the user-provided identifiers

    if(!is.list(id_raw)) id_raw = list(id = id_raw)

    id2bind = lapply(id_raw, function(x) x[res$obs])
    names(id2bind) = id_names

    res = cbind(res, id2bind)
  }

  if(dt){
    if(requireNamespace("data.table", quietly = TRUE)){
      data.table::setDT(res)
    } else {
      stop("To return a `data.table`, you need the `data.table` package to be installed. Currently this is not the case.",
           "\nUse install.packages('data.table')?")
    }
  }

  res
}

#' @describeIn string_split2df Splits a string vector and returns a `data.table`
string_split2dt = function(x, data = NULL, split = NULL, id = NULL, add.pos = FALSE,
                        id_unik = TRUE, fixed = FALSE){

  mc = match.call()
  string_split2df(x, data = data, split = split, id = id, add.pos = add.pos,
               id_unik = id_unik, fixed = fixed, mc = mc, dt = TRUE)
}

#' Paste a string vector conditionally 
#' 
#' Easily reconstruct a string vector that has been split with [string_split2df()].
#' 
#' @param x A character vector or a formula. If a vector: it represents the 
#' values to be pasted together.
#' If a formula, it must be of the form
#' `my_string ~ id1 + id2` with on the left the character vector and on the right
#' the (possibly many) identifiers. If a formula, the argument `id` must be a 
#' data frame.
#' @param id A vector of identifiers, a list of identifiers (can be a data frame),
#' or a `data.frame`. The identifiers can be a vector or a list of vectors. They represent
#' which elements of `x` are to be pasted together. 
#' 
#' When `x` is a formula, then `id` must be a data.frame containing the variables in 
#' the formula.
#' @param sep A character scalar, default is `" "`. The value used to paste together 
#' the elements of `x`.
#' @param names Logical scalar, default is `TRUE`. Whether to add, as names, the values of 
#' the identifiers.
#' @param sort Logical scalar, default is `TRUE`. Whether to sort the results according to 
#' the identifiers.
#' 
#' @return 
#' Returns a character vector. If the argument `names` is `TRUE` (default), the vector will have
#' names equal to the values of the identifiers.
#' 
#' @examples 
#' 
#' #
#' # let's paste together the letters of the alphabet
#' 
#' # first we create the identifier
#' id = rep(1:2, each = 13)
#' setNames(id, letters)
#' 
#' # now we conditionally paste together the letters
#' paste_conditional(letters, id, "")
#' 
#' #
#' # using a formula
#' 
#' # we create a small data set based on mtcars
#' base_cars = within(mtcars, carname <- row.names(mtcars))
#' base_cars = head(base_cars, 10)
#' 
#' # we use two identifiers
#' paste_conditional(carname ~ gear + carb, base_cars, sep = ", ")
#' 
#' 
#' 
paste_conditional = function(x, id, sep = " ", names = TRUE, sort = TRUE){
  
  check_logical(names, scalar = TRUE)
  check_logical(sort, scalar = TRUE)
  
  check_character(sep, scalar = TRUE)
  if(missing(x)) stop("The argument `x` must be provided. Problem: it is missing.")
  if(missing(id)) stop("The argument `id` must be provided. Problem: it is missing.")
  
  if(is.character(x)){
    n_x = length(x)
    # we expect id to be a vector or a list 
    if(is.atomic(id)){
      if(length(id) != n_x){
        stopi("The argument `id` must be of the same length as `x`.",
              "\nPROBLEM: `x` is {len?x} vs `id` is {len?id}.")
      }
      id = list(id)
    } else if(is.list(id)){
      n_id_all = lengths(id)
      if(any(n_id_all != n_x)){
        i_pblm = which(n_id_all != n_x)[1]
        stopi("The argument `id` must be of the same length as `x`.",
              "\nPROBLEM: `x` is {len?x} vs `id[[{i_pblm}]]` is {n?n_id_all[i_pblm]}.")
      }
    } else {
      stopi("When `x` is a vector, `id` must be i) a vector or ii) a list of vectors.",
            "\nPROBLEM: instead `id` is of class {enum.bs?class(id)}.")
    }
    
  } else if(inherits(x, "formula")){
    if(!is.data.frame(id)){
      stopi("When `x` is a vector, `id` must be a `data.frame`.",
            "\nPROBLEM: instead `id` is of class {enum.bs?class(id)}.")
    }
    
    fml = x
    df = id
    x = check_expr(eval(fml[[2]], df), "The left-hand-side of the formula, equal to `", 
                   deparse(fml[[2]]), 
                   "` could not be evaluated. See error below:")
    x = as.character(x)
    
    term_to_eval = attr(terms(fml), "term.labels")
    n_id = length(term_to_eval)
    id = vector("list", n_id)
    for(i in 1:n_id){
      term = term_to_eval[i]
      val = check_expr(eval(str2lang(term), df), "The right-hand-side of the formula, equal to `", 
                       term, "` could not be evaluated. See error below:")
      id[[i]] = val
    }
    names(id) = term_to_eval
    
  } else {
    stopi("The argument `x` must be a character vector or a formula.",
          "\nPROBLEM: instead it is of class {enum.bq?class(x)}.")
  }
  
  id_raw = id
  id = to_index(list = id, sorted = sort)
  
  # obs_first: used in names
  obs_first = which(!duplicated(id))
  obs_first_order = order(id[obs_first])
  obs_first = obs_first[obs_first_order]  
  
  # we need to order before applying the cpp function
  id_order = order(id)
  id = id[id_order]
  x = x[id_order]
  
  res = cpp_paste_conditional(x, id, sep = sep, sep_last = sep)
  
  # adding names
  if(is.atomic(id_raw) || (is.null(names(id_raw)) && length(id_raw) == 1)){
    if(!is.atomic(id_raw)){
      id_raw = id_raw[[1]]
    }
    nm = id_raw[obs_first]
  } else {
    id_raw = unclass(id_raw)
    id_names = names(id_raw)
    is_names = !is.null(id_names)
    
    for(i in 1:length(id_raw)){
      val = id_raw[[i]][obs_first]
      if(is_names){
        val = paste0(id_names[i], ": ", val) 
      }
      id_raw[[i]] = val
    }
    
    names(id_raw) = NULL
    id_raw$sep = ", "
    
    nm = do.call(paste, id_raw)
  }
  names(res) = nm
  
  res  
}




#' Cleans a character vector from multiple patterns
#'
#' Recursively cleans a character vector from several patterns. Quickly handle the 
#' tedious task of data cleaning by taking advantage of the syntax.
#' You can also apply all sorts of cleaning operations by summoning [string_ops()] operations.
#' 
#' @inheritParams string_is
#'
#' @param x A character vector.
#' @param ... Character scalars representing patterns. A pattern is of the form
#' "flags/pat1, pat2 => replacement". This means that patterns 'pat1' and 'pat2' will be replaced
#' with the string 'replacement'. By default patterns are comma separated and the replacement comes 
#' after a ' => ' (see args `split` and `pipe` to change this). By default the replacement is the empty string 
#' (so "pat1, pat2" *removes* the patterns).
#' 
#' Available regex flags are: 'word' (add word boundaries), 'ignore' (the case), 'fixed' (no regex), 
#' 'total', 'single' and 'magic'. 
#' The flag `total` leads to a *total replacement* of the string if the pattern is found. 
#' The flag 'magic' allows to interpolate variables within the pattern.  Use flags
#' with comma separation ("word, total/pat") or use only their initials ("wt/pat").
#' 
#' Starting with an '@' leads to operations in [string_ops()]. Ex: "@ascii, lower, ws" turns
#' the string into ASCII, lowers the case and normalizes white spaces (see help of [string_ops()]).
#' @param pipe Character scalar, default is `" => "`. If thevalue of `pipe` is found in a pattern,
#' then the string is split w.r.t. the pipe and anything after the pipe becomes the replacement.
#' 
#' For example in `string_clean(x, "e => a")` the default pipe is found in "e => a", so the pattern 
#' "e" will be replaced with "a". In other terms, this is equivalent to `string_clean(x, "e", replacement = "a")`.
#' Example changing the pipe: you can obtain the previous result with `string_clean(x, "e|>a", pipe = "|>")`.
#' @param split Character scalar, default is `",[ \t\n]+"` (which means a comma followed with spaces 
#' and/or new lines). By default the patterns to be replaced are comma separated, that is 
#' the pattern is split w.r.t. the argument `split` and a replacement is done for each sub-pattern.
#' 
#' Use `NULL` or the empty string to disable pattern separation.
#' 
#' For example: let's look at `string_clean(x, "w/one, two => three")`. First the flag "word" is extracted from
#' the pattern (see arg. `...`) as well as the replacement (see arg. `pipe`), leading to "one, two" the 
#' pattern to be replaced. Then the pattern is split w.r.t. `split`, leading 
#' to two patterns "one" and "two". Hence the words (thanks to the flag "w") "one" and "two" from
#' the string `x` will be replaced with "three".
#' @param replacement Character scalar, default is the empty string. It represents the default 
#' value by which the patterns found in the character strings will be replaced. For example
#' `string_clean(x, "e", replacement = "a")` turn all letters "e" in `x` into "a".
#' @param total Logical scalar, default is `FALSE`. If `TRUE`, then when a pattern is found 
#' in a string, the full string is replaced (instead of just the pattern). Note, *importantly*, 
#' that when `total = TRUE` you can use logical operators in the patterns.
#' 
#' Example: `string_clean(x, "wi/ & two, three & !four => ", total = TRUE)`
#' @param single Logical scalar, default is `FALSE`. Whether, in substitutions, to stop at 
#' the first match found. Ex: `string_clean("abc", "[[:alpha:]] => _", single = TRUE)` leads 
#' to `"_bc"`, while `string_clean("abc", "[[:alpha:]] => _")` leads to `"___"`.
#' @param namespace Character scalar or `NULL` (default). **Only useful for package developers.**
#' As a regular end-user you shouldn't care! If your package uses `string_magic`, you should care. 
#' It is useful **only** if your package uses 'custom' `string_magic` operations, set with 
#' [string_magic_register_fun()] or [string_magic_register_ops()].
#' 
#' If so pass the name of your package in this argument so that your function can access 
#' the new `string_magic` operations defined within your package.
#' @param pattern A character scalar containing a regular expression pattern to be replaced.
#' You can write the replacement directly in the string after a pipe: ' => ' (see arg. `pipe` to change this). 
#' By default the replacement is the empty string (so "pat1" *removes* the pattern).
#' 
#' Available regex flags are: 'word' (add word boundaries), 'ignore' (the case), 'fixed' (no regex), 
#' 'total', 'single' and 'magic'. 
#' The flag `total` leads to a *total replacement* of the string if the pattern is found. 
#' The flag 'magic' allows to interpolate variables within the pattern.  Use flags
#' with comma separation ("word, total/pat") or use only their initials ("wt/pat").
#' 
#' 
#'
#' @return
#' The main usage returns a character vector of the same length as the vector in input.
#' Note, however, that since you can apply arbitrary [string_ops()] operations, the length and type
#' of the final vector may depend on those (if they are used).
#' 
#' @inheritSection string_is Generic regular expression flags
#' 
#' @section Regular expression flags specific to replacement:
#' 
#' This function benefits from two specific regex flags: "total" and "single".
#' 
#' + "total" replaces the *complete string* if the pattern is found (remember that the 
#' default behavior is to replace just the pattern). 
#' 
#' + "single" performs a single substitution for each string element and stops there. 
#' Only the first match of each string is replaced. Technically we use [base::sub()]
#' internally instead of [base::gsub()].
#' 
#' @author 
#' Laurent R. Berge
#' 
#' @family tools with aliases
#' 
#' @seealso 
#' String operations: [string_is()], [string_get()], [string_clean()], [string_split2df()]. 
#' Chain basic operations with [string_ops()]. Clean character vectors efficiently
#' with [string_clean()].
#' 
#' Use [string_vec()] to create simple string vectors.
#' 
#' String interpolation combined with operation chaining: [string_magic()]. You can change `string_magic`
#' default values with [string_magic_alias()] and add custom operations with [string_magic_register_fun()].
#' 
#' Display messages while benefiting from `string_magic` interpolation with [cat_magic()] and [message_magic()].
#'
#' @examples
#'
#' x = c("hello world  ", "it's 5 am....")
#'
#' # we clean the o's and the points (we use 'fixed' to trigger fixed-search)
#' string_clean(x, "o", "f/.")
#' # equivalently
#' string_clean(x, "fixed/o, .")
#' # equivalently
#' string_clean(x, "o, .", fixed = TRUE)
#' # equivalently
#' string_clean(x, "o", ".", fixed = TRUE)
#' 
#' #
#' # chaining operations: example using cars
#' #
#' 
#' cars = row.names(mtcars)
#' new = string_clean(cars, 
#'            # replace strings containing "Maz" with Mazda
#'            "total/Maz => Mazda", 
#'            # replace the word 'Merc' with Mercedes
#'            "wi/merc => Mercedes",
#'            # replace strings containing "Merc" and a digit followed with an 'S'
#'            "t/Merc & \\dS => Mercedes S!",
#'            # put to lower case, remove isolated characters and normalize white spaces
#'            "@lower, ws.isolated")
#' 
#' cbind(cars, new)
#'
#'
string_clean = function(x, ..., replacement = "", pipe = " => ", split = ",[ \n\t]+",
                        ignore.case = FALSE, fixed = FALSE, word = FALSE, 
                        total = FALSE, single = FALSE, envir = parent.frame(), 
                        namespace = NULL){

  x = check_set_character(x, l0 = TRUE)
  if(length(x) == 0){
    return(x)
  }

  check_character(replacement, scalar = TRUE)
  check_character(pipe, scalar = TRUE)
  check_character(split, scalar = TRUE, null = TRUE)

  is_split = length(split) > 0 && nchar(split) > 0
  is_pipe = length(pipe) > 0 && nchar(pipe) > 0

  check_logical(ignore.case, scalar = TRUE)
  check_logical(fixed, scalar = TRUE)
  check_logical(word, scalar = TRUE)
  check_logical(total, scalar = TRUE)
  check_logical(single, scalar = TRUE)

  rep_main = replacement

  dots = list(...)
  warn_no_named_dots(dots)

  dots = unlist(dots)

  res = x
  for(i in seq_along(dots)){
    di = di_raw = dots[[i]]
    
    first_char = substr(di, 1, 1)
    is_string_ops = FALSE
    if(first_char == "\\" && substr(di, 2, 2) == "@"){
      di = substr(di, 2, nchar(di))
    } else if(first_char == "@" && nchar(di) > 1){
      is_string_ops = TRUE
      di = substr(di, 2, nchar(di))
    }

    if(is_string_ops){
      res = string_ops(res, di, namespace = namespace)
      next
    }
    
    pattern_pipe = FALSE
    if(is_pipe && grepl(pipe, di)){
      # we apply the pipe to 
      pattern_pipe = TRUE
      di_split = strsplit(di, pipe)[[1]]
      replacement = di_split[2]
      di = di_split[1]
    } else {
      replacement = rep_main
    }
    
    # we parse the special flags
    is_total = total
    di_parsed = parse_regex_pattern(di, c("ignore", "fixed", "word", "total", "single", "magic"), 
                                    parse_logical = FALSE, envir = envir)
    flags = di_parsed$flags
    patterns = di_parsed$patterns

    is_total = total || "total" %in% flags
    is_fixed = fixed || "fixed" %in% flags
    is_ignore = ignore.case || "ignore" %in% flags
    is_word = word || "word" %in% flags
    is_single = single || "single" %in% flags
    
    if(pattern_pipe && "magic" %in% flags){
      # we interpolate the replacement
      replacement_new = try(string_magic(replacement, .envir = envir), silent = TRUE)
      if(isError(replacement_new)){
        stop_hook("CONTEXT: concerns the pattern {bq?replacement}",
                  "\nINFO: The `magic` flag expands the pattern with `string_magic`.",
                  "\nPROBLEM: the evaluation with `string_magic` failed, see error below:",
                  "\n{'^[^\n]+\n'r?replacement_new}")
      }
      
      if(length(replacement_new) != 1){
        stop_hook("CONTEXT: concerns the pattern {bq?replacement}",
                  "\nINFO: The `magic` flag expands the pattern with `string_magic`. It must return a vector of length 1.",
                  "\nPROBLEM: the vector returned is of length {len?replacement_new}.")
      }
      
      replacement = replacement_new
    }

    if(is_split){
      all_patterns = strsplit(patterns, split = split)[[1]]
    } else {
      all_patterns = patterns
    }

    for(j in seq_along(all_patterns)){
      pat = all_patterns[j]

      if(is_total){
        # we allow logical operations when the replacement is in full
        pat_parsed = parse_regex_pattern(pat, NULL, parse_logical = TRUE, parse_flags = FALSE)
        pat = pat_parsed$patterns
        is_or = pat_parsed$is_or
        is_negate_all = pat_parsed$is_not
      }

      who_current = NULL
      for(k in seq_along(pat)){
        p = pat[k]

        if(is_total){
          negate = is_negate_all[k]
        }
        
        p = format_pattern(p, fixed = is_fixed, word = is_word, ignore = is_ignore)
        is_fixed = attr(p, "fixed")

        if(is_total){
          who_tmp = tryCatch(grepl(p, res, perl = !is_fixed, fixed = is_fixed), 
                         error = function(e) structure(conditionMessage(e), class = "try-error"),
                         warning = function(w) structure(conditionMessage(w), class = "try-warning"))
            
          is_warn = inherits(who_tmp, "try-warning")
          warn_msg = ""
          if(is_warn){
            # is there an error?
            warn_msg = who_tmp
            who_tmp = tryCatch(suppressWarnings(grepl(p, res, perl = !is_fixed, fixed = is_fixed)), 
                           error = function(e) structure(conditionMessage(e), class = "try-error"))
          }
          
          if(inherits(who_tmp, "try-error")){
            di_raw = escape_newline(di_raw)
            p = escape_newline(p)
            replacement = escape_newline(replacement)
            stopi("CONTEXT: evaluation of {Q?di_raw}",
                  "\n         pattern     = {Q?p} ",
                  "\n         replacement = {Q?replacement}",
                  "\nEXPECTATION: the pattern must be a valid regular expression",
                  "\nPROBLEM: `grepl` led to an error, see below:",
                  "\n{who_tmp}",
                  "{&nchar(warn_msg) > 0;\n{.}}")
          } else if(is_warn){
            di_raw = escape_newline(di_raw)
            p = escape_newline(p)
            replacement = escape_newline(replacement)
            warni("CONTEXT: evaluation of {Q?di_raw}",
                  "\n         pattern     = {Q?p} ",
                  "\n         replacement = {Q?replacement}",
                  "\nA warning was raised when evaluating the pattern:",
                  "\n{warn_msg}", immediate. = TRUE)
          }
          
          if(negate){
            who_tmp = !who_tmp
          }
          if(is.null(who_current)){
            who_current = who_tmp
          } else {
            if(is_or[k]){
              who_current = who_tmp | who_current
            } else {
              who_current = who_tmp & who_current
            }
          }
        } else {
          # shallow copy for error handling
          # I'm not sure it's useful: I think P(error|warning) = 100%, but well
          # you never know... 
          res_save = res
          my_sub = if(is_single) base::sub else base::gsub          
          
          res = tryCatch(my_sub(p, replacement, res, perl = !is_fixed, fixed = is_fixed), 
                         error = function(e) structure(conditionMessage(e), class = "try-error"),
                         warning = function(w) structure(conditionMessage(w), class = "try-warning"))
            
          is_warn = inherits(res, "try-warning")
          warn_msg = ""
          if(is_warn){
            # is there an error?
            warn_msg = res
            res = tryCatch(suppressWarnings(my_sub(p, replacement, res_save, perl = !is_fixed, fixed = is_fixed)), 
                           error = function(e) structure(conditionMessage(e), class = "try-error"))
          }
          
          if(inherits(res, "try-error")){
            di_raw = escape_newline(di_raw)
            p = escape_newline(p)
            replacement = escape_newline(replacement)
            stopi("CONTEXT: evaluation of {Q?di_raw}",
                  "\n         pattern     = {Q?p} ",
                  "\n         replacement = {Q?replacement}",
                  "\nEXPECTATION: the pattern must be a valid regular expression",
                  "\nPROBLEM: `{&is_single;sub;gsub}` led to an error, see below:",
                  "\n{res}",
                  "{&nchar(warn_msg) > 0;\n{.}}")
          } else if(is_warn){
            di_raw = escape_newline(di_raw)
            p = escape_newline(p)
            replacement = escape_newline(replacement)
            warni("CONTEXT: evaluation of {Q?di_raw}",
                  "\n         pattern     = {Q?p} ",
                  "\n         replacement = {Q?replacement}",
                  "\nA warning was raised when evaluating the pattern:",
                  "\n{warn_msg}", immediate. = TRUE)
          }
        }
      }
      
      if(is_total){
        res[who_current] = replacement
      }
    }

  }

  res
}

#' @describeIn string_clean Simplified version of `string_clean`
string_replace = function(x, pattern, replacement = "", pipe = " => ", ignore.case = FALSE, 
                          fixed = FALSE, word = FALSE, 
                          total = FALSE, single = FALSE, envir = parent.frame()){
  # this is a simplified version of string_clean
  # I think it's still useful bc of the flags, the pipe and the absence of splitting
  
  set_pblm_hook()
  
  string_clean(x, pattern, replacement = replacement, pipe = pipe, ignore.case = ignore.case,
               fixed = fixed, word = word, total = total, single = single, 
               envir = envir, split = "")
  
}

#' Efficient creation of string vectors with optional interpolation
#' 
#' Create string vectors in multiple ways: 1) add successive string elements (like in `c()`),
#' or 2) write a character string that will be broken with respect to commas 
#' (`"hi, there"` becomes `c("hi", "there")`), or 3) interpolate variables in 
#' character strings (`"x{1:2}"` becomes `c("x1", "x2")`) with full access to 
#' [string_magic()] operations, or any combination of the three.
#' 
#' @inheritParams string_magic
#' @inheritParams string_clean
#' 
#' @param ... Character vectors that will be vectorized. If commas are present in the 
#' character vector, it will be split with respect to commas and following blanks. 
#' The vectors can contain any interpolation in the form `"{var}"` and 
#' any [string_magic()] operation can be applied. To change the delimiters for interpolation,
#' see `.delim`. Named arguments are used in priority for variable substitution,
#' otherwise the value of the variables to be interpolated are fetched in the calling environment 
#' (see argument `.envir`).
#' @param .protect.vars Logical scalar, default is `FALSE`. If `TRUE`, then only
#' arguments equal to a "natural" character scalar are comma-split and interpolated, 
#' other arguments are not touched. Ex: `string_vec("x{1:5}")` will lead to a vector of 
#' length 5 ("x1" to "x5"), while `z = "x{1:5}"` followed by `string_vec(z)` leads 
#' to a vector of length 1: `"x{1:5}"`. If `FALSE`, comma-splitting and interpolation
#' is performed on all variables. 
#' @param .split Logical or a character symbol, default is `TRUE`. If `TRUE`, 
#' the character vectors are split with respect to commas (default). If `FALSE`, no 
#' splitting is performed. If a character symbol, the string vector will be split
#' according to this symbol. Note that any space after the symbol (including tabs and 
#' newlines) is discarded. You can escape the splitting symbol with a backslash right before it.
#' 
#' Ex: by default `string_vec("hi, there")` leads to the vector `c("hi", "there")`.
#' @param .sep Character scalar or `NULL` (default). If not `NULL`, the function
#' [base::paste()] is applied to the resulting vector with `sep = .sep`.
#' @param .collapse Character scalar or `NULL` (default). If not `NULL`, the function
#' [base::paste()] is applied to the resulting vector with `collapse = .collapse`.
#' 
#' If so, pass the name of your package in this argument so that your function can access 
#' the new `string_magic` operations defined within your package.
#' @param .cmat Logical scalar (default is `FALSE`), integer or complex with integer values.
#' If `TRUE`, we try to coerce the result into a **character** matrix. The number of rows and columns
#' is deduced from the look of the arguments. An integer indicates the number of rows.
#' If a complex, the imaginary part represents the number of columns. Ex: `5i` means 
#' 5 columns, `3 + 2i` means 3 rows and 2 columns.
#' 
#' The matrix is always filled by row.
#' @param .nmat Logical scalar (default is `FALSE`), integer or complex with integer values.
#' If `TRUE`, we try to coerce the result into a **numeric** matrix. The number of rows and columns
#' is deduced from the look of the arguments. An integer indicates the number of rows.
#' If a complex, the imaginary part represents the number of columns. Ex: `5i` means 
#' 5 columns, `3 + 2i` means 3 rows and 2 columns.
#' 
#' The matrix is always filled by row. Non numbers are silently turned to NA.
#' @param .df Logical scalar (default is `FALSE`), integer, complex with integer values, or
#' character vector.
#' If `TRUE`, we try to coerce the result into a `data.frame`. The number of rows and columns
#' is deduced from the look of the arguments. An integer indicates the number of rows.
#' If a complex, the imaginary part represents the number of columns. Ex: `5i` means 
#' 5 columns, `3 + 2i` means 3 rows and 2 columns. 
#' 
#' If a character vector: it should give the column names. Note that if the vector
#' is of length 1, its values are comma separated (i.e. the value `"id, name"` is
#' identical to `c("id", "name")`).
#' 
#' Note that the columns that can be converted to numeric are converted to numeric.
#' The other columns are in string form. Monitor this behavior with `.df.convert`.
#' 
#' @param .df.convert Logical scalar, default is `TRUE`. Only used when the result is 
#' to be converted to a data frame (with the argument `.df`). If `TRUE`, any column
#' looking like a numeric vector is converted to numeric. Otherwise all columns 
#' are character strings.
#' 
#' @details 
#' The main objective of this function is to simplify the creation of small character vectors.
#' By default, you can pass a character string of length 1 with values separated with commas
#' and a character vector will be returned. 
#' 
#' You can use interpolation using curly brackets (see `string_magic()`). You can pass
#' values for the interpolation directly in the arguments (this is why all 
#' arguments start with a dot).
#' 
#' By default character values containing commas are split with respect to the commas
#' to create vectors. To change this behavior, see the argument `.split`.
#' 
#' The default of the argument `.protect.vars` is `FALSE`. To avoid unwanted 
#' comma-splitting and interpolations on variables, set it to `TRUE`. 
#' The main use case of this function is
#' the creation of small string vectors, which can be written directly at
#' function call. 
#' 
#' Customize the default of this function with [string_vec_alias()].
#' 
#' @return 
#' By default this function returns a string vector, the length of which depends on the arguments.
#' 
#' This result can be processed with the arguments `cmat`, `nmat` and `.df` which will 
#' try to coerce the result into a character matrix, a numeric matrix , or a data frame, respectively.
#' 
#' @author 
#' Laurent Berge
#' 
#' @family tools with aliases
#' 
#' @inherit string_clean seealso
#' 
#' @examples 
#' 
#' # illustrating comma-splitting and interpolation
#' string_vec("x1, y2, z{4:5}")
#' 
#' # variable protection
#' x = "x{1:5}"
#' # without protection (default) => interpolation takes place
#' string_vec(x, "y{1:2}")
#' 
#' # with protection => no interpolation for x
#' string_vec(x, "y{1:2}", .protect.vars = TRUE)
#' 
#' # removing comma splitting
#' string_vec("Hi, said Charles.", "Hi, said {girl}.", girl = "Julia", .split = FALSE)
#' 
#' # changing the delimiters for interpolation
#' pkg = "\\usepackage[usenames,dvipsnames]{xcolor}"
#' string_vec("\\usepackage{.[S!graphicx, fourier, standalone]}", 
#'            pkg, .delim = ".[ ]")
#' 
#' #
#' # Customization
#' #
#' 
#' # Unhappy about the defaults? Create an alias!
#' 
#' # we create a "matrix generator"
#' matgen = string_vec_alias(.nmat = TRUE, .last = "'\n'split")
#' 
#' matgen("5, 4, 3
#'         8, 6, 2")
#' 
string_vec = function(..., .cmat = FALSE, .nmat = FALSE, .df = FALSE,
                      .df.convert = TRUE, .delim = c("{", "}"), 
                      .data  = list(), .envir = parent.frame(), 
                      .split = TRUE, .protect.vars = FALSE, .sep = NULL, 
                      .last = NULL, .check = TRUE, .help = NULL,
                      .collapse = NULL, .namespace = NULL){
  
  # checks
  set_pblm_hook()
  if(.check){
    .delim = check_set_delimiters(.delim)
    check_character(.sep, scalar = TRUE, null = TRUE)
    check_character(.collapse, scalar = TRUE, null = TRUE)
    check_character(.last, scalar = TRUE, null = TRUE)
    .split = check_set_split(.split)
    check_logical(.protect.vars, scalar = TRUE)
    check_envir(.envir)
    check_list(.data)
    # checking the dots + setting up the data
    dots = check_set_dots(..., mbt = TRUE, nofun = TRUE)
  } else {
    dots = list(...)
  }
  do_mat = check_set_mat(.cmat, .nmat, .df)
  
  # we check if some variables were passed in the arguments
  dot_names = names(dots)
  if(!is.null(dot_names)){
    is_var = dot_names != ""
    for(i in which(is_var)){
      .data[[dot_names[i]]] = dots[[i]]
    }
    dots[is_var] = NULL      
  }
  
  n = length(dots)
  if(n == 0){
    stop_hook("`string_vec` requires at least one character scalar to work.",
              "\nNamed arguments are only used as variables on which to apply interpolation.",
              "\nPROBLEM: all arguments are named.",
              "\nFIX: please provide at least one non-named argument.")
  }
  
  is_to_split = rep(TRUE, n)
  if(.protect.vars){
    mc_dots = match.call(expand.dots = FALSE)[["..."]]
    mc_dots_nm = names(mc_dots)
    if(!is.null(mc_dots_nm)){
      # we keep only the non named arguments
      mc_dots = mc_dots[nchar(mc_dots_nm) == 0]
    }
    
    is_to_split = sapply(mc_dots, is.character)
  }
  
  #
  # now the algorithm
  #
  
  BOX_OPEN = .delim[1]
  
  res_list = vector("list", n)
  for(i in 1:n){
    di = dots[[i]]
    if(!true_character(di)){
      di = as.character(di)
    }

    if(length(di) == 1 && is_to_split[i]){
    
      if(isFALSE(.split)){
        di_expanded = di
      } else {
        di_expanded = cpp_magic_split(di, .split, .delim)
      }      
      
      is_open = grepl(BOX_OPEN, di_expanded, fixed = TRUE)
      n_open = sum(is_open)
      if(n_open > 0){
        n_di_xpd = length(di_expanded)
        all_elements = vector("list", n_di_xpd)
        
        for(j in 1:n_di_xpd){
          if(is_open[j]){
            all_elements[[j]] = string_magic_internal(di_expanded[j], .delim = .delim, 
                                                      .envir = .envir, .is_root = TRUE,
                                                      .data = .data, .check = .check, 
                                                      .help = .help,
                                                      .namespace = .namespace)  
          } else {
            all_elements[[j]] = di_expanded[j]
          }        
        }
        
        di_expanded = do.call(base::c, all_elements)  
      }
      
      res_list[[i]] = di_expanded
    } else {
      res_list[[i]] = di
    }
  }
  
  if(!is.null(.sep) || !is.null(.collapse)){
    if(do_mat){
      stopi("The arguments `.sep` or `.collapse` are incompatible with the arguments ",
            "`.cmat` or `.nmat`. Please remove some of these arguments.")
    }
    res_list$sep = .sep
    res_list$collapse = .collapse
    res = do.call(base::paste, res_list)
  } else {
    res = do.call(base::c, res_list)
  }
  
  if(!is.null(.last)){
    res = string_ops(res, .last)
  }
  
  #
  # matrix formatting
  #
  
  if(do_mat){
    is_auto = isTRUE(attr(do_mat, "auto"))
    is_num = isTRUE(attr(do_mat, "num"))
    is_df = isTRUE(attr(do_mat, "df"))
    n_out = length(res)
    if(is_auto){
      done = TRUE
      if(n == 1){
        d = dots[[1]]
        nrows = length(strsplit(d, "\n")[[1]])
      } else {
        nrows = n
      }
      
      if((n_out / nrows) %% 1 != 0){
        done = FALSE
      }
      
      if(!done){
        nrows = floor(sqrt(n_out))
        while(nrows >= 1 && (n_out / nrows) %% 1 != 0){
          nrows = nrows - 1
        }
      }
      
      ncols = n_out / nrows
    } else {
      row_cols = attr(do_mat, "row_col")
      nrows = row_cols[1]
      ncols = row_cols[2]
      if(nrows == 0){
        nrows = n_out / ncols
        if(round(nrows) != nrows){
          arg = if(is_num) ".nmat" else if(is_df) ".df" else ".cmat"
          stopi("The number of columns in argument `{arg}` ", 
                "({ncols}) is not a divisor ",
                "of the final number of elements: {n?n_out}.")
        }
      } 
      
      if(ncols == 0){
        ncols = n_out / nrows
        if(round(ncols) != ncols){
          arg = if(is_num) ".nmat" else if(is_df) ".df" else ".cmat"
          stopi("The number of rows in argument `{arg}` ", 
                "({nrows}) is not a divisor ",
                "of the final number of elements: {n?n_out}.")
        }
      }
    }
    
    if(is_num){
      res = suppressWarnings(as.numeric(res))
    }
    
    res = matrix(res, nrows, ncols, byrow = TRUE)
    
    if(is_df){
      res = as.data.frame(res)
      colnames = attr(do_mat, "colnames")
      if(!is.null(colnames)){
        names(res) = colnames
      }
      # conversion of the columns to numeric
      check_logical(.df.convert, scalar = TRUE)
      if(.df.convert){
        for(i in 1:ncol(res)){
          if(is_numeric_in_char(res[[i]])){
            res[[i]] = as.numeric(res[[i]])
          }
        }
      }
    }
  }  
    
  res  
}



#' Fills a character string up to a size
#' 
#' Fills a character string up to a size and handles multibyte encodings 
#' (differently from sprintf).
#' 
#' @param x A character vector.
#' @param n Integer scalar, possibly equal to `NULL` (default). The size up to which the character 
#' vector will be filled. If `NULL` (default), it is set to the largest width in the character vector `x`.
#' To handle how the character is filled, see the arguments `symbol`, `right` and `center`.
#' @param symbol Character scalar of length 1, default is a space (" "). It is the symbol with which
#' the string will be filled.
#' @param right Logical scalar, default is `FALSE`. If `TRUE`, then the filling of the string is 
#' done from the left, leading to right-alignment.
#' @param center Logical scalar, default is `FALSE`. If `TRUE`, then the filling of the string will
#' be balanced so as to center the strings.
#' @param center.right Logical scalar, default is `TRUE`. Only used when `center = TRUE`, 
#' ignored otherwise. If `TRUE`, then when the width is odd and the number of characters of the 
#' string is even (or vice versa), the text is centered with one character on the right. 
#' If `FALSE`, this is one character on the left.
#' @param na Character scalar or `NA`. Default is "NA" (a character string!). What happens to NAs: by default 
#' they are replaced by the character string "NA".
#' 
#' @details 
#' If you use character filling of the form `sprintf("% 20s", x)` with `x``containing multibyte characters,
#' you may be suprised that all character strings do not end up at the same lenght (the occurrence of this problem
#' depends on many things: encodings are a mess). `string_fill`
#' uses only base R functions to compensate this. It is slightly slower but, in general, safer. 
#' 
#' It also looks a bit like [base::format()], but slightly different (and a bit faster, but more restrictive).
#' 
#' @return 
#' This functions returns a character vector of the same lenght as the vector in input.
#' 
#' @author 
#' Laurent R. Berge
#' 
#' @inherit string_clean seealso
#' 
#' @examples 
#' 
#' x = c("apple", "pineapple") 
#' 
#' # simple fill with blank
#' cat(paste0(string_fill(x), ":", c(3, 7), "€"), sep = "\n")
#' 
#' # center fill
#' cat(paste0(string_fill(x, center = TRUE), ":", c(3, 7), "€"), sep = "\n")
#' 
#' # changing the length of the fill and the symbol used for filling
#' cat(paste0(string_fill(x), ":", 
#'            string_fill(c(3, 7), 3, "0", right = TRUE), "€"), sep = "\n")
#' 
#' # na behavior: default/NA/other
#' x = c("hello", NA) 
#' string_fill(x)
#' string_fill(x, na = NA)
#' string_fill(x, na = "(missing)")
#' 
#' 
string_fill = function(x = "", n = NULL, symbol = " ", right = FALSE, 
                       center = FALSE, center.right = TRUE, na = "NA"){
  # Character vectors starting with " " are not well taken care of

  x = check_set_character(x, l0 = TRUE, null = TRUE)
  if(length(x) == 0) return(character(0))

  check_numeric(n, null = TRUE, scalar = TRUE, integer = TRUE)
  check_character(symbol, scalar = TRUE)
  check_logical(right, scalar = TRUE)

  if(anyNA(x)){
    x[is.na(x)] = na
  }  

  if(nchar(symbol) != 1){
    stop("Argument 'symbol' must be a single character ", 
         "(currenlty it is of length ", nchar(symbol), ").")
  }

  if(!is.null(n) && n == 0){
    return(x)
  }

  n_all = nchar(x)
  if(is.null(n)) n = max(n_all)

  n2fill = n - n_all
  qui = which(n2fill > 0)
  if(length(qui) == 0){
    return(x)
  }
  
  x_new = simple_string_fill(x[qui], n, symbol, right = right, 
                             center = center, center.right = center.right)

  res = x
  res[qui] = x_new
  res
}





#' Extracts a pattern from a character vector
#' 
#' Extracts the first, or several, patterns from a character vector.
#' 
#' @inheritParams string_is
#' 
#' @param x A character vector.
#' @param pattern A character scalar. It represents the pattern
#' to be extracted from `x`. By default 
#' this is a regular expression. You can use flags in the pattern in 
#' the form `flag1, flag2/pattern`.
#' Available flags are `ignore` (case), `fixed` (no regex), word (add word boundaries), 
#' single (select only the first element), and magic (add interpolation with `{}`) . Example:
#' if `"ignore/hello"` and `x = "Hello world` extracted text is `"Hello"`.
#' Shortcut: use the first letters of the flags. Ex: "iw/one" will extract the word 
#' "one" (flags 'ignore' + 'word').
#' @param single Logical scalar, default is `FALSE`. If `TRUE`, only the first pattern
#' that is detected will be returned. Note that in that case, a character vector is returned
#' of the same length as the vector in input. 
#' @param simplify Logical scalar, default is `TRUE`. If `TRUE`, then when the vector input `x`
#'  is of length 1, a character vector is returned instead of a list.
#' @param unlist Logical scalar, default is `FALSE`. If `TRUE`, the function `unlist` is applied
#' to the resulting list, leading to a character vector in output (instead of a list).
#' 
#' @inheritSection string_is Generic regular expression flags
#' 
#' @return 
#' The object returned by this functions can be a list or a character vector. 
#' 
#' If `single = TRUE`, a character vector is returned, containing the value of the first match.
#' If no match is found, an empty string is returned.
#' 
#' If `single = FALSE` (the default) and `simplify = TRUE` (default), the object returned is:
#' + a character vector if `x`, the vector in input, is of length 1: the character vector contains
#' all the matches and is of length 0 if no match is found.
#' + a list of the same length as `x`. The ith element of the list is a character vector
#' of the matches for the ith element of `x`.
#' 
#' If `single = FALSE` (default) and `simplify = FALSE`, the object returned is always a list.
#' 
#' @examples 
#' 
#' cars = head(row.names(mtcars))
#' 
#' # Let's extract the first word:
#' string_extract(cars, "\\w+", single = TRUE)
#' 
#' # same using flags
#' string_extract(cars, "s/\\w+")
#' 
#' # extract all words composed on only letters
#' # NOTE: we use the flag word (`w/`)
#' string_extract(cars, "w/[[:alpha:]]+")
#' 
#' # version without flag:
#' string_extract(cars, "\\b[[:alpha:]]+\\b")
#' 
#' # If a vector of length 1 => a vector is returned
#' greet = "Hi Tom, how's Mary doing?"
#' string_extract(greet, "w/[[:upper:]]\\w+")
#' 
#' # version with simplify = FALSE => a list is returned
#' string_extract(greet, "w/[[:upper:]]\\w+", simplify = FALSE)
#' 
string_extract = function(x, pattern, single = FALSE, simplify = TRUE, fixed = FALSE,
                          ignore.case = FALSE, word = FALSE, unlist = FALSE, 
                          envir = parent.frame()){
  
  x = check_set_character(x, mbt = TRUE, l0 = TRUE)

  check_logical(single, scalar = TRUE)
  check_logical(simplify, scalar = TRUE)
  check_logical(fixed, scalar = TRUE)
  check_logical(ignore.case, scalar = TRUE)
  check_logical(word, scalar = TRUE)
  check_logical(unlist, scalar = TRUE)
  
  pat_parsed = parse_regex_pattern(pattern, c("fixed", "word", "ignore", "magic", "single"),
                                   envir = envir, parse_logical = FALSE)
  
  flags = pat_parsed$flags
  
  is_single = single || "single" %in% flags
  is_fixed = fixed || "fixed" %in% flags
  is_word = word || "word" %in% flags
  is_ignore = ignore.case || "ignore" %in% flags
  is_magic = "magic" %in% flags
  
  info = format_simple_regex_flags(pat_parsed$patterns, 
                                   fixed = is_fixed, word = is_word, 
                                   ignore = is_ignore, magic = is_magic, envir = envir)
  
  pattern = info$pattern
  is_fixed = info$fixed
  
  if(length(x) == 0){
    if(is_single || simplify){
      return(character(0))
    } else {
      # simplify = FALSE and single = FALSE always return a list
      return(list())
    }    
  }
  
  if(is_single){
    x_pat = regexpr(pattern, x, fixed = is_fixed, perl = !is_fixed)
    
    res = substr(x, x_pat, x_pat - 1 + attr(x_pat, "match.length"))
  } else {
    res = regmatches(x, gregexpr(pattern, x, fixed = is_fixed, perl = !is_fixed))
    
    if(unlist || (simplify && length(x) == 1)){
      res = unlist(res)
    }
  }
  
  res
}


####
#### dedicated utilities ####
####



#' `stringmagic`'s regular expression parser
#' 
#' Parse regular expression with custom flags and obtain the final pattern
#' to be parsed as well as the vector of flags
#' 
#' @param pattern Character scalar, the regular expression pattern to parse.
#' @param authorized_flags Character vector representing the flags to be parsed.
#' Use the empty string if no flags are allowed.
#' @param parse_flags Logical scalar, default is `TRUE`. Whether to parse the 
#' optional regex flags.
#' @param parse_logical Logical scalar, default is `TRUE`. Whether to parse logical 
#' regex operations, similarly to [string_get()].
#' @param envir An environment, default is `parent.frame()`. Only used if the flag 
#' `magic` is present, it is used to find the variables to be interpolated.
#' 
#' @details 
#' This is an internal tool that is exposed in order to facilitate checking what's going on.
#' 
#' There is no error handling.
#' 
#' @return 
#' This function always returns a list of 4 elements:
#' 
#' + `flags`: the character vector of flags. If no flags were found, this is the empty string.
#' + `patterns`: the vector of regex patterns.
#' + `is_or`: logical vector of the same length as `patterns`. Indicates for each
#' pattern if it should be attached to the previous patterns with a logical OR (`FALSE`
#' means a logical `AND`).
#' + `is_not`: logical vector of the same length as `patterns`. Indicates for each pattern
#' if it should be negated.
#' 
#' @author 
#' Laurent Berge
#' 
#' @inherit string_clean seealso
#' 
#' @examples 
#' 
#' parse_regex_pattern("f/hello", c("fixed", "ignore"))
#' 
#' x = "john"
#' parse_regex_pattern("fm/{x} | Doe", c("fixed", "ignore", "magic"))
#' 
parse_regex_pattern = function(pattern, authorized_flags, parse_flags = TRUE, 
                               parse_logical = TRUE, envir = parent.frame()){
  # in: "fw/hey!, bonjour, a[i]"
  # common authorized_flags: c("fixed", "word", "ignore")

  info_pattern = cpp_parse_regex_pattern(pattern, parse_flags = parse_flags, 
                                         parse_logical = parse_logical)
  
  if(!is.null(info_pattern$error)){
    main_msg = .sma("Problem found in the regex pattern {'50|..'k, Q ? pattern}.",
                    "\nPROBLEM: ")
    extra = info_pattern$error_extra
    msg = switch(info_pattern$error,
                 "flag starting with space" = "flags cannot start with a space.", 
                 "flags list is too long" = 
                    .sma("the flag list is too long, ",
                         "most likely they are no flags.\nFIX? start with a slash or escape the slash",
                         "with a double backslash."), 
                 "non valid character at position" = 
                    .sma("the flag list contains a non ",
                        "valid character in position {extra} ",
                        "({`extra`firstchar, lastchar, bq ? pattern}). ",
                        "Flags must only consist of lower case letters followed by a comma or a slash."),
                "flag starts with !" = 
                    .sma("You cannot use the '!' operator before the flags."), 
                 "flag empty" = 
                    .sma("The {Nth ? extra} flag is empty. ", 
                         "{&extra==1;To use a regular slash, escape it with a double backslah.;",
                         "Flags must only consist of lower case letters followed by ",
                         "a comma or a slash.}"))
    
    note_rm = .sma("\n\nINFO: regex flags write 'flag1, flag2/pattern'",
                    " where a flag = lower case letters.",
                    "\n  Ex: 'ignore, fixed/dt[' leads",
                    " to the flags 'ignore' and 'fixed' for the pattern 'dt['.", 
                    "\n  Or simply collate the first letters of the flags:",
                    " 'if/dt['.", 
                    "\n  To use '/' without flag parsing, escape it. ",
                    "Ex: '\\\\/usr': leads to '/usr' without any flag on.")
    
    full_msg = .sma(main_msg, msg, note_rm)
    
    stop_hook(full_msg)
  }

  flags = info_pattern$flags
  if(length(flags) == 1){
    if(!flags %in% authorized_flags){
      # is it a flag made only of initials?
      flag_split = strsplit(flags, "")[[1]]
      authorized_initials = substr(authorized_flags, 1, 1)
      if(all(flag_split %in% authorized_initials)){
        flags = authorized_flags[authorized_initials %in% flag_split]
      }
    }
  }

  flags = check_set_options(flags, authorized_flags, free = TRUE)

  flag_pblm = setdiff(flags, authorized_flags)
  if(length(flag_pblm) > 0){
    info = setdiff(authorized_flags, flags)
    stop_hook("In the pattern {'20||...'k, bq?pattern} the flag {enum.bq?flag_pblm} is not authorized.",
              "\nFYI the authorized flags are: {enum.bq?authorized_flags}.",
              "\nNOTA: The syntax is \"flag1, flag2/regex\" or \"flag_initials/regex\".")
  }
  
  if("magic" %in% flags){
    for(i in seq_along(info_pattern$patterns)){
      p = info_pattern$patterns[i]
      p_escaped = gsub("(\\{( *\\d| *,))", "\\\\\\1", p)
      # we try the magic evaluation
      p_new = try(string_magic(p_escaped, .envir = envir), silent = TRUE)
      if(isError(p_new)){
        stop_hook("CONTEXT: concerns the pattern {bq?pattern}",
                  "{&len(info_pattern)>1;\n         when evaluating {bq?p}}",
                  "\nINFO: The `magic` flag expands the pattern with `string_magic`.",
                  "\nPROBLEM: the evaluation with `string_magic` failed, see error below:",
                  "\n{'^[^\n]+\n'r?p_new}")
      }
      
      if(length(p_new) != 1){
        stop_hook("CONTEXT: concerns the pattern {bq?pattern}",
                  "{&len(info_pattern)>1;\n         when evaluating {bq?p}}",
                  "\nINFO: The `magic` flag expands the pattern with `string_magic`. It must return a vector of length 1.",
                  "\nPROBLEM: the vector returned is of length {len?p_new}.")
      }
      
      info_pattern$patterns[i] = p_new
    }
  }

  res = list(flags = flags, patterns = info_pattern$patterns, is_or = info_pattern$is_or,
             is_not = info_pattern$is_not)

  res
}

format_simple_regex_flags = function(pattern, ignore = FALSE, word = FALSE, fixed = FALSE, 
                                     magic = FALSE, envir = parent.frame(2)){
  
  if(magic){
    new_regex = parse_regex_pattern(pattern, c("ignore", "word", "fixed", "magic"), 
                                    parse_logical = FALSE, envir = envir)
  } else {
    new_regex = parse_regex_pattern(pattern, c("ignore", "word", "fixed"), parse_logical = FALSE)
  }  
  
  pattern = new_regex$patterns
  is_ignore = "ignore" %in% new_regex$flags || ignore
  is_word = "word" %in% new_regex$flags || word
  is_fixed = "fixed" %in% new_regex$flags || fixed
  
  pattern = format_pattern(pattern, fixed = is_fixed, word = is_word, ignore = is_ignore)
  is_fixed = attr(pattern, "fixed")
    
  res = list(pattern = pattern, fixed = is_fixed)
  
  return(res)
}

format_pattern = function(pattern, fixed, word, ignore){

  if(word){
    items = strsplit(pattern, ",[ \t\n]+")[[1]]
    if(fixed){
      items = paste0("\\Q", items, "\\E")
      fixed = FALSE
    }
    pattern = paste0("\\b(", paste0(items, collapse = "|"), ")\\b")
  }

  if(ignore){
    if(fixed){
      fixed = FALSE
      pattern = paste0("(?i)\\Q", pattern, "\\E")
    } else {
      pattern = paste0("(?i)", pattern)
    }          
  }
  
  attr(pattern, "fixed") = fixed
  
  pattern
}


####
#### Aliases ####
####


#' @describeIn string_ops Alias to `string_ops`
stops = string_ops

#' @describeIn string_is Alias to `string_is`
stis = string_is

#' @describeIn string_is Alias to `string_any`
stany = string_any

#' @describeIn string_is Alias to `string_all`
stall = string_all

#' @describeIn string_is Alias to `string_which`
stwhich = string_which

#' @describeIn string_get Alias to `string_get`
stget = string_get

#' @describeIn string_split Alias to `string_split`
stsplit = string_split

#' @describeIn string_clean Alias to `string_clean`
stclean = string_clean

#' @describeIn string_clean Alias to `string_replace`
streplace = string_replace

#' @describeIn string_vec Alias to `string_vec`
stvec = string_vec

#' @describeIn string_extract Alias to `string_extract`
stextract = string_extract

#------------------------------------------------------------------------------#
# Author: Laurent R. Bergé
# Created: 2023-06-27
# ~: Alias generators
#------------------------------------------------------------------------------#


#' Create `string_magic` aliases with custom defaults
#' 
#' Utility to easily create `string_magic` aliases with custom default
#' 
#' @inheritParams string_magic
#' 
#' @param .local_ops Named list or `NULL` (default). If provided, it must be a list 
#' of the form `list(alias1 = ops1, alias2 = ops2)` where alias is the name of the newly defined 
#' operator an `ops` is a character scalar representing the associated string_magic operations.
#' Ex: `list(add = "' + 'collapse")` creates the operation `add` which collapses the 
#' string with pluses. All operations created here are only available to the
#' generated function.
#' 
#' @details 
#' 
#' Use this function if you want to change `string_magic` default values. For example,
#' if you want the interpolation to be done with `"{{}}"` (instead of `{}`) or if you want the 
#' default separation to be the space (instead of the empty string). See the example.
#' 
#' @return 
#' This function returns a function which will behave in the same way as [string_magic()]
#' 
#' 
#' @inheritSection string_magic_register_fun Writing a package using `string_magic`
#' 
#' @author 
#' Laurent Berge
#' 
#' @family related to string_magic
#' @family tools with aliases
#' 
#' @examples 
#' 
#' # we create the function sma2 with different defaults
#' sma2 = string_magic_alias(.delim = ".[ ]", .sep = " ", .class = "string_magic")
#' 
#' person = "john doe"
#' sma2("Hello", ".[title ? person]")
#' 
#' # you can use the arguments whose default has been changed
#' sma2("Hello", ".[title ? person]", .sep = ": ")
#' 
#' 
#' 
string_magic_alias = function(.sep = "", .vectorize = FALSE, 
                              .delim = c("{", "}"), .last = NULL, 
                              .post = NULL, .default = NULL, .nest = FALSE,
                              .invisible = FALSE, .local_ops = NULL,
                              .collapse = NULL,  .check = TRUE, 
                              .class = NULL, .namespace = NULL){
  
  # checks
  check_character(.sep, scalar = TRUE)
  check_logical(.vectorize, scalar = TRUE)
  .delim = check_set_delimiters(.delim)
  check_last(.last)
  check_function(.post, null = TRUE)
  check_character(.default, scalar = TRUE, null = TRUE)
  check_logical(.nest, scalar = TRUE)
  check_logical(.invisible, scalar = TRUE)
  # .local_ops, see below
  check_character(.collapse, scalar = TRUE, null = TRUE)
  check_logical(.check, scalar = TRUE)
  check_character(.class, no_na = TRUE, null = TRUE)
  check_character(.namespace, scalar = TRUE, null = TRUE)
  
  # .local_ops
  if(!missnull(.local_ops)){
    info = .sma("\nINFO: names of the list = alias. Content = string_magic operations.",
                "\nEx: .local_ops = list(\"plus\" = \"' + 'collapse\")")
    
    if(!is.list(.local_ops)){
      stopi("The argument `.local_ops` must be a list. ",
            "Currently it is of class {enum.bq?class(.local_ops)}.", info)
    }
    if(length(.local_ops) == 0){
      stopi("The argument `.local_ops` must be a non-empty list. ",
            "Currently it is empty.", info)
    }
    if(is.null(names(.local_ops))){
      stopi("In the argument `.local_ops` require names.",
            "\nPROBLEM: the list has no names.", info)
    }
    if(string_any(names(.local_ops), "[^[:lower:]]")){
      stopi("In the argument `.local_ops`, the aliases must be composed of ",
            "only lower case letters.",
            "\nPROBLEM: this is not the case for ",
            "{'[^[:lower:]]'get, enum.bq ? names(.local_ops)}.", info)
    }
    if(!all(sapply(.local_ops, is.character))){
      i_pblm = which(!sapply(.local_ops, is.character))[1]
      stopi("In the argument `.local_ops`, each element of the list must ",
            "be a character scalar.",
            "\nPROBLEM: the {nth?i_pblm} element is of class ",
            "{enum.bq?class(.local_ops[i_pblm]}.", info)
    }
    if(any(lengths(.local_ops) != 1)){
      i_pblm = which(lengths(.local_ops) != 1)[1]
      stopi("In the argument `.local_ops`, each element of the list must ",
            "be a character scalar.",
            "\nPROBLEM: the {nth?i_pblm} element is of length {len?.local_ops[i_pblm]}.", 
            info)
    }
    
    if(is.null(.namespace)){
      .namespace = tag_gen()
    }   
    
    for(i in seq_along(.local_ops)){
      alias = names(.local_ops)[[i]]
      content = .local_ops[[i]]
      string_magic_register_ops(content, alias, .namespace)
    }    
  }
  
  
  # forcing evaluations (INDISPENSABLE)
  sep = .sep
  vectorize = .vectorize
  delim = .delim
  last = .last
  post = .post
  default = .default
  nest = .nest
  invisible = .invisible
  collapse = .collapse
  check = .check
  class = .class
  namespace = .namespace
  
  if(.nest && .vectorize){
    stop("You cannot have `.nest` and `.vectorize` at the same time. One of the two must be set to `FALSE`.")
  }
  
  res = function(..., .envir = parent.frame(), .sep = sep, .vectorize = vectorize, 
                 .delim = delim, .last = last, .post = post, .default = default,
                 .nest = nest, .invisible = invisible, .collapse = collapse, 
                 .check = check, .class = class, .help = NULL, 
                 .namespace = namespace){
                    
    string_magic(..., .envir = .envir, .sep = .sep, .vectorize = .vectorize, 
                 .delim = .delim, .last = .last, .post = .post, .default = .default,
                 .nest = .nest, .invisible = .invisible, .collapse = .collapse,
                 .check = .check, .class = .class, .help = .help,
                 .namespace = .namespace)
  }
  
  res
}


#' @describeIn cat_magic Create an alias of `cat_magic` with custom defaults
cat_magic_alias = function(.sep = "", .end = "", .width = FALSE, .leader = "", 
                           .delim = c("{", "}"), .last = NULL, 
                           .trigger = TRUE, .check = TRUE, 
                           .namespace = NULL){
  
  # checks
  check_character(.sep, scalar = TRUE)
  check_character(.end, scalar = TRUE)
  check_character(.leader, scalar = TRUE)
  .delim = check_set_delimiters(.delim)
  check_character(.last, scalar = TRUE, null = TRUE)
  check_logical(.trigger, scalar = TRUE)
  check_logical(.check, scalar = TRUE)
  check_character(.namespace, scalar = TRUE, null = TRUE)
  
  # width is special
  is_call = isTRUE(try(is.call(.width), silent = TRUE))
  if(!is_call){
    .width = substitute(.width)
    if(!".sw" %in% all.vars(.width)){
      .width = eval(.width, parent.frame())
    }
  }
  
  sep = .sep
  end = .end
  leader = .leader
  last = .last
  trigger = .trigger
  check = .check
  namespace = .namespace
  width = .width
  delim = .delim
  
  res = function(..., .sep = sep, .end = end, .width = width, .leader = leader, 
                 .envir = parent.frame(), delim = delim, .last = last, 
                 .trigger = trigger, .check = check, .help = NULL, 
                 .namespace = namespace){
    cat_magic(..., .sep = .sep, .end = .end, .width = .width, .leader = .leader, 
              .envir = .envir, .delim = .delim, .last = .last, 
              .trigger = .trigger, .check = .check, .help = .help, 
              .namespace = .namespace)
  }
  
  res
}

#' @describeIn cat_magic Create an alias of `message_magic` with custom defaults
message_magic_alias = function(.sep = "", .end = "\n", .width = ~min(100, .sw), .leader = "", 
                               .delim = c("{", "}"), .last = NULL, .trigger = TRUE, 
                               .check = TRUE, .namespace = NULL){
  
  # checks
  check_character(.sep, scalar = TRUE)
  check_character(.end, scalar = TRUE)
  check_character(.leader, scalar = TRUE)
  .delim = check_set_delimiters(.delim)
  check_character(.last, scalar = TRUE, null = TRUE)
  check_logical(.check, scalar = TRUE)
  check_logical(.trigger, scalar = TRUE)
  check_character(.namespace, scalar = TRUE, null = TRUE)
  
  # width is special
  is_call = isTRUE(try(is.call(.width), silent = TRUE))
  if(!is_call){
    .width = substitute(.width)
    if(!".sw" %in% all.vars(.width)){
      .width = eval(.width, parent.frame())
    }
  }
  
  # forcing the eval
  sep = .sep
  end = .end
  leader = .leader
  last = .last
  trigger = .trigger
  check = .check
  namespace = .namespace
  width = .width
  delim = .delim
  
  res = function(..., .sep = sep, .end = end, .width = width, .leader = leader, 
                 .envir = parent.frame(), .delim = delim, .last = last, 
                 .trigger = trigger,
                 .check = check, .help = NULL, 
                 .namespace = namespace){

    message_magic(..., .sep = .sep, .end = .end, .width = .width, .leader = .leader, 
                  .envir = .envir, .delim = .delim, .last = .last, 
                  .trigger = .trigger, .check = .check, .help = .help, 
                  .namespace = .namespace)
  }
  
  res
}

#' @describeIn string_ops `string_ops` alias with custom defaults
string_ops_alias = function(op = NULL, pre_unik = NULL, namespace = NULL){
  #
  
  check_character(op, scalar = TRUE, null = TRUE)  
  check_logical(pre_unik, null = TRUE, scalar = TRUE)
  check_character(namespace, scalar = TRUE, null = TRUE)  
  
  # forcing evaluations
  .op = op
  .pre_unik = pre_unik
  .namespace = namespace
  
  res = function(x, op = .op, pre_unik = .pre_unik, namespace = .namespace){                  
    string_ops(x, op = op, pre_unik = pre_unik, namespace = namespace)
  }
  
  res
}

#' @describeIn string_clean Create a `string_clean` alias with custom defaults
string_clean_alias = function(replacement = "", pipe = " => ", split = ",[ \n\t]+", 
                              ignore.case = FALSE, fixed = FALSE, word = FALSE, 
                              total = FALSE, single = FALSE, 
                              namespace = NULL){
  
  check_character(replacement, scalar = TRUE)
  check_character(pipe, scalar = TRUE)
  check_character(split, scalar = TRUE, null = TRUE)
  check_logical(ignore.case, scalar = TRUE)
  check_logical(fixed, scalar = TRUE)
  check_logical(word, scalar = TRUE)
  check_logical(total, scalar = TRUE)
  check_logical(single, scalar = TRUE)
  check_character(namespace, scalar = TRUE, null = TRUE)
  
  # forcing the evaluations
  .replacement = replacement
  .pipe = pipe
  .split = split
  .ignore.case = ignore.case
  .fixed = fixed
  .word = word
  .total = total
  .single = single
  .namespace = namespace
  
  res = function(x, ..., replacement = .replacement, pipe = .pipe, split = .split, 
                 ignore.case = .ignore.case, fixed = .fixed, word = .word, 
                 total = .total, single = .single, envir = parent.frame(), 
                 namespace = .namespace){                  

    string_clean(x, ..., replacement = replacement, pipe = pipe, split = split, 
                 ignore.case = ignore.case, fixed = fixed, word = word, 
                 total = total, single = single, envir = envir, 
                 namespace = namespace)
  }
  
  res  
}

#' @describeIn string_vec Create `string_vec` aliases with custom defaults
string_vec_alias = function(.cmat = FALSE, .nmat = FALSE, .df = FALSE, .df.convert = TRUE, 
                            .last = NULL, .delim = c("{", "}"), .split = TRUE, 
                            .protect.vars = FALSE, .check = TRUE, .sep = NULL, 
                            .collapse = NULL, .namespace = NULL){
  
  .delim = check_set_delimiters(.delim)
  check_character(.sep, scalar = TRUE, null = TRUE)
  check_character(.collapse, scalar = TRUE, null = TRUE)
  check_character(.last, scalar = TRUE, null = TRUE)
  check_logical(.protect.vars, scalar = TRUE)
  check_logical(.check, scalar = TRUE)
  check_character(.namespace, scalar = TRUE, null = TRUE)
  .split = check_set_split(.split)
  check_set_mat(.cmat, .nmat, .df)
  check_logical(.df.convert, scalar = TRUE)
  
  # forcing evaluation
  nmat = .nmat
  cmat = .cmat
  df = .df
  df.convert = .df.convert
  last = .last
  delim = .delim
  split = .split
  protect.vars = .protect.vars
  check = .check
  sep = .sep
  collapse = .collapse
  namespace = .namespace  
  
  res = function(..., .cmat = cmat, .nmat = nmat, .df = df, .df.convert = df.convert,
                 .last = last, .delim = delim, .envir = parent.frame(), 
                 .split = split, .protect.vars = protect.vars, 
                 .check = check, .sep = sep, 
                 .collapse = collapse, .namespace = namespace){

    string_vec(..., .cmat = .cmat, .nmat = .nmat, .df = .df, .df.convert = .df.convert, 
               .last = last, .delim = .delim, .envir = .envir, 
               .split = .split, .protect.vars = .protect.vars, 
               .check = .check, .sep = .sep, 
               .collapse = .collapse, .namespace = .namespace)
  }
  
  res
}

#' String interpolation with operation chaining
#'
#' This is firstly a string interpolation tool. On top of this it can apply, and chain, over 50
#' basic string operations to the interpolated variables. Advanced support for pluralization.
#' 
#' @name string_magic
#'
#' @param ... Character scalars that will be collapsed with the argument `sep`. Note that 
#' named arguments are used for substitution.
#' 
#' To interpolate, you can 
#' use `"{x}"` within each character string to insert the value of `x` in the string. 
#' You can add string operations in each `"{}"` instance with the syntax `"'arg'op ? x"` 
#' (resp. `"'arg'op ! x"`) to apply the operation `'op'` with the argument `'arg'` to `x`
#'  (resp. the verbatim of `x`). Otherwise, what to say? Ah, nesting is enabled, and since 
#' there's over 50 operators, it's a bit complicated to sort you out in this small space. 
#' 
#' Use the argument `.help = "keyword"` (or `.help = TRUE`) to obtain a selective help 
#' from the main documentation.
#'
#' Note that in interpolations you have access to the special variables: `.now` and `.date`
#' to get the current time; and the special function `.now("format")` to 
#' format the time. Ex: `.now('%Y-%m %H:%m')`.
#' @param .envir An environment used to evaluate the variables in `"{}"`. By default the variables are
#' evaluated using the environment from where the function is called or using the named 
#' arguments passed to the function.
#' @param .data A list used to evaluate the variables in `"{}"`. Default is the empty list.
#' By default the variables are evaluated using the environment from where the 
#' function is called or using the named arguments passed to the function.
#' @param .sep Character scalar, default is the empty string `""`. It is used to collapse all
#'  the elements in `...` before applying any operation.
#' @param .vectorize Logical scalar, default is `FALSE`. If `TRUE`, Further, elements in `...` are 
#' NOT collapsed together, but instead vectorised.
#' @param .help Character scalar or `TRUE`, default is `NULL`. This argument
#' is used to generate a dynamic help on the console. If `TRUE`, the user can select which
#' topic to read from the main documentation, with the possibility to search for keywords and
#' navigate the help pages. If a character scalar, then a regex search is perfomed on the main
#' documentation and any section containining a match is displayed. The user can easily
#' navigate across matches.
#' @param .collapse Character scalar, default is `NULL`. If provided, the character vector
#' that should be returned is collapsed with the value of this argument. This leads
#' to return a string of length 1.
#' @param .check Logical scalar, default is `TRUE`. Whether to enable error-handling (i.e.
#' human readable error messages). 
#' Without error-handling you can save something of the order of 40us.Useful only
#' in long loops.
#' @param .delim Character vector of length 1 or 2. Default is `c("{", "}")`. Defines 
#' the opening and the closing delimiters for interpolation. 
#' 
#' If of length 1, it must be of the form: 1) the opening delimiter, 
#' 2) a single space, 3) the closing delimiter. Ex: `".[ ]"` is equivalent to `c(".[", "]")`.
#' The default value is equivalent to `"{ }"`.
#' @param .default Logical scalar, default is `TRUE`. Whether to use the global defaults 
#' set with the function [setstring_magic()]. If `FALSE`, then the default value of the arguments
#' is guaranteed to be ones of the function definition.
#' @param .class Character vector representing the class to give to the object returned. 
#' By default it is `NULL`. Note that the class `string_magic` has a specific `print` method, usually
#' nicer for small vectors (it [base::cat()]s the elements).
#' @param .last Character scalar, a function, or `NULL` (default). If provided and character: 
#' it must be an `string_magic` chain of operations of the form `"'arg1'op1, op2, etc"`. All these operations
#' are applied just before returning the vector. If a function, 
#' it will be applied to the resulting vector.
#' @param .post Function or `NULL` (default). If not `NULL`, this function will be applied 
#' after all the processing, just before returning the object. This function can have 
#' extra arguments which will be caught directly in the `...` argument of `string_magic`.
#' For example if `.post = head`, you can directly pass the argument `n = 3` to `string_magic`'s arguments.
#' @param .invisible Logical scalar, default is `FALSE`. Whether the object returned should be 
#' invisible (i.e. not printed on the console). 
#' @param .default Character scalar or `NULL` (default). If provided, it must be 
#' a sequence of `string_magic` operations. It will be applied as a default to any interpolation.
#' Ex: if `x = 1:2`, then `string_magic("x = {x}", .default = "enum")` leads to "x = 1 and 2", 
#' and is equivalent to `string_magic("x = {enum?x}")`. Note that this default operations does 
#' not apply to nested expressions. That is `string_magic("{!x{1:2}}", .default = "enum")` leads 
#' to `c("x1", "x2")` and NOT `"x1 and 2"`.
#' @param .namespace Character scalar or `NULL` (default). **Only useful for package developers.**
#' As a regular end-user you shouldn't care! If your package uses `string_magic`, you should care. 
#' It is useful **only** if your package uses 'custom' `string_magic` operations, set with 
#' [string_magic_register_fun()] or [string_magic_register_ops()].
#' @param .nest Logical, default is `FALSE`. If `TRUE`, it will nest the original string within 
#' interpolation delimiters, so that you can apply operations directly on the string. Example:
#' `string_magic("upper ! hello")` returns "upper ! hello", while `string_magic("upper ! hello", .nest = TRUE)`
#' returns `"HELLO"`.
#' @param .trigger Logical, default is `TRUE`. If `FALSE`, this function is not run. Can be 
#' useful in debugging situations where we want conditional evaluations.
#' 
#'
#' @details 
#' There are over 50 basic string operations, it supports pluralization, string operations can be 
#' nested, operations can be applied group-wise or conditionally and
#' operators have sensible defaults. 
#' You can also declare your own operations with [string_magic_register_fun()] or [string_magic_register_ops()]. 
#' They will be seamlessly integrated to `string_magic`.
#' 
#' The function `.string_magic` (prefixed with a dot) is a leaner version of the function `string_magic`.
#' It does the same operations but with the following differences:
#' 
#'  + there is no error handling: meaning that the error messages, if any, will be poor and
#' hard to understand
#'  + default options are not applied: hence the user must always explicitly provide the arguments
#' 
#' This leads to a faster processing time (of about 50 microseconds) at the cost of user experience.
#' 
#' If you want to change the default values of `string_magic` (like changing the delimiter), use
#' the function [string_magic_alias()].
#'
#' Use the argument `.help` to which
#' you can pass keywords or regular expressions and fecth select pieces from the main documentation.
#' 
#' 
#' @section Interpolation and string operations: Principle:
#' 
#' To interpolate a variable, say `x`, simply use `{x}`. For example `x = "world"; string_magic("hello {x}")` leads 
#' to "hello world".
#' 
#' To any interpolation you can add operations. Taking the previous example, say we want to display
#'  "hello W O R L D". This means upper casing all letters of the interpolated variable and adding a space between 
#' each of them. Do you think we can do that? Of course yes: 
#' `string_magic("hello {upper, ''s, c ? x}")`. And that's it.
#' 
#' Now let's explain what happened. Within the `{}` *box*, we first write a set of 
#' operations, here "upper, ''s, c", then add "?" and finally write 
#' the variable to interpolate, "x".  The operations (explained in more details 
#' below) are `upper`, upper-casing all letters, ''s: splitting
#' with the empty string, 'c': concatenating with spaces the vector of string that was just split.
#' The question mark means that the expression coming after it is to be evaluated 
#' (this is opposed to the exclamation mark presented next).
#' 
#' The syntax is always the same: `{operations ? expression}`, where the operations section
#' is a *comma separated* list of operations.
#' These operations are of the form `'arg'op`, with `arg` the argument to the operator 
#' code `op`. These operations are performed sequantially from left to right.
#' 
#' Some operations, like `upper`, accept options. You attach options to an operation 
#' with a dot followed by the option name. Formally: `op.option1.option2`, etc.
#' Example: `x = "hi there. what's up? fine." ; string_magic("He said: {upper.sentence, Q ? x}")`.
#' Leads to: `He said: "Hi there. What's up? Fine."`.
#' 
#' Both operators and options are partially matched. So `string_magic("He said: {up.s, Q ? x}")` would 
#' also work.
#' 
#' @section  Verbatim interpolation and nesting: Principle:
#' 
#' Instead of interpolating a variable, say `x`, with `{x}`, you can use an exclamation 
#' mark to trigger varbatim evaluation.
#' For example `string_magic("hello {!x}")` would lead to "hello x". It's a
#'  bit disappointing, right? What's the point of doing that? Wait until the next two paragraphs.
#' 
#' Verbatim evaluation is a powerful way to apply operations to plain text. For example:
#'  `string_magic("hello {upper, ''s, c ! world}")` leads to "hello W O R L D".
#' 
#' (A note in passing. The spaces surrounding the exclamation mark are non necessary,
#'  but when one space is present on both sides of the `!`, then the verbatim
#' expression only begins after it. Ex: `"{upper! hi}"` leads to " HI" while `"{upper ! hi}"` 
#' leads to "HI" and `"{upper !  hi}"` leads to " HI".)
#' 
#' The second advantage of verbatim evaluations is *nesting*. Anything in a verbatim 
#' expression is evaluated with the function `string_magic`.
#' This means that any *box* will be evaluated as previously described. Let's
#'  give an example. You want to write the expression of a polynomial of order n: a + bx + cx^2 + etc.
#' You can do that with nesting. Assume we have `n = 2`.
#' 
#' Then `string_magic("poly({n}): {' + 'c ! {letters[1 + 0:n]}x^{0:n}}")` leads to 
#' "poly(2): ax^0 + bx^1 + cx^2".
#' 
#' How does it work? The verbatim expression (the one following the exclamation mark),
#'  here `"{letters[1 + 0:n]}x^{0:n}"`, is evaluated with `string_magic`.
#' `string_magic("{letters[1 + 0:n]}x^{0:n}")` leads to the vector c("ax^0", "bx^1", "cx^2").
#' 
#' The operation `' + 'c` then concatenates (or collapses) that vector with ' + '.
#'  This value is then appended to the previous string.
#' 
#' We could refine by adding a cleaning operation in which we replace "x^0" and "^1" 
#' by the empty string. Let's do it:
#' 
#' `string_magic("poly({n}): {' + 'c, 'x\\^0|\\^1'r ! {letters[1 + 0:n]}x^{0:n}}")` leads to 
#' "poly(2): a + bx + cx^2", what we wanted.
#' 
#' You can try to write a function to express the polynomial as before: although it is 
#' a simple task, my guess is that it will require more typing.
#' 
#' @section Operations: General syntax:
#' 
#' As seen in the previous sections, within a *box* (i.e. `"{}"`), multiple operations
#'  can be performed.
#' We can do so by stacking the operations codes and in a comma separated enumeration.
#' Operations can have arguments, and operations can also have options. The general 
#' syntax, with argument and options, is:
#' 
#' `{'arg1'op1.optionA.optionB, arg2 op2.optionC, `arg3`op3, 51op4 ? x}`
#' 
#' The argument can appear in four forms: a) inside single or double quotes just 
#' before the operation name (`arg1` above), 
#' b) verbatim, separated with a space, just before the operation name (`arg2` above), 
#' c) inside bactick quotes the argument is evaluated from the environment (`arg3` above),
#' or d) when the argument is an integer it can be juxtaposed to the opeation name (like in `op4` above).
#' 
#' The options are always dot separated and attached to the operation name, they are 
#' specific to each operation.
#' 
#' Both the operation name and the option names are partially matched.
#' 
#' @section Basic string operations:
#' 
#' This section describes some of the most common string operations: extracting, replacing, collapsing, splitting, etc.
#' These functions accept generic flags ("ignore", "fixed", "word") in their patterns (syntax: "flags/pattern"). 
#' Please see the dedicated section for more information on flags.
#' 
#' + s, split, S, Split: splits the string according to a pattern. 
#' The operations have different defaults: `' '` 
#' for `s` and 'split', and `',[ \t\n]*'` for `S` and 'Split' (i.e. comma separation). 
#' Ex.1: `string_magic("{S ! romeo, juliet}")` leads to the vector c("romeo", "juliet"). 
#' Ex.2: `string_magic("{'f/+'s, '-'c ! 5 + 2} = 3")` leads to "5 - 2 = 3" (note the flag "fixed" in `s`'s pattern).
#' + c, C: to concatenate multiple strings into a single one. The two operations are 
#' identical, only their default change. c: default is `' '`, C: default is `', | and '`.
#'   The syntax of the argument is 's1' or 's1|s2'. s1 is the string used to concatenate 
#' (think `paste(x, collapse = s1)`). In arguments of the form `'s1|s2'`, `s2` will be used to concatenate the last two elements. 
#' Ex.1: `x = 1:4; string_magic("Et {' et 'c ? x}!")` leads to "Et 1 et 2 et 3 et 4!".
#' Ex.2: `string_magic("Choose: {', | or 'c ? 2:4}?")` leads to "Choose: 2, 3 or 4?".
#' + x, X: extracts patterns from a string. Both have the same default: `'[[:alnum:]]+'`. 
#' `x` extracts the first match while `X` extracts **all** the matches.
#'   Ex.1: `x = c("6 feet under", "mahogany") ; string_magic("{'\\w{3}'x ? x}")` leads to the vector c("fee", "mah").
#'   Ex2.: `x = c("6 feet under", "mahogany") ; string_magic("{'\\w{3}'X ? x}")` leads to the
#'  vector c("fee", "und", "mah", "oga").
#' + extract: extracts multiple patterns from a string, this is an alias to the operation `X` described above.
#' Use the option "first" to extract only the first match for each string (behavior becomes like `x`).
#' Ex: `x = c("margo: 32, 1m75", "luke doe: 27, 1m71") ; string_magic("{'^\\w+'extract ? x} is {'\\d+'extract.first ? x}")`
#' leads to c("margo is 32", "luke is 27").
#' + r, R: replacement within a string. The two operations are identical and have no default.
#'  The syntax is `'old'` or `'old => new'` with `'old'` the pattern to find and `new` the replacement. If `new` is missing, it is 
#' considered the empty string. This operation also accepts the flag "total" which instruct to 
#' replace the fulll string in case the pattern is found.
#' Ex.1: `string_magic("{'e'r ! Where is the letter e?}")` leads to "Whr is th lttr ?".
#' Ex.2: `string_magic("{'(?<!\\b)e => a'R ! Where is the letter e?}")` leads to "Whara is tha lattar e?".
#' Ex.3: `string_magic("{'t/e => here'r ! Where is the letter e?}")` leads to "here".
#' + clean: replacement with a string. Similar to the operation `r`, except that here the comma is
#' a pattern separator, see detailed explanations in [string_clean()]. Ex: `string_magic("{'f/[, ]'clean ! x[a]}")` 
#' leads to "xa".
#' + get: restricts the string only to values respecting a pattern. This operation has no default.
#' Accepts the options "equal" and "in".
#' By default it uses the same syntax as [string_get()] so that you can use regex flags and 
#' include logical operations with `' & '` and `' | '` to detect patterns.
#' If the option "equal" is used, a simple string equality with the argument is tested (hence
#' no flags are accepted). If the option "in" is used, the argument is first split with respect to commas
#' and then set inclusion is tested. 
#' Example: `x = row.names(mtcars) ; string_magic("Mercedes models: {'Merc & [[:alpha:]]$'get, '^.+ 'r, C ? x}")`
#' leads to "Mercedes models: 240D, 280C, 450SE, 450SL and 450SLC".
#' + is: detects if a pattern is present in a string, returns a logical vector. This operation has no default.
#' Accepts the options "equal" and "in".
#' By default it uses the same syntax as [string_is()] so that you can use regex flags and 
#' include logical operations with `' & '` and `' | '` to detect patterns.
#' If the option "equal" is used, a simple string equality with the argument is tested (hence
#' no flags are accepted). If the option "in" is used, the argument is first split with respect to commas
#' and then set inclusion is tested. 
#' Mostly useful as the final operation in a [string_ops()] call.
#' Example: `x = c("Mark", "Lucas") ; string_magic("Mark? {'i/mark'is, C ? x}")` leads to "Mark? TRUE and FALSE".
#' + which: returns the index of string containing a specified pattern. With no default, can be applied
#' to a logical vector directly. 
#' By default it uses the same syntax as string_which() so that you can use regex flags and 
#' include logical operations with `' & '` and `' | '` to detect patterns.
#' If the option "equal" is used, a simple string equality with the argument is tested (hence
#' no flags are accepted). If the option "in" is used, the argument is first split with respect to commas
#' and then set inclusion is tested. 
#' Mostly useful as the final operation in a [string_ops()] call.
#' Ex.1: `x = c("Mark", "Lucas") ; string_magic("Mark is number {'i/mark'which ? x}.")` leads to 
#' "Mark is number 1.".
#' 
#' @section Operations changing the length or the order:
#' 
#' + first: keeps only the first `n` elements. Example: `string_magic("First 3 numbers: {3 first, C ? mtcars$mpg}.")`
#' leads to "First 3 numbers: 21, 21 and 22.8.". Negative numbers as argument remove the 
#' first `n` values. You can add a second argument in the form `'n1|n2'first` in which case the first `n1` and last
#' `n2` values are kept; `n1` and `n2` must be positive numbers.
#' + K, Ko, KO: keeps only the first `n` elements; has more options than `first`. The syntax is `'n'K`, 
#' `'n|s'K`, `'n||s'K`. The values Ko and KO only accept the two first syntax (with `n` only).
#' `n` provides the number of elements to keep. If `s` is provided and the number of 
#' elements are greater than `n`, then in 'n|s' the string `s` is added at the end, and
#' if 'n||s' the string s replaces the nth element.
#'   The string `s` accepts specials values:
#'   + `:n:` or `:N:` which gives the total number of items in digits or letters (N)
#'   + `:rest:` or `:REST:` which gives the number of elements that have been truncated in digits or letters (REST)
#'   Ex: `string_magic("{'3|:rest: others'K ? 1:200}")` leads to the vector `c("1", "2", "3", "197 others")`.
#'   + The operator 'n'Ko is like `'n||:rest: others'K` and 'n'KO is like `'n||:REST: others'K`.
#' + last: keeps only the last `n` elements. Example: `string_magic("Last 3 numbers: {3 last, C ? mtcars$mpg}.")`
#' leads to "Last 3 numbers: 19.7, 15 and 21.4.". Negative numbers as argument remove the 
#' last `n` values.
#' + sort: sorts the vector in increasing order. Accepts optional arguments and the option "num". 
#' Example: `x = c("sort", "me") ; string_magic("{sort, c ? x}")` leads to "me sort". 
#' If an argument is provided, it must be a regex pattern that will be applied to
#' the vector using [string_clean()]. The sorting will be applied to the modified version of the vector
#' and the original vector will be ordered according to this sorting. 
#' Ex: `x = c("Jon Snow", "Khal Drogo")`; `string_magic("{'.+ 'sort, C?x}")` leads to 
#' "Khal Drogo and Jon Snow". The option "num" sorts over a numeric version 
#' (with silent conversion) of the vector and reorders the original vector accordingly. 
#' Values which could not be converted are last.
#' **Important note**: the sorting operation is applied before any character conversion.
#' If previous operations were applied, it is likely that numeric data were transformed to character.
#' Note the difference: `x = c(20, 100, 10); string_magic("{sort, ' + 'c ? x}")` leads to "10 + 20 + 100"
#' while `string_magic("{n, sort, ' + 'c ? x}")` leads to "10 + 100 + 20" because the operation "n"
#' first transformed the numeric vector into character.
#' + dsort: sorts the vector in decreasing order. It accepts an optional argument and 
#' the option "num". Example: `string_magic("5 = {dsort, ' + 'c ? 2:3}")` 
#' leads to "5 = 3 + 2". See the operation "sort" for a description of the argument and the option.
#' + rev: reverses the vector. Example: `string_magic("{rev, ''c ? 1:3}")` leads to "321".
#' + unik: makes the string vector unique. Example: `string_magic("Iris species: {unik, C ? iris$Species}.")`
#' leads to "Iris species: setosa, versicolor and virginica.".
#' + table: computes the frequency of each element and attaches each element to its frequency. 
#' Accepts an argument which must be a character string representing a `string_magic` interpolation
#' with the following variables: `x` (the element), `n` (its count) and `s` (its share). The default is `'{x} ({n ? n})'`. By default the resulting string vector is sorted by decreasing frequency. 
#' You can change how the vector is sorted with five options: `sort` (sorts on the elements), 
#' `dsort` (decreasing sort), `fsort` (sorts on frequency), 
#' `dfsort` (decreasing sort on freq. -- default), 
#' `nosort` (keeps the order of the first elements). Note that you can combine several sorts 
#' (to resolve the ties of elements with same frequencies).
#' Example: `string_magic("Freq. of months: {'{x} ({n})'table, enum ? month.name[airquality$Month]}.")`
#' + each: repeats each element of the vector `n` times. Option "c" then collapses the full vector 
#' with the empty string as a separator. Ex.1: `string_magic("{/x, y}{2 each ? 1:2}")` leads to the 
#' vector `c("x1", "y1", "x2", "y2")`. Ex.2: `string_magic("Large number: 1{5 each.c ! 0}")` leads to 
#' "Large number: 100000".
#' + times: repeats the vector sequence `n` times. Option "c" then collapses the full vector 
#' with the empty string as a separator. Example: `string_magic("What{6 times.c ! ?}")` leads to "What??????".
#' + rm: removes elements from the vector. Options: "empty", "blank", "noalpha", "noalnum", "all".
#' The *optional* argument represents the pattern used to detect strings to be deleted. 
#' Ex.1: `x = c("Luke", "Charles")`; `string_magic("{'i/lu'rm ? x}")` leads to "charles". By default it removes
#' empty strings. Option "blank" removes strings containing only blank characters (spaces, tab, newline).
#' Option "noalpha" removes strings not containing letters. Option "noalnum" removes strings not 
#' containing alpha numeric characters. Option "all" removes all strings (useful in conditions, see 
#' the dedicated section). If an argument is provided, only the options "empty" and "blank" are available.
#' Ex.2: `x = c("I want to enter.", "Age?", "21")`; `string_magic("Nightclub conversation: {rm.noalpha, c ! - {x}}")` 
#' leads to "Nightclub conversation: - I want to enter. - Age?"
#' + nuke: removes all elements, equivalent to `rm.all` but possibly more explicit (not sure). 
#' Useful in conditions, see the dedicated section.
#' Example: `x = c(5, 7, 453, 647); string_magic("Small numbers only: {if(.>20 ; nuke), C ? x}")` leads 
#' to "Small numbers only: 5 and 7";
#' + insert: inserts a new element to the vector. Options: "right" and "both". Option "right" adds
#' the new element to the right. Option "both" inserts the new element on the two sides of the vector.
#' Example: `string_magic("{'3'insert.right, ' + 'c ? 1:2}")` leads to "1 + 2 + 3".
#' + `dp` or `deparse`: It deparses an R object and collapses it into a single string 
#' using newlines. By default it keeps only the first 50 lines of the deparsed object.
#' The number of lines kept can be modified by using a numeric argument.
#' It accepts the options: `short` and `vec`. If `short`, multiline deparsed elements
#' are collapsed with `\\\\n` and the number of characters cannot exceed the value of the
#' argument, which by default is 50.
#' If `vec`, then the deparsed element is not collapsed into a single string, and only 
#' the first 50 (default) elements of the vector are kept. You can increase/reduce the 
#' number of elements kept with the argument.
#' Example: `fml = y ~ x1 + x2; string_magic("The estimated model is {dp ? fml}.")`
#' 
#' @section Formatting operations:
#'  
#' + lower: lower cases the full string.
#' + upper: upper cases the full string. Options: "first" and "sentence".
#' Option "first" upper cases only the first character. Option "sentence"
#' upper cases the first letter after punctuation. 
#' Ex: `x = "hi. how are you? fine." ; string_magic("{upper.sentence ? x}")` leads
#' to "Hi. How are you? Fine.".
#' + title: applies a title case to the string. Options: "force" and "ignore".
#' Option "force" first puts everything to lowercase before applying the title case. 
#' Option "ignore" ignores a few small prepositions ("a", "the", "of", etc).
#' Ex: `x = "bryan is in the KITCHEN" ; string_magic("{title.force.ignore ? x}")` leads to "Bryan Is in the Kitchen".
#' + ws: normalizes whitespaces (WS). It trims the whitespaces on the edges and transforms any succession 
#' of whitespaces into a single one. Can also be used to further clean the string with its options. 
#' Options: "punct", "digit", "isolated". Option "punct" cleans the punctuation. Option "digit" cleans digits.
#' Option "isolated" cleans isolated letters. WS normalization always come after any of these options.
#' **Important note:** punctuation (or digits) are replaced with WS and **not** 
#' the empty string. This means that `string_magic("ws.punct ! Meg's car")` will become "Meg s car".
#' + trimws: trims the white spaces on both ends of the strings.
#' + q, Q, bq: to add quotes to the strings. q: single quotes, Q: double quotes, bq: 
#' back quotes. `x = c("Mark", "Pam"); string_magic("Hello {q, C ? x}!")` leads to "Hello 'Mark' and 'Pam'!".
#' + format, Format: applies the base R's function [base::format()] to the string. 
#' By default, the values are left aligned, *even numbers* (differently from [base::format()]'s behavior).
#' The upper case command (`Format`) applies right alignment. Options: "0", "zero", "left", "right", "center".
#' Options "0" or "zero" fills the blanks with 0s: useful to format numbers. Option "right" right aligns,
#' and "center" centers the strings. Default is left alignment.
#' Ex: `x = c(1, 12345); string_magic("left: {format.0, q, C ? x}, right: {Format, q, C ? x}")` 
#' leads to "left: '000001' and '12,345', right: '     1' and '12,345'".
#' + %: applies [base::sprintf()] formatting. The syntax is 'arg'% with arg an sprintf formatting,
#' or directly the sprint formatting, e.g. `% 5s`. Example: `string_magic("pi = {%.3f ? pi}")` leads
#' to "pi = 3.142".
#' + stopwords: removes basic English stopwords (the snowball list is used). 
#' The stopwords are replaced with an empty space but the left and right WS are 
#' untouched. So WS normalization may be needed (see operation `ws`).
#'   `x = c("He is tall", "He isn't young"); string_magic("Is he {stop, ws, C ? x}?")` leads to "Is he tall and young?".
#' + ascii: turns all letters into ASCII with transliteration. Failed translations
#'  are transformed into question marks. Options: "silent", "utf8". By default, if some conversion fails
#' a warning is prompted. Option "silent" disables the warning in case of failed conversion. The conversion 
#' is done with [base::iconv()], option "utf8" indicates that the source endocing is UTF-8, can be useful 
#' in some cases.
#' + round, r0 to r6: formats numbers by rounding at a given level. Options:  `0` to `9`, `int`, `nocomma`, `s0`-`s9`. 
#' Option `0` to `9` controls the number of digits to round at. 
#' Option `int` is whether to preserve integers from formattting.
#' Option `nocomma` conrtols whether to drop the comma separating the thousands.
#' `s0`-`s9` also keeps a given number of significant digits.
#' + signif, s0 to s6: formats numbers by displaying a certain number of significant digits. Options:  `0` to `9`, `int`, `nocomma`, `r0`-`r9`. 
#' Option `0` to `9` controls the number of significant digits to display. 
#' Option `int` is whether to preserve integers from formattting.
#' Option `nocomma` conrtols whether to drop the comma separating the thousands.
#' `r0`-`r9` also rounds at a given number of decimals.
#' + n: formats integers by adding a comma to separate thousands. Options: "letter", "upper", "0", "zero".
#' The option "letter" writes the number in letters (large numbers keep their numeric format). The option
#' "upper" is like the option "letter" but uppercases the first letter. Options "0" or "zero" left pads
#' numeric vectors with 0s. Ex.1: `x = 5; string_magic("He's {N ? x} years old.")` leads to "He's five years old.".
#' Ex.2: `x = c(5, 12, 52123); string_magic("She owes {n.0, '$'paste, C ? x}.")` leads to 
#' "She owes $5, $12 and $52,123.".
#' + N: same as `n` but automatically adds the option "letter".
#' + nth: when applied to a number, these operators write them as a rank. Options: "letter", 
#' "upper", "compact".
#' Ex.1: `n = c(3, 7); string_magic("They finished {nth, enum ? n}!")` leads to "They finished 3rd and 7th!". 
#' Option "letter" tries to write the numbers in letters, but note that it stops at 20. Option "upper"
#' is the same as "letter" but uppercases the first letter. Option "compact" aggregates
#' consecutive sequences in the form "start_n_th to end_n_th". 
#' Ex.2: `string_magic("They arrived {nth.compact ? 5:20}.")` leads to "They arrived 5th to 20th.".
#' Nth: same as `nth`, but automatically adds the option "letter". Example:
#' `n = c(3, 7); string_magic("They finished {Nth, enum ? n}!")` leads to "They finished third and seventh!".
#' + ntimes: write numbers in the form `n` times. Options: "letter", "upper". Option 
#' "letter" writes the number in letters (up to 100). Option "upper" does the same as "letter" 
#' and uppercases the first letter. Example: `string_magic("They lost {C ! {ntimes ? c(1, 12)} against {S!Real, Barcelona}}.")`
#' leads to "They lost once against Real and 12 times against Barcelona.".
#' + Ntimes: same as `ntimes` but automatically adds the option "letter".
#' Example: `x = 5; string_magic("This paper was rejected {Ntimes ? x}...")` leads to
#' "This paper was rejected five times...".
#' + firstchar, lastchar: to select the first/last characters of each element. 
#'   Ex: `string_magic("{19 firstchar, 9 lastchar ! This is a very long sentence}")` leads to "very long".
#' Negative numbers remove the first/last characters.
#' + k: to keep only the first n characters (like `firstchar` but with more options). The
#'  argument can be of the form `'n'k`, `'n|s'k` or `'n||s'k` with `n` a number and `s` a string.
#'   `n` provides the number of characters to keep. Optionnaly, only for strings whose
#'  length is greater than `n`, after truncation, the string `s` can be appended at the end.
#'   The difference between 'n|s' and 'n||s' is that in the second case the strings
#'  will always be of maximum size `n`, while in the first case they can be of length `n + nchar(s)`.
#'   Ex: `string_magic("{4k ! long sentence}")` leads to "long",  `string_magic("{'4|..'k ! long sentence}") `
#' leads to "long..", `string_magic("{'4||..'k ! long sentence}")` leads to "lo..".
#' + fill, align (alias), width (alias): fills the character strings up to a size in order
#' to fit a given width. Options: "left", "right", "center".
#' Accepts arguments of the form `'n'` or `'n|s'`, with `n` a number and `s` a symbol. 
#' Default is left-alignment of the strings. 
#' Option "right" right aligns and "center" centers the strings. When using `'n|s'`, the symbol `s`
#' is used for the filling. By default if no argument is provided, the
#' maximum size of the character string is used. See help for [string_fill()] for more information.
#' Ex.1: `string_magic("Numbers: {'5|0'fill.right, C ? c(1, 55)}")` leads to "Numbers: 00001 and 00055".
#' + paste, append: pastes some character to all elements of the string. This operation has no default.
#' Options: "left", "both", "right", "front", "back", "delete". By default, a string is pasted on the left.
#' By default, it pastes on the left. Option "right" pastes on the right and "both" pastes on both sides. Option "front" only 
#' pastes on the first element while option "back" only pastes on the last element. Option "delete"
#' first replaces all elements with the empty string.
#' Example: `string_magic("6 = {'|'paste.both, ' + 'c ? -3:-1}")` leads to "6 = |-3| + |-2| + |-1|".
#' The argument can be of the form `s` or `s1|s2`. If of the second form, this is equivalent 
#' to chaining two `paste` operations, once on the left and once on the right: `'s1'paste, 's2'paste.right`.
#' + join: joins lines ending with a double backslash. Ex: `x = "the sun \\\n is shining"`; 
#' `string_magic("{join ? x}")` leads to "the sun is shining".
#' + escape: adds backslashes in front of specific characters. Options `"nl"`, `"tab"`. 
#' Option `"nl"` escapes the newlines (`\n`), leading them to be displayed as `"\\\\n"`.
#' Option `"tab"` does the same for tabs (`"\t"`). This is useful to make the value free
#' of space formatters. 
#' The default behavior is to escape both newlines and tabs. 
#' 
#' 
#' @section Other operations:
#' 
#' 
#' + num: converts to numeric. Options: "warn", "soft", "rm", "clear". By default, the conversion
#' is performed silently and elements that failed to convert are turned into NA. 
#' Option "warns" displays a warning if the conversion to numeric fails. 
#' Option "soft" does not convert if the conversion of at least one element fails. 
#' Option "rm" converts and removes the elements that could not be converted. 
#' Option "clear" turns failed conversions into the empty string, and hence lead to a character vector.
#' Example: `x = c(5, "six"); string_magic("Compare {num, C, q ? x} with {num.rm, C, q ? x}.")` leads to 
#' "Compare '5 and NA' with '5'.", and `string_magic("Compare {num.soft, C, q ? x} with {clear, C, q ? x}.")`
#' leads to "Compare '5 and six' with '5 and '.".
#' + enum: enumerates the elements. It creates a single string containing the comma 
#' separated list of elements.
#'   If there are more than 7 elements, only the first 6 are shown and the number of
#'  items left is written.
#'   For example `string_magic("enum ? 1:5")` leads to "1, 2, 3, 4, and 5".
#'   You can add the following options by appending the letter to enum after a dot:
#'   + q, Q, or bq: to quote the elements
#'   + or, nor: to finish with an 'or' (or 'nor') instead of an 'and'
#'   + comma: to finish the enumeration with ", " instead of ", and".
#'   + i, I, a, A, 1: to enumerate with this prefix, like in: i) one, and ii) two
#'   + a number: to tell the number of items to display
#'   Ex.1: `x = c("Marv", "Nancy"); string_magic("The main characters are {enum ? x}.")` leads to 
#' "The main characters are Marv and Nancy.".
#'   Ex.2: `x = c("orange", "milk", "rice"); string_magic("Shopping list: {enum.i.q ? x}.")` leads to
#'  "Shopping list: i) 'orange', ii) 'milk', and iii) 'rice'."
#' + len: gives the length of the vector. Options "letter", "upper", "num".
#' Option "letter" writes the length in words (up to 100). Option "upper" is the same 
#' as letter but uppercases the first letter. 
#' By default, commas are added to separate thousands. Use uption "num" to preserve
#' a regular numeric format.
#' Example: `string_magic("Size = {len ? 1:5000}")` leads to "Size = 5,000".
#' + swidth: stands for screen width. Formats the string to fit a given width 
#' by cutting at word boundaries and adding newlines appropriately. 
#' Accepts arguments of the form `'n'` or `'n|s'`, with `n` a number and `s` a string. 
#' An argument of the form `'n|s'` will add `s` at the beginning of each line. Further,
#' by default a trailing white space is added to `s`; to remove this 
#' behavior, add an underscore at the end of it. 
#' The argument `n` is either an integer giving the target character 
#' width (minimum is 15), or it can be a fraction expressing the 
#' target size as a fraction of the current screen. Finally it can be an expression that 
#' uses the variable `.sw` which will capture the value of the current screen width.
#' Ex.1: `string_magic("{15 swidth ! this is a long sentence}")` leads to "this is a long\\nsentence".
#' Ex.2: `string_magic("{15 swidth.#> ! this is a long sentence}")` leads to "#> this is a long\\n#> sentence".
#' + difftime: displays a formatted time difference. Option "silent" does not report a warning if the
#' operation fails. It accepts either objects of class `POSIXt` or `difftime`.
#' Example: `x = Sys.time() ; Sys.sleep(0.5) ; string_magic("Time: {difftime ? x}")` leads to something 
#' like "Time: 514ms".
#' 
#' @section Group-wise operations:
#' 
#' In `string_magic`, the splitting operation `s` (or `S`) keeps a memory of the strings 
#' that were split. Use the tilde operator, of the form `~(op1, op2)`, to apply operations
#' group-wise, to each of the split strings.
#' 
#' Better with an example. `x = c("Oreste, Hermione", "Hermione, Pyrrhus", "Pyrrhus, Andromaque") ;`
#' `string_magic("Troubles ahead: {S, ~(' loves 'c), C ? x}.")` leads to 
#' "Troubles ahead: Oreste loves Hermione, Hermione loves Pyrrhus and Pyrrhus loves Andromaque.".
#' 
#' Almost all operations can be applied group-wise (although only operations changing the order or 
#' the length of the strings really matter).
#' 
#' @section Conditional operations:
#' 
#' There are two operators to apply operations conditionally: `if` and `vif`, the latter
#' standing for *verbatim if*. 
#' 
#' The syntax of `if` is `if(cond ; ops_true ; ops_false)` with `cond` a
#' condition (i.e. logical operation) on the value being interpolated, `ops_true` a comma-separated
#' sequence of operations if the condition is `TRUE` and `ops_false` an *optional* a sequence of
#' operations if the condition is `FALSE`.
#' 
#' Ex.1: Let's take a sentence, delete words of less than 4 characters, and trim 
#' words of 7+ characters. 
#' x = "Songe Cephise a cette nuit cruelle qui fut pour tout un peuple une nuit eternelle"
#' `string_magic("{' 's, if(.nchar<=4 ; nuke ; '7|..'k), c ? x}")`.
#' Let's break it down. First the sentence is split w.r.t. spaces, leading to a vector
#' of words. Then we use the special variable `.nchar` in `if`'s condition to refer 
#' to the number of characters of the current vector (the words). The words with 
#' less than 4 characters are nuked (i.e. removed), and the other words are
#' trimmed at 7 characters. Finally the modified vector of words is collapsed with 
#' the function `c`, leading to the result.
#' 
#' The condition `cond` accepts the following special values: `.` (the dot), `.nchar`, `.C`, `.len`, `.N`.
#' The dot, `.`, refers to the current vector. `.nchar` represent the number of characters 
#' of the current vector (equivalent to `nchar(.)`). `.C` is an alias to `.nchar`.
#' `.len` represent the length of the current vector (equivalent to `length(.)`). 
#' `.N` is an alias to `.len`.
#' 
#' If a condition leads to a result of length 1, then the operations are applied to 
#' the full string vector and not element-wise (as was the case in Ex.1). Contrary to element-wise conditions
#' for which operations modifying the length of the vectors are forbidden (apart from nuking),
#' such operations are fine in full-string conditions.
#' 
#' Ex.2: `x = string_magic("x{1:10}")`; `string_magic("y = {if(.N>4 ; 3 first, '...'insert.right), ' + 'c ? x}")`
#' leads to "y = x1 + x2 + x3 + ...". the same opration applied to `x = string_magic("x{1:4}")`
#' leads to "y = x1 + x2 + x3 + x4".
#' 
#' For `vif`, the syntax is `vif(cond ; verb_true ; verb_false)` with `verb_true`
#' a verbatim value with which the vector will be replaced if the condition is `TRUE`. 
#' This is similar for `verb_false`. The condition works as in `if`.
#' 
#' Ex.3: `x = c(1, 25, 12, 6) ; string_magic("Values: {vif(.<10 ; <10), C ? x}")` leads to 
#' "Values: <10, 25, 12 and <10". As we can see values lower than 10 are replaced
#' with "<10" while other values are not modified.
#' 
#' Ex.4: `x = string_magic("x{1:10}")`; `string_magic("y = {vif(.N>4 ; {S!{x[1]}, ..., {last?x}}), ' + 'c ? x}")`
#' leads to "y = x1 + ... + x10".
#' Let's break it down. If the length of the vector is greater than 4 (here it's 10), then
#' the full string is replaced with `"{S!{x[1]}, ..., {last?x}}"`. Interpolation applies to
#' such string. Hence the split operation `S` breaks the string w.r.t.
#' the commas (default behavior), leading to the vector `c("{x[1]}", "...", "{last?x}")`. Since the 
#' string contains curly brackets, interpolation is applied again. This leads to 
#' the vector `c("x1", "...", "x10")`. Finally, this vector is collapsed with ' + ' leading
#' to the final string.
#' Note that there are many ways to get to the same result. Here is another example:
#' `string_magic("y = {vif(.N>4 ; {x[1]} + ... + {last?x} ; {' + 'c ? x}) ? x}")`.
#' 
#' The `vif` condition allows the use of '.' to refer to the current value in 
#' `verb_true` and `verb_false`, as illustrated by the last example:
#' 
#' Ex.5: `string_magic("{4 last, vif(. %% 2 ; x{.} ; y{rev?.}), C ? 1:11}")`
#' leads to "y10, x9, y8 and x11".
#' 
#' @section Special interpolation: if-else:
#' 
#' Using an ampersand ("&") as the first character of an interpolation leads to an *if-else* operation.
#' Using two ampersands ("&&") leads to a slightly different operation described at the end of this section.
#' 
#' The syntax is as follows: `{&cond ; verb_true ; verb_false}` with `cond` a
#' condition (i.e. logical operation) on the value being interpolated, `verb_true`
#' a verbatim value with which the vector will be replaced if the condition is `TRUE` and 
#' `verb_false` an *optional* verbatim value with which the vector will be replaced if the condition is `FALSE`. 
#' If not provided, `verb_false` is considered to be the empty string unless the operator is 
#' the double ampersand described at the end of this section.
#' 
#' Note that in `cond`, you can use the function `len`, an alias to `length`.
#' 
#' Ex.1: `x = 1:5`; \code{string_magic("x is {&len(x)<10 ; short ; {`log10(.N)-1`times, ''c ! very }long}")}
#' leads to "x is short". With `x = 1:50`, it leads to "x is long", and to "x is very very long"
#' if `x = 1:5000`.
#' 
#' If a condition leads to a result of length 1, the full string is replaced by the verbatim 
#' expression. Further, this expression will be interpolated if requested. This was the case
#' in Ex.1 where `verb_false` was interpolated.
#' 
#' If the condition's length is greater than 1, then each logical values equal to `TRUE` is replaced
#' by `verb_true`, and `FALSE` or `NA` values are replaced with `verb_false`. Note,
#' importantly, that **no interpolation is perfomed in that case**.
#' 
#' Ex.2: `x = 1:3 ; string_magic("x is {&x == 2 ; two ; not two}")` leads to the vector 
#' `c("x is not two", "x is two", "x is not two")`.
#' 
#' In that example, when x is odd, it is replaced with "odd", and when even it is
#' replaced with the elements of y.
#' 
#' Using the two ampersand operator ("&&") is like the simple ampersand version but the 
#' default for `verb_false` is the variable used in the condition itself. So the syntax is
#' `{&&cond ; verb_true}` and *it does not accept* `verb_false`.
#' 
#' Ex.3: `i = 3 ; string_magic("i = {&&i == 3 ; three}")` leads to "i = three", and to "i = 5" if `i = 5`. 
#' 
#' 
#' @section Special interpolation: Pluralization:
#' 
#' There is advanced support for pluralization which greatly facilitates the writing of messages 
#' in natural language.
#' 
#' There are two ways to pluralize: over length or over value. To trigger a "pluralization" interpolation
#' use as first character:
#' - `$` to pluralize over the length of a variable (see Ex.2)
#' - `#` to pluralize over the value of a variable (see Ex.1)
#' 
#' Ex.1: `x = 5; string_magic("I bought {N?x} book{#s}.")` leads to "I bought five books.". 
#' If `x = 1`, this leads to "I bought one book.".
#' 
#' The syntax is `{#plural_ops ? variable}` or `{#plural_ops}` where `plural_ops` are
#' specific pluralization operations which will be described below. 
#' The pluralization is perfomed *always* with respect to the value of a variable. 
#' You can either add the variable explicitly (`{#plural_ops ? variable}`) or refer
#' to it implicitly (`{#plural_ops}`). If implicit, then the algorithm will look at the 
#' previous variable that was interpolated and pluralize over it. This is exaclty what happens in
#' Ex.1 where `x` was interpolated in `{N?x}` and plural operation `s` in `{#s}` then applied to 
#' `x`. It was equivalent to have `{#s ? x}`. If a variable wasn't interpolated before, then
#' the next interpolated variable will be used (see Ex.2). If no variable is interpolated
#' at all, an error is thrown.
#' 
#' Ex.2: `x = c("J.", "M."); string_magic("My BFF{$s, are} {C?x}!")` leads to "My BFFs are J. and M.!".
#' If "x = "S.", this leads to "My BFF is S.!".
#' 
#' Pluralizing accepts the following operations:
#' - s, es: adds an "s" (or "es") if it is plural (> 1), nothing otherwise. Accepts the option `0` or `zero` which 
#' treats a 0-length or a 0-value as plural.
#' - y or ies: adds an 'y' if singular and 'ies' if plural (>1). Accepts the option `0` or `zero` which 
#' treats a 0-length or a 0-value as plural.
#' - enum: enumerates the elements (see help for the regular operation `enum`)
#' - n, N, len, Len: add the number of elements ("len") or the value ("n") of the variable as a formatted number or 
#' in letters (upper case versions). Accepts the options `letter` (to write in letter) 
#' and `upper` (to uppercase the first letter).
#' - nth, ntimes: writes the value of the variable as an order (nth) or a frequence (ntimes). Accepts the option `letter`
#' to write the numbers in letters (uppercase version of the operator does the same).
#' - is, or any verb: conjugates the verb appropriately
#' 
#' You can chain operations, in that case a whitespace is automatically added between them.
#' 
#'  Ex.3: `x = c(7, 3, 18); string_magic("The winning number{$s, is, enum ? sort(x)}.")`
#' leads to "The winning numbers are 3, 7 and 18.". With `x = 7` this leads to
#' "The winning number is 7.".
#' 
#' On top of the previous operations, there is a special operation allowing to add verbatim text depending on 
#' the situation. The syntax is as follows:
#' - `(s1;s2)`: adds verbatim 's1' if singular and 's2' if plural (>1)
#' - `(s1;s2;s3)`: adds verbatim 's1' if zero, 's2' if singular (=1) and 's3' if plural
#' - `(s1;;s3)`: adds verbatim 's1' if zero, 's3' if singular or plural (i.e. >=1)
#' 
#' These case-dependent verbatim values **are interpolated** (if appropriate). In these interpolations
#' you need not refer explicitly to the variable for pluralization interpolations.
#' 
#' Ex.4: `x = 3; string_magic("{#(Sorry, nothing found.;;{#N.upper} match{#es, were} found.)?x}")` leads to 
#' "Three matches were found.". If "x = 1", this leads to "One match was found." and if "x = 0" this leads
#' to "Sorry, nothing found.".
#' 
#' @section Escaping and special cases:
#' 
#' The opening and closing brakets, `{}`, are special characters and cannot be used as regular text. 
#' To bypass their special meaning, you need to escape them with a double backslash.
#' 
#' Ex.1: `string_magic("open = \\\\{, close = }")` leads to `"open = {, close = }"`.
#' Ex.2: `string_magic("many {5 times.c ! \\\\}}")` leads to `many }}}}}`.
#' 
#' You only need to escape the special delimiters which the algorithm is currently looking for.
#' As you can see, you don't need to escape the closing bracket in Ex.1 since no box
#' was open. On the other hand, you need to escape it in Ex.2.
#' 
#' Alternatively, use the argument `.delim` to set custom delimiters.
#' 
#' Ex.3: `string_magic("I {'can {write} {{what}} I want'}")` leads to `"I can {write} {{what}} I want"`.
#' 
#' Since `{expr}` evaluates `expr`, the stuff inside the *box*, you can pass a 
#' character string and it will stay untouched.
#' 
#' In the few operations expecting a semi-colon (if-else and pluralization), it can also be
#' escaped with a double backslash.
#' 
#' In interpolations, the exclamation mark (`!`) signals a verbatim expression. But what
#' if you use it to mean the logical operation *not* in an operation-free interpolation? 
#' In that case, you need a hack: use a question mark (`?`) first to indicate to the
#' algorithm that you want to evaluate the expression. 
#' 
#' Ex.4: `string_magic("{!TRUE} is {?!TRUE}")` leads to "TRUE is FALSE". The first expression is
#' taken verbatim while the second is evaluated.
#' 
#' @inheritSection string_is Generic regular expression flags
#'
#' @return
#' It returns a character vector whose length depends on the elements and operations in the interpolations.
#' 
#' @family string_magic functions
#' @family tools with aliases
#' 
#' @inherit str_clean seealso
#' 
#'
#' @examples
#'
#' #
#' # BASIC USAGE ####
#' #
#'
#' x = c("Romeo", "Juliet")
#'
#' # {x} inserts x
#' string_magic("Hello {x}!")
#'
#' # elements in ... are collapsed with "" (default)
#' string_magic("Hello {x[1]}, ",
#'     "how is {x[2]} doing?")
#'
#' # Splitting a comma separated string
#' # The mechanism is explained later
#' string_vec("J. Mills, David, Agnes, Dr Strong")
#'
#' # Nota: this is equivalent to (explained later)
#' string_magic("{', *'S ! J. Mills, David, Agnes, Dr Strong}")
#'
#' #
#' # Applying low level operations to strings
#' #
#'
#' # Two main syntax:
#'
#' # A) expression evaluation
#' # {operation ? x}
#' #            | |
#' #            |  \-> the expression to be evaluated
#' #             \-> ? means that the expression will be evaluated
#'
#' # B) verbatim
#' # {operation ! x}
#' #            | |
#' #            |  \-> the expression taken as verbatim (here 'x')
#' #             \-> ! means that the expression is taken as verbatim
#'
#' # operation: usually 'arg'op with op an operation code.
#'
#' # Example: splitting
#' x = "hello dear"
#' string_magic("{' 's ? x}")
#' # x is split by ' '
#'
#' string_magic("{' 's ! hello dear}")
#' # 'hello dear' is split by ' '
#' # had we used ?, there would have been an error
#'
#'
#' # There are 50+ string operators
#' # Operators usually have a default value
#' # Operations can have options
#' # Operations can be chained by separating them with a comma
#'
#' # Example: default of 's' is ' ' + chaining with collapse
#' string_magic("{s, ' my 'c ! hello dear}")
#'
#' #
#' # Nesting
#' #
#'
#' # {operations ! s1{expr}s2}
#' #             |   |
#' #             |    \-> expr will be interpolated then added to the string
#' #              \-> nesting requires verbatim evaluation: '!'
#'
#' string_magic("The variables are: {C ! x{1:4}}.")
#'
#' # This one is ugly but it shows triple nesting
#' string_magic("The variables are: {ws, C ! {2 times ! x{1:4}}{','s, 4 each !  ,_sq}}.")
#'
#' #
#' # Splitting
#' #
#'
#' # s: split with fixed pattern, default is ' '
#' string_magic("{s ! a b c}")
#' string_magic("{' b 's !a b c}")
#'
#' # S: same as 's' but default is ',[ \t\n]*'
#' string_magic("{S !a, b, c}")
#' string_magic("{'[[:punct:] ]+'S ! a! b; c}")
#' 
#' # add regex flags: e.g. fixed search
#' string_magic("{'f/.'s ! hi.there}")
#' 
#'
#' #
#' # Collapsing
#' #
#'
#' # c and C do the same, their default is different
#' # syntax: 's1|s2' with
#' # - s1 the string used for collapsing
#' # - s2 (optional) the string used for the last collapse
#'
#' # c: default is ' '
#' string_magic("{c ? 1:3}")
#'
#' # C: default is ', | and '
#' string_magic("{C ? 1:3}")
#'
#' string_magic("{', | or 'c ? 1:4}")
#'
#' #
#' # Extraction
#' #
#'
#' # extract: to extract patterns (option first)
#' # x: alias to extract.first
#' # X: alias to extract
#' # syntax: 'pattern'x
#' # Default is '[[:alnum:]]+'
#'
#' x = "This years is... 2020"
#' string_magic("{x ? x}") # similar to string_magic("{extract.first ? x}")
#' string_magic("{X ? x}") # similar to string_magic("{extract ? x}")
#'
#' string_magic("{'\\d+'x ? x}")
#'
#' #
#' # STRING FORMATTING ####
#' #
#'
#' #
#' # upper, lower, title
#'
#' # upper case the first letter
#' string_magic("{upper.first ! julia mills}")
#'
#' # title case
#' string_magic("{title ! julia mills}")
#'
#' # upper all letters
#' string_magic("{upper ! julia mills}")
#'
#' # lower case
#' string_magic("{lower ! JULIA MILLS}")
#'
#' #
#' # q, Q, bq: single, double, back quote
#'
#' string_magic("{S, q, C ! Julia, David, Wilkins}")
#' string_magic("{S, Q, C ! Julia, David, Wilkins}")
#' string_magic("{S, bq, C ! Julia, David, Wilkins}")
#'
#' #
#' # format, Format: formats the string to fit the same length
#'
#' # format: the right side is filled with blanks
#' # Format: the left side is filled with blanks, the string is right aligned
#'
#' score = c(-10, 2050)
#' nm = c("Wilkins", "David")
#' string_magic("Monopoly scores:\n{'\n'c ! - {format ? nm}: {Format ? score} US$}")
#'
#' # OK that example may have been a bit too complex,
#' # let's make it simple:
#'
#' string_magic("Scores: {format ? score}")
#' string_magic("Names: {Format ? nm}")
#'
#' #
#' # ws: white space normalization
#'
#' # ws: suppresses trimming white spaces + normalizes successive white spaces
#' # Add the following options in any order to:
#' # - punct: remove punctuation
#' # - digit: remove digits
#' # - isolated: remove isolated characters
#'
#' string_magic("{ws ! The   white  spaces are now clean.  }")
#'
#' string_magic("{ws.punct ! I, really -- truly; love punctuation!!!}")
#'
#' string_magic("{ws.digit ! 1, 2, 12, a microphone check!}")
#'
#' string_magic("{ws.i ! 1, 2, 12, a microphone check!}")
#'
#' string_magic("{ws.d.i ! 1, 2, 12, a microphone check!}")
#'
#' string_magic("{ws.p.d.i ! 1, 2, 12, a microphone check!}")
#'
#' #
#' # %: applies sprintf formatting
#'
#'  # add the formatting as a regular argument
#' string_magic("pi = {'.2f'% ? pi}")
#' # or right after the %
#' string_magic("pi = {%.2f ? pi}")
#'
#' #
#' # paste: appends text on each element
#' # Accepts the options: right, both, front and back
#' # It accepts the special values :1:, :i:, :I:, :a:, :A: to create enumerations
#'
#' # adding '|' on both sides
#' string_magic("{'|'paste.both, ' + 'c ! x{1:4}}")
#' 
#'
#' # Enumerations
#' acad = string_vec("you like admin, you enjoy working on weekends, you really love emails")
#' string_magic("Main reasons to pursue an academic career:\n {':i:) 'paste, C ? acad}.")
#' 
#' # You can also use the enum command
#' string_magic("Main reasons to pursue an academic career:\n {enum.i ? acad}.")
#'
#' #
#' # stopwords: removes basic English stopwords
#' # the list is from the Snowball project:
#' #  http://snowball.tartarus.org/algorithms/english/stop.txt
#'
#' string_magic("{stop, ws ! It is a tale told by an idiot, ",
#'                          "full of sound and fury, signifying nothing.}")
#'
#' #
#' # k: keeps the first n characters
#' # syntax: nk: keeps the first n characters
#' #         'n|s'k: same + adds 's' at the end of shortened strings
#' #         'n||s'k: same but 's' counts in the n characters kept
#'
#' words = string_vec("short, constitutional")
#' string_magic("{5k ? words}")
#'
#' string_magic("{'5|..'k ? words}")
#'
#' string_magic("{'5||..'k ? words}")
#'
#' #
#' # K: keeps the first n elements
#' # syntax: nK: keeps the first n elements
#' #         'n|s'K: same + adds the element 's' at the end
#' #         'n||s'K: same but 's' counts in the n elements kept
#' #
#' # Special values :rest: and :REST:, give the number of items dropped
#'
#' bx = string_vec("Pessac Leognan, Saint Emilion, Marguaux, Saint Julien, Pauillac")
#' string_magic("Bordeaux wines I like: {3K, ', 'C ? bx}.")
#'
#' string_magic("Bordeaux wines I like: {'3|etc..'K, ', 'C ? bx}.")
#'
#' string_magic("Bordeaux wines I like: {'3||etc..'K, ', 'C ? bx}.")
#'
#' string_magic("Bordeaux wines I like: {'3|and at least :REST: others'K, ', 'C ? bx}.")
#'
#' #
#' # Ko, KO: special operator which keeps the first n elements and adds "others"
#' # syntax: nKo
#' # KO gives the rest in letters
#'
#' string_magic("Bordeaux wines I like: {4KO, C ? bx}.")
#'
#' #
#' # r, R: string replacement 
#' # syntax: 's'R: deletes the content in 's' (replaces with the empty string)
#' #         's1 => s2'R replaces s1 into s2
#'
#' string_magic("{'e'r, ws ! The letter e is deleted}")
#'
#' # adding a perl look-behind
#' string_magic("{'(?<! )e'r !The letter e is deleted}")
#'
#' string_magic("{'e => a'r !The letter e becomes a}")
#'
#' string_magic("{'([[:alpha:]]{3})[[:alpha:]]+ => \\1.'r ! Trimming the words}")
#'
#' # Alternative way with simple operations: split, shorten, collapse
#' string_magic("{s, '3|.'k, c ! Trimming the words}")
#'
#' #
#' # times, each
#' # They accept the option c to collapse with the empty string
#'
#' string_magic("N{10 times.c ! o}!")
#'
#' string_magic("{3 times.c ? 1:3}")
#' string_magic("{3 each.c ? 1:3}")
#'
#' #
#' # erase: replaces the items by the empty string
#' # -> useful in conditions
#'
#' string_magic("{erase ! I am going to be annihilated}")
#'
#' #
#' # ELEMENT MANIPULATION ####
#' #
#'
#' #
#' # rm: removes the elements
#' # Its (optional) argument is a regular expression giving which element to remove
#' # Many options: "empty", "blank", "noalpha", "noalnum", "all" 
#'
#' x = c("Destroy", "All")
#' string_magic("{'A'rm ? x}")
#' 
#' string_magic("{rm.all ? x}")
#'
#' x = string_vec("1, 12, 123, 1234, 123456, 1234567")
#' # we delete elements whose number of characters is lower or equal to 3
#' # => see later section CONDITIONS
#' string_magic("{if(.nchar > 3 ; nuke) ? x}")
#'
#' #
#' # PLURALIZATION ####
#' #
#'
#' # Two ways to enable pluralization:
#' # {$ command}: means the plural is decuced from the length of the variable
#' # {# command}: means the plural is decuced from the value of the variable
#'
#' # Explanatory example
#' x = c("Eschyle", "Sophocle", "Euripide")
#' n = 37
#' string_magic("The author{$s, enum, have ? x} written {#N ? n} play{#s}.")
#'
#' x = "Laurent Berge"
#' n = 0
#' string_magic("The author{$s, enum, have ? x} written {#N ? n} play{#s}.")
#'
#' # How does it work?
#' # First is {$s, enum, have ? x}.
#' # The commands `s`, `enum` and `have` are applied to `x` which must come after a `?`
#' #    => there the plural (whether an s is added and how to conjugate the verb have) depends
#' #       on the **length** of the vector `x`
#' #
#' # Second comes {#N ? n}.
#' # The double dollar sign means that the command `N` will be applied to the **value** n.
#' # The value must come after the `?`
#' #
#' # Third is {#s}.
#' # The object to which `s` should be applied is missing (there is no `? n`).
#' # The default is to apply the command to the previous object. In this case,
#' #  this is `n`.
#'
#' # Another similar example illustrating that we need not express the object several times:
#' x = c("Eschyle", "Sophocle", "Euripide")
#' string_magic("The {Len ? x} classic author{$s, are, enum}.")
#'
#'
#'
#' #
#' # ARGUMENTS FROM THE ENVIRONMENT ####
#' #
#'
#' # Arguments can be evaluated from the calling environment.
#' # Simply use backticks instead of quotes.
#'
#' dollar = 6
#' reason = "glory"
#' string_magic("Why do you develop packages? For {`dollar`times.c ! $}?",
#'     "For money? No... for {upper,''s, c ? reason}!", .sep = "\n")
#' 
#' #
#' # Alias generation
#' #
#' 
#' # Let's create a formula filler
#' # - we use .local_ops to create the ad hoc operation "add" which adds variables
#' # - we transform into a formula ex post
#' 
#' fml = string_magic_alias(.post = as.formula, .local_ops = list(add = "' + 'collapse"))
#'
#' # example with mtcars
#' lhs = "mpg"
#' rhs = c("hp", "drat")
#' fml("{lhs} ~ {add?rhs} + am")
#'    
#'
NULL

#------------------------------------------------------------------------------#
# Author: Laurent R. Bergé
# Created: 2023-05-17
# ~: pluralization tests
#------------------------------------------------------------------------------#


chunk("plural")

dsb = string_magic_alias(.delim = ".[ ]")

#
# numbers ####
#

x = 5
txt = dsb("There .[#is, N ? x] cat.[#s] in the room.")
test(txt, "There are five cats in the room.")

x = 1
txt = dsb("There .[#is, N] cat.[#s ? x] in the room.")
test(txt, "There is one cat in the room.")

x = 7953
txt = dsb(".[#n ? x] observation.[#s, are] missing.")
test(txt, "7,953 observations are missing.")

x = 1
txt = dsb(".[#n ? x] observation.[#s, are] missing.")
test(txt, "1 observation is missing.")

#
# With 0 in () ####
#

x = 0
txt = dsb("There .[#(are no;;.[#is, N]) ? x] director.[#(ies;y;ies)].")
test(txt, "There are no directories.")

x = 1
txt = dsb("There .[#(are no;;.[#is, N]) ? x] director.[#(ies;y;ies)].")
test(txt, "There is one directory.")

x = 5
txt = dsb("There .[#(are no;;.[#is, N]) ? x] director.[#(ies;y;ies)].")
test(txt, "There are five directories.")

#
# plural (s, ies) ####
#

# without 0 option
n = 0
txt = string_magic("There {#is, n ? n} file{#s} and {#N} director{#y}.")
test(txt, "There is 0 file and zero directory.")

n = 1
txt = string_magic("There {#is, n ? n} file{#s} and {#N} director{#y}.")
test(txt, "There is 1 file and one directory.")

n = 5
txt = string_magic("There {#is, n ? n} file{#s} and {#N} director{#y}.")
test(txt, "There are 5 files and five directories.")

# with 0 option
n = 0
txt = string_magic("There {#is.0, n.letter.no ? n} file{#s.ze} and {#N} director{#y.zero}.")
test(txt, "There are no files and zero directories.")

n = 1
txt = string_magic("There {#is.0, n.letter.no ? n} file{#s.ze} and {#N} director{#y.zero}.")
test(txt, "There is one file and one directory.")

n = 5
txt = string_magic("There {#is.0, n.letter.no ? n} file{#s.ze} and {#N} director{#y.zero}.")
test(txt, "There are five files and five directories.")

#
# plural: nested evaluation ####
#

x = character(0)
txt = string_magic("The value `a` does not exist.{$(;; Maybe you meant: {$enum.or.bq}?) ? x}")
test(txt, "The value `a` does not exist.")

x = "append"
txt = string_magic("The value `a` does not exist.{$(;; Maybe you meant: {$enum.or.bq}?) ? x}")
test(txt, "The value `a` does not exist. Maybe you meant: `append`?")

x = c("append", "array")
txt = string_magic("The value `a` does not exist.{$(;; Maybe you meant: {$enum.or.bq}?) ? x}")
test(txt, "The value `a` does not exist. Maybe you meant: `append` or `array`?")


# With length
x = c("Charles", "Alice")
txt = dsb(".[$Is, enum ? x] crazy? Hmm... no .[$(he;they), aren't].")
test(txt, "Are Charles and Alice crazy? Hmm... no they aren't.")

txt = dsb(".[$Is, enum] crazy? Hmm... no .[$(he;they), aren't ? x[1]].") 
test(txt, "Is Charles crazy? Hmm... no he isn't.")

#
# enum, full force ####
# 

txt = dsb("I like the letter.[$s, enum.1.q ! .[S!u, v, w]].")
test(txt, "I like the letters 1) 'u', 2) 'v', and 3) 'w'.")

txt = dsb("I like the letter.[$s, enum.1.Q ! u].")
test(txt, 'I like the letter "u".')

txt = dsb("Choose one: .[$enum.bq.or ? 1:3]")
test(txt, "Choose one: `1`, `2` or `3`")

#
# multiple interpolations
#

a = 577
b = c("x", "y")
txt = dsb(".[#n ? a] observation.[#s, are] missing. It concerns the variable.[$s, enum.bq ? b].")
test(txt, "577 observations are missing. It concerns the variables `x` and `y`.")

a = 1
txt = dsb(".[#n ? a] observation.[#s, are] missing. It concerns the variable.[$s, enum.bq ? b].")
test(txt, "1 observation is missing. It concerns the variables `x` and `y`.")

txt = dsb(".[#n.u ? a] observation.[#s, are] missing. It concerns the variable.[$s, enum.bq ? b[1]].")
test(txt, "One observation is missing. It concerns the variable `x`.")

# combining the two
x = 1:5
txt = string_magic("{$n.u ? x} observation{$s}. He arrived {#nth.letter ? 3} and scored {#ntimes.le ? 1}.")
test(txt, "Five observations. He arrived third and scored once.")

#
# la conjugaison ####
#

pple = c("Francis", "Henry")
txt = string_magic("{$enum, is, (a;) ? pple} tall guy{$s}.",
          "\n{$(He;They), like} to eat donuts.",
          "\nWhen happy, at the pub {$(he;they), goes}!",
          "\n{$Don't, (he;they)} have wit, {$(he;they)} who {$try}?")
test(txt, "Francis and Henry are tall guys.\nThey like to eat donuts.\nWhen happy, at the pub they go!\nDon't they have wit, they who try?")

pple = "Francis"
txt = string_magic("{$enum, is, (a;) ? pple} tall guy{$s}.",
          "\n{$(He;They), like} to eat donuts.",
          "\nWhen happy, at the pub {$(he;they), goes}!",
          "\n{$Don't, (he;they)} have wit, {$(he;they)} who {$try}?")
test(txt, "Francis is a tall guy.\nHe likes to eat donuts.\nWhen happy, at the pub he goes!\nDoesn't he have wit, he who tries?")

#----------------------------------------------#
# Author: Laurent Berge
# Date creation: Fri Jul 10 08:53:10 2020
# ~: basic test function
#----------------------------------------------#

update_test_counter = function(reset = FALSE){
  n = getOption("TEST_COUNTER", default = 0)
  if(reset){
    options(TEST_COUNTER = 0)
  } else {
    options(TEST_COUNTER = n + 1)
  }  
}

get_test_counter = function(){
  getOption("TEST_COUNTER", default = 0)
}

test = function(x, y, type = "=", tol = 1e-6){
  
  update_test_counter()
  
  mc = match.call()
  IS_Y = TRUE
  if(missing(type) && length(y) == 1 && is.character(y) && y %in% c("err", "=", "~")){
    IS_Y = FALSE
    type = y
  }

  type = switch(type, "error" = "err", type)

  plural_core = function(PLURAL, type, s, verb = FALSE, past = FALSE){

    if(!missing(type)){
    args = strsplit(type, "\\.")[[1]]
    s = "s" %in% args
    past = "past" %in% args
    verb = setdiff(args, c("s", "past"))
    if(length(verb) == 0){
      verb = FALSE
    } else {
      verb = verb[1]
    }
    }

    if(isFALSE(verb)){
    res = ifelse(PLURAL, "s", "")
    } else {

    if(verb %in% c("be", "are")){
      verb = "is"
    } else if(verb == "have"){
      verb = "has"
    } else if(verb == "does"){
      verb = "do"
    } else if(verb %in% c("do not", "does not", "don't", "doesn't")){
      verb = "do not"
    } else if(verb %in% c("is not", "are not", "isn't", "aren't")){
      verb = "is not"
    } else if(verb %in% c("was", "were")){
      verb = "is"
      past = TRUE
    }

    if(past){
      if(verb %in% c("is", "is not", "has", "do", "do not")){
      verb_format = switch(verb, is = ifelse(!PLURAL, "was", "were"), "is not" = ifelse(!PLURAL, "wasn't", "weren't"), has = "had", do = "did", "do not" = "didn't")
      } else {
      verb_format = paste0(verb, "ed")
      }
    } else {
      if(verb %in% c("is", "is not", "has", "do", "do not")){
      verb_format = switch(verb, is = ifelse(!PLURAL, "is", "are"), "is not" = ifelse(!PLURAL, "isn't", "aren't"), has = ifelse(!PLURAL, "has", "have"), do = ifelse(!PLURAL, "does", "do"), "do not" = ifelse(!PLURAL, "doesn't", "don't"))
      } else {
      verb_format = ifelse(PLURAL, verb, paste0(verb, "s"))
      }
    }

    if(!missing(s) && isTRUE(s)){
      res = paste0(ifelse(PLURAL, "s ", " "), verb_format)
    } else {
      res = verb_format
    }

    }

    res
  }


  plural = function(x, type, s, verb = FALSE, past = FALSE){
    # adds s if x > 1, can alternatively add a verb
    PLURAL = x[1] > 1
    plural_core(PLURAL, type, s, verb, past)
  }

  n_th = function(n){
    if(length(n) == 0) return("")
    dict = c("first", "second", "third", "fourth", "fifth", "sixth", "seventh", "eighth", "nineth", "tenth", "eleventh", "twelfth", "thirteenth")
    res = as.character(n)
    qui = n <= 13
    res[qui] = dict[n[qui]]
    if(any(!is.na(qui) & !qui)){
    other = n[!qui]
    rest = other %% 10
    rest[rest == 0 | rest >= 4] = 4
    postfix = c("st", "nd", "rd", "th")
    other = paste0(other, postfix[rest])
    res[!qui] = other
    }

    res
  }

  if(type == "=" && is.numeric(x)){
    type = "~"
    if(missing(tol)) tol = 1e-12
  }

  if(type == "err"){
    # we expect an error
    m = tryCatch(x, error = function(e) structure(conditionMessage(e), class = "try-error"))
    if(!"try-error" %in% class(m)){
      stop("Expected an error that did not occur.")
    } else if(IS_Y && !grepl(tolower(y), tolower(m), fixed = TRUE)){
      stop("This is an error, but the messages don't match:\nEXPECTED: \n", y, "\n..ACTUAL: \n", x)
    }
  } else if(length(x) == 0){
    stop("Argument 'x' is of length 0. This is not allowed.")

  } else if(type %in% c("equal", "=")){
    if(length(x) != length(y)){
      stop("Lengths differ: ", 
         "\nEXPECTED: ", length(y), "\n",
         "\n..ACTUAL: ", length(x))

    } else if(!is.atomic(y) || !is.atomic(x)){
      if(!identical(x, y)){
        if(!identical(class(x), class(y))){
          stop("The two objects are of different class:",
               "\nEXPECTED: ", paste0(class(y), collapse = ", "),
               "\n..ACTUAL: ", paste0(class(x), collapse = ", "))
        }
        
        if(identical(class(x), "data.frame")){
          if(NCOL(x) != NCOL(y)){
            stop("The lengths of two objects differ:",
               "\nEXPECTED: ", NCOL(y), " variables",
               "\n..ACTUAL: ", NCOL(x), " variables")
          }
          
          same_name = names(x) == names(y)
          if(!all(same_name)){
            i = which(!same_name)
            stop("The names of two objects differ:",
               "\nEXPECTED: ", paste0(i, ': ', names(y)[i], collapse = ", "),
               "\n..ACTUAL: ", paste0(i, ': ', names(x)[i], collapse = ", "))
          }
          
          same_values = sapply(1:NCOL(x), function(i) identical(x[[i]] == y[[i]]))
          if(!all(same_values)){
            i_pblm = which(!same_name)
            text = c()
            for(i in i_pblm){
              pblm_out = capture.output(print(identical(y[[i]], x[[i]])))
              pblm_out = paste(pblm_out, collapse = "\n")
              text = c(text, pblm_out)
            }
            
            text = paste0(names(x)[i_pblm], ": see problem below\n", text, collapse = "\n\n")
            
            stop("Some variables in the two objects differ:", text)
          }
          
        } else {
          format_dp = function(expr){
            expr_dp = deparse(expr, nlines = 4)
            if(length(expr_dp) > 1){
              etc = if(length(expr_dp) == 4) "... etc..." else ""
              expr_dp = paste0("\n", paste0(head(expr_dp, 3), collapse = "\n"), etc)
            }
            expr_dp
          }
          
          y_dp = format_dp(y)
          x_dp = format_dp(x)
          
          stop("The two objects differ:",
              "\nEXPECTED: ", y_dp,
              "\n..ACTUAL: ", x_dp)
        }
        
      }
    } else if(anyNA(x)){
      if(!all(is.na(x) == is.na(y))){
        if(sum(is.na(x)) != sum(is.na(y))){
          stop("Number of NA values differ: ",
          "\nEXPECTED: ", sum(is.na(y)), "\n",
          "\n..ACTUAL: ", sum(is.na(x)))
        }

        i = which(is.na(x) != is.na(y))[1]

        na_y = 1 + is.na(y)[i]
        info = c("none", "one NA")

        stop("Position of the NA values differ. ", 
           "\nEXPECTED: ", info[na_y], " in position ", i, "\n",
           "\n..ACTUAL: ", info[3 - na_y], ".")

      } else if(!all(is_na <- is.na(x)) && any(qui_pblm <- x[!is_na] != y[!is_na])){

        if(all(qui_pblm)){
          if(length(x) == 1){
            stop("Non-NA values differ: ", 
               "\nEXPECTED: ", y[!is_na], "\n",
               "\n..ACTUAL: ", x[!is_na])
          } else {
            stop("All non-NA values differ: 1st elem.: ", 
               "\nEXPECTED: ", y[!is_na][1], "\n",
               "\n..ACTUAL: ", x[!is_na][1])
          }
        } else {
          n = sum(qui_pblm)
          i = which(qui_pblm)[1]
          stop(n, " non-NA value", plural(n, "s.differ"), ": ", n_th(i), " elem.: \nEXPECTED: ", y[!is_na][i], "\n..ACTUAL: ", x[!is_na][i])
        }
      }

    } else if(anyNA(y)){
      stop("Number of NA values differ: ", 
         "\nEXPECTED: ", sum(is.na(y)), "\n",
         "\n..ACTUAL: ", sum(is.na(x)))

    } else if(!all(x == y)){
      qui_pblm = x != y

      if(all(qui_pblm)){
        if(length(x) == 1){
          stop("Values differ: ", 
             "\nEXPECTED: ", y, "\n",
             "\n..ACTUAL: ", x)
        } else {
          stop("All values differ: 1st elem.: ", 
             "\nEXPECTED: ", y[1], "\n",
             "\n..ACTUAL: ", x[1])
        }
      } else {
        n = sum(qui_pblm)
        i = which(qui_pblm)[1]
        stop(n, " value", plural(n, "s.differ"), ": ", n_th(i), " elem.: \nEXPECTED: ", y[i], "\n..ACTUAL: ", x[i])
      }
    }
  } else if(type %in% c("~", "approx")){
    if(length(x) != length(y)){
      stop("Lengths differ: ", 
         "\nEXPECTED: ", length(y), "\n",
         "\n..ACTUAL: ", length(x))

    } else if(anyNA(x)){
      if(!all(is.na(x) == is.na(y))){
        if(sum(is.na(x)) != sum(is.na(y))){
          stop("Number of NA values differ: ", 
             "\nEXPECTED: ", sum(is.na(y)), "\n",
             "\n..ACTUAL: ", sum(is.na(x)))
        }

        i = which(is.na(x) != is.na(y))[1]

        na_y = 1 + is.na(y)[i]
        info = c("none", "one NA")

        stop("Position of the NA values differ. ", 
           "\nEXPECTED: ", info[na_y], " in position ", i, "\n",
           "\n..ACTUAL: ", info[3 - na_y], ".")

      } else if(max(abs(x - y), na.rm = TRUE) > tol){

        stop("Difference > tol: Max abs. diff: ", max(abs(x - y), na.rm = TRUE), " (in position ", which.max(abs(x - y)), ")")
      }
    } else if(anyNA(y)){
      stop("Number of NA values differ: ", 
         "\nEXPECTED: ", sum(is.na(y)), "\n",
         "\n..ACTUAL: ", sum(is.na(x)))

    } else if(max(abs(x - y)) > tol){
      stop("Difference > tol: Max abs. diff: ", max(abs(x - y)), " (in position ", which.max(abs(x - y)), ")")
    }
  }

}


chunk = function(x) cat(toupper(x), "\n\n")

test_contains = function(x, pattern){
  if(length(x) > 1){
  stop("Internal error: test_contains only works with vectors of length 1.")
  }
  
  if(!string_is(x, pattern)){
  stopi("The pattern {bq?pattern} was not found in the following string:", 
      "\n\n{x}")
  }
  
}

test_err_contains = function(x, pattern){
  err = try(x, silent = TRUE)
  if(!inherits(err, "try-error")){
  x_dp = deparse(substitute(x))[1]
  stopi("The expression should lead to an error.", 
      "\nPROBLEM: {bq?x_dp} is not an error.")
  }
  
  if(!string_is(err, pattern)){
  stopi("The pattern {bq?pattern} was not found in the following error message:", 
      "\n\n{err}")
  }
  
}

run_tests = function(chunk, from = 1, source = FALSE){
  
  wd = getwd()
  
  location = "tests"
  if(grepl("/tests$", wd)){
    location = "."
  }
  
  dir_tests = list.dirs(location, recursive = FALSE)
  if(length(dir_tests) > 0){
    # either all files in the dirs
    all_files = list.files(dir_tests, pattern = ".R$", full.names = TRUE)
  } else {
    # either the files in the main test dir
    all_files = list.files(location, pattern = ".R$", full.names = TRUE)
  }
  
  if(length(all_files) == 0){
    stop("No test file was found in the folder './tests/'.")
  }
  
  if(isTRUE(source)){
    assign("chunk", get("chunk", mode = "function"), parent.frame())
    assign("test", get("test", mode = "function"), parent.frame())
    assign("test_contains", get("test_contains", mode = "function"), parent.frame())
    assign("test_err_contains", test_err_contains, parent.frame())
    
    for(f in all_files){
      source(f)
    }
    return(NULL)
  }  

  test_code = c()
  lines_per_file = c()
  file_names = c()
  offset = 2
  for(f in all_files){
    file_names = c(file_names, f)
    f_open = file(f, encoding = "UTF-8")
    new_code = readLines(f_open)
    close(f_open)
    test_code = c(test_code, new_code)
    lines_per_file = c(lines_per_file, length(new_code))
  }

  pkg = gsub(".+/", "", normalizePath(".", "/"))

  # we don't load the package with library
  i_pkg = grepl(paste0("library(", pkg, ")"), test_code, fixed = TRUE)
  test_code[i_pkg] = ""

  # A) adding the line numbers

  lines = paste0("LINE_COUNTER = ", seq_along(test_code))

  code_LINE = rep(lines, each = 2)
  code_LINE[seq_along(test_code)  * 2] = test_code

  # We remove the counters for the lines with open parenthesis, like in
  # sum(x,
  #     y)
  # dt[, a :=
  #        32 + b]
  # since this leads to parsing errors

  i_open = which(grepl("[,=][[:blank:]]*$", code_LINE))

  if(length(i_open)) i_open = i_open + 1

  # We remove the counters just before closing brackets => otherwise equivalent to a return statement!
  i_closing_bracket = which(grepl("^[[:blank:]]*\\}", code_LINE))

  if(length(i_closing_bracket)) i_closing_bracket = i_closing_bracket - 1

  # removing
  i_rm = c(i_open, i_closing_bracket)
  if(length(i_rm) > 0){
    code_LINE = code_LINE[-(i_rm)]
  }

  # B) Adding the FOR loops

  bracket_close = which(grepl("^\\}", code_LINE))
  for_start = which(grepl("^for\\(", code_LINE))
  for_end = c()
  for(i in for_start){
    for_end = c(for_end, bracket_close[which.max(bracket_close > i)])
  }

  n = length(code_LINE)
  my_rep = rep(1, n)
  my_rep[c(for_start, for_end)] = 2

  code_LINE_FOR = rep(code_LINE, my_rep)
  n_loop = length(for_start)
  code_LINE_FOR[for_start + 2 * (0:(n_loop - 1))] = "INSIDE_LOOP = TRUE ; INDEX_LOOP = list()"
  code_LINE_FOR[for_end + 2 * 1:n_loop] = "INSIDE_LOOP = FALSE"

  qui_for = which(grepl("\\bfor\\(", code_LINE_FOR))
  index = gsub("^.*for\\(([^ ]+).+", "\\1", code_LINE_FOR[qui_for])

  n = length(code_LINE_FOR)
  my_rep = rep(1, n)
  my_rep[qui_for] = 2

  code_LINE_FOR = rep(code_LINE_FOR, my_rep)
  n_loop = length(qui_for)
  code_LINE_FOR[qui_for + 1 + (0:(n_loop - 1))] = paste0("INDEX_LOOP[[\"", index, "\"]] = ", index)

  test_code = code_LINE_FOR

  # C) Chunk selection
  
  if(!missing(chunk) || !missing(from)){

    qui = which(grepl("^chunk\\(", test_code))
    all_chunks = test_code[qui]
    chunk_names = tolower(gsub(".+\\(\"|\".*", "", all_chunks))
    n_chunks = length(qui)

    if(!missing(from)){
    if(is.character(from)){
      from = check_set_options(from, chunk_names, "test_fun")[1]
    } else {
      check_numeric(from, scalar = TRUE, integer = TRUE)
    }

    if(is.numeric(from)){
      if(any(from > n_chunks)){
      stop("There are maximum ", n_chunks, " chunks.")
      }
      chunk_select = from:n_chunks
    } else {
      chunk_select = which(chunk_names %in% from):n_chunks
    }

    } else {
    if(is.character(chunk)){
      chunk = check_set_options(chunk, chunk_names, "test_fun")
    } else {
      check_numeric(chunk, scalar = TRUE, integer = TRUE)
    }

      if(is.numeric(chunk)){
        if(any(chunk > n_chunks)){
          stop("There are maximum ", n_chunks, " chunks.")
        }
        chunk_select = sort(unique(chunk))
      } else {
        chunk_select = which(chunk_names %in% chunk)
      }
    }

    qui = c(qui, length(test_code))

    new_test_code = c()
    for(i in chunk_select){
      new_test_code = c(new_test_code, test_code[qui[i] : (qui[i + 1] - 1)])
    }

    # We add the preamble
    preamble = test_code[1:(qui[1] - 1)]

    test_code = c(preamble, new_test_code)
  }

  # D) Evaluation

  test_code = c("devtools::load_all(export_all = FALSE)", test_code)

  parsed_code = parse(text = test_code)

  env = new.env()
  assign("INSIDE_LOOP", FALSE, env)
  assign("chunk", get("chunk", mode = "function"), env)
  assign("test", test, env)
  assign("test_contains", get("test_contains", mode = "function"), parent.frame())
  assign("test_err_contains", test_err_contains, env)

  my_eval = try(eval(parsed_code, env))

  # E) Message

  if("try-error" %in% class(my_eval)){
    line_fail = get("LINE_COUNTER", env)
    inside_loop = get("INSIDE_LOOP", env)
    
    nfiles = length(all_files)
    line_start = c(0, cumsum(lines_per_file)[-nfiles])
    line_end = cumsum(lines_per_file)
    
    index = which(line_fail >= line_start & line_fail <= line_end)
    my_file_full = file_names[index]
    my_file_short = gsub(".+/", "", my_file_full)
    file_info = string_magic("In file {Q?my_file_short}\n==> ")
    
    line_in_file = line_fail - line_start[index]
    
    message("\n---------------------------\n")

    message(file_info, "Test failed at line: ", line_in_file)
    
    i = which(test_code == paste0("LINE_COUNTER = ", line_fail))
    my_line = test_code[i + 1]
    message("\nHere is the line:\n", my_line, "\n")
    if(inside_loop){
      index_loop = get("INDEX_LOOP", env)
      index_names = names(index_loop)
      index_values = unlist(index_loop)
      msg = paste0(index_names, ":", index_values, collapse = ", ")
      message("Loop values: ", msg, ".")
    }

    # We assign the variables to the global env to facilitate debugging
    for(var in names(env)){
      assign(var, get(var, env), parent.frame())
    }
    
    if(interactive()){
      command = .sma("rstudioapi::navigateToFile({Q?my_file_full}, {line_in_file})")
      command = str2lang(command)
      
      eval(command)
    }

  } else {
    n_tests = get_test_counter()
    message("tests performed successfully (", n_tests, ")")
  }
}


non_ascii = function(folder = "R"){
  # Finds non ascii characters lurking

  all_files = list.files(folder, full.names = TRUE)
  all_R_text = lapply(all_files, function(x) readLines(x, encoding = "UTF-8"))

  i_non_ascii = which(sapply(all_R_text, function(x) any(grepl("[^ -~\t]", x))))
  n = length(i_non_ascii)

  if(n == 0) return("No non-ASCII character found.")

  for(id in seq(n)){
    i = i_non_ascii[id]
    cat("File: ", gsub("R/", "", all_files[i]), "\n")

    text = all_R_text[[i]]
    all_lines = which(grepl("[^ -~\t]", text))
    for(line in all_lines){
      cat("-> line ", string_fill(line, max(nchar(all_lines))), ":\n===|", text[line], "\n")
      cat("===|", gsub("[^ -~\t]", "__HERE__", text[line]), "\n")

      if(line != tail(all_lines, 1)) cat("\n")
    }

    if(id < n) cat("\n ---------- \n\n")
  }
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cpp_paste_conditional <- function(x, id, sep = "", sep_last = "") {
    .Call(`_stringmagic_cpp_paste_conditional`, x, id, sep, sep_last)
}

cpp_normalize_ws <- function(Rstr) {
    .Call(`_stringmagic_cpp_normalize_ws`, Rstr)
}

cpp_normalize_string <- function(Rstr, clean_punct, clean_digit, clean_isolated) {
    .Call(`_stringmagic_cpp_normalize_string`, Rstr, clean_punct, clean_digit, clean_isolated)
}

cpp_trimws_in_place <- function(x) {
    .Call(`_stringmagic_cpp_trimws_in_place`, x)
}

cpp_which_empty <- function(Rstr) {
    .Call(`_stringmagic_cpp_which_empty`, Rstr)
}

cpp_find_first_index <- function(index, nb, is_last) {
    .Call(`_stringmagic_cpp_find_first_index`, index, nb, is_last)
}

cpp_group_rev_index <- function(index) {
    .Call(`_stringmagic_cpp_group_rev_index`, index)
}

cpp_recreate_index <- function(id) {
    .Call(`_stringmagic_cpp_recreate_index`, id)
}

cpp_parse_regex_pattern <- function(Rstr, parse_flags, parse_logical) {
    .Call(`_stringmagic_cpp_parse_regex_pattern`, Rstr, parse_flags, parse_logical)
}

cpp_create_pos <- function(index) {
    .Call(`_stringmagic_cpp_create_pos`, index)
}

cpp_table <- function(index, n_items) {
    .Call(`_stringmagic_cpp_table`, index, n_items)
}

cpp_format_numeric <- function(R_x, R_digits, R_signif, R_int_as_double, minus_sign, decimal, big_mark, small_mark, prefix, suffix) {
    .Call(`_stringmagic_cpp_format_numeric`, R_x, R_digits, R_signif, R_int_as_double, minus_sign, decimal, big_mark, small_mark, prefix, suffix)
}

cpp_split_newlines <- function(x) {
    .Call(`_stringmagic_cpp_split_newlines`, x)
}

cpp_string_magic_parser <- function(Rstr, Rdelimiters, only_last_parsed_section = FALSE) {
    .Call(`_stringmagic_cpp_string_magic_parser`, Rstr, Rdelimiters, only_last_parsed_section)
}

cpp_extract_quote_from_op <- function(Rstr) {
    .Call(`_stringmagic_cpp_extract_quote_from_op`, Rstr)
}

cpp_parse_operator <- function(Rstr) {
    .Call(`_stringmagic_cpp_parse_operator`, Rstr)
}

cpp_extract_pipe <- function(Rstr, check_double = FALSE) {
    .Call(`_stringmagic_cpp_extract_pipe`, Rstr, check_double)
}

cpp_parse_simple_operations <- function(Rstr, Rdelimiters) {
    .Call(`_stringmagic_cpp_parse_simple_operations`, Rstr, Rdelimiters)
}

cpp_magic_split <- function(Rstr, Rsymbol, Rdelimiters) {
    .Call(`_stringmagic_cpp_magic_split`, Rstr, Rsymbol, Rdelimiters)
}

cpp_find_closing_problem <- function(Rstr, Rdelimiters) {
    .Call(`_stringmagic_cpp_find_closing_problem`, Rstr, Rdelimiters)
}

cpp_to_index <- function(x) {
    .Call(`_stringmagic_cpp_to_index`, x)
}

#----------------------------------------------#
# Author: Laurent Berge
# Date creation: Wed Aug  3 09:40:02 2022
# ~: string.ops tests
#----------------------------------------------#

# Tests all string operations
# - should go through all the branches
# - I test the conditional operations inside the operators when relevant

chunk("string_magic")

dsb = string_magic_alias(.delim = ".[ ]")

#
# collapse ####
#


# regular
x = 1:3
test(string_magic("win = {'-'c ? x}"), "win = 1-2-3")

test(string_magic("win = {c ? x}"), "win = 1 2 3")

# with last element
test(string_magic("win = {C ? x}"), "win = 1, 2 and 3")

test(string_magic("win = {'|/'c ? x}"), "win = 12/3")

# default behavior
test(string_magic("win = {c ? x}"), "win = 1 2 3")

# empty collapse
test(string_magic("win = {''c ? x}"), "win = 123")

# conditional collapse
x = string_vec("bonjour les gens., comment ca va?, bien?, bien.")
txt = string_magic("{' 'S, ~(firstchar.3, '.'c) ? x}")
test(txt, c("bon.les.gen", "com.ca.va?", "bie", "bie"))

#
# split ####
#


x = "Cogito...Ergo...Sum"
txt = string_magic("He said: {'f/...'s ? x}")
test(txt, c("He said: Cogito", "He said: Ergo", "He said: Sum"))

txt = string_magic("He said: {'i/[cs]'s ? x}")
test(txt, c("He said: ", "He said: ogito...Ergo...", "He said: um"))



#
# replace ####
#


x = c("Blanche dit 'Bla bla blanc'.")
txt = string_magic("She said: {'bla'r, ws ? x}")
test(txt, "She said: Blanche dit 'Bla nc'.")

txt = string_magic("She said: {'bla'r.ignore, ws ? x}")
test(txt, "She said: nche dit ' nc'.")

txt = string_magic("She said: {'i/bla'r, ws ? x}")
test(txt, "She said: nche dit ' nc'.")

txt = string_magic("She said: {'bla'r.word, ws ? x}")
test(txt, "She said: Blanche dit 'Bla blanc'.")

txt = string_magic("She said: {'w/bla'r, ws ? x}")
test(txt, "She said: Blanche dit 'Bla blanc'.")

txt = string_magic("She said: {'bla'r.word.ignore, ws ? x}")
test(txt, "She said: Blanche dit ' blanc'.")

txt = string_magic("She said: {'wi/bla'r, ws ? x}")
test(txt, "She said: Blanche dit ' blanc'.")

txt = string_magic("She said: {'.'r.fixed, ws ? x}")
test(txt, "She said: Blanche dit 'Bla bla blanc'")

txt = string_magic("She said: {'fixed/.'r, ws ? x}")
test(txt, "She said: Blanche dit 'Bla bla blanc'")

txt = string_magic("She said: {'Bla, blanc => ??'r.word, ws ? x}")
test(txt, "She said: Blanche dit '?? bla ??'.")

txt = string_magic("She said: {'word/Bla, blanc => ??'r, ws ? x}")
test(txt, "She said: Blanche dit '?? bla ??'.")

txt = string_magic("She said: {'Bla, blanc => \\1\\1'r.w, ws ? x}")
test(txt, "She said: Blanche dit 'BlaBla bla blancblanc'.")

txt = string_magic("She said: {'w/Bla, blanc => \\1\\1'r, ws ? x}")
test(txt, "She said: Blanche dit 'BlaBla bla blancblanc'.")

# total replacement
x = string_vec("jingle bells, jingle bells, jingle, all the way")
txt = string_magic("{', 'c ! {'t/jing => [sound]'r ? x}}")
test(txt, "[sound], [sound], [sound], all the way")

txt = string_magic("{', 'c ! {'jing => [sound]'r.total ? x}}")
test(txt, "[sound], [sound], [sound], all the way")

txt = string_magic("{'(?<!\\b)e => a'replace.single ! Where is the letter e?}")
test(txt, "Whare is the letter e?")

# default
test(string_magic("a = {r ? 1:5}"), "err")
test(string_magic("a = {R ? 1:5}"), "err")

# other with replacement

x = c("[]", "Moon", "Sun")

txt = dsb("Stare at the .['f/[]'r ? x]")
test(txt, c("Stare at the ", "Stare at the Moon", "Stare at the Sun"))

txt = dsb("Stare at the .['f/[] => ...'r ? x]")
test(txt, c("Stare at the ...", "Stare at the Moon", "Stare at the Sun"))

txt = dsb("Stare at the .['[[:upper:]]'R ? x]")
test(txt, c("Stare at the []", "Stare at the oon", "Stare at the un"))

txt = dsb("Stare at the .['[[:upper:]] => P'R ? x]")
test(txt, c("Stare at the []", "Stare at the Poon", "Stare at the Pun"))


#
# x: single extraction ####
#


x = c("Blanche dit", " 'Bla bla blanc'.")
txt = string_magic("mots: {'bla'x ? x}")
test(txt, c("mots: ", "mots: bla"))

test(string_magic("mots: {'bla'x.ignore ? x}"), 
     c("mots: Bla", "mots: Bla"))
     
test(string_magic("mots: {'i/bla'x ? x}"), 
     c("mots: Bla", "mots: Bla"))

test(string_magic("mots: {'bla'x.i.w ? x}"), 
     c("mots: ", "mots: Bla"))

test(string_magic("mots: {'iw/bla'x ? x}"), 
     c("mots: ", "mots: Bla"))

# default
txt = string_magic("mots: {x ? x}")
test(txt, c("mots: Blanche", "mots: Bla"))

# alias
test(string_magic("mots: {'iw/bla'extract.first ? x}"), 
     c("mots: ", "mots: Bla"))


#
# X: multiple extractions ####
#


x = c("Blanche dit", " 'Bla bla blanc'.")
txt = string_magic("{'bla'X ? x}")
test(txt, c("bla", "bla"))

txt = string_magic("{'bla'X.i ? x}")
test(txt, c("Bla", "Bla", "bla", "bla"))

txt = string_magic("{'i/bla'X ? x}")
test(txt, c("Bla", "Bla", "bla", "bla"))

txt = string_magic("{'bla'X.i.w ? x}")
test(txt, c("Bla", "bla"))

txt = string_magic("{'ignore, word/bla'X ? x}")
test(txt, c("Bla", "bla"))

# conditional
x = string_vec("laura 57 and 26, charles 32 and 7, charly 29 and 8, june 55")
txt = string_magic("info: {'\\d+'X, ~('-'c), C ? x}")
test(txt, "info: 57-26, 32-7, 29-8 and 55")

#
# is, get, which ####
#


x = c("Hi Mary.", "Hi Charles!", "Are you OK?", "I think so Charles.", "Great to hear, Mary!")
txt = string_magic("A few quotes:\n{Q, 'mary | charles & \\!'get.ignore, '  'paste, '\n'c ? x}")
test(txt, "A few quotes:\n  \"Hi Charles!\"\n  \"Great to hear, Mary!\"")

txt = string_magic("A few quotes:\n{Q, 'i/mary | charles & \\!'get, '  'paste, '\n'c ? x}")
test(txt, "A few quotes:\n  \"Hi Charles!\"\n  \"Great to hear, Mary!\"")

x = c("It's me again.", "What do you mean a gain?", "Yes, again!", "Ah AGAIN, I thought A GAIN!")
index = string_ops(x, "'gain'which.w.ig")
test(index, c(2, 4))

index = string_ops(x, "'gain & .'is.fixed")
test(index, c(TRUE, FALSE, FALSE, FALSE))

# conditional
x = string_vec("laura 57 and 26, charles 32 and 7, charly 29 and 8, june 55")
txt = string_magic("info: {' 'S, '!\\d'get, ~('-'c), C ? x}")
test(txt, "info: laura-and, charles-and, charly-and and june")


#
# each/times ####
#


test(string_magic("I like {5 times.c ! ?} marks!"), "I like ????? marks!")

x = c("mary", "richard")
y = c("yes", "no")
txt = string_magic("The discussion: {', 'c ! {upper.first, 2 times ? x}: '{2 times ? y}'}...")
test(txt, "The discussion: Mary: 'yes', Richard: 'no', Mary: 'yes', Richard: 'no'...")

test(string_magic("values: {2 each.c ? c('a', 'b')}"), "values: aabb")

#
# case ####
#


x = "where is bryan? Bryan is in the KITCHEN."

txt = dsb(".[lower ? x]")
test(txt, "where is bryan? bryan is in the kitchen.")

txt = dsb(".[upper.first ? x]")
test(txt, "Where is bryan? Bryan is in the KITCHEN.")

txt = dsb(".[upper ? x]")
test(txt, "WHERE IS BRYAN? BRYAN IS IN THE KITCHEN.")

txt = dsb(".[lower, upper.sentence ? x]")
test(txt, "Where is bryan? Bryan is in the kitchen.")

txt = dsb(".[title ? x]")
test(txt, "Where Is Bryan? Bryan Is In The KITCHEN.")

txt = dsb(".[title.force ? x]")
test(txt, "Where Is Bryan? Bryan Is In The Kitchen.")

txt = dsb(".[title.force.ignore ? x]")
test(txt, "Where Is Bryan? Bryan Is in the Kitchen.")

x = "results from a new estimator: a new hope"
txt = dsb(".[title.i ? x]")
test(txt, "Results From a New Estimator: A New Hope")

#
# quotes ####
#


x = "siren"
txt = string_magic("Is it the song of the {q ? x}, of the {Q ? x} or of the {bq ? x}?")
test(txt, "Is it the song of the 'siren', of the \"siren\" or of the `siren`?")

#
# format ####
#


x = c(1, 123, 123456)
txt = string_magic("The numbers are:\n{'\n'c ! - {format ? x} | {rev, Format ? x}}")
test(txt, "The numbers are:\n- 1       | 123,456\n- 123     |     123\n- 123,456 |       1")

txt = string_magic("The numbers are:\n{'\n'c ! - {format.0 ? x} | {rev, Format.zero ? x}}")
test(txt, "The numbers are:\n- 0000001 | 123,456\n- 0000123 | 0000123\n- 123,456 | 0000001")

today = structure(19531, class = "Date")
txt = string_magic("Today is {'%d, %m, %Y'for ? today}")
test(txt, "Today is 23, 06, 2023")

txt = string_magic("Today is {'%d, %m, %Y'for ? .date}")
test_contains(txt, "\\d\\d, \\d\\d, \\d{4}")

txt = string_magic("Today is {.now}")
test_contains(txt, "\\d{4} & \\d:\\d")

txt = string_magic("We're {.now('%H:%M %Y')}")
test_contains(txt, "\\d:\\d & \\d{4}")

#
# sprintf ####
#


txt = string_magic("pi = {%.3f ? pi}")
test(txt, "pi = 3.142")

txt = string_magic("pi = {% 8.3f ? pi}")
test(txt, "pi =    3.142")

x = c("michael", "ana")
txt = string_magic("The winners are:\n{'\n'c ! - {% 10s ? x}}")
test(txt, "The winners are:\n-    michael\n-        ana")

txt = string_magic("The winners are:\n{'\n'c ! - {%- 10s ? x}!}")
test(txt, "The winners are:\n- michael   !\n- ana       !")

#
# ws ####
#


x = " &Hej! Welcome!! A\t\n 6-feet wide bed awaits!\t."

txt = dsb(".[ws ? x]")
test(txt, "&Hej! Welcome!! A 6-feet wide bed awaits! .")

txt = dsb(".[ws.p ? x]")
test(txt, "Hej Welcome A 6 feet wide bed awaits")

txt = dsb(".[ws.d ? x]")
test(txt, "&Hej! Welcome!! A -feet wide bed awaits! .")

txt = dsb(".[ws.i ? x]")
test(txt, "&Hej! Welcome!! 6-feet wide bed awaits!")

txt = dsb(".[ws.p.d ? x]")
test(txt, "Hej Welcome A feet wide bed awaits")

txt = dsb(".[ws.punct.isolated ? x]")
test(txt, "Hej Welcome feet wide bed awaits")

txt = dsb(".[ws.d.i ? x]")
test(txt, "&Hej! Welcome!! -feet wide bed awaits!")

txt = dsb(".[ws.p.d.i ? x]")
test(txt, "Hej Welcome feet wide bed awaits")

#
# tws ####
#


x = c("  bonjour les gens", "ahem \n  ", NA)
txt = dsb(".[tws ? x]")
test(txt, c("bonjour les gens", "ahem", NA))

#
# trim ####
#


x = c("bonjour", "ahem")
txt = dsb(".[4 trim ? x]")
test(txt, c("our", ""))

txt = dsb(".[-2 trim ? x]")
test(txt, c("bonjo", "ah"))

txt = dsb(".[5 trim.r ? x]")
test(txt, c("bo", ""))

txt = dsb(".[1 trim.b ? x]")
test(txt, c("onjou", "he"))

#
# k, keep char ####
#


x = c("this is a long sentence", "a short one")
txt = string_magic("{11k, ' and 'c ? x}")
test(txt, "this is a l and a short one")

txt = string_magic("{11 k, ' and 'c ? x}")
test(txt, "this is a l and a short one")

txt = string_magic("{'11|..'k, ' and 'c ? x}")
test(txt, "this is a l.. and a short one")

txt = string_magic("{11|.. shorten, ' and 'c ? x}")
test(txt, "this is a l.. and a short one")

txt = string_magic("{'11||..'k, ' and 'c ? x}")
test(txt, "this is a.. and a short one")

txt = string_magic("{11|.. shorten.inclu, ' and 'c ? x}")
test(txt, "this is a.. and a short one")

txt = string_magic("{11k.d, ' and 'c ? x}")
test(txt, "this is a .. and a short one")

txt = string_magic("{11 Shorten, ' and 'c ? x}")
test(txt, "this is a .. and a short one")

#
# K, keep elements ####
#


x = 1:5

txt = string_magic("{3K ? x}")
test(txt, 1:3)

txt = string_magic("{3KO ? x}")
test(txt, c("1", "2", "three others"))

txt = string_magic("{3Ko ? x}")
test(txt, c("1", "2", "3 others"))

txt = string_magic("{'3|le reste'K ? x}")
test(txt, c("1", "2", "3", "le reste"))

txt = string_magic("{'3||le reste'K ? x}")
test(txt, c("1", "2", "le reste"))

txt = string_magic("{'3||et :rest:/:n:'K ? x}")
test(txt, c("1", "2", "et 3/5"))

txt = string_magic("{'3||et :REST:/:N:'K ? x}")
test(txt, c("1", "2", "et three/five"))

# conditional
x = string_vec("laura 57 and 26, charles 32 and 7, charly 29 and 8, june 55")
test(string_magic("info: {' 'S, ~(2Ko), C ? x}"), "err")

#
# enum ####
#


x = 1:5
txt = string_magic("{enum.q.3.oxf ? x}")
test(txt, "'1', '2', and 3 others")

txt = string_magic("{enum.q.3.comm ? x}")
test(txt, "'1', '2', 3 others")

txt = string_magic("{enum.bq.or ! x{1:3}}")
test(txt, "`x1`, `x2` or `x3`")

txt = string_magic("{enum.Q.nor.1 ! x{1:3}}")
test(txt, '1) "x1", 2) "x2", nor 3) "x3"')

txt = string_magic("{enum.a ! x{1:3}}")
test(txt, "a) x1, b) x2, and c) x3")

txt = string_magic("{enum.i ! x{1:3}}")
test(txt, "i) x1, ii) x2, and iii) x3")

# conditional
x = c("123", "abc", "a1b2")
txt = string_magic("{''S, ~(enum), ' ; 'c ? x}")
test(txt, "1, 2 and 3 ; a, b and c ; a, 1, b and 2")

#
# first and last ###
#


x = 1:5
txt = string_magic("{2 first, ''c ? x}")
test(txt, "12")

txt = string_magic("{last.2, ''c ? x}")
test(txt, "45")

a = 2
txt = string_magic("{'3'first, `a`last, ''c ? x}")
test(txt, "23")

txt = string_magic("{'1|1'first, ''c ? x}")
test(txt, "15")

txt = string_magic("{'-3'first, ''c ? x}")
test(txt, "45")

# conditional
x = c("123", "abc", "a1b2")
txt = string_magic("{''S, ~(first), ' ; 'c ? x}")
test(txt, "1 ; a ; a")

txt = string_magic("{''S, ~(last), ' ; 'c ? x}")
test(txt, "3 ; c ; 2")

#
# firstchar, lastchar ####
#


x = c("bonjour", "les", "gens")
a = 2
txt = string_magic("{3 firstchar, `a`lastchar ? x}")
test(txt, c("on", "es", "en"))

#
# rev ####
#


test(string_magic("{rev ? 1:3}"), 3:1)

# conditional
x = c("123", "abc", "a1b2")
txt = string_magic("{''S, ~(rev, ''c), ' ; 'c ? x}")
test(txt, "321 ; cba ; 2b1a")

#
# sort, dsort ####
#


x = c(5, 3, 8, 1)
test(string_magic("{sort ? x}"), c(1, 3, 5, 8))

test(string_magic("{dsort ? x}"), c(8, 5, 3, 1))

# with preprocessing
x = "Mark is 34, Bianca is 55, Odette is 101, Julie is 21 and Frank is 5"
txt = string_magic("{', | and 's, '\\D'sort, C ? x}")
test(txt, "Odette is 101, Julie is 21, Mark is 34, Frank is 5 and Bianca is 55")

txt = string_magic("{', | and 's, '\\D'sort.num, C ? x}")
test(txt, "Frank is 5, Julie is 21, Mark is 34, Bianca is 55 and Odette is 101")

# conditional
x = c("521", "aebc")
txt = string_magic("{''S, ~(sort, ''c), ' ; 'c ? x}")
test(txt, "125 ; abce")

txt = string_magic("{''S, ~(dsort, ''c), ' ; 'c ? x}")
test(txt, "521 ; ecba")

#
# paste ####
#


x = 1:2
txt = string_magic("{'x'paste, ', 'c ? x}")
test(txt, "x1, x2")

txt = string_magic("{'*'paste.both, ', 'c ? x}")
test(txt, "*1*, *2*")

txt = string_magic("{_ paste.right, ', 'c ? x}")
test(txt, "1_, 2_")

x = 1:3
txt = string_magic("The number{if(.N>1 ; 's are 'paste.front ; ' is'paste.front), C ? x}.")
test(txt, "The numbers are 1, 2 and 3.")

x = 5
txt = string_magic("The number{if(.N>1 ; 's are 'paste.front ; ' is 'paste.front), C ? x}.")
test(txt, "The number is 5.")

#
# insert ####
#


x = 1:2
txt = string_magic("{'0'insert, ''c ? x}")
test(txt, "012")

txt = string_magic("{'0'insert.both, ''c ? x}")
test(txt, "0120")

txt = string_magic("{'0'insert.right, ''c ? x}")
test(txt, "120")

# conditional
x = c("521", "aebc")
test(string_magic("{''S, ~('0'insert, ''c), ' ; 'c ? x}"), "err")

#
# fill ####
#


x = c("bon", "bonjour les gens")
txt = string_magic("{fill, q, C ? x}")
test(txt, "'bon             ' and 'bonjour les gens'")

txt = string_magic("{fill.right, q, C ? x}")
test(txt, "'             bon' and 'bonjour les gens'")

txt = string_magic("{fill.center, q, C ? x}")
test(txt, "'       bon      ' and 'bonjour les gens'")

x = c(5, 15)
txt = string_magic("{3 fill.right, q, C ? x}")
test(txt, "'  5' and ' 15'")

txt = string_magic("{1 fill.right, q, C ? x}")
test(txt, "'5' and '15'")

txt = string_magic("{'3|0'fill.right, q, C ? x}")
test(txt, "'005' and '015'")

#
# join ####
#


x = "\n    bonjour \\\n   les gens\n est-ce que ca va?"
txt = string_magic("Text: {tws, join ? x}")
test(txt, "Text: bonjour les gens\n est-ce que ca va?")

#
# escape ####
#


x = "bon\njour \t les \t gens"
txt = string_magic(x, .last = "escape")
test(txt, "bon\\njour \\t les \\t gens")

#
# unik ####
#


x = c(1, 1, 2, 3, 3, 3) 
test(string_magic("{unik ? x}"), 1:3)

# conditional
x = c("11211125564454", "aggafsgaffasg")
txt = string_magic("{''S, ~(unik, ''c), ' ; 'c ? x}")
test(txt, "12564 ; agfs")

#
# table ####
#

x = strsplit("aggafsggaffasg", "")[[1]]

# default
txt = sma("letters: {table, enum ? x}.")
test(txt, "letters: g (5), a (4), f (3) and s (2).")

txt = sma("letters: {table.sort, enum ? x}.")
test(txt, "letters: a (4), f (3), g (5) and s (2).")

txt = sma("letters: {table.dsort, enum ? x}.")
test(txt, "letters: s (2), g (5), f (3) and a (4).")

txt = sma("letters: {table.fsort, enum ? x}.")
test(txt, "letters: s (2), f (3), a (4) and g (5).")

txt = sma("letters: {table.nosort, enum ? x}.")
test(txt, "letters: a (4), g (5), f (3) and s (2).")

# with argument
txt = sma("letters: {'{q ? x} {round(s * 100)}%'table, enum ? x}.")
test(txt, "letters: 'g' 36%, 'a' 29%, 'f' 21% and 's' 14%.")

txt = sma("letters: {'{x}: {n}'table.sort, ', 'c ? x}.")
test(txt, "letters: a: 4, f: 3, g: 5, s: 2.")

# error
test(sma("letters: {'x = n'table, ', 'c ? x}."), "no interpolated", "err")

test(sma("letters: {'{zz}'table, ', 'c ? x}."), "following invalid", "err")

# double sort
x = sample(stsplit("aacceebbbuuu", ""))

txt = string_ops(x, "table.dfsort.sort, ', 'c")
test(txt, "b (3), u (3), a (2), c (2), e (2)")

txt = string_ops(x, "table.fsort.sort, ', 'c")
test(txt, "a (2), c (2), e (2), b (3), u (3)")

txt = string_ops(x, "table.dfsort.dsort, ', 'c")
test(txt, "u (3), b (3), e (2), c (2), a (2)")

txt = string_ops(x, "table.fsort.dsort, ', 'c")
test(txt, "e (2), c (2), a (2), u (3), b (3)")

#
# nth ####
#


x = c(1, 6)
txt = string_magic("They arrived {nth, C ? x}.")
test(txt, "They arrived 1st and 6th.")

txt = string_magic("They arrived {Nth, C ? x}.")
test(txt, "They arrived first and sixth.")

#
# ntimes ####
#


x = c(1, 6)
txt = string_magic("They won {ntimes, C ? x}.")
test(txt, "They won once and 6 times.")

txt = string_magic("They won {Ntimes, C ? x}.")
test(txt, "They won once and six times.")

#
# n ####
#


x = c(45546, "bonjour")
txt = string_magic("A = {n ? x} ; B = {n.l ? 55}")
test(txt, c("A = 45,546 ; B = fifty-five", "A = bonjour ; B = fifty-five"))

txt = string_magic("{N.up ? 8} times he won!")
test(txt, "Eight times he won!")

year = 2023
txt = string_magic("This test was written in {n.R?year}.")
test(txt, "This test was written in MMXXIII.")

txt = string_magic("This test was written in {n.r?year}.")
test(txt, "This test was written in mmxxiii.")

#
# len ####
#


x = 1:5
txt = string_magic("{Len.up ? x} numbers.")
test(txt, "Five numbers.")

x = 1:2
txt = string_magic("x is of length {len ? x}, or {Len ? x}.")
test(txt, "x is of length 2, or two.")

x = 1:1e5
txt = string_magic("x is of length {len ? x}, or {len.num ? x}.")
test(txt, "x is of length 100,000, or 100000.")

# conditionaly
x = c("bonjour les gens", "la pluie", "est drue, je rentre")
txt = string_magic("Number of words: {' 'S, ~(len), C ? x}.")
test(txt, "Number of words: 3, 2 and 4.")

num = string_ops(x, "' 'S, ~(len), num")
test(num, c(3, 2, 4))

#
# width ####
#


x = "Rome, l'unique objet de mon ressentiment, Rome a qui vient ton bras d'immoler mon amant"
txt = string_magic("Voici le texte a apprendre:\n{'40|>'swidth ? x}.")
test(txt, c("Voici le texte a apprendre:\n> Rome, l'unique objet de mon\n> ressentiment, Rome a qui vient ton\n> bras d'immoler mon amant."))

#
# difftime ####
#


x = 3654
txt = string_magic("Time since last check: {difftime ? x}.")
test(txt, "Time since last check: 1 hour 00 min.")

x = structure(1680294984.14505, class = c("POSIXct", "POSIXt")) - structure(1680292481.19258, class = c("POSIXct", "POSIXt"))
txt = string_magic("Time since last check: {difftime ? x}.")
test(txt, "Time since last check: 41 min 42 sec.")

#
# erase ####
#


x = 1:5
txt = string_magic("{if(. <= 3 ; erase), '.'c ? x}")
test(txt, "...4.5")

#
# nuke ####
#


x = 1:5
txt = string_magic("nothing = {nuke ? x}")
test(txt, "nothing = ")

#
# rm ####
#


x = c("", "    ", "556", ":!", "pour qui sont ces", "serpents qui sifflent sur nos tetes?")
txt = string_magic("{5 firstchar, rm, ', 'c ? x}")
test(txt, "    , 556, :!, pour , serpe")

txt = string_magic("{5 firstchar, rm.blank, ', 'c ? x}")
test(txt, "556, :!, pour , serpe")

txt = string_magic("{5 firstchar, rm.noalpha, ', 'c ? x}")
test(txt, "pour , serpe")

txt = string_magic("{5 firstchar, rm.noalnum, ', 'c ? x}")
test(txt, "556, pour , serpe")

txt = string_magic("{5 firstchar, rm.all, ', 'c ? x}")
test(txt, "")

txt = string_magic("{5 firstchar, 'pour'rm.blank, ', 'c ? x}")
test(txt, "556, :!, serpe")

# conditional
x = string_vec("laura 57 and 26, charles 32 and 7, charly 29 and 8, june 55")
txt = string_magic("{' 'S, rm.noalpha, ~('-'c), C ? x}")
test(txt, "laura-and, charles-and, charly-and and june")

#
# num ####
#


x = "parse55this"
txt = string_ops(x, "'\\d+'x, num")
test(txt, 55)

x = c("55", "abc")
txt = string_magic("{num ? x}")
test(txt, c(55, NA))

txt = string_magic("{num.soft ? x}")
test(txt, c(55, "abc"))

txt = string_magic("{num.clear ? x}")
test(txt, c(55, ""))

txt = string_magic("{num.rm ? x}")
test(txt, 55)

# conditional
x = string_vec("laura 57 and 26, charles 32 and 7, charly 29 and 8, june 55")
txt = string_magic("{' 'S, num.rm, ~('-'c), C ? x}")
test(txt, "57-26, 32-7, 29-8 and 55")

#
# ascii ####
#
# No ascii tests bc result will depend on file encoding
# test(dsb("laurent .[ascii.utf8 ! bergé]"), "laurent berge")

#
# stop ####
#


x = "Hi I'm Laurent and I'm trying to remove stop-words."
txt = string_magic("Before: {x}\nAfter: {stop, ws ? x}")
test(txt, "Before: Hi I'm Laurent and I'm trying to remove stop-words.\nAfter: Hi Laurent trying remove stop-words.")

#
# if-else ####
#


# elementwise
x = c(15, 550)
txt = dsb("The value is .[& x > 50 ; > 50 ; <= 50]")
test(txt, c("The value is <= 50", "The value is > 50"))

txt = dsb("The value is .[& x > 50 ; > 50]")
test(txt, c("The value is ", "The value is > 50"))

# testing the quote
txt = dsb("The value is .[& x > 50 ; > 50 ; '   .']")
test(txt, c("The value is    .", "The value is > 50"))

# whole string
x = c(15, 550)
txt = dsb("The max value is .[& max(x) > 50 ; large ; small].")
test(txt, "The max value is large.")

x = c(-5, 15)
txt = dsb("The max value is .[& max(x) > 50 ; large ; small].")
test(txt, "The max value is small.")

x = c(-5, 15)
txt = dsb("The max value is .[& max(x) > 50 ; large].")
test(txt, "The max value is .")

# nesting
x = c(15, 550)
txt = string_magic("The values are{& length(x) < 5 ; ': {C ? .}' ; ' too many'}.")
test(txt, "The values are: 15 and 550.")

data = data.frame(x = x, stringsAsFactors = FALSE)
txt = string_magic("The values are{& length(data$x) < 5 ; ': {C ? .}' ; ' too many'}.")
test(txt, "The values are: 15 and 550.")

# special if else
x = c(15, 550)
txt = dsb("The value is .[&& x > 50 ; > 50]")
test(txt, c("The value is 15", "The value is > 50"))

# long interpolated true and false
x = 1:3
y = letters[1:3]
txt = string_magic("{C ! {&&x %% 2 == 1;{y}}}")
test(txt, "a, 2 and c")

z = tail(letters, 3)
txt = string_magic("{C ! {&x %% 2 == 1 ; {y} ; {bq?z}}}")
test(txt, "a, `y` and c")

#
# if ####
#


# special values: ., .len (or .N), .nchar (or .C)

x = c(5, 25, 30, 7)
txt = string_magic("The large numbers are {if(.<10 ; nuke), C ? x}.")
test(txt, "The large numbers are 25 and 30.")

x = c(5, 25, 30, 7)
txt = string_magic("The numbers are {if(.C < 2 ; 'x'pa ; 1k), C ? x}.")
test(txt, "The numbers are x5, 2, 3 and x7.")

txt = string_magic("The numbers are {if(.nchar >= 2 ; 'x'pa ; 1k), C ? x}.")
test(txt, "The numbers are 5, x25, x30 and 7.")

x = c(5, 25, 30, 7)
txt = string_magic("The hash is {if(.N < 4 ; ':'c ; '-'c), C ? x}.")
test(txt, "The hash is 5-25-30-7.")

x = c(5, 25, 30)
txt = string_magic("The hash is {if(.N < 4 ; ':'c ; '-'c), C ? x}.")
test(txt, "The hash is 5:25:30.")

# conditional
x = c(5, 25, 30)
txt = string_magic("Example: {if(.C < 2 ; 'x'pa ; ''s, ~('/'c)), C ? x}.")
test(txt, "Example: x5, 2/5 and 3/0.")


#
# vif ####
#


x = c(5, 25, 30, 7)
txt = string_magic("There are {vif(.len<3 ; not many ; many), C ? x} numbers.")
test(txt, "There are many numbers.")

x = c(5, 25)
txt = string_magic("There are {vif(.N<3 ; not many ; many), C ? x} numbers.")
test(txt, "There are not many numbers.")

# elementwise
x = c(5, 25, 30, 7)
txt = string_magic("Numbers: {vif(.<10 ; small ; large), C ? x}.")
test(txt, "Numbers: small, large, large and small.")

# nesting
x = c(5, 25, 30, 7)
txt = string_magic("Numbers: {vif(.N > 5 ; sum = {n?sum(x)} ; prod = {n?prod(x)}), C ? x}.")
test(txt, "Numbers: prod = 26,250.")

x = c(5, 25, 30, 7, 5, 4)
txt = string_magic("Numbers: {vif(.N > 5 ; sum = {n?sum(x)} ; prod = {n?prod(x)}), C ? x}.")
test(txt, "Numbers: sum = 76.")

# '.' as a placeholder
x = c(5, 12, 20, 35)
txt = string_magic("y = {3 first, vif(.N<2 ; short ; {' + 'c ? .}) ? x}")
test(txt, "y = 5 + 12 + 20")

#
# escaping ####
#


txt = string_magic("\\{} it's some braces")
test(txt, "{} it's some braces")

txt = dsb("\\.[] it's some dsb")
test(txt, ".[] it's some dsb")

txt = string_magic("I'm saying {q!ha \\} ha}!")
test(txt, "I'm saying 'ha } ha'!")

txt = dsb("I'm saying .[q!ha \\] ha]!")
test(txt, "I'm saying 'ha ] ha'!")

#
# user-defined ops ####
#


# without scope
xplode = function(x, argument, options, ...){
  unlist(strsplit(as.character(x), ""))
}

string_magic_register_fun(xplode, "xplode")

txt = string_magic("bon{xplode!jour}")
test(txt, c("bonj", "bono", "bonu", "bonr"))

string_magic_register_ops("'50|-'fill", "h2")
txt = string_magic("my header ", .last = "h2")


# with scope
string_magic_register_ops("''s, x paste, ' + 'c", "equick", namespace = "test_ns")

sma_test = string_magic_alias(.namespace = "test_ns")
txt = sma_test("y = {equick!173ab}")
test(txt, "y = x1 + x7 + x3 + xa + xb")

test_err_contains(string_magic("y = {equick!173ab}"), "valid operator")
test_err_contains(sma_test("y = {aezgfaffs!173ab}"), "!xplode")

#
# errors ####
#

test_err_contains(string_magic("hi {there"), "bracket")

test_err_contains(string_magic("hi {upper ! there"), "bracket")

test_err_contains(string_magic("hi {S, ''S, ~(first ! bon, jour}"), "parenthesis")

test_err_contains(string_magic("hi {$(bon) ! jour}"), "if/plurali & (v1")

test_err_contains(string_magic("hi {&bonjour}"), "semi-colon")

test_err_contains(string_magic("hi {&bonjour ; jour}"), "bonjour & evaluated")

test_err_contains(string_magic("hi {&TRUE ; {upper ! jour ; voir}"), "closing bracket")

test_err_contains(string_magic("hi {&TRUE ; {upperest ! jour} ; voir}"), "w/upper")

test_err_contains(string_magic("hi {&TRUE ; {upperest ? jour} ; voir}"), "jour & evaluated")

test_err_contains(string_magic("hi {&bonjour ; jour}"), "bonjour & not found")

test_err_contains(string_magic("error operation: {if(1:5 ; upper ; lower) ! ohohoh}"), "w/if & length")

test_err_contains(string_magic("error operation: {if(fakfak; upper ; lower) ! ohohoh}"), "w/if & evaluated & not found")

test_err_contains(string_magic("error operation: {if(fakfak%m; upper ; lower) ! ohohoh}"), "w/if & parsed")

test_err_contains(string_magic("error operation: {S, ~(sekg) ! ohohoh, hihihi}"), "not a valid op")

test_err_contains(string_vec("hi, {there"), "bracket & matched & escape")

#
# multi line expressions ####
#


txt = string_magic("The solution to x + 32 = 5 is {
     y = 32
     z = 5
     z - y
}")
test(txt, "The solution to x + 32 = 5 is -27")

txt = string_magic("First letters: {
     x = head(letters)
     fun = function(z){
          paste(z, collapse = '')
     }
     fun(x)
}")
test(txt, "First letters: abcdef")

#
# variables as arguments ####
#


txt = string_magic("Hi {you}!", you = "Omer")
test(txt, "Hi Omer!")

txt = string_magic("m * n = {m * n}", m = 5, n = 4)
test(txt, "m * n = 20")

# with nesting
txt = string_magic("vars = {' + 'c ! {v1}_{suffix}}", v1 = c("x1", "z5"), suffix = "post")
test(txt, "vars = x1_post + z5_post")

# with ifelse
txt = string_magic("value = {&v1 > 0 ; {.} ; {v2}}", v1 = 55, v2 = "error")
test(txt, "value = 55")

txt = string_magic("value = {&v1 > 0 ; {.} ; {v2}}", v1 = -55, v2 = "error")
test(txt, "value = error")

# with plural
txt = string_magic("The number{$s, is, ({v};{$enum}) ? x}", x = 1:3, v = "error")
test(txt, "The numbers are 1, 2 and 3")

txt = string_magic("The number{$s, is, ({v};{$enum}) ? x}", x = 3, v = "error")
test(txt, "The number is error")

#
# default options ####
#


sma2 = string_magic_alias(.delim = "$[ ]")
txt = sma2("x$[1:2]", .last = "C")
test(txt, "x1 and x2")

#
# .post + .last ####
#


x = string_magic("{1:5}0", .last = "num", .post = sum)
test(x, 150)

#
# .default ####
#


x = string_magic("{1:5}0", .default = "'1|1'first, enum")
test(x, "1 and 50")

#------------------------------------------------------------------------------#
# Author: Laurent R. Bergé
# Created: 2023-05-17
# ~: test for auxilliary tools
#------------------------------------------------------------------------------#

chunk("tools")


####
#### string_ops ####
####

cars = row.names(mtcars)
val = string_ops(cars, "'\\d+'x, rm, unik, num, dsort, 3 first")
test(val, c(914, 710, 450))

val = string_ops(cars, "'\\d+'x, rm", "unik, num", "dsort, 3 first")
test(val, c(914, 710, 450))

all_months = month.name[airquality$Month]
val = string_ops(all_months, "lower", "5 firstchar", pre_unik = TRUE)
val_bis = tolower(substr(all_months, 1, 5))
test(val, val_bis)

####
#### string_is ####
####


x = string_vec("one, two, one... two, microphone, check")

val = string_is(x, "^...$")
test(val, c(TRUE, TRUE, FALSE, FALSE, FALSE))


val = string_is(x, "f/...")
test(val, c(FALSE, FALSE, TRUE, FALSE, FALSE))

val = string_is(x, "...", fixed = TRUE)
test(val, c(FALSE, FALSE, TRUE, FALSE, FALSE))


val = string_is(x, "f/!...")
test(val, c(TRUE, TRUE, FALSE, TRUE, TRUE))

val = string_is(x, "!...", fixed = TRUE)
test(val, c(TRUE, TRUE, FALSE, TRUE, TRUE))


val = string_is(x, "iw/one")
test(val, c(TRUE, FALSE, TRUE, FALSE, FALSE))

val = string_is(x, "one", ignore.case = TRUE, word = TRUE)
test(val, c(TRUE, FALSE, TRUE, FALSE, FALSE))


val = string_is(x, "iw/!one")
test(val, c(FALSE, TRUE, FALSE, TRUE, TRUE))

val = string_is(x, "!one", ignore.case = TRUE, word = TRUE)
test(val, c(FALSE, TRUE, FALSE, TRUE, TRUE))


val = string_is(x, "i/one & .{5,}")
test(val, c(FALSE, FALSE, TRUE, TRUE, FALSE))


val = string_is(x, "one", "c", fixed = TRUE)
test(val, c(FALSE, FALSE, FALSE, TRUE, FALSE))

val = string_is(x, "one & c", fixed = TRUE)
test(val, c(FALSE, FALSE, FALSE, TRUE, FALSE))

val = string_is(x, "one | c", fixed = TRUE)
test(val, c(TRUE, FALSE, TRUE, TRUE, TRUE))

# magic
my_words = "one, two"
val = string_is(x, "wm/{my_words}", fixed = TRUE)
test(val, c(TRUE, TRUE, TRUE, FALSE, FALSE))

# escaping
x = string_vec("hey!, /twitter, !##!, a & b, c | d")

val = string_which(x, "\\/t")
test(val, 2)

val = string_which(x, "!\\/")
test(val, c(1, 3:5))

val = string_which(x, "\\!")
test(val, c(1, 3))

val = string_which(x, "a \\& b")
test(val, 4)

val = string_which(x, "f/twi |  \\|  | \\!")
test(val, c(1:3, 5))

# other
test(string_is("bonjour", "fixed/---|"), FALSE)



####
#### string_get ####
####


x = rownames(mtcars)

val = string_get(x, "Mazda")
test(val, c("Mazda RX4", "Mazda RX4 Wag"))

val = string_get(x, "!\\d & u")
test(val, c("Hornet Sportabout", "Lotus Europa"))

val = string_get(x, "Mazda", "Volvo", seq = TRUE)
test(val, c("Mazda RX4", "Mazda RX4 Wag", "Volvo 142E"))

val = string_get(x, "volvo", "mazda", seq = TRUE, ignore.case = TRUE)
test(val, c("Volvo 142E", "Mazda RX4", "Mazda RX4 Wag"))

cars = string_clean(rownames(mtcars)," .+")
val = string_get(cars, "i/^h", "i/^m", seq = TRUE)
test(val, c("Hornet", "Hornet", "Honda", "Mazda", 
            "Mazda", "Merc", "Merc", 
            "Merc", "Merc", "Merc", "Merc", "Merc", "Maserati"))

val = string_get(cars, "i/^h", "i/^m", seq.unik = TRUE)
test(val, c("Hornet", "Honda", "Mazda", "Merc", "Maserati"))

####
#### string_clean ####
####


x = c("hello world  ", "it's 5 am....")

val = string_clean(x, "o", "f/.")
test(val, c("hell wrld  ", "it's 5 am"))

val = string_clean(x, "f/o, .")
test(val, c("hell wrld  ", "it's 5 am"))

val = string_clean(x, "@ws")
test(val, c("hello world", "it's 5 am...."))

val = string_clean(x, "o(?= )", "@ws.p.i")
test(val, c("hell world", "it am"))

# replacement
val = string_clean(x, "o => a", "f/. => !")
test(val, c("hella warld  ", "it's 5 am!!!!"))

val = string_clean(x, "w/hello, it => _")
test(val, c("_ world  ", "_'s 5 am...."))

# total + string_ops
val = string_clean(x, "it/Hell => Heaven", "@'f/...'rm")
test(val, "Heaven")

# single
val = string_clean(x, "single/[aeiou] => _")
test(val, c("h_llo world  ", "_t's 5 am...."))

# magic
vowels = "aeiou"
val = string_clean(x, "m/[{vowels}] => _")
test(val, c("h_ll_ w_rld  ", "_t's 5 _m...."))

val = string_clean("bonjour les gens", "m/[{vowels}]{2,} => _")
test(val, "bonj_r les gens")

vowels_split = strsplit(vowels, "")[[1]]
val = string_clean("bonjour les gens", "m/{'|'c?vowels_split} => _")
test(val, "b_nj__r l_s g_ns")

vowels = "aeiou"
replace = "."
val = string_clean(x, "m/[{vowels}] => {replace}")

####
#### string_split2df ####
####


x = c("Nor rain, wind", "When my information")
id = c("ws", "jmk")

val = string_split2df(x, "[[:punct:] ]+")
test(val$x, c("Nor", "rain", "wind", "When", "my", "information"))
test(val$obs, c(1L, 1L, 1L, 2L, 2L, 2L))

val = string_split2df(x, "[[:punct:] ]+", id = id)
test(val$x, c("Nor", "rain", "wind", "When", "my", "information"))
test(val$obs, c(1L, 1L, 1L, 2L, 2L, 2L))
test(val$id, c("ws", "ws", "ws", "jmk", "jmk", "jmk"))

base = data.frame(text = x, my_id = id, stringsAsFactors = FALSE)
val = string_split2df(text ~ my_id, base, "[[:punct:] ]+")
test(val$text, c("Nor", "rain", "wind", "When", "my", "information"))
test(val$obs, c(1L, 1L, 1L, 2L, 2L, 2L))
test(val$my_id, c("ws", "ws", "ws", "jmk", "jmk", "jmk"))

base = data.frame(text = x, my_id = id, stringsAsFactors = FALSE)
val = string_split2df(text ~ my_id, base, "[[:punct:] ]+", add.pos = TRUE)
test(val$pos, c(1L, 2L, 3L, 1L, 2L, 3L))

x = c("One two, one two", "Microphone check")
val = string_split2df(x, "w/one")
test(val$x, c("One two, ", " two", "Microphone check"))

####
#### string_fill ####
####

x = c("bon", "bonjour les gens")
txt = string_fill(x)
test(txt, c("bon             ", "bonjour les gens"))

txt = string_fill(x, right = TRUE)
test(txt, c("             bon", "bonjour les gens"))

txt = string_fill(x, center = TRUE)
test(txt, "       bon      ", "bonjour les gens")

x = c(5, 15)
txt = string_fill(x, n = 3, right = TRUE)
test(txt, c("  5", " 15"))

txt = string_fill(x, n = 1, right = TRUE)
test(txt, c("5", "15"))

txt = string_fill(x, n = 3, right = TRUE, symbol = "0")
test(txt, c("005", "015"))

x = c("  hey  ", "de bas en haut")
txt = string_fill(x)

#
# string_vec ####
#

# regular
txt = string_vec("bon, jour, les gens, 
          je suis, la bas")
test(txt, c("bon", "jour", "les gens", "je suis", "la bas"))

# with nesting
y = c("Ana", "Charles")
z = c("Romeo Montaigu", "Juliette Capulet")
txt = string_vec("Jules, {y}, Francis, {'\\s+'S, ~(firstchar, ''c) ? z}")
test(txt, c("Jules", "Ana", "Charles", "Francis", "RM", "JC"))

# with/without variable protection
x = "x{{1:2}}, xx"
txt = string_vec(x, "y{{1:3}}", .delim = "{{ }}", .protect.vars = TRUE)
test(txt, c("x{{1:2}}, xx", "y1", "y2", "y3"))

txt = string_vec(x, "y{{1:3}}", .delim = "{{ }}")
test(txt, c("x1", "x2", "xx", "y1", "y2", "y3"))

txt = string_vec(x, "y{{1:3}}", .delim = "{{ }}", .split = FALSE)
test(txt, c("x1, xx", "x2, xx", "y1", "y2", "y3"))

# split
txt = string_vec("x1,x2\\,x3,   x4")
test(txt, c("x1", "x2,x3", "x4"))

txt = string_vec("x1;x2\\;x3;   x4", .split = ";")
test(txt, c("x1", "x2;x3", "x4"))

# matrix
mat = string_vec("1, 2, 3",
                 "4, 9, 9", 
                 "1, 2, 3", 
                 "7, 7, 9", .nmat = TRUE)
test(mat, structure(c(1, 4, 1, 7, 2, 9, 2, 7, 3, 9, 3, 9), dim = 4:3))

mat = string_vec("1, 2, 3, \n 4, 9, 9, \n 1, 2, 3", .nmat = TRUE)
test(mat, structure(c(1, 4, 1, 2, 9, 2, 3, 9, 3), dim = c(3L, 3L)))

mat = string_vec("1, 2, 3, 4, 9, 9, 1, 2, 3", .nmat = 3)
test(mat, structure(c(1, 4, 1, 2, 9, 2, 3, 9, 3), dim = c(3L, 3L)))

mat = string_vec("1, 2, 3, 4, 9, 9, 1, 2", .nmat = 2i)
test(mat, structure(c(1, 3, 9, 1, 2, 4, 9, 2), dim = c(4L, 2L)))

mat = string_vec("1, 2, 3 \n 4, 9, 9 \n 1, 2, 3", .nmat = TRUE, .last = "'\n'S")
test(mat, structure(c(1, 4, 1, 2, 9, 2, 3, 9, 3), dim = c(3L, 3L)))

mat = string_vec("1, 2, 3", .nmat = 2 + 3i, .last = "'\n'S")
test(mat, structure(c(1, 1, 2, 2, 3, 3), dim = 2:3))

mat = string_vec("1, john,\n 3, marie,\n 5, harry", .cmat = TRUE)
test(mat, structure(c("1", "3", "5", "john", "marie", "harry"), dim = 3:2))

df = string_vec("1, john,\n 3, marie,\n 5, harry", .df = TRUE)
target = data.frame(V1 = c(1, 3, 5), V2 = c("john", "marie", "harry"), 
                    stringsAsFactors = FALSE)
test(df, target)

df = string_vec("1, john,\n 3, marie,\n 5, harry", .df = "id, name")
names(target) = c("id", "name")
test(df, target)
test(names(df), c("id", "name"))

df = string_vec("1, john,\n 3, marie,\n 5, harry", .df = c("id", "name"))
test(names(df), c("id", "name"))

#
# paste_conditional ####
#

id = rep(1:2, each = 13)
x = paste_conditional(letters, id, "", names = FALSE)
test(x, c("abcdefghijklm", "nopqrstuvwxyz"))

x = paste_conditional(letters, 3 - id, "", names = FALSE)
test(x, c("nopqrstuvwxyz", "abcdefghijklm"))

base_cars = within(mtcars, carname <- row.names(mtcars))
base_cars = head(base_cars)
cars = paste_conditional(carname ~ gear + carb, base_cars, sep = ", ")
test(cars, c("Hornet 4 Drive, Valiant", "Hornet Sportabout", "Datsun 710", "Mazda RX4, Mazda RX4 Wag"))

test(names(cars), c("gear: 3, carb: 1", "gear: 3, carb: 2", "gear: 4, carb: 1", "gear: 4, carb: 4"))


####
#### string_extract ####
####

cars = row.names(mtcars)
val = string_extract(cars, "s/^\\w+")
val_bis = gsub(" .+", "", cars)
test(val, val_bis)

val = unlist(string_extract(cars, "^\\w+"))
test(val, val_bis)

val = string_extract(cars, "^\\w+", unlist = TRUE)
test(val, val_bis)

pat = "\\w+"
val = string_extract(cars, "m/^{pat}", unlist = TRUE)
test(val, val_bis)


####
#### string_split ####
####


time = "This is the year 2024."

val = string_split(time, " ")
test(val, c("This", "is", "the", "year", "2024."))

val = string_split(time, " ", simplify = FALSE)
test(val, list(c("This", "is", "the", "year", "2024.")))

val = string_split(time, "is")
test(val, c("Th", " ", " the year 2024."))

val = string_split(time, "w/is")
test(val, c("This ", " the year 2024."))

pat = "is"
val = string_split(time, "wm/{pat}")
test(val, c("This ", " the year 2024."))

# 1771597780
# DO NOT EDIT BY HAND: generated with generate_help_extensive() in help.R

setup_help_extensive = function(){
txt = c("", "# Interpolation and string operations: Principle ----|", 
"", "To interpolate a variable, say `x`, simply use `{x}`. For example `x = \"world\"; string_magic(\"hello {x}\")` leads to \"hello world\".", 
"", "To any interpolation you can add operations. Taking the previous example, say we want to display \"hello W O R L D\". This means upper casing all letters of the interpolated variable and adding a space between each of them. Do you think we can do that? Of course yes: `string_magic(\"hello {upper, ''s, c ? x}\")`. And that's it.", 
"", "Now let's explain what happened. Within the `{}` *box*, we first write a set of operations, here \"upper, ''s, c\", then add \"?\" and finally write the variable to interpolate, \"x\".  The operations (explained in more details below) are `upper`, upper-casing all letters, ''s: splitting with the empty string, 'c': concatenating with spaces the vector of string that was just split. The question mark means that the expression coming after it is to be evaluated (this is opposed to the exclamation mark presented next).", 
"", "The syntax is always the same: `{operations ? expression}`, where the operations section is a *comma separated* list of operations. These operations are of the form `'arg'op`, with `arg` the argument to the operator code `op`. These operations are performed sequantially from left to right.", 
"", "Some operations, like `upper`, accept options. You attach options to an operation with a dot followed by the option name. Formally: `op.option1.option2`, etc. Example: `x = \"hi there. what's up? fine.\" ; string_magic(\"He said: {upper.sentence, Q ? x}\")`. Leads to: `He said: \"Hi there. What's up? Fine.\"`.", 
"", "Both operators and options are partially matched. So `string_magic(\"He said: {up.s, Q ? x}\")` would also work.", 
"", "# Verbatim interpolation and nesting: Principle ----|", 
"", "Instead of interpolating a variable, say `x`, with `{x}`, you can use an exclamation mark to trigger varbatim evaluation. For example `string_magic(\"hello {!x}\")` would lead to \"hello x\". It's a bit disappointing, right? What's the point of doing that? Wait until the next two paragraphs.", 
"", "Verbatim evaluation is a powerful way to apply operations to plain text. For example: `string_magic(\"hello {upper, ''s, c ! world}\")` leads to \"hello W O R L D\".", 
"", "(A note in passing. The spaces surrounding the exclamation mark are non necessary, but when one space is present on both sides of the `!`, then the verbatim expression only begins after it. Ex: `\"{upper! hi}\"` leads to \" HI\" while `\"{upper ! hi}\"` leads to \"HI\" and `\"{upper !  hi}\"` leads to \" HI\".)", 
"", "The second advantage of verbatim evaluations is *nesting*. Anything in a verbatim expression is evaluated with the function `string_magic`. This means that any *box* will be evaluated as previously described. Let's give an example. You want to write the expression of a polynomial of order n: a + bx + cx^2 + etc. You can do that with nesting. Assume we have `n = 2`.", 
"", "Then `string_magic(\"poly({n}): {' + 'c ! {letters[1 + 0:n]}x^{0:n}}\")` leads to \"poly(2): ax^0 + bx^1 + cx^2\".", 
"", "How does it work? The verbatim expression (the one following the exclamation mark), here `\"{letters[1 + 0:n]}x^{0:n}\"`, is evaluated with `string_magic`. `string_magic(\"{letters[1 + 0:n]}x^{0:n}\")` leads to the vector c(\"ax^0\", \"bx^1\", \"cx^2\").", 
"", "The operation `' + 'c` then concatenates (or collapses) that vector with ' + '. This value is then appended to the previous string.", 
"", "We could refine by adding a cleaning operation in which we replace \"x^0\" and \"^1\" by the empty string. Let's do it:", 
"", "`string_magic(\"poly({n}): {' + 'c, 'x\\\\^0|\\\\^1'r ! {letters[1 + 0:n]}x^{0:n}}\")` leads to \"poly(2): a + bx + cx^2\", what we wanted.", 
"", "You can try to write a function to express the polynomial as before: although it is a simple task, my guess is that it will require more typing.", 
"", "# Operations: General syntax ----|", "", "As seen in the previous sections, within a *box* (i.e. `\"{}\"`), multiple operations can be performed. We can do so by stacking the operations codes and in a comma separated enumeration. Operations can have arguments, and operations can also have options. The general syntax, with argument and options, is:", 
"", "`{'arg1'op1.optionA.optionB, arg2 op2.optionC, `arg3`op3, 51op4 ? x}`", 
"", "The argument can appear in four forms: a) inside single or double quotes just before the operation name (`arg1` above), b) verbatim, separated with a space, just before the operation name (`arg2` above), c) inside bactick quotes the argument is evaluated from the environment (`arg3` above), or d) when the argument is an integer it can be juxtaposed to the opeation name (like in `op4` above).", 
"", "The options are always dot separated and attached to the operation name, they are specific to each operation.", 
"", "Both the operation name and the option names are partially matched.", 
"", "# Basic string operations ----|", "", "This section describes some of the most common string operations: extracting, replacing, collapsing, splitting, etc. These functions accept generic flags (\"ignore\", \"fixed\", \"word\") in their patterns (syntax: \"flags/pattern\"). Please see the dedicated section for more information on flags.", 
"", "+ s, split, S, Split: splits the string according to a pattern. The operations have different defaults: `' '` for `s` and 'split', and `',[ \\t\\n]*'` for `S` and 'Split' (i.e. comma separation). Ex.1: `string_magic(\"{S ! romeo, juliet}\")` leads to the vector c(\"romeo\", \"juliet\"). Ex.2: `string_magic(\"{'f/+'s, '-'c ! 5 + 2} = 3\")` leads to \"5 - 2 = 3\" (note the flag \"fixed\" in `s`'s pattern).", 
"  + c, C: to concatenate multiple strings into a single one. The two operations are identical, only their default change. c: default is `' '`, C: default is `', | and '`. The syntax of the argument is 's1' or 's1|s2'. s1 is the string used to concatenate (think `paste(x, collapse = s1)`). In arguments of the form `'s1|s2'`, `s2` will be used to concatenate the last two elements. Ex.1: `x = 1:4; string_magic(\"Et {' et 'c ? x}!\")` leads to \"Et 1 et 2 et 3 et 4!\". Ex.2: `string_magic(\"Choose: {', | or 'c ? 2:4}?\")` leads to \"Choose: 2, 3 or 4?\".", 
"  + x, X: extracts patterns from a string. Both have the same default: `'[[:alnum:]]+'`. `x` extracts the first match while `X` extracts **all** the matches. Ex.1: `x = c(\"6 feet under\", \"mahogany\") ; string_magic(\"{'\\\\w{3}'x ? x}\")` leads to the vector c(\"fee\", \"mah\"). Ex2.: `x = c(\"6 feet under\", \"mahogany\") ; string_magic(\"{'\\\\w{3}'X ? x}\")` leads to the vector c(\"fee\", \"und\", \"mah\", \"oga\").", 
"  + extract: extracts multiple patterns from a string, this is an alias to the operation `X` described above. Use the option \"first\" to extract only the first match for each string (behavior becomes like `x`). Ex: `x = c(\"margo: 32, 1m75\", \"luke doe: 27, 1m71\") ; string_magic(\"{'^\\\\w+'extract ? x} is {'\\\\d+'extract.first ? x}\")` leads to c(\"margo is 32\", \"luke is 27\").", 
"  + r, R: replacement within a string. The two operations are identical and have no default. The syntax is `'old'` or `'old => new'` with `'old'` the pattern to find and `new` the replacement. If `new` is missing, it is considered the empty string. This operation also accepts the flag \"total\" which instruct to replace the fulll string in case the pattern is found. Ex.1: `string_magic(\"{'e'r ! Where is the letter e?}\")` leads to \"Whr is th lttr ?\". Ex.2: `string_magic(\"{'(?<!\\\\b)e => a'R ! Where is the letter e?}\")` leads to \"Whara is tha lattar e?\". Ex.3: `string_magic(\"{'t/e => here'r ! Where is the letter e?}\")` leads to \"here\".", 
"  + clean: replacement with a string. Similar to the operation `r`, except that here the comma is a pattern separator, see detailed explanations in [string_clean()]. Ex: `string_magic(\"{'f/[, ]'clean ! x[a]}\")` leads to \"xa\".", 
"  + get: restricts the string only to values respecting a pattern. This operation has no default. Accepts the options \"equal\" and \"in\". By default it uses the same syntax as [string_get()] so that you can use regex flags and include logical operations with `' & '` and `' | '` to detect patterns. If the option \"equal\" is used, a simple string equality with the argument is tested (hence no flags are accepted). If the option \"in\" is used, the argument is first split with respect to commas and then set inclusion is tested. Example: `x = row.names(mtcars) ; string_magic(\"Mercedes models: {'Merc & [[:alpha:]]$'get, '^.+ 'r, C ? x}\")` leads to \"Mercedes models: 240D, 280C, 450SE, 450SL and 450SLC\".", 
"  + is: detects if a pattern is present in a string, returns a logical vector. This operation has no default. Accepts the options \"equal\" and \"in\". By default it uses the same syntax as [string_is()] so that you can use regex flags and include logical operations with `' & '` and `' | '` to detect patterns. If the option \"equal\" is used, a simple string equality with the argument is tested (hence no flags are accepted). If the option \"in\" is used, the argument is first split with respect to commas and then set inclusion is tested. Mostly useful as the final operation in a [string_ops()] call. Example: `x = c(\"Mark\", \"Lucas\") ; string_magic(\"Mark? {'i/mark'is, C ? x}\")` leads to \"Mark? TRUE and FALSE\".", 
"  + which: returns the index of string containing a specified pattern. With no default, can be applied to a logical vector directly. By default it uses the same syntax as string_which() so that you can use regex flags and include logical operations with `' & '` and `' | '` to detect patterns. If the option \"equal\" is used, a simple string equality with the argument is tested (hence no flags are accepted). If the option \"in\" is used, the argument is first split with respect to commas and then set inclusion is tested. Mostly useful as the final operation in a [string_ops()] call. Ex.1: `x = c(\"Mark\", \"Lucas\") ; string_magic(\"Mark is number {'i/mark'which ? x}.\")` leads to \"Mark is number 1.\".", 
"", "# Operations changing the length or the order ----|", "", 
"+ first: keeps only the first `n` elements. Example: `string_magic(\"First 3 numbers: {3 first, C ? mtcars$mpg}.\")` leads to \"First 3 numbers: 21, 21 and 22.8.\". Negative numbers as argument remove the first `n` values. You can add a second argument in the form `'n1|n2'first` in which case the first `n1` and last `n2` values are kept; `n1` and `n2` must be positive numbers.", 
"  + K, Ko, KO: keeps only the first `n` elements; has more options than `first`. The syntax is `'n'K`, `'n|s'K`, `'n||s'K`. The values Ko and KO only accept the two first syntax (with `n` only). `n` provides the number of elements to keep. If `s` is provided and the number of elements are greater than `n`, then in 'n|s' the string `s` is added at the end, and if 'n||s' the string s replaces the nth element. The string `s` accepts specials values:", 
"  + `:n:` or `:N:` which gives the total number of items in digits or letters (N)", 
"  + `:rest:` or `:REST:` which gives the number of elements that have been truncated in digits or letters (REST) Ex: `string_magic(\"{'3|:rest: others'K ? 1:200}\")` leads to the vector `c(\"1\", \"2\", \"3\", \"197 others\")`.", 
"  + The operator 'n'Ko is like `'n||:rest: others'K` and 'n'KO is like `'n||:REST: others'K`.", 
"  + last: keeps only the last `n` elements. Example: `string_magic(\"Last 3 numbers: {3 last, C ? mtcars$mpg}.\")` leads to \"Last 3 numbers: 19.7, 15 and 21.4.\". Negative numbers as argument remove the last `n` values.", 
"  + sort: sorts the vector in increasing order. Accepts optional arguments and the option \"num\". Example: `x = c(\"sort\", \"me\") ; string_magic(\"{sort, c ? x}\")` leads to \"me sort\". If an argument is provided, it must be a regex pattern that will be applied to the vector using [string_clean()]. The sorting will be applied to the modified version of the vector and the original vector will be ordered according to this sorting. Ex: `x = c(\"Jon Snow\", \"Khal Drogo\")`; `string_magic(\"{'.+ 'sort, C?x}\")` leads to \"Khal Drogo and Jon Snow\". The option \"num\" sorts over a numeric version (with silent conversion) of the vector and reorders the original vector accordingly. Values which could not be converted are last. **Important note**: the sorting operation is applied before any character conversion. If previous operations were applied, it is likely that numeric data were transformed to character. Note the difference: `x = c(20, 100, 10); string_magic(\"{sort, ' + 'c ? x}\")` leads to \"10 + 20 + 100\" while `string_magic(\"{n, sort, ' + 'c ? x}\")` leads to \"10 + 100 + 20\" because the operation \"n\" first transformed the numeric vector into character.", 
"  + dsort: sorts the vector in decreasing order. It accepts an optional argument and the option \"num\". Example: `string_magic(\"5 = {dsort, ' + 'c ? 2:3}\")` leads to \"5 = 3 + 2\". See the operation \"sort\" for a description of the argument and the option.", 
"  + rev: reverses the vector. Example: `string_magic(\"{rev, ''c ? 1:3}\")` leads to \"321\".", 
"  + unik: makes the string vector unique. Example: `string_magic(\"Iris species: {unik, C ? iris$Species}.\")` leads to \"Iris species: setosa, versicolor and virginica.\".", 
"  + table: computes the frequency of each element and attaches each element to its frequency. Accepts an argument which must be a character string representing a `string_magic` interpolation with the following variables: `x` (the element), `n` (its count) and `s` (its share). The default is `'{x} ({n ? n})'`. By default the resulting string vector is sorted by decreasing frequency. You can change how the vector is sorted with five options: `sort` (sorts on the elements), `dsort` (decreasing sort), `fsort` (sorts on frequency), `dfsort` (decreasing sort on freq. -- default), `nosort` (keeps the order of the first elements). Note that you can combine several sorts (to resolve the ties of elements with same frequencies). Example: `string_magic(\"Freq. of months: {'{x} ({n})'table, enum ? month.name[airquality$Month]}.\")`", 
"  + each: repeats each element of the vector `n` times. Option \"c\" then collapses the full vector with the empty string as a separator. Ex.1: `string_magic(\"{/x, y}{2 each ? 1:2}\")` leads to the vector `c(\"x1\", \"y1\", \"x2\", \"y2\")`. Ex.2: `string_magic(\"Large number: 1{5 each.c ! 0}\")` leads to \"Large number: 100000\".", 
"  + times: repeats the vector sequence `n` times. Option \"c\" then collapses the full vector with the empty string as a separator. Example: `string_magic(\"What{6 times.c ! ?}\")` leads to \"What??????\".", 
"  + rm: removes elements from the vector. Options: \"empty\", \"blank\", \"noalpha\", \"noalnum\", \"all\". The *optional* argument represents the pattern used to detect strings to be deleted. Ex.1: `x = c(\"Luke\", \"Charles\")`; `string_magic(\"{'i/lu'rm ? x}\")` leads to \"charles\". By default it removes empty strings. Option \"blank\" removes strings containing only blank characters (spaces, tab, newline). Option \"noalpha\" removes strings not containing letters. Option \"noalnum\" removes strings not containing alpha numeric characters. Option \"all\" removes all strings (useful in conditions, see the dedicated section). If an argument is provided, only the options \"empty\" and \"blank\" are available. Ex.2: `x = c(\"I want to enter.\", \"Age?\", \"21\")`; `string_magic(\"Nightclub conversation: {rm.noalpha, c ! - {x}}\")` leads to \"Nightclub conversation: - I want to enter. - Age?\"", 
"  + nuke: removes all elements, equivalent to `rm.all` but possibly more explicit (not sure). Useful in conditions, see the dedicated section. Example: `x = c(5, 7, 453, 647); string_magic(\"Small numbers only: {if(.>20 ; nuke), C ? x}\")` leads to \"Small numbers only: 5 and 7\";", 
"  + insert: inserts a new element to the vector. Options: \"right\" and \"both\". Option \"right\" adds the new element to the right. Option \"both\" inserts the new element on the two sides of the vector. Example: `string_magic(\"{'3'insert.right, ' + 'c ? 1:2}\")` leads to \"1 + 2 + 3\".", 
"  + `dp` or `deparse`: It deparses an R object and collapses it into a single string using newlines. By default it keeps only the first 50 lines of the deparsed object. The number of lines kept can be modified by using a numeric argument. It accepts the options: `short` and `vec`. If `short`, multiline deparsed elements are collapsed with `\\\\\\\\n` and the number of characters cannot exceed the value of the argument, which by default is 50. If `vec`, then the deparsed element is not collapsed into a single string, and only the first 50 (default) elements of the vector are kept. You can increase/reduce the number of elements kept with the argument. Example: `fml = y ~ x1 + x2; string_magic(\"The estimated model is {dp ? fml}.\")`", 
"", "# Formatting operations ----|", "", "+ lower: lower cases the full string.", 
"  + upper: upper cases the full string. Options: \"first\" and \"sentence\". Option \"first\" upper cases only the first character. Option \"sentence\" upper cases the first letter after punctuation. Ex: `x = \"hi. how are you? fine.\" ; string_magic(\"{upper.sentence ? x}\")` leads to \"Hi. How are you? Fine.\".", 
"  + title: applies a title case to the string. Options: \"force\" and \"ignore\". Option \"force\" first puts everything to lowercase before applying the title case. Option \"ignore\" ignores a few small prepositions (\"a\", \"the\", \"of\", etc). Ex: `x = \"bryan is in the KITCHEN\" ; string_magic(\"{title.force.ignore ? x}\")` leads to \"Bryan Is in the Kitchen\".", 
"  + ws: normalizes whitespaces (WS). It trims the whitespaces on the edges and transforms any succession of whitespaces into a single one. Can also be used to further clean the string with its options. Options: \"punct\", \"digit\", \"isolated\". Option \"punct\" cleans the punctuation. Option \"digit\" cleans digits. Option \"isolated\" cleans isolated letters. WS normalization always come after any of these options. **Important note:** punctuation (or digits) are replaced with WS and **not** the empty string. This means that `string_magic(\"ws.punct ! Meg's car\")` will become \"Meg s car\".", 
"  + trimws: trims the white spaces on both ends of the strings.", 
"  + q, Q, bq: to add quotes to the strings. q: single quotes, Q: double quotes, bq: back quotes. `x = c(\"Mark\", \"Pam\"); string_magic(\"Hello {q, C ? x}!\")` leads to \"Hello 'Mark' and 'Pam'!\".", 
"  + format, Format: applies the base R's function [base::format()] to the string. By default, the values are left aligned, *even numbers* (differently from [base::format()]'s behavior). The upper case command (`Format`) applies right alignment. Options: \"0\", \"zero\", \"left\", \"right\", \"center\". Options \"0\" or \"zero\" fills the blanks with 0s: useful to format numbers. Option \"right\" right aligns, and \"center\" centers the strings. Default is left alignment. Ex: `x = c(1, 12345); string_magic(\"left: {format.0, q, C ? x}, right: {Format, q, C ? x}\")` leads to \"left: '000001' and '12,345', right: '     1' and '12,345'\".", 
"  + %: applies [base::sprintf()] formatting. The syntax is 'arg'% with arg an sprintf formatting, or directly the sprint formatting, e.g. `% 5s`. Example: `string_magic(\"pi = {%.3f ? pi}\")` leads to \"pi = 3.142\".", 
"  + stopwords: removes basic English stopwords (the snowball list is used). The stopwords are replaced with an empty space but the left and right WS are untouched. So WS normalization may be needed (see operation `ws`). `x = c(\"He is tall\", \"He isn't young\"); string_magic(\"Is he {stop, ws, C ? x}?\")` leads to \"Is he tall and young?\".", 
"  + ascii: turns all letters into ASCII with transliteration. Failed translations are transformed into question marks. Options: \"silent\", \"utf8\". By default, if some conversion fails a warning is prompted. Option \"silent\" disables the warning in case of failed conversion. The conversion is done with [base::iconv()], option \"utf8\" indicates that the source endocing is UTF-8, can be useful in some cases.", 
"  + round, r0 to r6: formats numbers by rounding at a given level. Options:  `0` to `9`, `int`, `nocomma`, `s0`-`s9`. Option `0` to `9` controls the number of digits to round at. Option `int` is whether to preserve integers from formattting. Option `nocomma` conrtols whether to drop the comma separating the thousands. `s0`-`s9` also keeps a given number of significant digits.", 
"  + signif, s0 to s6: formats numbers by displaying a certain number of significant digits. Options:  `0` to `9`, `int`, `nocomma`, `r0`-`r9`. Option `0` to `9` controls the number of significant digits to display. Option `int` is whether to preserve integers from formattting. Option `nocomma` conrtols whether to drop the comma separating the thousands. `r0`-`r9` also rounds at a given number of decimals.", 
"  + n: formats integers by adding a comma to separate thousands. Options: \"letter\", \"upper\", \"0\", \"zero\". The option \"letter\" writes the number in letters (large numbers keep their numeric format). The option \"upper\" is like the option \"letter\" but uppercases the first letter. Options \"0\" or \"zero\" left pads numeric vectors with 0s. Ex.1: `x = 5; string_magic(\"He's {N ? x} years old.\")` leads to \"He's five years old.\". Ex.2: `x = c(5, 12, 52123); string_magic(\"She owes {n.0, '$'paste, C ? x}.\")` leads to \"She owes $5, $12 and $52,123.\".", 
"  + N: same as `n` but automatically adds the option \"letter\".", 
"  + nth: when applied to a number, these operators write them as a rank. Options: \"letter\", \"upper\", \"compact\". Ex.1: `n = c(3, 7); string_magic(\"They finished {nth, enum ? n}!\")` leads to \"They finished 3rd and 7th!\". Option \"letter\" tries to write the numbers in letters, but note that it stops at 20. Option \"upper\" is the same as \"letter\" but uppercases the first letter. Option \"compact\" aggregates consecutive sequences in the form \"start_n_th to end_n_th\". Ex.2: `string_magic(\"They arrived {nth.compact ? 5:20}.\")` leads to \"They arrived 5th to 20th.\". Nth: same as `nth`, but automatically adds the option \"letter\". Example: `n = c(3, 7); string_magic(\"They finished {Nth, enum ? n}!\")` leads to \"They finished third and seventh!\".", 
"  + ntimes: write numbers in the form `n` times. Options: \"letter\", \"upper\". Option \"letter\" writes the number in letters (up to 100). Option \"upper\" does the same as \"letter\" and uppercases the first letter. Example: `string_magic(\"They lost {C ! {ntimes ? c(1, 12)} against {S!Real, Barcelona}}.\")` leads to \"They lost once against Real and 12 times against Barcelona.\".", 
"  + Ntimes: same as `ntimes` but automatically adds the option \"letter\". Example: `x = 5; string_magic(\"This paper was rejected {Ntimes ? x}...\")` leads to \"This paper was rejected five times...\".", 
"  + firstchar, lastchar: to select the first/last characters of each element. Ex: `string_magic(\"{19 firstchar, 9 lastchar ! This is a very long sentence}\")` leads to \"very long\". Negative numbers remove the first/last characters.", 
"  + k: to keep only the first n characters (like `firstchar` but with more options). The argument can be of the form `'n'k`, `'n|s'k` or `'n||s'k` with `n` a number and `s` a string. `n` provides the number of characters to keep. Optionnaly, only for strings whose length is greater than `n`, after truncation, the string `s` can be appended at the end. The difference between 'n|s' and 'n||s' is that in the second case the strings will always be of maximum size `n`, while in the first case they can be of length `n + nchar(s)`. Ex: `string_magic(\"{4k ! long sentence}\")` leads to \"long\",  `string_magic(\"{'4|..'k ! long sentence}\") ` leads to \"long..\", `string_magic(\"{'4||..'k ! long sentence}\")` leads to \"lo..\".", 
"  + fill, align (alias), width (alias): fills the character strings up to a size in order to fit a given width. Options: \"left\", \"right\", \"center\". Accepts arguments of the form `'n'` or `'n|s'`, with `n` a number and `s` a symbol. Default is left-alignment of the strings. Option \"right\" right aligns and \"center\" centers the strings. When using `'n|s'`, the symbol `s` is used for the filling. By default if no argument is provided, the maximum size of the character string is used. See help for [string_fill()] for more information. Ex.1: `string_magic(\"Numbers: {'5|0'fill.right, C ? c(1, 55)}\")` leads to \"Numbers: 00001 and 00055\".", 
"  + paste, append: pastes some character to all elements of the string. This operation has no default. Options: \"left\", \"both\", \"right\", \"front\", \"back\", \"delete\". By default, a string is pasted on the left. By default, it pastes on the left. Option \"right\" pastes on the right and \"both\" pastes on both sides. Option \"front\" only pastes on the first element while option \"back\" only pastes on the last element. Option \"delete\" first replaces all elements with the empty string. Example: `string_magic(\"6 = {'|'paste.both, ' + 'c ? -3:-1}\")` leads to \"6 = |-3| + |-2| + |-1|\". The argument can be of the form `s` or `s1|s2`. If of the second form, this is equivalent to chaining two `paste` operations, once on the left and once on the right: `'s1'paste, 's2'paste.right`.", 
"  + join: joins lines ending with a double backslash. Ex: `x = \"the sun \\\\\\n is shining\"`; `string_magic(\"{join ? x}\")` leads to \"the sun is shining\".", 
"  + escape: adds backslashes in front of specific characters. Options `\"nl\"`, `\"tab\"`. Option `\"nl\"` escapes the newlines (`\\n`), leading them to be displayed as `\"\\\\\\\\n\"`. Option `\"tab\"` does the same for tabs (`\"\\t\"`). This is useful to make the value free of space formatters. The default behavior is to escape both newlines and tabs.", 
"", "# Other operations ----|", "", "+ num: converts to numeric. Options: \"warn\", \"soft\", \"rm\", \"clear\". By default, the conversion is performed silently and elements that failed to convert are turned into NA. Option \"warns\" displays a warning if the conversion to numeric fails. Option \"soft\" does not convert if the conversion of at least one element fails. Option \"rm\" converts and removes the elements that could not be converted. Option \"clear\" turns failed conversions into the empty string, and hence lead to a character vector. Example: `x = c(5, \"six\"); string_magic(\"Compare {num, C, q ? x} with {num.rm, C, q ? x}.\")` leads to \"Compare '5 and NA' with '5'.\", and `string_magic(\"Compare {num.soft, C, q ? x} with {clear, C, q ? x}.\")` leads to \"Compare '5 and six' with '5 and '.\".", 
"  + enum: enumerates the elements. It creates a single string containing the comma separated list of elements. If there are more than 7 elements, only the first 6 are shown and the number of items left is written. For example `string_magic(\"enum ? 1:5\")` leads to \"1, 2, 3, 4, and 5\". You can add the following options by appending the letter to enum after a dot:", 
"  + q, Q, or bq: to quote the elements", "  + or, nor: to finish with an 'or' (or 'nor') instead of an 'and'", 
"  + comma: to finish the enumeration with \", \" instead of \", and\".", 
"  + i, I, a, A, 1: to enumerate with this prefix, like in: i) one, and ii) two", 
"  + a number: to tell the number of items to display Ex.1: `x = c(\"Marv\", \"Nancy\"); string_magic(\"The main characters are {enum ? x}.\")` leads to \"The main characters are Marv and Nancy.\". Ex.2: `x = c(\"orange\", \"milk\", \"rice\"); string_magic(\"Shopping list: {enum.i.q ? x}.\")` leads to \"Shopping list: i) 'orange', ii) 'milk', and iii) 'rice'.\"", 
"  + len: gives the length of the vector. Options \"letter\", \"upper\", \"num\". Option \"letter\" writes the length in words (up to 100). Option \"upper\" is the same as letter but uppercases the first letter. By default, commas are added to separate thousands. Use uption \"num\" to preserve a regular numeric format. Example: `string_magic(\"Size = {len ? 1:5000}\")` leads to \"Size = 5,000\".", 
"  + swidth: stands for screen width. Formats the string to fit a given width by cutting at word boundaries and adding newlines appropriately. Accepts arguments of the form `'n'` or `'n|s'`, with `n` a number and `s` a string. An argument of the form `'n|s'` will add `s` at the beginning of each line. Further, by default a trailing white space is added to `s`; to remove this behavior, add an underscore at the end of it. The argument `n` is either an integer giving the target character width (minimum is 15), or it can be a fraction expressing the target size as a fraction of the current screen. Finally it can be an expression that uses the variable `.sw` which will capture the value of the current screen width. Ex.1: `string_magic(\"{15 swidth ! this is a long sentence}\")` leads to \"this is a long\\\\nsentence\". Ex.2: `string_magic(\"{15 swidth.#> ! this is a long sentence}\")` leads to \"#> this is a long\\\\n#> sentence\".", 
"  + difftime: displays a formatted time difference. Option \"silent\" does not report a warning if the operation fails. It accepts either objects of class `POSIXt` or `difftime`. Example: `x = Sys.time() ; Sys.sleep(0.5) ; string_magic(\"Time: {difftime ? x}\")` leads to something like \"Time: 514ms\".", 
"", "# Group-wise operations ----|", "", "In `string_magic`, the splitting operation `s` (or `S`) keeps a memory of the strings that were split. Use the tilde operator, of the form `~(op1, op2)`, to apply operations group-wise, to each of the split strings.", 
"", "Better with an example. `x = c(\"Oreste, Hermione\", \"Hermione, Pyrrhus\", \"Pyrrhus, Andromaque\") ;` `string_magic(\"Troubles ahead: {S, ~(' loves 'c), C ? x}.\")` leads to \"Troubles ahead: Oreste loves Hermione, Hermione loves Pyrrhus and Pyrrhus loves Andromaque.\".", 
"", "Almost all operations can be applied group-wise (although only operations changing the order or the length of the strings really matter).", 
"", "# Conditional operations ----|", "", "There are two operators to apply operations conditionally: `if` and `vif`, the latter standing for *verbatim if*.", 
"", "The syntax of `if` is `if(cond ; ops_true ; ops_false)` with `cond` a condition (i.e. logical operation) on the value being interpolated, `ops_true` a comma-separated sequence of operations if the condition is `TRUE` and `ops_false` an *optional* a sequence of operations if the condition is `FALSE`.", 
"", "Ex.1: Let's take a sentence, delete words of less than 4 characters, and trim words of 7+ characters. x = \"Songe Cephise a cette nuit cruelle qui fut pour tout un peuple une nuit eternelle\" `string_magic(\"{' 's, if(.nchar<=4 ; nuke ; '7|..'k), c ? x}\")`. Let's break it down. First the sentence is split w.r.t. spaces, leading to a vector of words. Then we use the special variable `.nchar` in `if`'s condition to refer to the number of characters of the current vector (the words). The words with less than 4 characters are nuked (i.e. removed), and the other words are trimmed at 7 characters. Finally the modified vector of words is collapsed with the function `c`, leading to the result.", 
"", "The condition `cond` accepts the following special values: `.` (the dot), `.nchar`, `.C`, `.len`, `.N`. The dot, `.`, refers to the current vector. `.nchar` represent the number of characters of the current vector (equivalent to `nchar(.)`). `.C` is an alias to `.nchar`. `.len` represent the length of the current vector (equivalent to `length(.)`). `.N` is an alias to `.len`.", 
"", "If a condition leads to a result of length 1, then the operations are applied to the full string vector and not element-wise (as was the case in Ex.1). Contrary to element-wise conditions for which operations modifying the length of the vectors are forbidden (apart from nuking), such operations are fine in full-string conditions.", 
"", "Ex.2: `x = string_magic(\"x{1:10}\")`; `string_magic(\"y = {if(.N>4 ; 3 first, '...'insert.right), ' + 'c ? x}\")` leads to \"y = x1 + x2 + x3 + ...\". the same opration applied to `x = string_magic(\"x{1:4}\")` leads to \"y = x1 + x2 + x3 + x4\".", 
"", "For `vif`, the syntax is `vif(cond ; verb_true ; verb_false)` with `verb_true` a verbatim value with which the vector will be replaced if the condition is `TRUE`. This is similar for `verb_false`. The condition works as in `if`.", 
"", "Ex.3: `x = c(1, 25, 12, 6) ; string_magic(\"Values: {vif(.<10 ; <10), C ? x}\")` leads to \"Values: <10, 25, 12 and <10\". As we can see values lower than 10 are replaced with \"<10\" while other values are not modified.", 
"", "Ex.4: `x = string_magic(\"x{1:10}\")`; `string_magic(\"y = {vif(.N>4 ; {S!{x[1]}, ..., {last?x}}), ' + 'c ? x}\")` leads to \"y = x1 + ... + x10\". Let's break it down. If the length of the vector is greater than 4 (here it's 10), then the full string is replaced with `\"{S!{x[1]}, ..., {last?x}}\"`. Interpolation applies to such string. Hence the split operation `S` breaks the string w.r.t. the commas (default behavior), leading to the vector `c(\"{x[1]}\", \"...\", \"{last?x}\")`. Since the string contains curly brackets, interpolation is applied again. This leads to the vector `c(\"x1\", \"...\", \"x10\")`. Finally, this vector is collapsed with ' + ' leading to the final string. Note that there are many ways to get to the same result. Here is another example: `string_magic(\"y = {vif(.N>4 ; {x[1]} + ... + {last?x} ; {' + 'c ? x}) ? x}\")`.", 
"", "The `vif` condition allows the use of '.' to refer to the current value in `verb_true` and `verb_false`, as illustrated by the last example:", 
"", "Ex.5: `string_magic(\"{4 last, vif(. %% 2 ; x{.} ; y{rev?.}), C ? 1:11}\")` leads to \"y10, x9, y8 and x11\".", 
"", "# Special interpolation: if-else ----|", "", "Using an ampersand (\"&\") as the first character of an interpolation leads to an *if-else* operation. Using two ampersands (\"&&\") leads to a slightly different operation described at the end of this section.", 
"", "The syntax is as follows: `{&cond ; verb_true ; verb_false}` with `cond` a condition (i.e. logical operation) on the value being interpolated, `verb_true` a verbatim value with which the vector will be replaced if the condition is `TRUE` and `verb_false` an *optional* verbatim value with which the vector will be replaced if the condition is `FALSE`. If not provided, `verb_false` is considered to be the empty string unless the operator is the double ampersand described at the end of this section.", 
"", "Note that in `cond`, you can use the function `len`, an alias to `length`.", 
"", "Ex.1: `x = 1:5`; \\code{string_magic(\"x is {&len(x)<10 ; short ; {`log10(.N)-1`times, ''c ! very }long}\")} leads to \"x is short\". With `x = 1:50`, it leads to \"x is long\", and to \"x is very very long\" if `x = 1:5000`.", 
"", "If a condition leads to a result of length 1, the full string is replaced by the verbatim expression. Further, this expression will be interpolated if requested. This was the case in Ex.1 where `verb_false` was interpolated.", 
"", "If the condition's length is greater than 1, then each logical values equal to `TRUE` is replaced by `verb_true`, and `FALSE` or `NA` values are replaced with `verb_false`. Note, importantly, that **no interpolation is perfomed in that case**.", 
"", "Ex.2: `x = 1:3 ; string_magic(\"x is {&x == 2 ; two ; not two}\")` leads to the vector `c(\"x is not two\", \"x is two\", \"x is not two\")`.", 
"", "In that example, when x is odd, it is replaced with \"odd\", and when even it is replaced with the elements of y.", 
"", "Using the two ampersand operator (\"&&\") is like the simple ampersand version but the default for `verb_false` is the variable used in the condition itself. So the syntax is `{&&cond ; verb_true}` and *it does not accept* `verb_false`.", 
"", "Ex.3: `i = 3 ; string_magic(\"i = {&&i == 3 ; three}\")` leads to \"i = three\", and to \"i = 5\" if `i = 5`.", 
"", "# Special interpolation: Pluralization ----|", "", "There is advanced support for pluralization which greatly facilitates the writing of messages in natural language.", 
"", "There are two ways to pluralize: over length or over value. To trigger a \"pluralization\" interpolation use as first character:", 
"  - `$` to pluralize over the length of a variable (see Ex.2)", 
"  - `#` to pluralize over the value of a variable (see Ex.1)", 
"", "Ex.1: `x = 5; string_magic(\"I bought {N?x} book{#s}.\")` leads to \"I bought five books.\". If `x = 1`, this leads to \"I bought one book.\".", 
"", "The syntax is `{#plural_ops ? variable}` or `{#plural_ops}` where `plural_ops` are specific pluralization operations which will be described below. The pluralization is perfomed *always* with respect to the value of a variable. You can either add the variable explicitly (`{#plural_ops ? variable}`) or refer to it implicitly (`{#plural_ops}`). If implicit, then the algorithm will look at the previous variable that was interpolated and pluralize over it. This is exaclty what happens in Ex.1 where `x` was interpolated in `{N?x}` and plural operation `s` in `{#s}` then applied to `x`. It was equivalent to have `{#s ? x}`. If a variable wasn't interpolated before, then the next interpolated variable will be used (see Ex.2). If no variable is interpolated at all, an error is thrown.", 
"", "Ex.2: `x = c(\"J.\", \"M.\"); string_magic(\"My BFF{$s, are} {C?x}!\")` leads to \"My BFFs are J. and M.!\". If \"x = \"S.\", this leads to \"My BFF is S.!\".", 
"", "Pluralizing accepts the following operations:", "  - s, es: adds an \"s\" (or \"es\") if it is plural (> 1), nothing otherwise. Accepts the option `0` or `zero` which treats a 0-length or a 0-value as plural.", 
"  - y or ies: adds an 'y' if singular and 'ies' if plural (>1). Accepts the option `0` or `zero` which treats a 0-length or a 0-value as plural.", 
"  - enum: enumerates the elements (see help for the regular operation `enum`)", 
"  - n, N, len, Len: add the number of elements (\"len\") or the value (\"n\") of the variable as a formatted number or in letters (upper case versions). Accepts the options `letter` (to write in letter) and `upper` (to uppercase the first letter).", 
"  - nth, ntimes: writes the value of the variable as an order (nth) or a frequence (ntimes). Accepts the option `letter` to write the numbers in letters (uppercase version of the operator does the same).", 
"  - is, or any verb: conjugates the verb appropriately", "", 
"You can chain operations, in that case a whitespace is automatically added between them.", 
"", "Ex.3: `x = c(7, 3, 18); string_magic(\"The winning number{$s, is, enum ? sort(x)}.\")` leads to \"The winning numbers are 3, 7 and 18.\". With `x = 7` this leads to \"The winning number is 7.\".", 
"", "On top of the previous operations, there is a special operation allowing to add verbatim text depending on the situation. The syntax is as follows:", 
"  - `(s1;s2)`: adds verbatim 's1' if singular and 's2' if plural (>1)", 
"  - `(s1;s2;s3)`: adds verbatim 's1' if zero, 's2' if singular (=1) and 's3' if plural", 
"  - `(s1;;s3)`: adds verbatim 's1' if zero, 's3' if singular or plural (i.e. >=1)", 
"", "These case-dependent verbatim values **are interpolated** (if appropriate). In these interpolations you need not refer explicitly to the variable for pluralization interpolations.", 
"", "Ex.4: `x = 3; string_magic(\"{#(Sorry, nothing found.;;{#N.upper} match{#es, were} found.)?x}\")` leads to \"Three matches were found.\". If \"x = 1\", this leads to \"One match was found.\" and if \"x = 0\" this leads to \"Sorry, nothing found.\".", 
"", "# Escaping and special cases ----|", "", "The opening and closing brakets, `{}`, are special characters and cannot be used as regular text. To bypass their special meaning, you need to escape them with a double backslash.", 
"", "Ex.1: `string_magic(\"open = \\\\\\\\{, close = }\")` leads to `\"open = {, close = }\"`. Ex.2: `string_magic(\"many {5 times.c ! \\\\\\\\}}\")` leads to `many }}}}}`.", 
"", "You only need to escape the special delimiters which the algorithm is currently looking for. As you can see, you don't need to escape the closing bracket in Ex.1 since no box was open. On the other hand, you need to escape it in Ex.2.", 
"", "Alternatively, use the argument `.delim` to set custom delimiters.", 
"", "Ex.3: `string_magic(\"I {'can {write} {{what}} I want'}\")` leads to `\"I can {write} {{what}} I want\"`.", 
"", "Since `{expr}` evaluates `expr`, the stuff inside the *box*, you can pass a character string and it will stay untouched.", 
"", "In the few operations expecting a semi-colon (if-else and pluralization), it can also be escaped with a double backslash.", 
"", "In interpolations, the exclamation mark (`!`) signals a verbatim expression. But what if you use it to mean the logical operation *not* in an operation-free interpolation? In that case, you need a hack: use a question mark (`?`) first to indicate to the algorithm that you want to evaluate the expression.", 
"", "Ex.4: `string_magic(\"{!TRUE} is {?!TRUE}\")` leads to \"TRUE is FALSE\". The first expression is taken verbatim while the second is evaluated."
)

  options(string_magic_help_extensive = txt)
}

---
title: "`string_magic`: Magical String Interpolation"
author: "Laurent R. Bergé"
date: "`r Sys.Date()`"
output:
  html_document:
    theme: journal
    highlight: haddock
    number_sections: yes
    toc: yes
    toc_depth: 3
    toc_float:
      collapsed: no
      smooth_scroll: no
  pdf_document:
    toc: yes
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{string_magic_guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

# we preload it to avoid ugly (was compiled with R.x.x) warnings in the doc
library(stringmagic)
```

String interpolation is a common feature of many programming languages and is
used to insert the value of variables directly into character strings.
This is useful to compose dynamic messages in a natural way. For example
`"Hello {x}"` will be interpolated as "Hello John" if `x = "John"`.

What if you want to apply some modifications to the variable before insertion?
Like normalizing the case, trimming at 40 characters, or concatenating? 
In general, you can't. Well, actually you can but have to use the language's own native tools. 
And it can be surprising how performing a few basic operations can quickly amount to lot of typing.

The aim of the package `stringmagic` is to simplify and empower string interpolation. 
With the `string_magic` function, you can apply 50+ basic operations to interpolated variables, 
interpolations can be nested, there is advanced support for pluralization, and much more.

# Basic interpolation {#sec_basic}

`string_magic` behaves in a similar way to the well known function [glue](https://glue.tidyverse.org/).
Almost anything `glue` can do, `string_magic` can do.
Use curly brackets to interpolate variables: i.e to insert their value directly in the string:
```{r}
library(stringmagic)
x = "John" ; y = "Mary"
string_magic("Hi {x}! How's {y} doing?")
```

You can either interpolate a variable that already exists, as above, or provide the variable as an argument:
```{r}
string_magic("Hi {x}! How's {y} doing?", y = "Jane")
```

So far, so good. Now let's come to `string_magic`'s originality: the ability to add operations to interpolations.

# How to add operations to interpolations {#sec_interpol}

In `string_magic` you can apply any arbitrary chain of operations to the interpolated variables. 
The syntax is as follows:

![](images/operation-template.png)

The `operations` are a comma separated sequence of keywords, each keyword being bound 
to a specific function. Here is an example in which we apply two operations:
```{r}
lovers = c("romeo", "juliet")
string_magic("Famous lovers: {title, enum ? lovers}.")
```

![](images/example-simple_operation.png)

As shown, `title` puts to title case and `enum` creates an enumeration (i.e. turns
the elements of a vector into a single character string). These two operations are applied 
sequentially from left to right to the variable `lovers`. 
You can stack as many operations as you wish.

Here is a short list of common operations: `enum` to enumerate; `lower`, `upper` and `title` to change the case;
`unik` to create the vector of unique elements; `n`, `N` to format integers;
`split` to break the string with respect to a pattern; `replace` to make pattern replacements; `collapse`
to concatenate the strings; `first` to select the first elements; `q`, `Q` and `bq` for quoting; `align` to
align the elements.

The question mark defines the separation between the operations and the expression. After 
the question mark, you can use any valid R expression. For example:
```{r}
string_magic("The max of each variable is: {enum ? sapply(iris[, 1:4], max)}.")
```

## Adding arguments to operations

Many operations use arguments. Arguments are passed using quoted text
just before the operation. The syntax is:

![](images/argument.png)

Let's take the example of splitting an email address and keeping the text before the domain:
```{r}
email = "John@Doe.com"
string_magic("This message comes from {'@'split, first ? email}.")
```

![](images/example-argument.png)

In this example, `split` breaks `"John@Doe.com"` with respect to `"@"`, then `first` keeps 
only the first element. Almost all operations have default values. For `split`, the 
default value is a single space. The operation `first` also accepts arguments and its 
default is 1. Let's change it to keep the first two values:

```{r}
string_magic("The first two species are: {unik, '2'first, q, enum ? iris$Species}.")
```

### Alternative ways to pass arguments 

There are three ways to pass arguments to operations:
![](images/passing-arguments.png)

The first, canonical way, is using quotes like in the examples of the previous section. 

The second way uses a space to separate the 
argument from the operation and hence forbids any space in the argument.
```{r}
string_magic("The first five sepal lengths are: {5 first, enum ? iris$Sepal.Length}.")
```

The third way uses backticks. Passing arguments with backticks requests an evaluation
of the argument, using the variables defined in the calling environment. Let's
redo the previous example with argument evaluation:
```{r}
n = 5
string_magic("The first {N?n} sepal lengths are: {`n`first, enum ? iris$Sepal.Length}.")
```


## Adding options to operations

Options are keywords working like flags (i.e. things
that can be turned on) and change the behavior of the current operation. 
Many operations acccept options.
Add options using a dot separated sequence of keywords attached to the operation:

![](images/options.png)


We have seen the `enum` operation in an earlier example, let's add a couple of options to it.
```{r}
fields = c("maths", "physics")
string_magic("This position requires a PhD in either: {enum.i.or ? fields}.")
```

![](images/example-options.png)

As just seen, the option `i` triggers an enumeration with roman lower case numbers, and option
`or` finishes the enumeration with an `or` instead of an `and`. The order of the options
does not matter. Note that there is partial matching on the options, so that long 
option names can be shortened. 

Let's have another example where we wish to add upper cases but only for the first letter
of each sentence. To do this, we use the option `sentence` of the operation `upper`:
```{r}
oversight = "hey, you forgot. forgot what? forgot the capital letters!"
string_magic("{upper.sentence ? oversight}")
```

**Wrapping up.** With `string_magic` you can perform string interpolations *à la* `glue` and 
also empower these interpolations by adding operations on-the-fly (see the [operations reference](https://lrberge.github.io/stringmagic/articles/ref_operations.html)). 
This can take us a long way already. We are presenting next how you can nest interpolations and follow with
an introduction to pluralization.

# Nesting interpolations

In `string_magic`, you have two types of interpolations: regular interpolations (the ones we have covered so far),
and verbatim evaluation. The syntax of vertatim interpolations is as follows:

![](images/verbatim.png)

Anything after the exclamation mark is treated as verbatim: i.e. anything coming 
before the closing bracket is treated as a simple character string. Let's give an example:

```{r}
string_magic("Hello {upper ! world}.")
```

In that case, since `world` is placed after the excalamation mark (and not a question mark!), 
the character string `"world"` is passed to the operation `upper`, leading to the result.

The main interest of verbatim evaluations is nesting. Indeed, the function `string_magic` is always applied 
to the verbatim character string *before* any operation is applied. This means that 
interpolations within the verbatim text will be resolved. For example, let's create the 
sum from `x1` to `x3`:
```{r}
string_magic("y = {' + 'collapse ! x{1:3}}")
```

![](images/example-verbatim.png)

Now let's tackle a more challenging example: write a polynomial of order `n`. It should be
of the form `a + bx + cx^2 + etc`.

```{r}
n = 2
string_magic("poly({n}): {' + 'collapse ! {letters[1 + 0:n]}x^{0:n}}")
```

Let's break it down. The verbatim expression (the one following the exclamation mark),
here `"{letters[1 + 0:n]}x^{0:n}"`, is evaluated with `string_magic`.
`string_magic("{letters[1 + 0:n]}x^{0:n}")` leads to the vector `c("ax^0", "bx^1", "cx^2")`.

The operation `' + 'collapse` then concatenates that vector with `' + '`.
This value is then appended to the previous string.

We can refine by adding a cleaning operation in which we replace `"x^0"` and `"^1"` 
by the empty string. Let's do it:

```{r}
n = 4
string_magic("poly({n}): {' + 'c, 'f/x^0, ^1'clean ! {letters[1 + 0:n]}x^{0:n}}")
```

To achieve our goal, we used the operation `clean` which replaces
comma separated regular expression patterns (of the form `'flags/pat1, pat2, etc'`) with the empty
string. We also used the flag `f` (standing for 'fixed') to suppress the regular expression
meaning of the patterns (because `'^'` has a specific meaning in regexes) -- see the 
[vignette on flags](https://lrberge.github.io/stringmagic/articles/ref_regex_flags.html) for more information.
Finally, you can notice that the operation `collapse` has a single letter alias.


# Pluralization: A brief introduction

Pluralization is about the automatic writing of a message in singular or plural form, depending on context.
This feature is especially important for developers who often need to write human readable 
messages to inform the user. `string_magic` has an extensive support for pluralization, which we 
briefly introduce here. 

The syntax for pluralization is as follows:

![](images/plural.png)

After the opening bracket, to trigger pluralization use an hashtag (`#`) or a dollar sign (`$`) followed
by plural operations. There is two kinds of pluralization: either on the value of the expresison (the `#`),
either on its length (the `$`). Here is an example of pluralization on length with three operations:
```{r}
friends = c("Piglet", "Eeyore")
string_magic("My best friend{$s, are, enum ? friends}. Who am I?")

friends = "Mercutio"
string_magic("My best friend{$s, are, enum ? friends}. Who am I?")
```
As you can see, the name of the operations is intuitive: `s` to add an "s" when plural, 
`are` to write `is` when singular and `are` when plural (the operation `is` is identical), 
and `enum` to create an enumeration. 

Note that in plural interpolations the expression over which to pluralize is optional. It is 
automaically deduced from the previous interpolations. 
Now let's illustrate this behavior with a pluralization on value:
```{r}
nFiles = 6
string_magic("Warning: {N?nFiles} file{#s, is} corrupted.")

nFiles = 1
string_magic("Warning: {N?nFiles} file{#s, is} corrupted.")
```

The first interpolation is a regular one. The second one, `{#s, is}`, deduces the 
expression over which to pluralize from this previous interpolation.

This was only an introduction. To have more information on pluralization,
please see the [dedicated section](https://lrberge.github.io/stringmagic/articles/ref_string_magic_special_operations.html#sec_pluralization).


# Escaping and special cases {#sec_special_cases}

We end this introduction with a few special cases. 

### Escape brackets

The opening and closing brakets, `{}`, are special characters 
and cannot be used as regular text. 
To bypass their special meaning, you need to escape them with a double backslash.

You only need to escape the special delimiters which the algorithm is currently looking for.
For example, if we escape the opening delimiter, we don't need to escape the closing one:
```{r}
# Here opening delimiter is escaped: the closing delimiter has no special meaning
string_magic("open = \\{, close = }")
```

You need to escape the closing delimiter only when the algorithm is looking for it. 
In the following example, we need to escape the closing curly bracket to treat it as a regular character:
```{r}
string_magic("Here are closing brackets: {5 times.c ! \\}}")
```

Alternatively, use the argument `.delim` to change the delimiters:
```{r}
string_magic("Here I {interpolate} with .[this] ", .delim = ".[ ]", this = ".[]")
```

When using delimiters composed of two or more characters, like in the previous example, 
to escape it you need to add a double backslash before the first character (i.e. right before the
delimiter):
```{r}
string_magic("Here I {interpolate} with \\.[] ", .delim = ".[ ]")

# another example with {{ }}  as delimiter
string_magic("I {{what}} with \\{{ }} ", .delim = "{{ }}", what = "interpolate")
```

Another way is to pass a character string in an expression.
Indeed, since `{expr}` evaluates `expr`, if `expr` is a character string, it will stay untouched. 
```{r}
string_magic("I {'can {write} {{what}} I want'}")
```

### Other escapes 

In the few operations expecting a semi-colon (*if-else* and certain 
operations in *pluralization*), it can also be escaped with a double backslash.

```{r}
# if-else statement with semi-colon
is_c = TRUE
string_magic("{&is_c ; int i = 1\\; ; i = 1}")

is_c = FALSE
string_magic("{&is_c ; int i = 1\\; ; i = 1}")
```

### How to use a regular `!` in interpolations?

In interpolations, the exclamation mark (`!`) signals a verbatim expression. 
But what if you use it to mean the logical operation *not* in an operation-free interpolation? 
In that case, you need a hack: use a question mark (`?`) first to indicate to the
algorithm that you want to evaluate the expression. 

Ex.4: using the `?` hack to use `!` as a logical operation.
```{r}
string_magic("{!TRUE} is {?!TRUE}")
```

---
title: "String tools: magic edition"
author: "Laurent R. Bergé"
date: "`r Sys.Date()`"
output:
  html_document:
    theme: journal
    highlight: haddock
    number_sections: yes
    toc: yes
    toc_depth: 2
    toc_float:
      collapsed: no
      smooth_scroll: no
  pdf_document:
    toc: yes
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{string_tools}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

# we preload it to avoid ugly (was comiled with R.x.x) warnings in the doc
library(stringmagic)

# Option to allow caching in non interactive mode
options("string_magic_string_get_forced_caching" = TRUE)
```

This vignette describes `stringmagic` tools for handling character vectors. 
It details:

- how to [detect complex regex (=regular expression) patterns](#sec_detect)
- how to easily, and clearly, [chain multiple string operations](#sec_ops)
- how to efficiently, and clearly, [clean character strings](#sec_clean)
- how to quickly [create character vectors](#sec_vec)
- how to [split a vector and turn the result into a data frame, and vice versa](#sec_split)

# Detection of regex patterns {#sec_detect}

Detecting a single regex pattern is pretty straightforward with regular tools
like `base::grepl` or `stringr::string_detect`. Things become more complicated
when we want to detect the presence of multiple patterns. 

`stringmagic` offers three functions with an intuitive syntax to deal with 
complex pattern detection:


# Pattern detection with `string_is`, `string_which` and `string_get` {#detect_funs}

Use `string_is`, `string_which` and `string_get` to detect patterns in character vectors and obtain either 
a logical vector, an integer vector, or the values.

In this section we give examples for `string_get` which hopefully will be explicit 
enough to illustrate how it works. For the record, `string_get` uses `string_is` internally so
these examples are equivalent with `string_is` or `string_which`. 

Ex.1: series of examples using the *recommended syntax*.
```{r}
cars = row.names(mtcars)
cat_magic("All cars from mtcars:\n{C, 60 swidth ? cars}")

# cars with an 'a', an 'e', an 'i', and an 'o', all in lower case
string_get(cars, "a & e & i & o")

# cars with no 'e' and at least one digit
string_get(cars, "!e & \\d")

# flags apply to all
# contains the 'words' 2, 9 or l
# alternative syntax for flags: "wi/2 | 9 | l"
string_get(cars, "word, ignore/2 | 9 | l")
```

The default syntax is `string_get(x, ...)` (same for `string_is` and `string_which`), 
where `...` contains any number of patterns
to detect. By default the results of these pattern detections are combined with 
a logical AND. To combine them with a logical OR, you need to use the argument
`or = TRUE`. You can also pass the flags as regular function arguments. They then
apply to all patterns.

Ex.2: replication of Ex.1 using an alternative syntax.
```{r}
# string_get(cars, "a & e & i & o")
# cars with an 'a', an 'e', an 'i', and an 'o', all in lower case
string_get(cars, "a", "e", "i", "o")

# string_get(cars, "!e & \\d")
# cars with no 'e' and at least one digit
string_get(cars, "!e", "\\d")

# string_get(cars, "!/e & \\d")
# This example cannot be replicated directly, we need to apply logical equivalence
string_get(cars, "!e", "!\\d", or = TRUE)

# string_get(cars, "wi/2 | 9 | l")
# contains the 'words' 2, 9 or l
string_get(cars, "2", "9", "l", or = TRUE, word = TRUE, ignore.case = TRUE)
```

### Specificities of `srt_get` {#detect_get}

On top of the detection previously described, the function `srt_get` changes its 
behavior with the arguments `seq` or `seq.unik`. It also supports [automatic caching](#get_caching).

#### Sequentially appending results

As seen previously, patterns in `...` are combined with a logical AND. If you set
`seq = TRUE`, this behavior changes. The results of each pattern becomes stacked 
sequentially. Schematically, you obtain the vector `c(x_that_contains_pat1, x_that_contains_pat2, etc)`
with `pat1` the first pattern in `...`, `pat2` the second pattern, etc.

Using `seq.unik = TRUE` is like `seq` but applies the function `unique()` at the end.

Ex: sequentially combining results.
```{r}
# cars without digits, then cars with 2 'a's or 2 'e's and a digit
string_get(cars, "!\\d", "i/a.+a | e.+e & \\d", seq = TRUE)

# let's get the first word of each car name
car_first = string_ops(cars, "extract.first")
# we select car brands ending with 'a', then ending with 'i'
string_get(car_first, "a$", "i$", seq = TRUE)
# seq.unik is similar to seq but applies unique()
string_get(car_first, "a$", "i$", seq.unik = TRUE)
```

#### Caching {#get_caching}

At the exploration stage, we often run the same command with a few
variations on the same data set. Acknowledging this, `string_get` supports the caching
of the data argument in interactive use. This means that the user can concentrate in
the pattern to find and need not bother to write the data from where to fectch the
values. Note that `string_get` is the only `stringmagic` function to have this
ability.

Caching is always enabled, you don't need to do anything.

Ex: caching of the data.
```{r}
# Since we used `car_first` in the previous example, we don't need to provide
# it explicitly now
# => brands containing 'M' and ending with 'a' or 'i'; brands containing 'M'
string_get("M & [ai]$", "M", seq.unik = TRUE)
```

# Chaining string operations with `string_ops` {#sec_ops}

Formatting text data often requires applying many functions
(be it for parsing, text analysis, etc). 
Even for simple tasks, the number of operations can quickly balloon, adding many
lines of code, reducing readability, and all this for basic processing.

The function `string_ops` tries to solve this problem. 
It has access to all (50+) [`string_magic` operations](https://lrberge.github.io/stringmagic/articles/ref_operations.html), allowing for
a compact and readable way to chain basic operations on character strings.

Below are a few motivating examples.

Ex.1: Parsing data.
```{r}
# parsing an input: extracting the numbers
input = "8.5in, 5.5, .5 cm"
string_ops(input, "','split, tws, '^\\. => 0.'replace, '^\\D+|\\D+$'replace, num")


# Explanation------------------------------------------------------------------|
# ','split: splitting w.r.t. ','                                               |
# tws: trimming the whitespaces                                                |
# '^\\. => 0.'replace: adds a 0 to strings starting with '.'                   |
# '^\\D+|\\D+$'replace: removes non-digits on both ends of the string          |
# num: converts to numeric                                                     |


# now extracting the units
string_ops(input, "','split, '^[ \\d.]+'replace, tws")


# Explanation------------------------------------------------------------------|
# ','split: splitting w.r.t. ','                                               |
# '^[ \\d.]+'replace: removes the ' ', digit                                   |
#                     and '.' at the beginning of the string                   |
# tws: trimming the whitespaces                                                |
```

Ex.2: extracing information from text.
```{r}
# Now using the car data
cars = row.names(mtcars)

# let's get the brands starting with an "m"
string_ops(cars, "'i/^m'get, x, unik")


# Explanation------------------------------------------------------------------|
# 'i/^m'get: keeps only the elements starting with an m,                       |
#            i/ is the 'regex-flag' "ignore" to ignore the case                |
#            ^m means "starts with an m" in regex language                     |
# x: extracts the first pattern. The default pattern is "[[:alnum:]]+"         |
#    which means an alpha-numeric word                                         |
# unik: applies unique() to the vector                                         |


# let's get the 3 largest numbers appearing in the car models
string_ops(cars, "'\\d+'x, rm, unik, num, dsort, 3 first")


# Explanation------------------------------------------------------------------|
# '\\d+'x: extracts the first pattern, the pattern meaning "a succession"      |
#          of digits in regex language                                         |
# rm: removes elements equal to the empty string (default behavior)            |
# unik: applies unique() to the vector                                         |
# num: converts to numeric                                                     |
# dsort: sorts in decreasing order                                             |
# 3 first: keeps only the first three element                                  |
```

As you can see, an operation that would take multiple lines to read and understand
now can be read from left to right in a single line.

# `string_clean`: One function to clean them all {#sec_clean}

The function `string_clean` streamlines the cleaning of character vectors by providing:

- i) a specialized syntax to replace multiple regex patterns, 
- ii) a direct access to many low level string operations, and 
- iii) the ability to chain these two operations. 

## Cleaning syntax

This function is of the form `string_clean(x, ...)` with `x` the vector to clean and
`...` any number of cleaning operations which can be of two types:

1. use `"pat1, pat2 => replacement"` to replace the regex patterns `pat1` and `pat2` 
with the value `replacement`.
1. use `"@op1, op2"` to perform any arbitrary sequence of [`string_magic` operation](https://lrberge.github.io/stringmagic/articles/ref_operations.html)

In the operation `"pat1, pat2 => replacement"`, the pattern is first split with respect
to the pipe, `" => "` (change it with argument `pipe`), to get `replacement`. 
Then the pattern is split with respect to commas (i.e. `",[ \t\n]+"`, change it with argument `sep`) 
to get `pat1` and `pat2`. A sequence of `base::gsub` calls is performed to replace each
`patx` with `replacement`. 

By default the replacement is the empty string. This means that writting `"pat1, pat2"`
will lead to erasing these two patterns.

If a pattern starts with an `"@"`, the subsequent character string is sent to `string_ops`.
For example `"@ascii, lower"` is equivalent to `string_ops(x, "ascii, lower")` which 
turns `x` to ASCII and lowers the case. 

## Example of text cleaning {#clean_example}

```{r}
monologue = c("For who would bear the whips and scorns of time",
              "Th' oppressor's wrong, the proud man's contumely,",
              "The pangs of despis'd love, the law's delay,",
              "The insolence of office, and the spurns",
              "That patient merit of th' unworthy takes,",
              "When he himself might his quietus make",
              "With a bare bodkin? Who would these fardels bear,",
              "To grunt and sweat under a weary life,",
              "But that the dread of something after death-",
              "The undiscover'd country, from whose bourn",
              "No traveller returns- puzzles the will,",
              "And makes us rather bear those ills we have",
              "Than fly to others that we know not of?")

# Cleaning a text
string_clean(monologue, 
          # use string_magic to: lower the case and remove basic stopwords
          "@lower, stopword",
          # remove a few extra stopwords(we use the flag word 'w/')
          "w/th, 's",
          # manually stem some verbs
          "despis'd => despise", "undiscover'd => undiscover", "(m|t)akes => \\1ake",
          # still stemming: dropping the ending 's' for words of 4+ letters, except for quietus
          "(\\w{3,}[^u])s\\b => \\1",
          # normalizing the whitespaces + removing punctuation
          "@ws.punct")

```


# Create simple character vectors with `string_vec` {#sec_vec}


The function `string_vec` is dedicated to the creation of small character vectors.
You feed it a comma separated list of values in a string and it will turn it 
into a vector. 

Ex.1: creating a simple vector.
```{r}
fruits = string_vec("orange, apple, pineapple, strawberry")
fruits
```

Within the enumeration, you can use interpolation, with curly brackets (`{}`), to 
insert the elements from a vector into the current string. 

Ex.2: adding a vector into an enumeration.
```{r}
more_fruits = string_vec("lemon, {fruits}, peach")
more_fruits
```

The interpolation is performed with [`string_magic`](https://lrberge.github.io/stringmagic/articles/guide_string_magic.html). This means that any [`string_magic` operation](https://lrberge.github.io/stringmagic/articles/ref_operations.html)
can be applied on-the-fly.

Ex.3: replicating Ex.2 but shortening long fruit names.
```{r}
more_fruits = string_vec("lemon, {6 Shorten ? fruits}, peach")
more_fruits
```

Since interpolations are resolved with `string_magic`, you can add any text before/after the 
interpolation:

Ex.4: adding text before the interpolation.
```{r}
pkgs = string_vec("pandas, os, time, re")
imports = string_vec("import numpy as np, import {pkgs}")
imports
```

### Creating small matrices or data frames

You can transform the returned vector into a matrix or a data frame using the 
arguments `.cmat`, `.nmat` (**c**haracter or **n**umeric matrix) or `.df`.

Ex.5: returning a matrix.
```{r}
string_vec("1, 5,
            3, 2,
            5, 12", .nmat = TRUE)
```

The number of rows is guessed from the number of newlines in the string. You can avoid using
character strings, but in that case you need to explicitly give the number of rows.

Ex.5-bis: returning a numeric matrix, giving `.nmat` the number of rows.
```{r}
string_vec(1, 5,
           3, 2,
           5, 12, .nmat = 3)
```

If you want to return a data.frame, you can add the column names in the .df argument: either
in a regular vector, either in a comma separated list. Note that columns looking like
numeric values are always converted.

Ex.6: returning a data frame.
```{r}
# you can add the column names directly in the argument .df
df = string_vec("1, john,
                 3, marie,
                 5, harry", .df = "id, name")
df

# automatic conversion of numeric values
df$id * 5
```

# Split vectors and turn the result into a data frame, and vice versa {#sec_split}

The function `string_split2df` (and `string_split2dt`) splits a vector using a 
regular expression pattern and turns it into a data frame, remembering the 
original identifiers. You can get the original vectors back (almost) with the 
function `paste_conditional`.

Ex.1: breaking up two sentences with respect to punctuation and spaces; then merging them back.
```{r}
x = c("Nor rain, wind, thunder, fire are my daughters.",
      "When my information changes, I alter my conclusions.")

# we split at each word
sentences_split = string_split2df(x, "[[:punct:] ]+")
sentences_split

# recovering the original vectors (we only lose the punctuation)
paste_conditional(sentences_split$x, sentences_split$obs)
```

If identifiers are associated to the elements of the vector, you can provide them
so that the data frame returned contains them.

Ex.2: splitting with identifiers and merging back with a formula.
```{r}
id = c("ws", "jmk")
# we add the identifier
base_words = string_split2df(x, "[[:punct:] ]+", id = list(author = id))

# merging back using a formula
paste_conditional(x ~ author, base_words)
```

---
title: "Regular expression flags"
author: "Laurent R. Bergé"
date: "`r Sys.Date()`"
output:
  html_document:
    theme: journal
    highlight: haddock
    number_sections: yes
    toc: yes
    toc_depth: 3
    toc_float:
      collapsed: no
      smooth_scroll: no
  pdf_document:
    toc: yes
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{ref_regex_flags}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(stringmagic)
```

All functions in `stringmagic` accept optional regular expressions (regex) flags when regular expressions
are expected. The idea is similar to 
[regular regex flags](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_expressions#advanced_searching_with_flags),
but the flags are different in name and effect.

# Regex flags: Syntax

Use `"flag1, flag2/regex"` to add the flags `flag1` and `flag2` to the regular expression `regex`.
For example `"ignore, fixed/dt["` will add the flags `ignore` and `fixed` to the regex `dt[`.

Alternatively, use only the initials of the flags. Hence, `"if/dt["` would also add the 
flags `ignore` and `fixed`.

If the regex does not contain a slash (`/`), no flags are added. If your regex should 
contain a slash, see the [section on escaping](#flag_escaping).

Ex: let's find lines containing `"dt["`:
```{r}
code = c("DT = as.data.table(iris)", 
         "DT[, .(pl_sl = string_magic('PL/SL = {Petal.Length / Sepal.Length}')]")

string_get(code, "if/dt[")
```

# Regex flags: Reference

There are 6 flags:

- [ignore](#flag_ignore): always available
- [fixed](#flag_fixed): always available
- [word](#flag_word): always available
- [magic](#flag_magic): always available
- [total](#flag_total): only available in functions performing a replacement
- [single](#flag_single): only available in functions performing a replacement

### ignore {#flag_ignore}

The flag `"ignore"` leads to a case-insensitive search. 

Ex: let's extract words starting with the last letters of the alphabet.
```{r}
unhappy = "Rumble thy bellyful! Spit, fire! spout, rain!
Nor rain, wind, thunder, fire are my daughters.
I tax not you, you elements, with unkindness.
I never gave you kingdom, call'd you children,
You owe me no subscription. Then let fall
Your horrible pleasure. Here I stand your slave,
A poor, infirm, weak, and despis'd old man."

# the ignore flag allows to retain words starting with the
# upper cased letters
# ex: getting words starting with the letter 'r' to 'z'
cat_magic("{'ignore/\\b[r-z]\\w+'extract, c, 60 swidth ? unhappy}")
```

*Technically*, the [perl](https://www.pcre.org/) expression `"(?i)"` is added at the 
beginning of the pattern.

### fixed {#flag_fixed}

The flag `"fixed"` removes any special regular expression meaning from the pattern, 
and treats it as verbatim.

Ex: let's fix the equation by changing the operators. 
```{r}
x = "50 + 5 * 5 = 40"
string_clean(x, "f/+", "f/*", replacement = "-")

# Without the fixed flag, we would have gotten an error since '+' or '*'
# have a special meaning in regular expressions (it is a quantifier)
# and expects something before

# Here's the error
try(string_clean(x, "+", "*", replacement = "-"))
```

*Technically*, if `"fixed"` is the only flag, then the functions `base::grepl` or `base::gsub`
are run with the argument `fixed = TRUE`. If there are also the flags `"ignore"` or `"word"`, 
the pattern is nested into the perl boundaries `\\Q` and `\\E` which strip any special meaning
from the pattern.

### word {#flag_word}

The flag `"word"`:

- adds word boundaries to the pattern
- accepts comma-separated enumerations of words which are concatenated with a logical 'or'

The logic of accepting comma-separated enumerations is to increase readability.
For example, with the flag `"word"`, `"is, are, were"` is equivalent to `"\\b(is|are|were)\\b"`.

Ex: we hide a few words from Alfred de Vigny's poem.
```{r}
le_mont_des_oliviers = "S'il est vrai qu'au Jardin sacré des Écritures,
Le Fils de l'homme ai dit ce qu'on voit rapporté ;
Muet, aveugle et sourd au cri des créatures,
Si le Ciel nous laissa comme un monde avorté,
Alors le Juste opposera le dédain à l'absence
Et ne répondra plus que par un froid silence
Au silence éternel de la Divinité."

# we hide a few words from this poem
string_magic("{'wi/et, le, il, au, des?, ce => _'replace ? le_mont_des_oliviers}")
```

*Technically*, first the pattern is split with respect to `",[ \t\n]+"`, then all elements
are collapsed with `"|"`. If the flag `"fixed"` was also present, each element is first wrapped 
into `"\\Q"` and `"\\E"`. Finally, we add parentheses (to enable capture) and word 
boundaries (`"\\b"`) on both sides.

### magic {#flag_magic}

Use the `"magic"` flag to interpolate variables inside the regular expression
before the regex is evaluated.

Ex: interpolating variables inside regular expressions.
```{r}
vowels ="aeiouy"
# let's keep only the vowels
# we want the pattern: "[^aeiouy]"
lmb = "'Tis safer to be that which we destroy
Than by destruction dwell in doubtful joy."
string_replace(lmb, "magic/[^{vowels}]", "_")

#
# Illustration of `string_magic` operations before regex application
#

cars = row.names(mtcars)
# Which of these models contain a digit?
models = c("Toyota", "Hornet", "Porsche")
# we want the pattern "(Toyota|Hornet|"Porsche).+\\d"
# we collapse the models with a pipe using '|'c
string_get(cars, "m/({'|'c ? models}).+\\d")

# alternative: same as above but we first comma-split the vector
models_comma = "Toyota, Hornet, Porsche"
string_get(cars, "m/({S, '|'c ? models_comma}).+\\d")

#
# Interpolation does not apply to regex-specific curly brackets
#

# We delete only successions of 2+ vowels
# {2,} has a rexex meaning and is not interpolated:
string_replace(lmb, "magic/[{vowels}]{2,}", "_")
```

*Technically*, the algorithm does not interpolate curly brackets having a 
regular expression meaning. The expression of the form `"{a, b}"` with `"a"` and `"b"`
digits means a repetition of the previous symbol of at least `"a"` times and at most `"b"` times.
The variables are fetched in the calling environment. To fetch them from a different location,
you can use the argument `envir`.

### total {#flag_total}

The flag `"total"` is only available to functions performing a replacement. 
In that case, if a pattern is detected, *the full character string* is replaced 
(instead of just the pattern). 

Ex: let's replace a few car models.
```{r}
cars_small = head(row.names(mtcars))
print(cars_small)

string_replace(cars_small, "ti/mazda", "Mazda: sold out!")
```

On top of this, the `"total"` flag allows to perform logical operations across
several regex patterns. You have more information on this in the [dedicated vignette](https://lrberge.github.io/stringmagic/articles/ref_regex_logic.html).
In a nutshell, you can write `"pat1 & !pat2 | pat3"` with `"patx"` regular 
expresion patterns. This means: contains `pat1` and does not contain `pat2`,
or contains `pat3`.

Ex: detect car brands with a digit and no 'e'.
```{r}
cars_small = head(row.names(mtcars))
print(cars_small)

string_replace(cars_small, "total, ignore/\\d & !e", "I don't like that brand!")
```

*Technically*, instead of using `gsub` to replace the pattern, [`string_is`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_detect) is used
to detect which element contains the pattern. Each element with the pattern
is then substituted with the replacement.

### single {#flag_single}

The flag `"single"` is only available to functions performing a replacement.
It allows only a single substitution to take place. Said differently, only the 
first replacement is performed.

Ex: single substitutions.
```{r}
encounter = string_vec("Hi Cyclops., Hi you. What's your name?, Odysseus is my name.")
# we only remove the first word
string_replace(encounter, "single/\\w+", "...")
```

*Technically*, the function `base::sub` is used instead of `base::gsub`.

## Escaping flags: How to, and a word of caution with paths {#flag_escaping}

If your regular expression contains a slash (`"/"`), this will come in conflict
with the parsing of the optional flags. 
At the moment a `/` is present in a pattern, the algorithm will throw an error
if the expected flags are not written correctly.

To use a slash in the regex without adding flags there is only one solutions:

- escape the first `"/"` with a double backslash

Ex: let's invert the numerator and denominator of a division.
```{r}
eq = "5/x = 3/2"
# escaping with backslashes
string_replace(eq, "(\\w)\\/(\\w)", "\\2/\\1")
```

**Warning:** when applying regular expressions on file paths, to avoid unexpected behavior,
the flags algorithm is very strict. Everytime a pattern contains a slash that is not 
associated with valid flags, an error will be thrown unless that slash is escaped.

```{r}
path = "my/path/to/the/file.tex"
# we keep the directory only

# first try: an error bc flags are expected before the first '/'
try(string_replace(path, "/[^/]+$"))

# after escaping: works (only the first slash requires escaping)
string_replace(path, "\\/[^/]+$")

# if we did add a flag, we would need to double the slash
# compare...
string_replace(path, "i//[^/]+$")
# to...
string_replace(path, "i/[^/]+$")
```

Hence if you need to write path related regexes, you very likely need to escape the first slash.

---
title: "`string_magic`'s operations: The reference"
author: "Laurent R. Bergé"
date: "`r Sys.Date()`"
output:
  html_document:
    theme: journal
    highlight: haddock
    number_sections: yes
    toc: yes
    toc_depth: 3
    toc_float:
      collapsed: no
      smooth_scroll: no
  pdf_document:
    toc: yes
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{operations_reference}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(stringmagic)
```

This document references all regular `string_magic` operations. They can be used within 
`string_magic` or can be accessed from `string_ops` or `string_clean`. 

The operations are divided into four groups:

- [basic string operations](#sec_basic)
- [operations changing the length or order](#sec_op_length)
- [formatting operations](#sec_format)
- [other operations](#sec_other)

By default the function `string_magic` returns a plain character vector. In this 
vignette it is sometimes nicer to apply the function `base::cat` to display 
`string_magic` results containing newlines. Ths function `cat_magic` does exactly 
that and we will use it from time to time.


# Basic string operations {#sec_basic}

This section describes some of the most common string operations: extracting, 
replacing, collapsing, splitting, etc. Because they are so common, many of these 
operations have a one letter alias.
These functions accept regex flags in their patterns. For more information on regex flags, 
see the [dedicated section](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html).

## `s`, `S`, `split`, `Split`: Split strings

Splits the string according to a pattern. The four operations have different defaults: `' '` 
for `s` and `split`, and `',[ \t\n]*'` for `S` and `Split` (i.e. comma separation). 

When character strings are split, their identity is kept in memory 
so that group-wise operations can be applied. See the [section on group-wise operations](https://lrberge.github.io/stringmagic/articles/ref_string_magic_special_operations.html#sec_group_wise).
```{r}
# 'Split' with its default (comma separation)
string_magic("{Split ! romeo, juliet}")

# result with 's' is different
string_magic("{split ! romeo, juliet}")

# with argument: 's' and 'S' are identical
# note the flag 'fixed' (`f/`) to remove regex interpretation
string_magic("{'f/+'split, '-'collapse ! 5 + 2} = 3")

# group wise operations (here `~(sort, collapse)`, see dedicated section)
prince_talk = c("O that this too too solid flesh would melt",
                "Thaw, and resolve itself into a dew!",
                "Or that the Everlasting had not fix'd",
                "His canon 'gainst self-slaughter!")
cat_magic("Order matters:\n{split, ~(sort, collapse), align.center, lower, upper.sentence,
                            Q, '\n'collapse ? prince_talk}")
```

## `c`, `C`, `collapse`, `Collapse`: Collapse strings

To collapse (or concatenate) multiple strings into a single one. The four operations are 
identical, only their defaults change. The default is `' '` for `c` and `collapse`, 
and `', | and '` for `C` and `Collapse`.
The syntax of the argument is `'s1'` or `'s1|s2'`. `s1` is the string used to concatenate 
(think `paste(x, collapse = s1)`). In arguments of the form `'s1|s2'`, `s2` will be 
used to concatenate the last two elements. 

```{r}
# regular way
x = 1:4
string_magic("And {' and 'collapse ? x}!")

# with s2
string_magic("Choose: {', | or 'collapse ? 2:4}?")

# default of Collapse: enumeration (similar to operation enum)
wines = c("Saint-Estephe", "Margaux")
string_magic("I like {Collapse ? wines}.")

# default of collapse: space concatenation
string_magic("{split, '.{5,}'get, collapse ! I don't like short words}")
```

## `extract`, `x`, `X`: Extract patterns

Extracts the first or multiple patterns from a string. Default argument is `'[[:alnum:]]+'`.
Command `"extract"` accepts the option `"first"`, and 
`"x"` and `"X"` accept no option. `x` is an alias for `extract.first` and `X` for `extract`.
Use the option `"first"` to extract only the first match for each string.

When patterns are extracted, the identity of each original character string is kept in memory 
so that group-wise operations can be applied. See the [section on group-wise operations](https://lrberge.github.io/stringmagic/articles/ref_string_magic_special_operations.html#sec_group_wise).
```{r}
x = c("margo: 32, 1m75", "luke doe: 27, 1m71")
string_magic("{'^\\w+'extract ? x} is {'\\d+'extract.first ? x}")

# illustrating multiple extractions
# group-wise operation (~()) is detailed in its own section
x = c("Combien de marins, combien de capitaines.",
      "Qui sont partis joyeux pour des courses lointaines,",
      "Dans ce morne horizon se sont évanouis !")
string_magic("Endings with i: {'i\\w*'extract, ~(', 'collapse), enum.1 ? x}.")


x = c("6 feet under", "mahogany")
# single extraction
string_magic("{'\\w{3}'x ? x}")

# multiple extraction
string_magic("{'\\w{3}'X ? x}")
```

## `r`, `R`, `replace`: Replace patterns

Replaces a pattern with a string. The three operators `r`, `R` and `replace` are identical and have no default.
The syntax is `'flags/old'` or `'old => new'` with `'old'` the pattern to find and `new` the 
replacement. `flags/` are optional regex flags. The default for `new` is 
the empty string. On top of regular regex flags, this operation also accepts 
the flags `"total"` and `single`. `total` instructs to replace the fulll string in case 
the pattern is found.

The flag `"single"` leads to only a single substitution per string (the first pattern is replaced). 
That is, the function `base::sub` is used instead of `base::gsub`.
```{r}
# regex without replacement (ie removing)
string_magic("{'e'replace ! Where is the letter e?}")

# regex with replacement
string_magic("{'(?<!\\b)e => a'replace ! Where is the letter e?}")

# with option "single"
string_magic("{'single/(?<!\\b)e => a'replace ! Where is the letter e?}")

# we replace the full string with the flag total (`t/`)
x = c("Where is the letter e?", "Not this way!")
string_magic("{'total/e => here!'r ? x}")
```

## `clean`: Clean string patterns {#op_clean}

Replaces a pattern with a string. Similar to the operation `r`, except that here the comma is
a pattern separator. The argument is of the form `"flags/pattern1, pattern2 => replacement"`. 
See detailed explanations in [`string_clean()`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_clean). 

```{r}
# we use the fixed pattern to remove the regex meaning
string_magic("{'f/[, ]'clean ! x[a]}")
```

## `get`: Get selected strings

Restricts the string vector to only the values respecting a pattern. This operation has no default.
Accepts the options `"equal"` and `"in"`.
By default it uses the same syntax as [`string_get()`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#detect_funs) 
so that you can use regex flags and 
include logical operations between regex patterns with `' & '` and `' | '`.
If the option `"equal"` is used, a simple string equality with the argument is tested (hence
no flags are accepted). If the option `"in"` is used, the argument is first split with respect to commas
and then set inclusion is tested. 

```{r}
x = row.names(mtcars)
#  we only keep models containing "Merc" and ending with a letter ([[:alpha:]]$)
string_magic("Mercedes models: {'Merc & [[:alpha:]]$'get, '^.+ 'r, enum ? x}.")

models = c("Merc 230", "Merc 450SE", "Merc 480")
# we only ekep the ones in the set
string_magic("Mercedes models: {`models`get.in, enum ? x}.")
```

## `is`: Detect patterns in strings

Detects if a pattern is present in a string, returns a logical vector. This operation has no default.
Mostly useful as the final operation in a [`string_ops()`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_ops) call.
By default it uses the same syntax as [`string_is()`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#detect_funs) 
so that you can use regex flags and 
include logical operations between regex patterns with `' & '` and `' | '`.
If the option `"equal"` is used, a simple string equality with the argument is tested (hence
no flags are accepted). If the option `"in"` is used, the argument is first split with respect to commas
and then set inclusion is tested. 
```{r}
x = c("Mark", "Lucas")
# note that we use the flag `i/` to ignore the case
string_magic("Mark? {'i/mark'is, enum ? x}")
```

## `which`: Get the index of the strings containing a pattern

Returns the index of string containing a specified pattern. With no default, can be applied
to a logical vector directly. 
By default it uses the same syntax as [`string_which()`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#detect_funs) 
so that you can use regex flags and 
include logical operations between regex patterns with `' & '` and `' | '`.
If the option `"equal"` is used, a simple string equality with the argument is tested (hence
no flags are accepted). If the option `"in"` is used, the argument is first split with respect to commas
and then set inclusion is tested. 
Mostly useful as the final operation in a [`string_ops()`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_ops) call.
```{r}
x = c("Mark", "Lucas", "Markus")
# note that we use the flag `i` to ignore the case and `w` to add word boundaries
string_magic("Mark is number {'iw/mark'which ? x}.")
```

# Operations changing the length or the order {#sec_op_length}

## `first`: Keep only the first elements

Keeps only the first `n` elements. 

```{r}
string_magic("First 3 mpg values: {3 first, enum ? mtcars$mpg}.")

# you could have done the same with regular R in the expression...
string_magic("First 3 mpg values: {enum ? head(mtcars$mpg, 3)}.")

# ...but not in the middle of an operations chain
string_magic("First 3 integer mpg values: {'f/!.'get, 3 first, enum ? mtcars$mpg}.")
```

Negative numbers as argument remove the enum
first `n` values. You can add a second argument in the form `'n1|n2'first` in which case the first `n1` and last
`n2` values are kept; `n1` and `n2` must be positive numbers.

```{r}
string_magic("Letters in the middle: {13 first, 5 last, enum ? letters}.")

string_magic("First and last letters: {'3|3'first, enum ? letters}.")

string_magic("Last letters: {-21 first, enum ? letters}.")
```

## `K`: Keep only the first elements (alternative)

Keeps only the first `n` elements; has more options than `first`. The syntax is `'n'K`, 
`'n|s'K`, `'n||s'K`. `n` provides the number of elements to keep. If `s` is provided and the number of 
elements are greater than `n`, then in `'n|s'` the string `s` is added at the end, and
if `'n||s'` the string `s` replaces the nth element.
The string `s` accepts specials values:
  + `:n:` or `:N:` which gives the total number of items in digits or letters (N)
  + `:rest:` or `:REST:` which gives the number of elements that have been truncated in digits or letters (REST)

```{r}
# basic use
string_magic("First 3 letters: {'3'K, q, enum ? letters}.")

# advanced use: using the extra argument
string_magic("The letters are: {q, '3|:rest: others'K, enum ? letters}.")
```

## `last`: Keep only the last elements

Keeps only the last `n` elements. Negative numbers as argument remove the 
last `n` values.
```{r}
string_magic("Last 3 mpg values: {3 last, enum ? mtcars$mpg}.")

string_magic("Removing the 3 last elements leads to {-3 last, enum ! x{1:5}}.")
```

## `sort`: Sort the vector

Sorts the vector in increasing order. Accepts an optional argument and the option `"num"`. 
```{r}
x = c("sort", "me")
# basic use
string_magic("{sort, collapse ? x}")
```
If an argument is provided, it must be a regex pattern that will be applied to
the vector using [`string_clean()`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_clean). 
The sorting will be applied to the modified version of the vector
and the original vector will be ordered according to this sorting. 
```{r}
# first modifying the string before sorting
# here the regex first removes the first word, meaning that we sort on the last names
x = c("Jon Snow", "Khal Drogo")
string_magic("{'.+ 'sort, enum?x}")
```

The option `"num"` sorts over a numeric version 
(with silent conversion) of the vector and reorders the original vector accordingly. 
Values which could not be converted are last.
```{r}
x = "Mark is 34, Bianca is 55, Odette is 101, Julie is 21 and Frank is 5"
# sort on the "character string" number
string_magic("{', | and 'split, '\\D'sort, enum ? x}")

# we extract the numbers, then convert to numeric, then sort
string_magic("{', | and 'split, '\\D'sort.num, enum ? x}")
```
**Important note**: the sorting operation is applied before any character conversion.
If previous operations were applied, it is likely that numeric data were transformed to character.
```{r}
# note the difference
x = c(20, 100, 10)

# sorting on numeric
string_magic("{sort, ' + 'collapse ? x}")

# sorting on character since 'n' operation transformed the vector to character
string_magic("{n, sort, ' + 'collapse ? x}")
```

## `dsort`: Sort the vector in decreasing order

Sorts the vector in decreasing order. It accepts an optional argument and 
the option `"num"`. See the operation `"sort"` for a description of the argument and the option.
```{r}
string_magic("5 = {dsort, ' + 'collapse ? 2:3}")
```

## `rev`: Reverse the vector

Reverses the vector.
```{r}
string_magic("{rev, ''collapse ? 1:3}")
```

## `unik`: Keep only unique elements

Makes the string vector unique. 
```{r}
string_magic("Iris species: {unik, upper.first, enum ? iris$Species}.")
```

## `table`: Attach unique elements to their frequencies

Computes the frequency of each element and attaches each element to its frequency. Accepts an argument which must be a character string representing a `string_magic` interpolation
with the following variables: `x` (the element), `n` (its count) and `s` (its share). The default argument is `'{x} ({n ? n})'`. 

By default the resulting string vector is sorted by decreasing frequency. You can change how the vector is sorted with five options: `sort` (sorts on the elements), `dsort` (decreasing sort), `fsort` (sorts on frequency), `dfsort` (decreasing sort on freq. -- default), `nosort` (keeps the order of the first elements). Note that you can combine several sorts (to resolve the ties of elements with same frequencies).

```{r}
dna = string_split("atagggagctacctgcgcgtcgcccaaaagcaggg", "")
cat_magic("Letters in the DNA seq. {''c, Q? dna}: ",
          "  -      default: {table, enum ? dna}",
          "  - value sorted: {table.sort, enum ? dna}",
          "  -       shares: {'{x} [{round(s * 100)}%]' table, enum ? dna}",
          # `fsort` sorts by **increasing** frequency
          "  - freq. sorted: {'{q ? x}' table.fsort, enum ? dna}",
          .sep = "\n")
```

## `each`: Repeat each elements of the vector

Repeats each element of the vector `n` times. Option `"c"` then collapses the full vector 
with the empty string as a separator. 
```{r}
# note: operation `S` splits splits wrt to commas (default behavior)
string_magic("{S!x, y}{2 each ? 1:2}")

# illustrating collapsing
string_magic("Large number: 1{5 each.c ! 0}")
```

## `times`: Repeats the vector

Repeats the vector sequence `n` times. Option `"c"` then collapses the full vector 
with the empty string as a separator. 
```{r}
string_magic("What{6 times.c ! ?}")
```

## `rm`: Remove specific values

Removes elements from the vector. Options: `"empty"`, `"blank"`, `"noalpha"`, `"noalnum"`, `"all"`.
The *optional* argument represents the pattern used to detect strings to be deleted. 
```{r}
x = c("Luke", "Charles")
string_magic("{'i/lu'rm ? x}")
```

By default it removes empty strings. 
Option `"blank"` removes strings containing only blank characters (spaces, tab, newline).
Option `"noalpha"` removes strings not containing letters. Option `"noalnum"` removes strings not 
containing alpha numeric characters. Option `"all"` removes all strings (useful in conditions, see 
the dedicated section). If an argument is provided, only the options `"empty"` and `"blank"` are available.

```{r}
x = c("I want to enter.", "Age?", "21.")
string_magic("Nightclub conversation: {rm.noalpha, c ! - {x}}")
```

## `nuke`: Remove all value

Removes all elements, equivalent to `rm.all` but possibly more explicit. 
Useful in conditions, see the dedicated section.
```{r}
x = c(5, 7, 453, 647)
# here we use a condition: see the dedicated section for more information
string_magic("Small numbers only: {if(.>20 ; nuke), enum ? x}.")
```

## `insert`: Insert a character string

Inserts a new element to the vector. Options: `"right"` and `"both"`. Option `"right"` adds
the new element to the right. Option `"both"` inserts the new element on the two sides of the vector.

```{r}
string_magic("{'3'insert.right, ' + 'collapse ? 1:2}")
```

## `dp`, `deparse`: Deparse an object

It deparses an R object and collapses it into a single string using newlines. 
By default it keeps only the first 50 lines of the deparsed object.
The number of lines kept can be modified by using a *numeric argument*.

It accepts the options: `short` and `vec`. If `short`, multiline deparsed elements
are collapsed with `\\\\n` and the number of characters cannot exceed the value of the
argument, which by default is 50.

If `vec`, then the deparsed element is not collapsed into a single string, and only 
the first 50 (default) elements of the vector are kept. You can increase/reduce the 
number of elements kept with the argument.

```{r}
fml = y ~ x1 + x2
# the full formula
string_magic("The estimated model is {dp ? fml}.")
# only the beginning of the formula
string_magic("The estimated model is {10 dp.short ? fml}.")

# deparsing a long function, we keep only the first 5 lines:
cat_magic("The function lm:\n{5 dp ? lm}.")

# using vec to keep the deparsed vector, we add line numbers, then 
# collapse it with newlines
cat_magic("The function lm:\n{'\n'c ! {1:5}: {5 dp.vec ? lm}}")
```

# Formatting operations {#sec_format}
 
## `lower`: Change the case

Lower cases the full string.
```{r}
x = "MesSeD uP CaSe"
string_magic("from a {x} to {lower?x}")
```

## `upper`: Change the case

Upper cases the full string. Options: `"first"` and `"sentence"`.
Option `"first"` upper cases only the first character. Option `"sentence"`
upper cases the first letter after punctuation. 

```{r}
x = "hi. how are you? fine."
string_magic("{upper.sentence ? x}")
```

## `title`: Change the case

Applies a title case to the string. Options: `"force"` and `"ignore"`.
Option `"force"` first puts everything to lowercase before applying the title case. 
Option `"ignore"` ignores a few small prepositions ("a", "the", "of", etc).

```{r}
x = "bryan is in the KITCHEN"

# default: respects upper cases
string_magic("{title ? x}")

# force: force to title case
string_magic("{title.force ? x}")

# ignore: ignores small prepositions
string_magic("{title.force.ignore ? x}")
```

## `ws`: Normalize white spaces

Normalizes whitespaces (WS). It trims the whitespaces on the edges and transforms any succession 
of whitespaces into a single one. Can also be used to further clean the string with its options. 
Options: `"punct"`, `"digit"`, `"isolated"`. Option `"punct"` cleans the punctuation. Option `"digit"` cleans digits.
Option `"isolated"` cleans isolated letters. WS normalization always come after any of these options.
**Important note:** punctuation (or digits) are replaced with WS and **not** 
the empty string. This means that `string_magic("ws.punct ! Meg's car")` will become `"Meg s car"`.

```{r}
x = "   I    should? review 85 4 this text!!"
cat_magic("v0: {x}", 
          "v1: {ws ? x}",
          "v2: {ws.punct ? x}",
          "v3: {ws.punct.digit ? x}",
          "v4: {ws.punct.digit.isolated ? x}", .sep = "\n")
```

## `tws`: Trim white spaces

Trims the white spaces on both ends of the strings.
```{r}
x = "  too much space \t\n"
string_magic("With trim: {tws, Q ? x}")
```

## `q`, `Q`, `bq`: Add various type of quotes

To add quotes to the strings. q: single quotes, Q: double quotes, bq: 
back quotes.
```{r}
x = c("Mark", "Pam")
string_magic("Hello {q, enum ? x}!")
```

## `format`, `Format`: Format the values with `base::format`

Applies the base R's function `base::format()` to the string. 
By default, the values are left aligned, *even numbers* (differently from `base::format()`'s behavior).
The upper case command (`Format`) applies right alignment. Options: `"0"`, `"left"`, `"zero"`, `"right"`, `"center"`.
Options `"0"` or `"zero"` fills the blanks with 0s: useful to format numbers. Option `"right"` right aligns,
and `"center"` centers the strings.
```{r}
x = c(1, 12345) 
cat_magic("left  : {format, q, enum ? x}", 
          "right : {Format, q, enum ? x}",
          "center: {format.center, q, enum ? x}",
          "zero  : {format.0, q, enum ? x}", .sep = "\n")
```

## `%:` Apply `sprintf` formatting

Applies `base::sprintf()` formatting. The syntax is `'arg'%` with arg an sprintf formatting,
or directly the sprint formatting.
```{r}
string_magic("pi = {%.3f ? pi}")
```

## `stopwords`: Remove stop words

Removes basic English stopwords (the snowball list is used). 
The stopwords are replaced with an empty space but the left and right WS are 
untouched. So WS normalization may be needed (see operation `ws`).
```{r}
x = c("He is tall", "He isn't young")
string_magic("Is he {stop, ws, enum ? x}?")
```

## `ascii`: Turn the string to ASCII

Turns all letters into ASCII with transliteration. Failed translations are transformed 
into question marks. Options: `"silent"`, `"utf8"`. By default, if some conversion fails
a warning is prompted. Option `"silent"` disables the warning in case of failed conversion. The conversion 
is done with `base::iconv()`, option `"utf8"` indicates that the source endocing is UTF-8, can be useful 
in some cases.
```{r}
author = "Laurent Bergé"
string_magic("This package has been developped by {ascii ? author}.")
```

## `round`, `signif`, `r0`-`r6`, `s0`-`s6`: Formatting numbers

Formats numbers by rounding at a given level or displaying a certain number of significant digits.

Arguments:
- `'suffix'`: adds a suffix
- `'prefix|suffix'`: adds a prefix and a suffix, the two are separated with a pipe.

Options:
- `0`-`9`: the number of digits to round at (`round`) or the number of significant digits (`signif`)
- `int`: whether to preserve integers from formattting. By default decimals are appended to integers (unless rounding at 0 digits is requested)
- `nocomma`: whether to drop the comma separating the thousands
- `s0`-`s9` (options for `round`): additionnaly the number of significant digits to preserve
- `r0`-`r9` (options for `signif`): additionnaly the number of digits to round at

Algorithm:
`signif` displays `d` significant digits (default is `1`). If the `d`th significant digit is not a decimal, the digits up to the first decimal are displayed. Example if `d = 2`, `153.12` will be displayed as `153.1`; `0.0153` will be displayed as `0.015`; `153` as `153.0`.

`round` displays up to the `d`th decimal (default is `0`). Example if `d = 2`: `153.154` will be displayed as `153.15`; `153`  as `153.00`.

When `signif` and `round` are combined, round provides the *minimum* number of decimals displayed.

Note that non-numeric vectors are converted to numeric beforehand, non-numeric values that could not be converted are preserved. To have more control on the conversion, you can apply the [`num` operation](#subsec_num) beforehand.


```{r}
# rounding at 2 digits
cat_magic("pi = {r2 ? pi}\n",
          # same as above with a different syntax
          "pi = {round.2 ? pi}")

x = c(153, 207.256, 0.00254, 15231312.2)
# keeping one significant digit
cat_magic("v1: {s1, align ? x} ",
          # removing the comma for large numbers and preserving ints
          "v2: {s1.int.nocomma ? x}", .collapse = "\n")
          
# combining signif with round
y = c(pi, 0.00125)
cat_magic("raw: {align ? y} ",
          "s1: {s1, align ? y} ", 
          "r2: {r2, align ? y} ",
          "s1.r2: {s1.r2 ? y}", .collapse = "\n")

# prefix and prefix: use and argument
z = c(55, 22.5, 21)
cat_magic("Costs in euros  : {' €'r1, enum ? z}.",
          "\nCosts in dollars: {'USD |'r1, enum ? z}.")

# non numeric values are preserved
xraw = c("55", "five")
string_magic("Valuess {r1, enum ? xraw}.")
# use the `num` command for more control
string_magic("Values: {num.rm, r1, enum ? xraw}.")
```

## `n`, `N`: Formatting integers

Formats integers by adding a comma to separate thousands. 

Options: 
- `"letter"`: writes the number in letters (large numbers keep their numeric format)
- `"upper"`: like the option `"letter"` but uppercases the first letter
- `"0"`, `"zero"`: left pads numeric vectors with 0s
- `"roman"`, `"Roman"`: write the integer in Roman with `utils::as.roman`. The lower case version writes them in lower case

Variants:
- the upper case command `N` adds the option `"letter"` (i.e. is equiv. to `n.letter`)


```{r}
x = c(5, 12, 52123)
string_magic("She owes {n, '$'paste, enum ? x}.")

# option 0: all same width, no ',' for thousands
cat_magic("|---|\n{n.0, '\n'collapse ? x}")

# option "upper"
n = 5
string_magic("{n.upper ? n} is my favourite number.")

# N: like "n.letter"
x = 5
string_magic("He's {N ? x} years old.")

# roman
string_magic("What's nicer? {collapse?11:13}, {n.roman, c?11:13} or {n.Roman, c?11:13}?")
```

## `nth`, `Nth`: Numbered position

When applied to a number, this operator writes them as a rank. Options: `"letter"`, 
`"upper"`, `"compact"`. 
```{r}
n = c(3, 7)
string_magic("They finished {nth, enum ? n}!")
```
Option `"letter"` tries to write the numbers in letters, but note that it stops at 20. Option `"upper"`
is the same as `"letter"` but uppercases the first letter. Option `"compact"` aggregates
consecutive sequences in the form `"start_n_th to end_n_th"`. 
```{r}
string_magic("They arrived {nth.compact ? 5:20}.")
```
The upper case command (`Nth`) adds the option `"letter"`.
```{r}
n = c(3, 7)
string_magic("They finished {Nth, enum ? n}!")
```

## `ntimes`, `Ntimes`: Number of times

Write numbers in the form `n` times. Options: `"letter"`, `"upper"`. Option 
`"letter"` writes the number in letters (up to 100). Option `"upper"` does the same as `"letter"` 
and uppercases the first letter. 
```{r}
string_magic("They lost {enum ! {ntimes ? c(1, 12)} against {S!Real, Barcelona}}.")
```
The upper case command (`Ntimes`) adds the option `"letter"`.
```{r}
x = 5
string_magic("This paper was rejected {Ntimes ? x}...")
```

## `firstchar`, `lastchar`: Keep only the first, last characters

To select the first/last characters of each element. 
Negative numbers remove the first/last characters.
```{r}
string_magic("{19 firstchar, 9 lastchar ! This is a very long sentence}")

string_magic("delete 3 = {-3 firstchar ! delete 3}")
```

## `k`, `shorten`, `Shorten`: Shortens character strings

To keep only the first `n` characters (like `firstchar` but with more options). 
(Note that `k` stands for "keep" and exists for historical reasons.)
Available options: `"include"`, `"dots"`. The argument can be of the form `'n'` or `'n|s'` 
with `n` a number and `s` a string. `n` provides the number of characters to keep. 
Optionnaly, only for strings *whose length is greater than `n`*, 
after truncation, the string `s` is appended at the end.

By default, if argument `s` is provided, strings longer than `n` end up at size `n + nchar(s)`.
If option `"include"` is provided, the strings are guaranteed to be of maximum size `n`,
even after the string `s` has been appended. Example: if `n=4` and `s=".."`, then "hello"
becomes "hell.." without `"include"`, and "he.." with it.

Option `"dots"`: if strings are longer than `n+1`, they are truncated at `n-1` and two dots are 
appended. For example if `n = 3`, then "hello" becomes "he..". Disregards the argument `s`.
The operation `"Shorten"` (upper case), is with the option `"dots"`.

Note that another way to add the option `"include"` is to use a double pipe
for the argument `s`, like in `'n||s'`.

```{r}
x = "long sentence"
cat_magic("v0: {x}", 
          "v1: {4 shorten ? x}", 
          "v2: {'4|..'shorten ? x}", 
          "v3: {'4|..'shorten.include ? x}", 
          "v4: {4 shorten.dots ? x}", .sep = "\n")
```

## `fill`, `align`, `width`: Fill character strings

Fills the character strings up to a size in order to fit a given width. `align` and `width` are aliases to `fill`. 
Options: `"right"` or `"center"`. Default is left-alignment of the strings. 

The argument is optional and can be of the form `'n'` or `'n|s'`. By default if no argument
is provided, of if `n=0`,`n` is equal to the maximum character length of the vector. The optional
argument `s` is a symbol used to fill the blanks. By default `s` is equal to a white space.

Option `"right"` right aligns and `"center"` centers the strings.

See help for [`string_fill()`](https://lrberge.github.io/stringmagic/reference/string_fill.html) 
for more information.
```{r}
life = "full of sound and fury, Signifying nothing"
cat_magic("{'[ ,]+'split, upper.first, fill.center, q, '\n'collapse ? life}")

# fixing the length and filling with 0s
string_magic("{'5|0'fill.right, enum ? c(1, 55)}")
```

## `paste`, `append`: Append text

Pastes a custom character string to all elements of the string. The operations `paste` and `append` are equivalent.
This operation has no default. 
Options: `"left"`, `"both"`, `"right"`, `"front"`, `"back"`, `"delete"`. By default, a string is pasted on the left.
Option `"right"` pastes on the right and `"both"` pastes on both sides. Option `"front"` only 
pastes on the first element while option `"back"` only pastes on the last element. Option `"delete"`
first replaces all elements with the empty string.
```{r}
string_magic("y = {'x'paste, ' + 'collapse ? 1:3}")
```

The argument can be of the form `s1` or `s1|s2`. If of the second form, this is equivalent 
to chaining two `paste` operations, once on the left and once on the right: `'s1'paste, 's2'paste.right`.

```{r}
string_magic("y = {'x|0'paste, ' + 'collapse ? 1:3}")
```

## `join`: Join lines

Joins lines ending with a double backslash. 

```{r}
sun = "The sun \\
is shining."
string_magic("How is the sun? {join ? sun}")
```

## `escape`: Escape special characters

Adds backslashes in front of specific characters. Options `"nl"`, `"tab"`. 
Option `"nl"` escapes the newlines (`\n`), leading them to be displayed as `"\\\\n"`.
Option `"tab"` does the same for tabs (`"\t"`). This is useful to make the value free
of space formatters. 
The default behavior is to escape both newlines and tabs.

```{r}
input = "yes \n\n no"
msg = string_magic("Your input, equal to {escape, bq ? input} is incorrect.")
cat(msg, sep = "\n")
```


# Other operations {#sec_other}

## `num`: Convert to numeric {#subsec_num}

Converts to numeric. Options: `"warn"`, `"soft"`, `"rm"`, `"clear"`. By default, the conversion
is performed silently and elements that fail to convert are turned into NA. 
Option `"warn"` displays a warning if the conversion to numeric fails. 
Option `"soft"` does not convert if the conversion of at least one element fails, leading to 
a character vector. 
Option `"rm"` converts and removes the elements that could not be converted. 
Option `"clear"` turns failed conversions into the empty string, and hence lead to a character vector.
```{r}
x = c(5, "six")
cat_magic("   origin: {enum, q ? x}", 
          "      num: {num, enum, q ? x}", 
          "   num.rm: {num.rm, enum, q ? x}", 
          " num.soft: {num.soft, enum, q ? x}", 
          "num.clear: {num.clear, enum, q ? x}", .sep = "\n")
```

## `enum`: Create an enumeration

Enumerates the elements. It creates a single string containing the comma 
separated list of elements.
If there are more than 7 elements, only the first 6 are shown and the number of
items left is written.
```{r}
string_magic("{enum ? 1:5}")
```
You can add the following options:

  + `q`, `Q`, or `bq`: to quote the elements
  + `or`, `nor`: to finish with an 'or' (or 'nor') instead of an 'and'
  + `comma`: to finish the enumeration with ", " instead of ", and".
  + `i`, `I`, `a`, `A`, `1`: to enumerate with this prefix, like in: i) one, and ii) two
  + a number: to tell the number of items to display
```{r}
x = c("Marv", "Nancy")
string_magic("The murderer must be {enum.or ? x}.")

x = c("oranges", "milk", "rice")
string_magic("Shopping list: {enum.i.q ? x}.")

# enum is made for display: when vectors are too long, they are truncated
# default is at 7
x = string_magic("x{sample(100, 30)}")
string_magic("The problematic variables are {'x'sort.num, enum ? x}.")

# you can augment or reduce the numbers to display with an option
string_magic("The problematic variables are {'x'sort.num, enum.3 ? x}.")
```


## `len`, `Len`: Formatted length

Gives the length of the vector. Options `"letter"`, `"upper"`, `"num"`.
Option `"letter"` writes the length in words (up to 100). Option `"upper"` is the same 
as letter but uppercases the first letter. By default, the number is formatted:
commas are inserted to separate thousands.
```{r}
cat_magic("The length of 1:5000:", 
          " - len     = {len ? 1:5000}", 
          " - len.num = {len.num ? 1:5000}", .sep = " \n")
```
The upper case command (`Len`) adds the option `"letter"` (only for small numbers).
```{r}
string_magic("Its size is {Len ? 1:8}")
```

## `swidth`: Add newlines to force the string to fit a given width

Thie operation stands for "screen width". It formats the string to fit a given width by cutting at word boundaries and adding newlines appropriately. 
Accepts arguments of the form `'n'` or `'n|s'`, with `n` a number and `s` a string. 
An argument of the form `'n|s'` will add `s` at the beginning of each line. Further,
by default a trailing white space is added to `s`; to remove this 
behavior, add an underscore at the end of it. 
The argument `n` is either 
an integer giving the target character width (minimum is 15), or it can be a fraction expressing the 
target size as a fraction of the current screen. Finally it can be an expression that 
uses the variable `.sw` which will capture the value of the current screen width.
```{r}
x = "this is a long sentence"
cat_magic("------ version 0 ------\n{x}", 
          "------ version 1 ------\n{15 swidth ? x}", 
          "------ version 2 ------\n{'15|#>'swidth ? x}",
          "------ version 3 ------\n{'15|#>_'swidth ? x}", .sep = "\n")
```

## `difftime`: Formatted time difference

Displays a formatted time difference. Option `"silent"` does not report a warning if the
operation fails. It accepts either objects of class `POSIXt` or `difftime`.
```{r}
x = Sys.time()
Sys.sleep(0.15) 
string_magic("Time: {difftime ? x}")
```

---
title: "Regular expression logic"
author: "Laurent R. Bergé"
date: "`r Sys.Date()`"
output:
  html_document:
    theme: journal
    highlight: haddock
    number_sections: yes
    toc: yes
    toc_depth: 3
    toc_float:
      collapsed: no
      smooth_scroll: no
  pdf_document:
    toc: yes
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{ref_regex_logic}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(stringmagic)
```

In `stringmagic`, any time you use a regular expression (regex) to *detect a pattern* in
a character string, you can use regex logic. The syntax to logically combine
regular expressions is intuitive: simply use regular logical operators and it will work!

The functions for whcih regex logic is available are: a) pattern detection functions 
(`string_is`, `string_get`, etc), and b) string replacement functions (`string_clean`, 
`string_replace`) with the `total` flag (see the [vignette on regex flags](https://lrberge.github.io/stringmagic/articles/ref_regex_flags.html)).


# Logically combining regex patterns {#detect_logic}

Assume `"pat1"` and `"pat2"` are two regular expression patterns and we want 
to test whether the string `x` contains a combination of these patterns. 
Then:

- `"pat1 & pat2"` = `x` contains `pat1` AND `x` contains `pat2`
- `"pat1 | pat2"` = `x` contains `pat1` OR `x` contains `pat2`
- `"!pat1"` = `x` does not contain `pat1`
- `"!pat1 & pat2"` = `x` does not contain `pat1` AND `x` contains `pat2`

Hence the three logial operators are:

- `" & "`: logical AND, it **must** be a space + an ampersand + a space 
(just the `&` *does not work*)
- `" | "`: logical OR, it **must** be a space + a pipe + a space 
(just the `|` *does not work*)
- `"!"`: logical NOT, it works only when it is the first character of the pattern.
Note that anything after it (including spaces and other `!`) *is part of the regular expression*

The parsing of the logical elements is done before any regex interpretation. 
The logical evaluations are done from left to right and are sequentially combined. 

Ex: selecting cars.
```{r}
cars = row.names(mtcars)
print(cars)

# which one...
# ... contains all letters 'a', 'e', 'i' AND 'o'?
string_get(cars, "a & e & i & o")

# ... does NOT contain any digit?
string_get(cars, "!\\d")
```

You **cannot** combine logical statements with parentheses. 

For example: `"hello | (world & my lady)"` leads to: `x` contains `"hello"` or contains `"(world"`, 
and contains `"my lady)"`. The two latter are invalid regexes but can make sense if you
have the flag "fixed" turned on. To escape the meaning of the logical operators, 
see the [dedicated section](#logical_escape).

The logical `"not"` always apply to a single pattern and **not** to the full pattern.

### Escaping the meaning of the logical operators {#logical_escape}

To escape the meaning of the logical operators, there are two solutions to 
escape them:

- use two backslashes just before the operator: `"a \\& b"` means `x`
contains `"a & b"`
- use a regex hack: the previous example is equivalent to `"a [&] b"` in regex parlance
and won't be parsed as a logical AND

The two solutions work for the three operators: `" & "`, `" | "` and `"!"`.

### How do regex flags work with logically combined regexes? {#logical_flags}

All `stringmagic` regexes accept optional flags. Please see the [associated vignette](https://lrberge.github.io/stringmagic/articles/ref_regex_flags.html).

When you add flags to a pattern, these apply to *all* regex sub-patterns. 
This means that `"f/( | )"` treats the two parentheses as "fixed". 
*You cannot add flags specific to a single sub-pattern.*

---
title: "`string_magic`'s sepcial operations"
author: "Laurent R. Berge"
date: "`r Sys.Date()`"
output:
  html_document:
    theme: journal
    highlight: haddock
    number_sections: yes
    toc: yes
    toc_depth: 3
    toc_float:
      collapsed: yes
      smooth_scroll: no
  pdf_document:
    toc: yes
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{ref_string_magic_special_ops}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(stringmagic)
```

This vingette presents advanced operations included in `string_magic`. There are four parts:

- [group-wise operations](#sec_group_wise)
- [conditional operations](#sec_conditional)
- [compact if-else operator](#sec_ifelse)
- [pluralization](#sec_pluralization)

By default the function `string_magic` returns a plain character vector. In this 
vignette it is sometimes nicer to apply the function `base::cat` to display 
`string_magic` results containing newlines. Ths function `cat_magic` does exactly 
that and we will use it from time to time.

# Group-wise operations {#sec_group_wise}

In `string_magic`, the operations `split` and `extract` keep a memory of the strings 
that were split (i.e. they provide multiple results for each initial string element). 

Use the tilde operator, of the form `~(op1, op2)`, to apply operations
group-wise, to each of the split strings. Better with an example.

```{r}
x = c("Oreste, Hermione", "Hermione, Pyrrhus", "Pyrrhus, Andromaque")
string_magic("Troubles ahead: {', 'split, ~(' loves 'collapse), enum ? x}.")
```

Almost all operations can be applied group-wise (although only operations changing the order or 
the length of the strings really matter, see the [dedicated section](https://lrberge.github.io/stringmagic/articles/ref_operations.html#sec_op_length)).

# Conditional operations {#sec_conditional}

There are two operators to apply operations conditionally: `if` and `vif`, the latter
standing for *verbatim if*.

## *if* statement

The syntax of `if` is `if(cond ; ops_true ; ops_false)` with `cond` a
condition (i.e. logical operation) on the value being interpolated, `ops_true` a comma-separated
sequence of operations if the condition is `TRUE` and `ops_false` an *optional* a sequence of
operations if the condition is `FALSE`.

The condition `cond` accepts the following special values: 

- `.` (the dot): refers to the current vector
- `.nchar` or `.C`: represent the number of characters of the current vector (equivalent to `nchar(.)`)
- `.len` or `.N`.: represent the length of the current vector (equivalent to `length(.)`)

Ex.1: Let's take a sentence, delete words of less than 4 characters, and trim 
words of 7+ characters. 
```{r}
x = "Songe Céphise à cette nuit cruelle qui fut pour tout un peuple une nuit éternelle"
string_magic("{' 'split, if(.nchar<=4 ; nuke ; 7 shorten), collapse ? x}")
```

Let's break it down. First the sentence is split w.r.t. spaces (command `' 'split`), leading to a vector
of words. Then we use the special variable `.nchar` in `if`'s condition to refer 
to the number of characters of the current vector (the words). The words with 
less than 4 characters are nuked (i.e. removed), and the other words are
trimmed at 7 characters (`7 Shorten`). Finally the modified vector of words is collapsed with 
the command `collapse`, leading to the result.

In Ex.1 the condition led to a vector of length greater than 1 (length = number of words), 
triggerring element-wise operations. 

If a condition leads to a result of length 1, then the operations are applied to 
the full string vector. 
Contrary to element-wise conditions
for which operations modifying the length of the vectors are **forbidden** (apart from nuking),
such operations are fine in full-string conditions.

Ex.2: we write the sum of several elements, if the vector is longer than 4, we replace all
remaining elements with an ellispsis.
```{r}
# same expression for two values of x give different results
x_short = string_magic("x{1:4}")
# the false statement is missing: it means that nothing is done is .N<=4
string_magic("y = {if(.N>4 ; 3 first, '...'insert.right), ' + 'c ? x_short}")

x_long = string_magic("x{1:10}")
string_magic("y = {if(.N>4 ; 3 first, '...'insert.right), ' + 'c ? x_long}")
```

In this example, the operations applied are:

- short vector: `string_magic("y = {' + 'c ? x_short}")`
- long vector: `string_magic("y = {3 first, '...'insert.right, ' + 'c ? x_long}")`

## *Verbatim if* statement

For `vif`, the syntax is `vif(cond ; verb_true ; verb_false)` with `verb_true`
a verbatim value with which the vector will be replaced if the condition is `TRUE`. 
This is similar for `verb_false`. The condition works as in `if`.

As for the if operator, you can use the special values `"."`, `.len`, `.N`, `.nchar` 
and `.C` in the condition.
On top of this, you can use `'.'` to refer to the current value in `verb_true` 
and `verb_false`, as illustrated by the following example.

Ex.3: we want to replace all values lower than 10 by the the string "<10", and then
create an enumeration.
```{r}
pval = c(1e-20, 0.15, 0.5)
cat_magic("pvalues: {vif(.<1e-16 ; <1e-16 ; {%05f ? .}), align.right ? pval}", 
          .sep = "\n")
```

In this example, the condition is of the same length as the vector, so an element-wise 
operation is triggered. Note that we use `'.'` to refer to `pval` in the condition.
Elements lower than 1e-16 are replaced with the string `"<1e-16"`.
Other elements are replaced with `{%05f?.}`. This string contains the interpolation delimiters,
interpolation applies. The dot, `'.'`, now refers to the values of `pval` respecting the condition.
And `sprintf` formatting is applied (`%05f`). After this, we right align the results.

Conditions of length 1 apply the replacement 
to the full vector. Knowing this, let's redo Ex.2 differently:
```{r}
x = string_magic("x{1:10}")
string_magic("y = {vif(.N>4 ; {first?x} + ... + {last?x} ; {' + 'c ? x}) ? x}")
```

Let's break it down. If the length of the vector is greater than 4 (here it's 10), then
the full string is replaced with `"{first?x} + ... + {last?x}"`. Since this string 
contains curly brackets, interpolation applies. 
Hence we obtain the string `"x1 + ... + x10"`. Finally, this is collated to `"y = "` 
leading to the result. 

If the vector were of length lower than 4, it would have been replaced with `"{' + 'collapse?x}"`,
which will be interpolated.

# Special interpolation: if-else {#sec_ifelse}

Using an ampersand (`"&"`) as the first character of an interpolation leads to an *if-else* operation.
Using two ampersands (`"&&"`) leads to a slightly different operation 
described at the end of this section.

## Regular if-else: "&"

The syntax is as follows: `{&cond ; verb_true ; verb_false}` with `cond` a
condition (i.e. logical operation), `verb_true`
a verbatim value with which the vector will be replaced if the condition is `TRUE` and 
`verb_false` an *optional* verbatim value with which the vector will be replaced 
if the condition is `FALSE`. 
If not provided, `verb_false` is considered to be the *empty string* unless the operator is 
the double ampersand described (`&&`) at the end of this section.

Note that in `cond`, you can use the function `len`, an alias to `length`.

Ex.1: we take a vector and compose a message depending on its length. If its length is
lower than 10 then we write the message `"x is short"`, otherwise we write `"x is long"`,
and insert `"very"` depending on the number of digits.
```{r}
x = 1:5
string_magic("x is {&len(x)<10 ; short ; {`log10(.N)-1`times, ''c ! very }long}")

x = 1:50
string_magic("x is {&len(x)<10 ; short ; {`log10(.N)-1`times, ''c ! very }long}")

x = 1:5000
string_magic("x is {&len(x)<10 ; short ; {`log10(.N)-1`times, ''c ! very }long}")
```

If a condition leads to a result of length 1, the full string is replaced by the verbatim 
expression. Further, this expression will be interpolated if requested. This was the case
in Ex.1 where `verb_false` was interpolated.

If the condition uses a variable, in later interpolations you can refer to the 
first variable present in the condition with '.', and use '.len' or '.N' to refer
to its length.

If the condition's length is greater than 1, then each logical values equal to `TRUE` is replaced
by `verb_true`, and `FALSE` values are replaced with `verb_false`. There can be interpolation
on the values `verb_true` and `verb_false`. In that case the interpolation must
result into a vector of either length 1 or a length equal to the condition. Then element-wise
replacements are made, a la `base::ifelse`.

Ex.2: illustration of element-wise replacements.
```{r}
x = 1:4
y = letters[1:4]
string_magic("{&x %% 2 ; odd ; {y}}")
```

In that example, when x is odd, it is replaced with `"odd"`, and when even it is
replaced with the elements of y.

## *if-else* with automatic filling

Using the two ampersands operator (`&&`) is like the simple ampersand version but the 
default for `verb_false` is the variable used in the condition itself. So the syntax is
`{&&cond ; verb_true}` and *it does not accept* `verb_false`.

Ex.3: let's write the integer `i` in letters when equal to 3 only.
```{r}
i = 3 
string_magic("i = {&&i == 3 ; three}")

i = 5
string_magic("i = {&&i == 3 ; three}")
```

When the condition is of length 1: the full vector is replaced. When the condition is 
of the length of the vector, an element-wise replacement is triggered, like in example 3.

# Special interpolation: Pluralization {#sec_pluralization}

There is advanced support for pluralization which greatly facilitates the writing of messages 
in natural language.

## Pluralization: Principles

There are two ways to pluralize: over length or over value. To trigger a "pluralization" interpolation
use as first character:

- `#` to pluralize over the value of a variable (see Ex.1)
- `$` to pluralize over the length of a variable (see Ex.2)

Ex.1: we add an ending 's' based on a number.
```{r}
x = 5
string_magic("I bought {N?x} book{#s}.")

x = 1
string_magic("I bought {N?x} book{#s}.")
```

The syntax is `{#plural_ops ? variable}` or `{#plural_ops}` where `plural_ops` are
specific pluralization operations which will be described below. 
The pluralization is perfomed *always* with respect to the value of a variable. 

You can either add the variable explicitly (`{#plural_ops ? variable}`) or refer
to it implicitly (`{#plural_ops}`). If implicit, then the algorithm will look at the 
previous variable that was interpolated and pluralize over it. This is exaclty what happens in
Ex.1 where `x` was interpolated in `{N?x}` and the plural operation `s` (in `{#s}`) applies to 
`x`. It would have been equivalent to have `{#s ? x}`. If a variable wasn't interpolated before, then
the next interpolated variable will be used (see Ex.2). If no variable is interpolated
at all, an error is thrown.

Ex.2: we add an ending 's' and conjugate the verb 'be' based on the length of a vector.
```{r}
x = c("J.", "M.")
string_magic("My BFF{$s, are} {enum?x}!")

x = "J."
string_magic("My BFF{$s, are} {enum?x}!")
```

As you can notice in Ex.2, you can chain operations (here `'s'` and `'are'`). 
In that case a whitespace is automatically added between them.

Now let's come to the specific pluralization operations, which are different 
from regular operations.

## Pluralization: Regular operations

### s, es

Adds an `"s"` (or `"es"`) if it is plural (> 1), nothing otherwise. Accepts the option `0` or `zero` which 
treats a 0-length or a 0-value as plural.

```{r}
nfiles = 1
string_magic("We've found {#n.no ? nfiles} file{#s}.")

nfiles = 0
string_magic("We've found {#n.no ? nfiles} file{#s}.")

nfiles = 0
string_magic("We've found {#n.no ? nfiles} file{#s.0}.")

nfiles = 4
string_magic("We've found {#n.no ? nfiles} file{#s.0}.")
```

### y or ies

Adds an 'y' if singular and 'ies' if plural (>1). Accepts the option `0` or `zero` which 
treats a 0-length or a 0-value as plural.

```{r}
ndir = 1
string_magic("We've found {ndir} director{#y}.")

ndir = 5
string_magic("We've found {ndir} director{#y}.")

ndir = 1
string_magic("We've found {ndir} director{#ies}.")
```

### enum

Enumerates the elements (see help for the [regular operation `enum`](https://lrberge.github.io/stringmagic/articles/ref_operations.html#create-an-enumeration-enum)).

```{r}
fruits = c("apples", "oranges")
string_magic("The fruit{$s ? fruits} I love {$are, enum}.")

fruits = "apples"
string_magic("The fruit{$s ? fruits} I love {$are, enum}.")
```

### n, N, len, Len

Add the number of elements (`"len"`) or the value (`"n"`) of the variable as a formatted number or 
in letters (upper case versions). Accepts the options `letter` (to write in letter) 
and `upper` (to uppercase the first letter).

You can also pass the options `no` or `No`, which replace the 0 values with "no"/"No". Alternatively, 
pass a free-form argument to be used in lieu of 0 values.

```{r}
nfiles = 5
string_magic("{#N.upper.No ? nfiles} file{#s, are} compromised.")

nfiles = 1
string_magic("{#N.upper.No ? nfiles} file{#s, are} compromised.")

nfiles = 0
string_magic("{#N.upper.No ? nfiles} file{#s, are} compromised.")

# Using free-form arguments
nfiles = 5
string_magic("{#'Absolutely no'N.upper ? nfiles} file{#s, are} compromised.")

nfiles = 0
string_magic("{#'Absolutely no'N.upper ? nfiles} file{#s, are} compromised.")
```

### nth, ntimes

Writes the value of the variable as an order (nth) or a frequence (ntimes). Accepts the option `letter`
to write the numbers in letters (uppercase version of the operator does the same).

```{r}
n = 2
string_magic("Writing the same sentence {#Ntimes ? n} is unnecessary.")
```

### is, or any verb

Conjugates any English verb appropriately depending on context. Any command that
is not recognized as one of the commands previously described is treated as a verb.

Simply add an upper case first to upper case the conjugated verb.

Multiple verbs are illustrated in the example below. It also anticipaed the 
conditional statements described in the next section.

Ex.3: multiple verbs and conditional statements.
```{r}
pple = c("Francis", "Henry")
cat_magic("{$enum, is, (a;) ? pple} tall guy{$s}.",
        "{$(He;They), like} to eat donuts.",
        "When happy, at the pub {$(he;they), goes}!",
        "{$Don't, (he;they)} have wit, {$(he;they)} who {$try}?", .sep = "\n")

pple = "Francis"
cat_magic("{$enum, is, (a;) ? pple} tall guy{$s}.",
        "{$(He;They), like} to eat donuts.",
        "When happy, at the pub {$(he;they), goes}!",
        "{$Don't, (he;they)} have wit, {$(he;they)} who {$try}?", .sep = "\n")

```

## Pluralization: Conditional statements

On top of the previous operations, there is a special operation allowing to add 
verbatim text depending on the situation. The syntax is as follows:

- `(s1;s2)`: adds verbatim 's1' if singular and 's2' if plural (>1)
- `(s1;s2;s3)`: adds verbatim 's1' if zero, 's2' if singular (=1) and 's3' if plural
- `(s1;;s3)`: adds verbatim 's1' if zero, 's3' if singular or plural (i.e. >=1)

These case-dependent verbatim values **are interpolated** (if appropriate). In these interpolations
you need not refer explicitly to the variable for pluralization interpolations.

```{r}
x = 0
string_magic("{#(Sorry, nothing found.;;{#N.upper} match{#es, were} found.)?x}")

x = 1
string_magic("{#(Sorry, nothing found.;;{#N.upper} match{#es, were} found.)?x}")

x = 3
string_magic("{#(Sorry, nothing found.;;{#N.upper} match{#es, were} found.)?x}")
```


# Direct access to the current time and a timer facility {#sec_timer}

### Current date and time

You can refer to the current date or the current time with the special variables 
.date and .now:

- `.date` is equivalent to `Sys.date()`
- `.now` is equivalent to `Sys.time()`

Here is an example where we display the day:
```{r}
string_magic("This message has been written on {.date}.")
```

On top of this, you can use `.now` as a function whose sole argument
provides the format of the time. The format follows 
`base::strptime` (see `?strptime`).
In the following example, we display the day, month and the hour:

```{r}
string_magic("This message has been written on {.now('%A %B at %Hh%M')}.")
```

### Timer

Timers can be very useful inside code to find choke points and debug. `stringmagic` offers
a simple system:

- first set the timer anywhere with `timer_magic()`
- within `*_magic` functions, display the elapsed time in three possible special variables:
  - `.timer`: displays the elapsed time and resets the timer
  - `.timer_lap`: displays the elapsed time and *does not* reset the timer
  - `.timer_total`: displays the elapsed time since the `timer_magic()` call

Here is an example where we time a few computations within a function:

```{r}
rnorm_crossprod = function(n, mean = 0, sd = 1){
  # we set the timer
  timer_magic()
  # we compute some stuff
  x = rnorm(n, mean, sd)
  # we can report the time with .timer
  message_magic("{10 align ! Generation}: {.timer}")
  
  res = x %*% x
  message_magic("{10 align ! Product}: {.timer}",
                "{10 align ! Total}: {.timer_total}", .sep = "\n")
  res
}

rnorm_crossprod(1e5)
```

Note that timer is precise at +/- 1ms (this is due to its very simple interface), 
and hence should not be used to time code chunks with very short execution times.

We could refine by trigerring the messages only when debugging. We can do that by
using the argument `.trigger`:

```{r}
rnorm_crossprod = function(n, mean = 0, sd = 1, debug = FALSE){
  # we set the timer
  timer_magic()
  # we compute some stuff
  x = rnorm(n, mean, sd)
  # we can report the time with .timer
  message_magic("{10 align ! Generation}: {.timer}", .trigger = debug)
  
  res = x %*% x
  message_magic("{10 align ! Product}: {.timer}",
                "{10 align ! Total}: {.timer_total}", 
                .sep = "\n", .trigger = debug)
                
  res
}

# timer not shown
rnorm_crossprod(1e5)

# timers shown thanks to the argument
rnorm_crossprod(1e5, debug = TRUE)
```

---
title: "Tailor-made functions and operations"
author: "Laurent R. Bergé"
date: "`r Sys.Date()`"
output:
  html_document:
    theme: journal
    highlight: haddock
    number_sections: yes
    toc: yes
    toc_depth: 3
    toc_float:
      collapsed: no
      smooth_scroll: no
  pdf_document:
    toc: yes
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{guide_customization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  collapse = TRUE, 
  comment = "#>"
)

library(stringmagic)
```

This vignette presents how to customize `stringmagic` to (better) suit your needs. It covers:

- how to [create custom functions with alias generators](#sec_alias)
- how to [create new string operations](#sec_creation) available within 
[`string_magic`](https://lrberge.github.io/stringmagic/articles/guide_string_magic.html)
- how to [use `stringmagic` in a package when employing custom operations](#sec_package) (this is advanced)

# Creating new functions with alias generators {#sec_alias}

Several `stringmagic` functions dispose of alias generators, which end with the suffix
`_alias`. They generate a copy of the function with different default values (and also take
care of setting up the environment correctly). Some functions have many arguments,
by changing the default arguments you can create completely new functions. 

This section will guide you through alias generation, and how they can be useful, using examples.

### Creating a formula builder

This is a common R problem: *how to turn a character string into a formula, in a handy way*? 
You can use a few arguments of `string_magic` to make it work:

- `.default`: to apply a default sequence of operations to interpolations
- `.post`: to appy a custom function right before returning the object

The objective here is to inject variable names into a character vector and turn it into a formula.
Since variables in a formula are separated with a `"+"`, we need to collapse several variables
with `"+"`. This will be achieved with the `.default` argument. Then we turn it into a formula
by passing `as.formula` in the `.post` argument.

```{r}
y = "Petal.Length"
x = c("Sepal.Length", "Petal.Width", "Species")
string_magic("{y} ~ {x}", .default = "' + 'collapse", .post = as.formula)
```

Now that we see that our builder works with `string_magic`, we create a dedicated function with an alias.

```{r}
fml_builder = string_magic_alias(.default = "' + 'collapse", .post = as.formula)
```

The function `fml_builder` works as `string_magic` but with different default values. 
Now we can apply it directly.

```{r}
fml_builder("{y} ~ {x}")
```

Since this function is just a call to `string_magic`, you can apply anything you want. Let's
scale the variables on the right-hand-side by using nesting:

```{r}
x = c("Sepal.Length", "Petal.Width")
fml_builder("{y} ~ {'+'collapse ! scale({x})}")
```

### Changing `str_clean`

The function [`str_clean`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_clean) specialses in cleaning character vectors. To do so it uses a specific syntax
to transform various regular expressions at once. For example `"pat1, pat2 => replacement"` will turn the 
regular expression `pat1` and `pat2` into the `replacement`. The syntax is: i) a comma separated
list of regular expressions, 2) a pipe (`" => "`), 3) the replacement. 
If you are not happy with the fact that regexes are separated with commas, or of the look of the pipe, no problem!

Let's change the default values, let's: i) use a semi-colon separation, ii) use `">>"` instead of the regular pipe.
```{r}
my_clean = string_clean_alias(split = "; ", pipe = " >> ")

x = "My name is Bond, James Bond"
# old way
string_clean(x, "e, o => a")

# new way
my_clean(x, "e; o >> a")
```

### Creating small numeric matrices

The function [`string_vec`](https://lrberge.github.io/stringmagic/articles/guide_string_tools.html#sec_vec) facilitates the creation of small character vectors.
With its arguments `.nmat`, you can turn the vector into a numeric matrix. Hence, let us write
a function to create small numeric matrices.

We will use:

- `.nmat = TRUE` to ask to transform the result into a numeric matrix. The number of 
rows will be deduced from the number of newlines.
-  `.last = "'[\n ,]+'split"` (`.last` means *last* operation) to split the resulting character vector with respect to
newlines, commas and space, so that numbers can be separated by any succession of these

```{r}
num_mat = string_vec_alias(.nmat = TRUE, .last = "'[\n ,]+'split")
num_mat("1, 2, 3
         7, 5, 0
         0, 0, 1")
```

# Creating your own string operations {#sec_creation}

You can add any arbitraty operation to `string_magic`. There are two main ways:

- registering a custom sequence of regular `string_magic` operations (`string_magic_register_ops`) 
- registering a custom function (`string_magic_register_fun`)

### New operations as a sequence of existing operations {#sec_reg_ops}

Create new sequence of operations with `string_magic_register_ops`. It takes two arguments:

1. the sequence of `string_magic` operations, 
2. the name of the new operation.

For example, let's create a new operation `h1` which formats a string into a header. 
It adds an hypen before the text and adds hyphens after the text up to the 40th column.

Ex.1: creating a header operation.
```{r}
# 1) we register the sequence of regular string_magic operations
string_magic_register_ops("'- | 'paste, '40|-'fill", "h1")
# 2) we use it
string_magic("h1 ! That's my header", .nest = TRUE)
```

### New operations using a custom function

To implement new operations using functions, you have two steps: 
1. create the custom function,
2. register the function an provide an alias to the operation.

First you need to create a function that will be applied to a character vector. 

That function **must** have at least the arguments `x` and `...`. 

Additionnaly, it can have the optional arguments: `argument`, `options`, `group`, `group_flag`.
This function must return a vector. Optionnally, and only if relevant (see the last example),
you can add an attribute `"group"` to the returned object which will be used in grouped operations.

Second, you need to register the function with `string_magic_register_fun` and assign an alias to it. 
Optionnally, you can provide a list of valid options. 

Let's create an example in which we add markdown emphasis to words.

Ex.1: a new operation adding markdown emphasis.
```{r}
library(stringmagic)
# A) define the function
fun_emph = function(x, ...) paste0("*", x, "*")
 
# B) register it
string_magic_register_fun(fun_emph, "emph")

# C) use it
x = string_vec("right, now")
string_magic("Take heed, {emph, collapse ? x}.")
```

More generally, the function taken by `string_magic_register_fun` is called internally 
by `string_magic` in the form `fun(x, argument, options, group, group_flag)`.
Here is the meaning of the arguments:

- `x`: the value to which the operation applies. 
- `argument`: the quoted `string_magic` argument (always character). 
- `options`: a character vector of `string_magic` options. 
- `group`: an index of the group to which belongs each observation (integer). 
- `group_flag`: value between 0 and 2; 0: no grouping operation requested; 
1: keep track of groups; 2: apply grouping.

The two last arguments, `group` and `group_flag`, are of use
only in group-wise operations only if `fun` changes the length or the order of vectors. 

Let's add an argument and an option to the `"emph"` operation that we defined in Ex.1.

Ex.2: new operation with argument and option.
```{r}
fun_emph = function(x, argument, options, ...){
  arg = argument
  if(nchar(arg) == 0) arg = "*"
  
  if("strong" %in% options){
    arg = paste0(rep(arg, 3), collapse = "")
  }
  
  paste0(arg, x, arg)
}

string_magic_register_fun(fun_emph, "emph", "strong")

x = string_vec("right, now")
string_magic("Take heed, {'_'emph.s, c? x}.")

# In string_magic_register_fun, the valid_option argument is used to validate them.
try(string_magic("Take heed, {'_'emph.aaa, c? x}."))
```

Finally let's illustrate an example with group-wise awareness. This is somewhat advanced and should
be of concern only when you regularly use [group-wise operations](https://lrberge.github.io/stringmagic/articles/ref_string_magic_special_operations.html#sec_group_wise).

Ex.3: we create a function that only keeps variable names (ex: x5, is_num, etc).
```{r}

keep_varnames = function(x, group, group_flag, ...){
  is_ok = grepl("^[[:alpha:].][[:alnum:]._]*$", x)
  
  if(group_flag != 0){
    group = group[is_ok]
    # recreating the index
    group = unclass(as.factor(group))
  } 
  
  res = x[is_ok]
  # we add the group in an attribute (this is the way)
  attr(res, "group") = group
  
  return(res)
}

string_magic_register_fun(keep_varnames, "keepvar")

expr = c("x1 + 52", "73 %% 5 == x", "y[y > .z_5]")
string_magic("All vars: {'[^[:alnum:]_.]+'split, keepvar, unik, enum.bq ? expr}.")

# thanks to the group flag, we can apply group-wise operations
# we apply cat after the function (using .post) to have a nice display of the newlines 
string_magic("Vars in each expr:\n",
             "{'\n'c ! - {1:3}) {'[^[:alnum:]_.]+'split, ",
                                 "keepvar, ~(unik, enum.bq) ? expr}}", .post = cat)
```


# Using `stringmagic` with custom operations as a dependency {#sec_package}

This is section is only relevant if: 

1. you use `stringmagic` as a dependency in your package, and 
2. you use custom `stringmagic` operations. 

If you answer "no" to any of these 
two points, do not read this section, it's only about details. *Otherwise it's a must read.*

To use custom `stringmagic` operations within a package, you need to explicitly register
the operations in a specific namespace, and, when using the `stringmagic` functions, 
you need to add the argument `.namespace` telling where the new operations are located.

## Why do I need a namespace?

There are two reasons: i) to ensure compatibility with future versions of the `stringmagic` package,
and ii) to avoid conflicts with user-created operations. 
Let's take an example. You create a package an use `stringmagic` with the custom operation
`h1`, detailed [here](#sec_reg_ops), to create headers. Now let's say a future version of 
`stringmagic` also introduces a `h1` function, that works differently from yours. This means
that your package will work with the old `stringmagic` version but will lead to a bug with 
the new version. Not great!

Same story if the user defines her own `h1` operation: we end up with two operations with the 
same name, hence a conflict. 

We need a mechanism to ensure that the `h1` operation in
your package always work, irrespecive of the doings of the user and of the version of `stringmagic`.
We now detail how to do it.

## Using custom operations in a package

To use custom operations in a package:

1. add the argument `.namespace = "my_package_name"` to the calls to `string_magic_register_ops` 
and `string_magic_register_fun`
2. add the argument `.namespace = "my_package_name"` to any call to `stringmagic`'s functions using 
custom operation. Achive this simply by creating aliases to `stringmagic` function.

### Example

You develop the package `superpack` which uses `stringmagic` to display messages to the user 
and you want to register the `header` operation (behaving [similarly to `h1`](#sec_reg_ops)).
```{r}
string_magic_register_ops("'- | 'paste, '70|-'fill", 
                          alias = "header", 
                          namespace = "superpack")
```

You can now summon your new operation to write a message to the user. Just remenber that 
you need the `.namespace` argument:

```{r}
time = 0.7
cat_magic("{header!Important message to you, user}",
          "The algorithm converged in {time}s.", 
          .sep = "\n", .namespace = "superpack")
```

Without the namespace argument, this leads to an error:
```{r, error = TRUE}
time = 0.7
cat_magic("{header!Important message to you, user}",
          "The algorithm converged in {time}s.", 
          .sep = "\n")
```

Using the `.namsepace` argument is cumbersome. That is why `stringmagic` offers alias generators to
easily create aliases of the `stringmagic` functions using custom operations.

### Using aliases

To avoid providing the argument `.namespace` at each call (which makes the function completely useless),
use the alias genetaors, as described in [the first section](#sec_alias). Let's continue on the
previous example but this time we avoid the argument `.namespace` by creating an alias.

To create the alias, we use `cat_magic_alias` which creates a copy of `cat_magic` for which
the default values have been modified. Here we override the function name (of course you can use 
any other name, like `cmagic`, `magcat`, whatever!):
```{r}
cat_magic = stringmagic::cat_magic_alias(.namespace = "superpack")
```

An now we are able to access our previously defined function without error:
```{r, error = TRUE}
time = 0.7
cat_magic("{header!Important message to you, user}",
          "The algorithm converged in {time}s.", 
          .sep = "\n")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_magic_main.R
\name{string_magic_register_fun}
\alias{string_magic_register_fun}
\alias{string_magic_register_ops}
\title{Register custom operations to apply them in \code{string_magic}}
\usage{
string_magic_register_fun(fun, alias, valid_options = NULL, namespace = NULL)

string_magic_register_ops(ops, alias, namespace = NULL)
}
\arguments{
\item{fun}{A function which must have at least the arguments 'x' and '...'.
Additionnaly, it can have the arguments: 'argument', 'options', 'group', 'group_flag'.
This function must return a vector.
This function will be internally called by \code{string_magic} in the form
\code{fun(x, argument, options, group, group_flag)}.\code{x}: the value to which the
operation applies. \code{argument}: the quoted \code{string_magic} argument (always character).
\code{options}: a character vector of \code{string_magic} options. The two last arguments are of use
only in group-wise operations if \code{fun} changes the lengths of vectors. \code{group}: an index of
the group to which belongs each observation (integer). \code{group_flag}: value between 0
and 2; 0: no grouping operation requested; 1: keep track of groups; 2: apply grouping.}

\item{alias}{Character scalar, the name of the operation.}

\item{valid_options}{A character vector or NULL (default). Represents a list of
valid options for the operation. This is used: a) to enable auto-completion,
b) for error-handling purposes.}

\item{namespace}{Character scalar or \code{NULL} (default). \strong{Only useful for package developers.}
As a regular end-user you shouldn't care! If your package uses \code{string_magic}, you should care.
If the function \verb{string_magic_register_*} is located in the \code{onLoad} function (see \code{help("onLoad")}),
there is nothing to do. Otherwise, pass the name of your package in this argument to
make all the new operation definitions scoped (i.e. only your package can access it and
it can't be messed up by end users).}

\item{ops}{Character scalar representing a valid chain of \code{string_magic} operations. It should
be of the form \code{"op1, 'arg'op2, etc"}. For example \code{"'80|-'fill"} fills the line
with dashes up to 80 characters.}
}
\value{
These function do not return anything. They register new operations to be used in the
\code{string_magic} family of functions by placing them in the options (later fetched by
\code{string_magic()} at run-time).
}
\description{
Extends the capabilities of \code{\link[=string_magic]{string_magic()}} by adding any custom operation
}
\details{
We try to strongly check the new operations since it's always better to find out problems
sooner than later. This means that when the function is defined, it is also
tested.

If you pass a function, note that it should work for non-character arguments in \code{x}.
}
\section{Functions}{
\itemize{
\item \code{string_magic_register_ops()}: Create new combinations of \code{string_magic} operations

}}
\section{Writing a package using \code{string_magic}}{


If you want to use \code{string_magic} in your package and want to make use of custom operations:
\itemize{
\item place any \code{string_magic_register_fun} and \code{string_magic_register_ops} in your \code{.onLoad} function
(see \code{help("onLoad")}). The .onLoad function is run whenever the package is loaded
for the first time. It's a function that you can place anywhere in your \verb{R/*} files
and which looks like this:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{.onLoad = function(libname, pkgname)\{
  # string_magic custom operations
  string_magic_register_ops("'80|-'fill", "h1")

  invisible()
\}
}\if{html}{\out{</div>}}
\itemize{
\item if you don't want to place the \verb{string_magic_register_*} functions in the .onLoad function,
you can, but then you \strong{must} provide the argument \code{namespace}:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{string_magic_register_ops("'80|-'fill", "h1", namespace = "myPackageName")
}\if{html}{\out{</div>}}
\itemize{
\item you must create an \code{\link[=string_magic_alias]{string_magic_alias()}} to create an alias to \code{\link[=string_magic]{string_magic()}} and use the
argument \code{.namespace = "myPackageName"}. Use this opportunity to change the
defaults if you wish. You can even override the \code{string_magic} function:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# creating an alias with the same name + changing the delimiter
string_magic = stringmagic::string_magic_alias(.namespace = "myPackageName", .delim = "\{\{ \}\}")
}\if{html}{\out{</div>}}
}

\examples{

# let's create an operation that adds markdown emphasis
# to elements of a vector

# A) define the function
fun_emph = function(x, ...) paste0("*", x, "*")

# B) register it
string_magic_register_fun(fun_emph, "emph")

# C) use it
x = string_vec("right, now")
string_magic("Take heed, {emph, c? x}.")

#
# now let's add the option "strong"
fun_emph = function(x, options, ...) {
  if("strong" \%in\% options){
    paste0("***", x, "***")
  } else {
    paste0("*", x, "*")
  }
}

string_magic_register_fun(fun_emph, "emph", "strong")

x = string_vec("right, now")
string_magic("Take heed, {emph.strong, c? x}.")

#
# now let's add an argument
fun_emph = function(x, argument, options, ...){
  arg = argument
  if(nchar(arg) == 0) arg = "*"
  
  if("strong" \%in\% options){
    arg = paste0(rep(arg, 3), collapse = "")
  }
  
  paste0(arg, x, arg)
}

string_magic_register_fun(fun_emph, "emph", "strong")

x = string_vec("right, now")
string_magic("Take heed, {'_'emph.s, c? x}.")

#
# using string_magic_register_ops
#

# create a 'header' maker
string_magic_register_ops("tws, '# 'paste, ' 'paste.right, '40|-'fill", "h1")
cat_magic("{h1 ! My title}\n my content")




}
\seealso{
Other related to string_magic: 
\code{\link{string_magic_alias}()}
}
\author{
Laurent R. Berge
}
\concept{related to string_magic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_tools.R
\name{paste_conditional}
\alias{paste_conditional}
\title{Paste a string vector conditionally}
\usage{
paste_conditional(x, id, sep = " ", names = TRUE, sort = TRUE)
}
\arguments{
\item{x}{A character vector or a formula. If a vector: it represents the
values to be pasted together.
If a formula, it must be of the form
\code{my_string ~ id1 + id2} with on the left the character vector and on the right
the (possibly many) identifiers. If a formula, the argument \code{id} must be a
data frame.}

\item{id}{A vector of identifiers, a list of identifiers (can be a data frame),
or a \code{data.frame}. The identifiers can be a vector or a list of vectors. They represent
which elements of \code{x} are to be pasted together.

When \code{x} is a formula, then \code{id} must be a data.frame containing the variables in
the formula.}

\item{sep}{A character scalar, default is \code{" "}. The value used to paste together
the elements of \code{x}.}

\item{names}{Logical scalar, default is \code{TRUE}. Whether to add, as names, the values of
the identifiers.}

\item{sort}{Logical scalar, default is \code{TRUE}. Whether to sort the results according to
the identifiers.}
}
\value{
Returns a character vector. If the argument \code{names} is \code{TRUE} (default), the vector will have
names equal to the values of the identifiers.
}
\description{
Easily reconstruct a string vector that has been split with \code{\link[=string_split2df]{string_split2df()}}.
}
\examples{

#
# let's paste together the letters of the alphabet

# first we create the identifier
id = rep(1:2, each = 13)
setNames(id, letters)

# now we conditionally paste together the letters
paste_conditional(letters, id, "")

#
# using a formula

# we create a small data set based on mtcars
base_cars = within(mtcars, carname <- row.names(mtcars))
base_cars = head(base_cars, 10)

# we use two identifiers
paste_conditional(carname ~ gear + carb, base_cars, sep = ", ")



}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aliases.R
\name{string_magic_alias}
\alias{string_magic_alias}
\title{Create \code{string_magic} aliases with custom defaults}
\usage{
string_magic_alias(
  .sep = "",
  .vectorize = FALSE,
  .delim = c("{", "}"),
  .last = NULL,
  .post = NULL,
  .default = NULL,
  .nest = FALSE,
  .invisible = FALSE,
  .local_ops = NULL,
  .collapse = NULL,
  .check = TRUE,
  .class = NULL,
  .namespace = NULL
)
}
\arguments{
\item{.sep}{Character scalar, default is the empty string \code{""}. It is used to collapse all
the elements in \code{...} before applying any operation.}

\item{.vectorize}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, Further, elements in \code{...} are
NOT collapsed together, but instead vectorised.}

\item{.delim}{Character vector of length 1 or 2. Default is \code{c("{", "}")}. Defines
the opening and the closing delimiters for interpolation.

If of length 1, it must be of the form: 1) the opening delimiter,
2) a single space, 3) the closing delimiter. Ex: \code{".[ ]"} is equivalent to \code{c(".[", "]")}.
The default value is equivalent to \code{"{ }"}.

[ ]: R:\%20
[", "]: R:\%22,\%20\%22}

\item{.last}{Character scalar, a function, or \code{NULL} (default). If provided and character:
it must be an \code{string_magic} chain of operations of the form \code{"'arg1'op1, op2, etc"}. All these operations
are applied just before returning the vector. If a function,
it will be applied to the resulting vector.}

\item{.post}{Function or \code{NULL} (default). If not \code{NULL}, this function will be applied
after all the processing, just before returning the object. This function can have
extra arguments which will be caught directly in the \code{...} argument of \code{string_magic}.
For example if \code{.post = head}, you can directly pass the argument \code{n = 3} to \code{string_magic}'s arguments.}

\item{.default}{Character scalar or \code{NULL} (default). If provided, it must be
a sequence of \code{string_magic} operations. It will be applied as a default to any interpolation.
Ex: if \code{x = 1:2}, then \code{string_magic("x = {x}", .default = "enum")} leads to "x = 1 and 2",
and is equivalent to \code{string_magic("x = {enum?x}")}. Note that this default operations does
not apply to nested expressions. That is \code{string_magic("{!x{1:2}}", .default = "enum")} leads
to \code{c("x1", "x2")} and NOT \code{"x1 and 2"}.}

\item{.nest}{Logical, default is \code{FALSE}. If \code{TRUE}, it will nest the original string within
interpolation delimiters, so that you can apply operations directly on the string. Example:
\code{string_magic("upper ! hello")} returns "upper ! hello", while \code{string_magic("upper ! hello", .nest = TRUE)}
returns \code{"HELLO"}.}

\item{.invisible}{Logical scalar, default is \code{FALSE}. Whether the object returned should be
invisible (i.e. not printed on the console).}

\item{.local_ops}{Named list or \code{NULL} (default). If provided, it must be a list
of the form \code{list(alias1 = ops1, alias2 = ops2)} where alias is the name of the newly defined
operator an \code{ops} is a character scalar representing the associated string_magic operations.
Ex: \code{list(add = "' + 'collapse")} creates the operation \code{add} which collapses the
string with pluses. All operations created here are only available to the
generated function.}

\item{.collapse}{Character scalar, default is \code{NULL}. If provided, the character vector
that should be returned is collapsed with the value of this argument. This leads
to return a string of length 1.}

\item{.check}{Logical scalar, default is \code{TRUE}. Whether to enable error-handling (i.e.
human readable error messages).
Without error-handling you can save something of the order of 40us.Useful only
in long loops.}

\item{.class}{Character vector representing the class to give to the object returned.
By default it is \code{NULL}. Note that the class \code{string_magic} has a specific \code{print} method, usually
nicer for small vectors (it \code{\link[base:cat]{base::cat()}}s the elements).}

\item{.namespace}{Character scalar or \code{NULL} (default). \strong{Only useful for package developers.}
As a regular end-user you shouldn't care! If your package uses \code{string_magic}, you should care.
It is useful \strong{only} if your package uses 'custom' \code{string_magic} operations, set with
\code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or \code{\link[=string_magic_register_ops]{string_magic_register_ops()}}.}
}
\value{
This function returns a function which will behave in the same way as \code{\link[=string_magic]{string_magic()}}
}
\description{
Utility to easily create \code{string_magic} aliases with custom default
}
\details{
Use this function if you want to change \code{string_magic} default values. For example,
if you want the interpolation to be done with \code{"{{}}"} (instead of \code{{}}) or if you want the
default separation to be the space (instead of the empty string). See the example.
}
\section{Writing a package using \code{string_magic}}{


If you want to use \code{string_magic} in your package and want to make use of custom operations:
\itemize{
\item place any \code{string_magic_register_fun} and \code{string_magic_register_ops} in your \code{.onLoad} function
(see \code{help("onLoad")}). The .onLoad function is run whenever the package is loaded
for the first time. It's a function that you can place anywhere in your \verb{R/*} files
and which looks like this:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{.onLoad = function(libname, pkgname)\{
  # string_magic custom operations
  string_magic_register_ops("'80|-'fill", "h1")

  invisible()
\}
}\if{html}{\out{</div>}}
\itemize{
\item if you don't want to place the \verb{string_magic_register_*} functions in the .onLoad function,
you can, but then you \strong{must} provide the argument \code{namespace}:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{string_magic_register_ops("'80|-'fill", "h1", namespace = "myPackageName")
}\if{html}{\out{</div>}}
\itemize{
\item you must create an \code{\link[=string_magic_alias]{string_magic_alias()}} to create an alias to \code{\link[=string_magic]{string_magic()}} and use the
argument \code{.namespace = "myPackageName"}. Use this opportunity to change the
defaults if you wish. You can even override the \code{string_magic} function:
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# creating an alias with the same name + changing the delimiter
string_magic = stringmagic::string_magic_alias(.namespace = "myPackageName", .delim = "\{\{ \}\}")
}\if{html}{\out{</div>}}
}

\examples{

# we create the function sma2 with different defaults
sma2 = string_magic_alias(.delim = ".[ ]", .sep = " ", .class = "string_magic")

person = "john doe"
sma2("Hello", ".[title ? person]")

# you can use the arguments whose default has been changed
sma2("Hello", ".[title ? person]", .sep = ": ")



}
\seealso{
Other related to string_magic: 
\code{\link{string_magic_register_fun}()}

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_clean_alias}()},
\code{\link{string_magic}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent Berge
}
\concept{related to string_magic}
\concept{tools with aliases}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aliases.R, R/string_tools.R
\name{string_vec_alias}
\alias{string_vec_alias}
\alias{string_vec}
\alias{stvec}
\title{Efficient creation of string vectors with optional interpolation}
\usage{
string_vec_alias(
  .cmat = FALSE,
  .nmat = FALSE,
  .df = FALSE,
  .df.convert = TRUE,
  .last = NULL,
  .delim = c("{", "}"),
  .split = TRUE,
  .protect.vars = FALSE,
  .check = TRUE,
  .sep = NULL,
  .collapse = NULL,
  .namespace = NULL
)

string_vec(
  ...,
  .cmat = FALSE,
  .nmat = FALSE,
  .df = FALSE,
  .df.convert = TRUE,
  .delim = c("{", "}"),
  .data = list(),
  .envir = parent.frame(),
  .split = TRUE,
  .protect.vars = FALSE,
  .sep = NULL,
  .last = NULL,
  .check = TRUE,
  .help = NULL,
  .collapse = NULL,
  .namespace = NULL
)

stvec(
  ...,
  .cmat = FALSE,
  .nmat = FALSE,
  .df = FALSE,
  .df.convert = TRUE,
  .delim = c("{", "}"),
  .data = list(),
  .envir = parent.frame(),
  .split = TRUE,
  .protect.vars = FALSE,
  .sep = NULL,
  .last = NULL,
  .check = TRUE,
  .help = NULL,
  .collapse = NULL,
  .namespace = NULL
)
}
\arguments{
\item{.cmat}{Logical scalar (default is \code{FALSE}), integer or complex with integer values.
If \code{TRUE}, we try to coerce the result into a \strong{character} matrix. The number of rows and columns
is deduced from the look of the arguments. An integer indicates the number of rows.
If a complex, the imaginary part represents the number of columns. Ex: \code{5i} means
5 columns, \code{3 + 2i} means 3 rows and 2 columns.

The matrix is always filled by row.}

\item{.nmat}{Logical scalar (default is \code{FALSE}), integer or complex with integer values.
If \code{TRUE}, we try to coerce the result into a \strong{numeric} matrix. The number of rows and columns
is deduced from the look of the arguments. An integer indicates the number of rows.
If a complex, the imaginary part represents the number of columns. Ex: \code{5i} means
5 columns, \code{3 + 2i} means 3 rows and 2 columns.

The matrix is always filled by row. Non numbers are silently turned to NA.}

\item{.df}{Logical scalar (default is \code{FALSE}), integer, complex with integer values, or
character vector.
If \code{TRUE}, we try to coerce the result into a \code{data.frame}. The number of rows and columns
is deduced from the look of the arguments. An integer indicates the number of rows.
If a complex, the imaginary part represents the number of columns. Ex: \code{5i} means
5 columns, \code{3 + 2i} means 3 rows and 2 columns.

If a character vector: it should give the column names. Note that if the vector
is of length 1, its values are comma separated (i.e. the value \code{"id, name"} is
identical to \code{c("id", "name")}).

Note that the columns that can be converted to numeric are converted to numeric.
The other columns are in string form. Monitor this behavior with \code{.df.convert}.}

\item{.df.convert}{Logical scalar, default is \code{TRUE}. Only used when the result is
to be converted to a data frame (with the argument \code{.df}). If \code{TRUE}, any column
looking like a numeric vector is converted to numeric. Otherwise all columns
are character strings.}

\item{.last}{Character scalar, a function, or \code{NULL} (default). If provided and character:
it must be an \code{string_magic} chain of operations of the form \code{"'arg1'op1, op2, etc"}. All these operations
are applied just before returning the vector. If a function,
it will be applied to the resulting vector.}

\item{.delim}{Character vector of length 1 or 2. Default is \code{c("{", "}")}. Defines
the opening and the closing delimiters for interpolation.

If of length 1, it must be of the form: 1) the opening delimiter,
2) a single space, 3) the closing delimiter. Ex: \code{".[ ]"} is equivalent to \code{c(".[", "]")}.
The default value is equivalent to \code{"{ }"}.

[ ]: R:\%20
[", "]: R:\%22,\%20\%22}

\item{.split}{Logical or a character symbol, default is \code{TRUE}. If \code{TRUE},
the character vectors are split with respect to commas (default). If \code{FALSE}, no
splitting is performed. If a character symbol, the string vector will be split
according to this symbol. Note that any space after the symbol (including tabs and
newlines) is discarded. You can escape the splitting symbol with a backslash right before it.

Ex: by default \code{string_vec("hi, there")} leads to the vector \code{c("hi", "there")}.}

\item{.protect.vars}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then only
arguments equal to a "natural" character scalar are comma-split and interpolated,
other arguments are not touched. Ex: \code{string_vec("x{1:5}")} will lead to a vector of
length 5 ("x1" to "x5"), while \code{z = "x{1:5}"} followed by \code{string_vec(z)} leads
to a vector of length 1: \code{"x{1:5}"}. If \code{FALSE}, comma-splitting and interpolation
is performed on all variables.}

\item{.check}{Logical scalar, default is \code{TRUE}. Whether to enable error-handling (i.e.
human readable error messages).
Without error-handling you can save something of the order of 40us.Useful only
in long loops.}

\item{.sep}{Character scalar or \code{NULL} (default). If not \code{NULL}, the function
\code{\link[base:paste]{base::paste()}} is applied to the resulting vector with \code{sep = .sep}.}

\item{.collapse}{Character scalar or \code{NULL} (default). If not \code{NULL}, the function
\code{\link[base:paste]{base::paste()}} is applied to the resulting vector with \code{collapse = .collapse}.

If so, pass the name of your package in this argument so that your function can access
the new \code{string_magic} operations defined within your package.}

\item{.namespace}{Character scalar or \code{NULL} (default). \strong{Only useful for package developers.}
As a regular end-user you shouldn't care! If your package uses \code{string_magic}, you should care.
It is useful \strong{only} if your package uses 'custom' \code{string_magic} operations, set with
\code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or \code{\link[=string_magic_register_ops]{string_magic_register_ops()}}.}

\item{...}{Character vectors that will be vectorized. If commas are present in the
character vector, it will be split with respect to commas and following blanks.
The vectors can contain any interpolation in the form \code{"{var}"} and
any \code{\link[=string_magic]{string_magic()}} operation can be applied. To change the delimiters for interpolation,
see \code{.delim}. Named arguments are used in priority for variable substitution,
otherwise the value of the variables to be interpolated are fetched in the calling environment
(see argument \code{.envir}).}

\item{.data}{A list used to evaluate the variables in \code{"{}"}. Default is the empty list.
By default the variables are evaluated using the environment from where the
function is called or using the named arguments passed to the function.}

\item{.envir}{An environment used to evaluate the variables in \code{"{}"}. By default the variables are
evaluated using the environment from where the function is called or using the named
arguments passed to the function.}

\item{.help}{Character scalar or \code{TRUE}, default is \code{NULL}. This argument
is used to generate a dynamic help on the console. If \code{TRUE}, the user can select which
topic to read from the main documentation, with the possibility to search for keywords and
navigate the help pages. If a character scalar, then a regex search is perfomed on the main
documentation and any section containining a match is displayed. The user can easily
navigate across matches.}
}
\value{
By default this function returns a string vector, the length of which depends on the arguments.

This result can be processed with the arguments \code{cmat}, \code{nmat} and \code{.df} which will
try to coerce the result into a character matrix, a numeric matrix , or a data frame, respectively.
}
\description{
Create string vectors in multiple ways: 1) add successive string elements (like in \code{c()}),
or 2) write a character string that will be broken with respect to commas
(\code{"hi, there"} becomes \code{c("hi", "there")}), or 3) interpolate variables in
character strings (\code{"x{1:2}"} becomes \code{c("x1", "x2")}) with full access to
\code{\link[=string_magic]{string_magic()}} operations, or any combination of the three.
}
\details{
The main objective of this function is to simplify the creation of small character vectors.
By default, you can pass a character string of length 1 with values separated with commas
and a character vector will be returned.

You can use interpolation using curly brackets (see \code{string_magic()}). You can pass
values for the interpolation directly in the arguments (this is why all
arguments start with a dot).

By default character values containing commas are split with respect to the commas
to create vectors. To change this behavior, see the argument \code{.split}.

The default of the argument \code{.protect.vars} is \code{FALSE}. To avoid unwanted
comma-splitting and interpolations on variables, set it to \code{TRUE}.
The main use case of this function is
the creation of small string vectors, which can be written directly at
function call.

Customize the default of this function with \code{\link[=string_vec_alias]{string_vec_alias()}}.
}
\section{Functions}{
\itemize{
\item \code{string_vec_alias()}: Create \code{string_vec} aliases with custom defaults

\item \code{stvec()}: Alias to \code{string_vec}

}}
\examples{

# illustrating comma-splitting and interpolation
string_vec("x1, y2, z{4:5}")

# variable protection
x = "x{1:5}"
# without protection (default) => interpolation takes place
string_vec(x, "y{1:2}")

# with protection => no interpolation for x
string_vec(x, "y{1:2}", .protect.vars = TRUE)

# removing comma splitting
string_vec("Hi, said Charles.", "Hi, said {girl}.", girl = "Julia", .split = FALSE)

# changing the delimiters for interpolation
pkg = "\\\\usepackage[usenames,dvipsnames]{xcolor}"
string_vec("\\\\usepackage{.[S!graphicx, fourier, standalone]}", 
           pkg, .delim = ".[ ]")

#
# Customization
#

# Unhappy about the defaults? Create an alias!

# we create a "matrix generator"
matgen = string_vec_alias(.nmat = TRUE, .last = "'\n'split")

matgen("5, 4, 3
        8, 6, 2")

}
\seealso{
Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_clean_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()}
}
\author{
Laurent Berge
}
\concept{tools with aliases}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_magic_doc.R, R/string_magic_main.R
\name{string_magic}
\alias{string_magic}
\alias{.string_magic}
\alias{sma}
\title{String interpolation with operation chaining}
\usage{
string_magic(
  ...,
  .envir = parent.frame(),
  .data = list(),
  .sep = "",
  .vectorize = FALSE,
  .delim = c("{", "}"),
  .last = NULL,
  .post = NULL,
  .nest = FALSE,
  .collapse = NULL,
  .invisible = FALSE,
  .default = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .class = NULL,
  .help = NULL,
  .namespace = NULL
)

.string_magic(
  ...,
  .envir = parent.frame(),
  .data = list(),
  .sep = "",
  .vectorize = FALSE,
  .delim = c("{", "}"),
  .collapse = NULL,
  .last = NULL,
  .nest = FALSE,
  .trigger = TRUE,
  .namespace = NULL
)

sma(
  ...,
  .envir = parent.frame(),
  .data = list(),
  .sep = "",
  .vectorize = FALSE,
  .delim = c("{", "}"),
  .last = NULL,
  .post = NULL,
  .nest = FALSE,
  .collapse = NULL,
  .invisible = FALSE,
  .default = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .class = NULL,
  .help = NULL,
  .namespace = NULL
)
}
\arguments{
\item{...}{Character scalars that will be collapsed with the argument \code{sep}. Note that
named arguments are used for substitution.

To interpolate, you can
use \code{"{x}"} within each character string to insert the value of \code{x} in the string.
You can add string operations in each \code{"{}"} instance with the syntax \code{"'arg'op ? x"}
(resp. \code{"'arg'op ! x"}) to apply the operation \code{'op'} with the argument \code{'arg'} to \code{x}
(resp. the verbatim of \code{x}). Otherwise, what to say? Ah, nesting is enabled, and since
there's over 50 operators, it's a bit complicated to sort you out in this small space.

Use the argument \code{.help = "keyword"} (or \code{.help = TRUE}) to obtain a selective help
from the main documentation.

Note that in interpolations you have access to the special variables: \code{.now} and \code{.date}
to get the current time; and the special function \code{.now("format")} to
format the time. Ex: \code{.now('\%Y-\%m \%H:\%m')}.}

\item{.envir}{An environment used to evaluate the variables in \code{"{}"}. By default the variables are
evaluated using the environment from where the function is called or using the named
arguments passed to the function.}

\item{.data}{A list used to evaluate the variables in \code{"{}"}. Default is the empty list.
By default the variables are evaluated using the environment from where the
function is called or using the named arguments passed to the function.}

\item{.sep}{Character scalar, default is the empty string \code{""}. It is used to collapse all
the elements in \code{...} before applying any operation.}

\item{.vectorize}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, Further, elements in \code{...} are
NOT collapsed together, but instead vectorised.}

\item{.delim}{Character vector of length 1 or 2. Default is \code{c("{", "}")}. Defines
the opening and the closing delimiters for interpolation.

If of length 1, it must be of the form: 1) the opening delimiter,
2) a single space, 3) the closing delimiter. Ex: \code{".[ ]"} is equivalent to \code{c(".[", "]")}.
The default value is equivalent to \code{"{ }"}.

[ ]: R:\%20
[", "]: R:\%22,\%20\%22}

\item{.last}{Character scalar, a function, or \code{NULL} (default). If provided and character:
it must be an \code{string_magic} chain of operations of the form \code{"'arg1'op1, op2, etc"}. All these operations
are applied just before returning the vector. If a function,
it will be applied to the resulting vector.}

\item{.post}{Function or \code{NULL} (default). If not \code{NULL}, this function will be applied
after all the processing, just before returning the object. This function can have
extra arguments which will be caught directly in the \code{...} argument of \code{string_magic}.
For example if \code{.post = head}, you can directly pass the argument \code{n = 3} to \code{string_magic}'s arguments.}

\item{.nest}{Logical, default is \code{FALSE}. If \code{TRUE}, it will nest the original string within
interpolation delimiters, so that you can apply operations directly on the string. Example:
\code{string_magic("upper ! hello")} returns "upper ! hello", while \code{string_magic("upper ! hello", .nest = TRUE)}
returns \code{"HELLO"}.}

\item{.collapse}{Character scalar, default is \code{NULL}. If provided, the character vector
that should be returned is collapsed with the value of this argument. This leads
to return a string of length 1.}

\item{.invisible}{Logical scalar, default is \code{FALSE}. Whether the object returned should be
invisible (i.e. not printed on the console).}

\item{.default}{Character scalar or \code{NULL} (default). If provided, it must be
a sequence of \code{string_magic} operations. It will be applied as a default to any interpolation.
Ex: if \code{x = 1:2}, then \code{string_magic("x = {x}", .default = "enum")} leads to "x = 1 and 2",
and is equivalent to \code{string_magic("x = {enum?x}")}. Note that this default operations does
not apply to nested expressions. That is \code{string_magic("{!x{1:2}}", .default = "enum")} leads
to \code{c("x1", "x2")} and NOT \code{"x1 and 2"}.}

\item{.trigger}{Logical, default is \code{TRUE}. If \code{FALSE}, this function is not run. Can be
useful in debugging situations where we want conditional evaluations.}

\item{.check}{Logical scalar, default is \code{TRUE}. Whether to enable error-handling (i.e.
human readable error messages).
Without error-handling you can save something of the order of 40us.Useful only
in long loops.}

\item{.class}{Character vector representing the class to give to the object returned.
By default it is \code{NULL}. Note that the class \code{string_magic} has a specific \code{print} method, usually
nicer for small vectors (it \code{\link[base:cat]{base::cat()}}s the elements).}

\item{.help}{Character scalar or \code{TRUE}, default is \code{NULL}. This argument
is used to generate a dynamic help on the console. If \code{TRUE}, the user can select which
topic to read from the main documentation, with the possibility to search for keywords and
navigate the help pages. If a character scalar, then a regex search is perfomed on the main
documentation and any section containining a match is displayed. The user can easily
navigate across matches.}

\item{.namespace}{Character scalar or \code{NULL} (default). \strong{Only useful for package developers.}
As a regular end-user you shouldn't care! If your package uses \code{string_magic}, you should care.
It is useful \strong{only} if your package uses 'custom' \code{string_magic} operations, set with
\code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or \code{\link[=string_magic_register_ops]{string_magic_register_ops()}}.}
}
\value{
It returns a character vector whose length depends on the elements and operations in the interpolations.
}
\description{
This is firstly a string interpolation tool. On top of this it can apply, and chain, over 50
basic string operations to the interpolated variables. Advanced support for pluralization.
}
\details{
There are over 50 basic string operations, it supports pluralization, string operations can be
nested, operations can be applied group-wise or conditionally and
operators have sensible defaults.
You can also declare your own operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or \code{\link[=string_magic_register_ops]{string_magic_register_ops()}}.
They will be seamlessly integrated to \code{string_magic}.

The function \code{.string_magic} (prefixed with a dot) is a leaner version of the function \code{string_magic}.
It does the same operations but with the following differences:
\itemize{
\item there is no error handling: meaning that the error messages, if any, will be poor and
hard to understand
\item default options are not applied: hence the user must always explicitly provide the arguments
}

This leads to a faster processing time (of about 50 microseconds) at the cost of user experience.

If you want to change the default values of \code{string_magic} (like changing the delimiter), use
the function \code{\link[=string_magic_alias]{string_magic_alias()}}.

Use the argument \code{.help} to which
you can pass keywords or regular expressions and fecth select pieces from the main documentation.
}
\section{Functions}{
\itemize{
\item \code{string_magic()}: String interpolation with operation chaining

\item \code{.string_magic()}: A simpler version of \code{string_magic} without any error handling to save a few micro seconds

\item \code{sma()}: Alias to \code{string_magic}

}}
\section{Interpolation and string operations}{
 Principle:

To interpolate a variable, say \code{x}, simply use \code{{x}}. For example \verb{x = "world"; string_magic("hello \{x\}")} leads
to "hello world".

To any interpolation you can add operations. Taking the previous example, say we want to display
"hello W O R L D". This means upper casing all letters of the interpolated variable and adding a space between
each of them. Do you think we can do that? Of course yes:
\code{string_magic("hello {upper, ''s, c ? x}")}. And that's it.

Now let's explain what happened. Within the \code{{}} \emph{box}, we first write a set of
operations, here "upper, ''s, c", then add "?" and finally write
the variable to interpolate, "x".  The operations (explained in more details
below) are \code{upper}, upper-casing all letters, ''s: splitting
with the empty string, 'c': concatenating with spaces the vector of string that was just split.
The question mark means that the expression coming after it is to be evaluated
(this is opposed to the exclamation mark presented next).

The syntax is always the same: \code{{operations ? expression}}, where the operations section
is a \emph{comma separated} list of operations.
These operations are of the form \verb{'arg'op}, with \code{arg} the argument to the operator
code \code{op}. These operations are performed sequantially from left to right.

Some operations, like \code{upper}, accept options. You attach options to an operation
with a dot followed by the option name. Formally: \code{op.option1.option2}, etc.
Example: \verb{x = "hi there. what's up? fine." ; string_magic("He said: \{upper.sentence, Q ? x\}")}.
Leads to: \verb{He said: "Hi there. What's up? Fine."}.

Both operators and options are partially matched. So \code{string_magic("He said: {up.s, Q ? x}")} would
also work.
}

\section{Verbatim interpolation and nesting}{
 Principle:

Instead of interpolating a variable, say \code{x}, with \code{{x}}, you can use an exclamation
mark to trigger varbatim evaluation.
For example \code{string_magic("hello {!x}")} would lead to "hello x". It's a
bit disappointing, right? What's the point of doing that? Wait until the next two paragraphs.

Verbatim evaluation is a powerful way to apply operations to plain text. For example:
\code{string_magic("hello {upper, ''s, c ! world}")} leads to "hello W O R L D".

(A note in passing. The spaces surrounding the exclamation mark are non necessary,
but when one space is present on both sides of the \code{!}, then the verbatim
expression only begins after it. Ex: \code{"{upper! hi}"} leads to " HI" while \code{"{upper ! hi}"}
leads to "HI" and \code{"{upper !  hi}"} leads to " HI".)

The second advantage of verbatim evaluations is \emph{nesting}. Anything in a verbatim
expression is evaluated with the function \code{string_magic}.
This means that any \emph{box} will be evaluated as previously described. Let's
give an example. You want to write the expression of a polynomial of order n: a + bx + cx^2 + etc.
You can do that with nesting. Assume we have \code{n = 2}.

Then \code{string_magic("poly({n}): {' + 'c ! {letters[1 + 0:n]}x^{0:n}}")} leads to
"poly(2): ax^0 + bx^1 + cx^2".

How does it work? The verbatim expression (the one following the exclamation mark),
here \code{"{letters[1 + 0:n]}x^{0:n}"}, is evaluated with \code{string_magic}.
\code{string_magic("{letters[1 + 0:n]}x^{0:n}")} leads to the vector c("ax^0", "bx^1", "cx^2").

The operation \verb{' + 'c} then concatenates (or collapses) that vector with ' + '.
This value is then appended to the previous string.

We could refine by adding a cleaning operation in which we replace "x^0" and "^1"
by the empty string. Let's do it:

\code{string_magic("poly({n}): {' + 'c, 'x\\\\^0|\\\\^1'r ! {letters[1 + 0:n]}x^{0:n}}")} leads to
"poly(2): a + bx + cx^2", what we wanted.

You can try to write a function to express the polynomial as before: although it is
a simple task, my guess is that it will require more typing.
}

\section{Operations}{
 General syntax:

As seen in the previous sections, within a \emph{box} (i.e. \code{"{}"}), multiple operations
can be performed.
We can do so by stacking the operations codes and in a comma separated enumeration.
Operations can have arguments, and operations can also have options. The general
syntax, with argument and options, is:

\verb{\{'arg1'op1.optionA.optionB, arg2 op2.optionC, }arg3\verb{op3, 51op4 ? x\}}

The argument can appear in four forms: a) inside single or double quotes just
before the operation name (\code{arg1} above),
b) verbatim, separated with a space, just before the operation name (\code{arg2} above),
c) inside bactick quotes the argument is evaluated from the environment (\code{arg3} above),
or d) when the argument is an integer it can be juxtaposed to the opeation name (like in \code{op4} above).

The options are always dot separated and attached to the operation name, they are
specific to each operation.

Both the operation name and the option names are partially matched.
}

\section{Basic string operations}{


This section describes some of the most common string operations: extracting, replacing, collapsing, splitting, etc.
These functions accept generic flags ("ignore", "fixed", "word") in their patterns (syntax: "flags/pattern").
Please see the dedicated section for more information on flags.
\itemize{
\item s, split, S, Split: splits the string according to a pattern.
The operations have different defaults: \code{' '}
for \code{s} and 'split', and \code{',[ \\t\\n]*'} for \code{S} and 'Split' (i.e. comma separation).
Ex.1: \code{string_magic("{S ! romeo, juliet}")} leads to the vector c("romeo", "juliet").
Ex.2: \code{string_magic("{'f/+'s, '-'c ! 5 + 2} = 3")} leads to "5 - 2 = 3" (note the flag "fixed" in \code{s}'s pattern).
\item c, C: to concatenate multiple strings into a single one. The two operations are
identical, only their default change. c: default is \code{' '}, C: default is \code{', | and '}.
The syntax of the argument is 's1' or 's1|s2'. s1 is the string used to concatenate
(think \code{paste(x, collapse = s1)}). In arguments of the form \code{'s1|s2'}, \code{s2} will be used to concatenate the last two elements.
Ex.1: \verb{x = 1:4; string_magic("Et \{' et 'c ? x\}!")} leads to "Et 1 et 2 et 3 et 4!".
Ex.2: \code{string_magic("Choose: {', | or 'c ? 2:4}?")} leads to "Choose: 2, 3 or 4?".
\item x, X: extracts patterns from a string. Both have the same default: \code{'[[:alnum:]]+'}.
\code{x} extracts the first match while \code{X} extracts \strong{all} the matches.
Ex.1: \verb{x = c("6 feet under", "mahogany") ; string_magic("\{'\\\\w\{3\}'x ? x\}")} leads to the vector c("fee", "mah").
Ex2.: \verb{x = c("6 feet under", "mahogany") ; string_magic("\{'\\\\w\{3\}'X ? x\}")} leads to the
vector c("fee", "und", "mah", "oga").
\item extract: extracts multiple patterns from a string, this is an alias to the operation \code{X} described above.
Use the option "first" to extract only the first match for each string (behavior becomes like \code{x}).
Ex: \verb{x = c("margo: 32, 1m75", "luke doe: 27, 1m71") ; string_magic("\{'^\\\\w+'extract ? x\} is \{'\\\\d+'extract.first ? x\}")}
leads to c("margo is 32", "luke is 27").
\item r, R: replacement within a string. The two operations are identical and have no default.
The syntax is \code{'old'} or \code{'old => new'} with \code{'old'} the pattern to find and \code{new} the replacement. If \code{new} is missing, it is
considered the empty string. This operation also accepts the flag "total" which instruct to
replace the fulll string in case the pattern is found.
Ex.1: \code{string_magic("{'e'r ! Where is the letter e?}")} leads to "Whr is th lttr ?".
Ex.2: \code{string_magic("{'(?<!\\\\b)e => a'R ! Where is the letter e?}")} leads to "Whara is tha lattar e?".
Ex.3: \code{string_magic("{'t/e => here'r ! Where is the letter e?}")} leads to "here".
\item clean: replacement with a string. Similar to the operation \code{r}, except that here the comma is
a pattern separator, see detailed explanations in \code{\link[=string_clean]{string_clean()}}. Ex: \code{string_magic("{'f/[, ]'clean ! x[a]}")}
leads to "xa".
\item get: restricts the string only to values respecting a pattern. This operation has no default.
Accepts the options "equal" and "in".
By default it uses the same syntax as \code{\link[=string_get]{string_get()}} so that you can use regex flags and
include logical operations with \code{' & '} and \code{' | '} to detect patterns.
If the option "equal" is used, a simple string equality with the argument is tested (hence
no flags are accepted). If the option "in" is used, the argument is first split with respect to commas
and then set inclusion is tested.
Example: \verb{x = row.names(mtcars) ; string_magic("Mercedes models: \{'Merc & [[:alpha:]]$'get, '^.+ 'r, C ? x\}")}
leads to "Mercedes models: 240D, 280C, 450SE, 450SL and 450SLC".
\item is: detects if a pattern is present in a string, returns a logical vector. This operation has no default.
Accepts the options "equal" and "in".
By default it uses the same syntax as \code{\link[=string_is]{string_is()}} so that you can use regex flags and
include logical operations with \code{' & '} and \code{' | '} to detect patterns.
If the option "equal" is used, a simple string equality with the argument is tested (hence
no flags are accepted). If the option "in" is used, the argument is first split with respect to commas
and then set inclusion is tested.
Mostly useful as the final operation in a \code{\link[=string_ops]{string_ops()}} call.
Example: \verb{x = c("Mark", "Lucas") ; string_magic("Mark? \{'i/mark'is, C ? x\}")} leads to "Mark? TRUE and FALSE".
\item which: returns the index of string containing a specified pattern. With no default, can be applied
to a logical vector directly.
By default it uses the same syntax as string_which() so that you can use regex flags and
include logical operations with \code{' & '} and \code{' | '} to detect patterns.
If the option "equal" is used, a simple string equality with the argument is tested (hence
no flags are accepted). If the option "in" is used, the argument is first split with respect to commas
and then set inclusion is tested.
Mostly useful as the final operation in a \code{\link[=string_ops]{string_ops()}} call.
Ex.1: \verb{x = c("Mark", "Lucas") ; string_magic("Mark is number \{'i/mark'which ? x\}.")} leads to
"Mark is number 1.".
}
}

\section{Operations changing the length or the order}{

\itemize{
\item first: keeps only the first \code{n} elements. Example: \code{string_magic("First 3 numbers: {3 first, C ? mtcars$mpg}.")}
leads to "First 3 numbers: 21, 21 and 22.8.". Negative numbers as argument remove the
first \code{n} values. You can add a second argument in the form \verb{'n1|n2'first} in which case the first \code{n1} and last
\code{n2} values are kept; \code{n1} and \code{n2} must be positive numbers.
\item K, Ko, KO: keeps only the first \code{n} elements; has more options than \code{first}. The syntax is \verb{'n'K},
\verb{'n|s'K}, \verb{'n||s'K}. The values Ko and KO only accept the two first syntax (with \code{n} only).
\code{n} provides the number of elements to keep. If \code{s} is provided and the number of
elements are greater than \code{n}, then in 'n|s' the string \code{s} is added at the end, and
if 'n||s' the string s replaces the nth element.
The string \code{s} accepts specials values:
\itemize{
\item \verb{:n:} or \verb{:N:} which gives the total number of items in digits or letters (N)
\item \verb{:rest:} or \verb{:REST:} which gives the number of elements that have been truncated in digits or letters (REST)
Ex: \code{string_magic("{'3|:rest: others'K ? 1:200}")} leads to the vector \code{c("1", "2", "3", "197 others")}.
\item The operator 'n'Ko is like \verb{'n||:rest: others'K} and 'n'KO is like \verb{'n||:REST: others'K}.
}
\item last: keeps only the last \code{n} elements. Example: \code{string_magic("Last 3 numbers: {3 last, C ? mtcars$mpg}.")}
leads to "Last 3 numbers: 19.7, 15 and 21.4.". Negative numbers as argument remove the
last \code{n} values.
\item sort: sorts the vector in increasing order. Accepts optional arguments and the option "num".
Example: \verb{x = c("sort", "me") ; string_magic("\{sort, c ? x\}")} leads to "me sort".
If an argument is provided, it must be a regex pattern that will be applied to
the vector using \code{\link[=string_clean]{string_clean()}}. The sorting will be applied to the modified version of the vector
and the original vector will be ordered according to this sorting.
Ex: \code{x = c("Jon Snow", "Khal Drogo")}; \code{string_magic("{'.+ 'sort, C?x}")} leads to
"Khal Drogo and Jon Snow". The option "num" sorts over a numeric version
(with silent conversion) of the vector and reorders the original vector accordingly.
Values which could not be converted are last.
\strong{Important note}: the sorting operation is applied before any character conversion.
If previous operations were applied, it is likely that numeric data were transformed to character.
Note the difference: \verb{x = c(20, 100, 10); string_magic("\{sort, ' + 'c ? x\}")} leads to "10 + 20 + 100"
while \code{string_magic("{n, sort, ' + 'c ? x}")} leads to "10 + 100 + 20" because the operation "n"
first transformed the numeric vector into character.
\item dsort: sorts the vector in decreasing order. It accepts an optional argument and
the option "num". Example: \code{string_magic("5 = {dsort, ' + 'c ? 2:3}")}
leads to "5 = 3 + 2". See the operation "sort" for a description of the argument and the option.
\item rev: reverses the vector. Example: \code{string_magic("{rev, ''c ? 1:3}")} leads to "321".
\item unik: makes the string vector unique. Example: \code{string_magic("Iris species: {unik, C ? iris$Species}.")}
leads to "Iris species: setosa, versicolor and virginica.".
\item table: computes the frequency of each element and attaches each element to its frequency.
Accepts an argument which must be a character string representing a \code{string_magic} interpolation
with the following variables: \code{x} (the element), \code{n} (its count) and \code{s} (its share). The default is \code{'{x} ({n ? n})'}. By default the resulting string vector is sorted by decreasing frequency.
You can change how the vector is sorted with five options: \code{sort} (sorts on the elements),
\code{dsort} (decreasing sort), \code{fsort} (sorts on frequency),
\code{dfsort} (decreasing sort on freq. -- default),
\code{nosort} (keeps the order of the first elements). Note that you can combine several sorts
(to resolve the ties of elements with same frequencies).
Example: \code{string_magic("Freq. of months: {'{x} ({n})'table, enum ? month.name[airquality$Month]}.")}
\item each: repeats each element of the vector \code{n} times. Option "c" then collapses the full vector
with the empty string as a separator. Ex.1: \code{string_magic("{/x, y}{2 each ? 1:2}")} leads to the
vector \code{c("x1", "y1", "x2", "y2")}. Ex.2: \code{string_magic("Large number: 1{5 each.c ! 0}")} leads to
"Large number: 100000".
\item times: repeats the vector sequence \code{n} times. Option "c" then collapses the full vector
with the empty string as a separator. Example: \code{string_magic("What{6 times.c ! ?}")} leads to "What??????".
\item rm: removes elements from the vector. Options: "empty", "blank", "noalpha", "noalnum", "all".
The \emph{optional} argument represents the pattern used to detect strings to be deleted.
Ex.1: \code{x = c("Luke", "Charles")}; \code{string_magic("{'i/lu'rm ? x}")} leads to "charles". By default it removes
empty strings. Option "blank" removes strings containing only blank characters (spaces, tab, newline).
Option "noalpha" removes strings not containing letters. Option "noalnum" removes strings not
containing alpha numeric characters. Option "all" removes all strings (useful in conditions, see
the dedicated section). If an argument is provided, only the options "empty" and "blank" are available.
Ex.2: \code{x = c("I want to enter.", "Age?", "21")}; \code{string_magic("Nightclub conversation: {rm.noalpha, c ! - {x}}")}
leads to "Nightclub conversation: - I want to enter. - Age?"
\item nuke: removes all elements, equivalent to \code{rm.all} but possibly more explicit (not sure).
Useful in conditions, see the dedicated section.
Example: \verb{x = c(5, 7, 453, 647); string_magic("Small numbers only: \{if(.>20 ; nuke), C ? x\}")} leads
to "Small numbers only: 5 and 7";
\item insert: inserts a new element to the vector. Options: "right" and "both". Option "right" adds
the new element to the right. Option "both" inserts the new element on the two sides of the vector.
Example: \code{string_magic("{'3'insert.right, ' + 'c ? 1:2}")} leads to "1 + 2 + 3".
\item \code{dp} or \code{deparse}: It deparses an R object and collapses it into a single string
using newlines. By default it keeps only the first 50 lines of the deparsed object.
The number of lines kept can be modified by using a numeric argument.
It accepts the options: \code{short} and \code{vec}. If \code{short}, multiline deparsed elements
are collapsed with \verb{\\\\\\\\n} and the number of characters cannot exceed the value of the
argument, which by default is 50.
If \code{vec}, then the deparsed element is not collapsed into a single string, and only
the first 50 (default) elements of the vector are kept. You can increase/reduce the
number of elements kept with the argument.
Example: \verb{fml = y ~ x1 + x2; string_magic("The estimated model is \{dp ? fml\}.")}
}
}

\section{Formatting operations}{

\itemize{
\item lower: lower cases the full string.
\item upper: upper cases the full string. Options: "first" and "sentence".
Option "first" upper cases only the first character. Option "sentence"
upper cases the first letter after punctuation.
Ex: \verb{x = "hi. how are you? fine." ; string_magic("\{upper.sentence ? x\}")} leads
to "Hi. How are you? Fine.".
\item title: applies a title case to the string. Options: "force" and "ignore".
Option "force" first puts everything to lowercase before applying the title case.
Option "ignore" ignores a few small prepositions ("a", "the", "of", etc).
Ex: \verb{x = "bryan is in the KITCHEN" ; string_magic("\{title.force.ignore ? x\}")} leads to "Bryan Is in the Kitchen".
\item ws: normalizes whitespaces (WS). It trims the whitespaces on the edges and transforms any succession
of whitespaces into a single one. Can also be used to further clean the string with its options.
Options: "punct", "digit", "isolated". Option "punct" cleans the punctuation. Option "digit" cleans digits.
Option "isolated" cleans isolated letters. WS normalization always come after any of these options.
\strong{Important note:} punctuation (or digits) are replaced with WS and \strong{not}
the empty string. This means that \code{string_magic("ws.punct ! Meg's car")} will become "Meg s car".
\item trimws: trims the white spaces on both ends of the strings.
\item q, Q, bq: to add quotes to the strings. q: single quotes, Q: double quotes, bq:
back quotes. \verb{x = c("Mark", "Pam"); string_magic("Hello \{q, C ? x\}!")} leads to "Hello 'Mark' and 'Pam'!".
\item format, Format: applies the base R's function \code{\link[base:format]{base::format()}} to the string.
By default, the values are left aligned, \emph{even numbers} (differently from \code{\link[base:format]{base::format()}}'s behavior).
The upper case command (\code{Format}) applies right alignment. Options: "0", "zero", "left", "right", "center".
Options "0" or "zero" fills the blanks with 0s: useful to format numbers. Option "right" right aligns,
and "center" centers the strings. Default is left alignment.
Ex: \verb{x = c(1, 12345); string_magic("left: \{format.0, q, C ? x\}, right: \{Format, q, C ? x\}")}
leads to "left: '000001' and '12,345', right: '     1' and '12,345'".
\item \%: applies \code{\link[base:sprintf]{base::sprintf()}} formatting. The syntax is 'arg'\% with arg an sprintf formatting,
or directly the sprint formatting, e.g. \verb{\% 5s}. Example: \code{string_magic("pi = {\%.3f ? pi}")} leads
to "pi = 3.142".
\item stopwords: removes basic English stopwords (the snowball list is used).
The stopwords are replaced with an empty space but the left and right WS are
untouched. So WS normalization may be needed (see operation \code{ws}).
\verb{x = c("He is tall", "He isn't young"); string_magic("Is he \{stop, ws, C ? x\}?")} leads to "Is he tall and young?".
\item ascii: turns all letters into ASCII with transliteration. Failed translations
are transformed into question marks. Options: "silent", "utf8". By default, if some conversion fails
a warning is prompted. Option "silent" disables the warning in case of failed conversion. The conversion
is done with \code{\link[base:iconv]{base::iconv()}}, option "utf8" indicates that the source endocing is UTF-8, can be useful
in some cases.
\item round, r0 to r6: formats numbers by rounding at a given level. Options:  \code{0} to \code{9}, \code{int}, \code{nocomma}, \code{s0}-\code{s9}.
Option \code{0} to \code{9} controls the number of digits to round at.
Option \code{int} is whether to preserve integers from formattting.
Option \code{nocomma} conrtols whether to drop the comma separating the thousands.
\code{s0}-\code{s9} also keeps a given number of significant digits.
\item signif, s0 to s6: formats numbers by displaying a certain number of significant digits. Options:  \code{0} to \code{9}, \code{int}, \code{nocomma}, \code{r0}-\code{r9}.
Option \code{0} to \code{9} controls the number of significant digits to display.
Option \code{int} is whether to preserve integers from formattting.
Option \code{nocomma} conrtols whether to drop the comma separating the thousands.
\code{r0}-\code{r9} also rounds at a given number of decimals.
\item n: formats integers by adding a comma to separate thousands. Options: "letter", "upper", "0", "zero".
The option "letter" writes the number in letters (large numbers keep their numeric format). The option
"upper" is like the option "letter" but uppercases the first letter. Options "0" or "zero" left pads
numeric vectors with 0s. Ex.1: \verb{x = 5; string_magic("He's \{N ? x\} years old.")} leads to "He's five years old.".
Ex.2: \verb{x = c(5, 12, 52123); string_magic("She owes \{n.0, '$'paste, C ? x\}.")} leads to
"She owes $5, $12 and $52,123.".
\item N: same as \code{n} but automatically adds the option "letter".
\item nth: when applied to a number, these operators write them as a rank. Options: "letter",
"upper", "compact".
Ex.1: \verb{n = c(3, 7); string_magic("They finished \{nth, enum ? n\}!")} leads to "They finished 3rd and 7th!".
Option "letter" tries to write the numbers in letters, but note that it stops at 20. Option "upper"
is the same as "letter" but uppercases the first letter. Option "compact" aggregates
consecutive sequences in the form "start_n_th to end_n_th".
Ex.2: \code{string_magic("They arrived {nth.compact ? 5:20}.")} leads to "They arrived 5th to 20th.".
Nth: same as \code{nth}, but automatically adds the option "letter". Example:
\verb{n = c(3, 7); string_magic("They finished \{Nth, enum ? n\}!")} leads to "They finished third and seventh!".
\item ntimes: write numbers in the form \code{n} times. Options: "letter", "upper". Option
"letter" writes the number in letters (up to 100). Option "upper" does the same as "letter"
and uppercases the first letter. Example: \code{string_magic("They lost {C ! {ntimes ? c(1, 12)} against {S!Real, Barcelona}}.")}
leads to "They lost once against Real and 12 times against Barcelona.".
\item Ntimes: same as \code{ntimes} but automatically adds the option "letter".
Example: \verb{x = 5; string_magic("This paper was rejected \{Ntimes ? x\}...")} leads to
"This paper was rejected five times...".
\item firstchar, lastchar: to select the first/last characters of each element.
Ex: \code{string_magic("{19 firstchar, 9 lastchar ! This is a very long sentence}")} leads to "very long".
Negative numbers remove the first/last characters.
\item k: to keep only the first n characters (like \code{firstchar} but with more options). The
argument can be of the form \verb{'n'k}, \verb{'n|s'k} or \verb{'n||s'k} with \code{n} a number and \code{s} a string.
\code{n} provides the number of characters to keep. Optionnaly, only for strings whose
length is greater than \code{n}, after truncation, the string \code{s} can be appended at the end.
The difference between 'n|s' and 'n||s' is that in the second case the strings
will always be of maximum size \code{n}, while in the first case they can be of length \code{n + nchar(s)}.
Ex: \code{string_magic("{4k ! long sentence}")} leads to "long",  \code{string_magic("{'4|..'k ! long sentence}") }
leads to "long..", \code{string_magic("{'4||..'k ! long sentence}")} leads to "lo..".
\item fill, align (alias), width (alias): fills the character strings up to a size in order
to fit a given width. Options: "left", "right", "center".
Accepts arguments of the form \code{'n'} or \code{'n|s'}, with \code{n} a number and \code{s} a symbol.
Default is left-alignment of the strings.
Option "right" right aligns and "center" centers the strings. When using \code{'n|s'}, the symbol \code{s}
is used for the filling. By default if no argument is provided, the
maximum size of the character string is used. See help for \code{\link[=string_fill]{string_fill()}} for more information.
Ex.1: \code{string_magic("Numbers: {'5|0'fill.right, C ? c(1, 55)}")} leads to "Numbers: 00001 and 00055".
\item paste, append: pastes some character to all elements of the string. This operation has no default.
Options: "left", "both", "right", "front", "back", "delete". By default, a string is pasted on the left.
By default, it pastes on the left. Option "right" pastes on the right and "both" pastes on both sides. Option "front" only
pastes on the first element while option "back" only pastes on the last element. Option "delete"
first replaces all elements with the empty string.
Example: \code{string_magic("6 = {'|'paste.both, ' + 'c ? -3:-1}")} leads to "6 = |-3| + |-2| + |-1|".
The argument can be of the form \code{s} or \code{s1|s2}. If of the second form, this is equivalent
to chaining two \code{paste} operations, once on the left and once on the right: \verb{'s1'paste, 's2'paste.right}.
\item join: joins lines ending with a double backslash. Ex: \code{x = "the sun \\\\\\n is shining"};
\code{string_magic("{join ? x}")} leads to "the sun is shining".
\item escape: adds backslashes in front of specific characters. Options \code{"nl"}, \code{"tab"}.
Option \code{"nl"} escapes the newlines (\verb{\\n}), leading them to be displayed as \code{"\\\\\\\\n"}.
Option \code{"tab"} does the same for tabs (\code{"\\t"}). This is useful to make the value free
of space formatters.
The default behavior is to escape both newlines and tabs.
}
}

\section{Other operations}{

\itemize{
\item num: converts to numeric. Options: "warn", "soft", "rm", "clear". By default, the conversion
is performed silently and elements that failed to convert are turned into NA.
Option "warns" displays a warning if the conversion to numeric fails.
Option "soft" does not convert if the conversion of at least one element fails.
Option "rm" converts and removes the elements that could not be converted.
Option "clear" turns failed conversions into the empty string, and hence lead to a character vector.
Example: \verb{x = c(5, "six"); string_magic("Compare \{num, C, q ? x\} with \{num.rm, C, q ? x\}.")} leads to
"Compare '5 and NA' with '5'.", and \code{string_magic("Compare {num.soft, C, q ? x} with {clear, C, q ? x}.")}
leads to "Compare '5 and six' with '5 and '.".
\item enum: enumerates the elements. It creates a single string containing the comma
separated list of elements.
If there are more than 7 elements, only the first 6 are shown and the number of
items left is written.
For example \code{string_magic("enum ? 1:5")} leads to "1, 2, 3, 4, and 5".
You can add the following options by appending the letter to enum after a dot:
\itemize{
\item q, Q, or bq: to quote the elements
\item or, nor: to finish with an 'or' (or 'nor') instead of an 'and'
\item comma: to finish the enumeration with ", " instead of ", and".
\item i, I, a, A, 1: to enumerate with this prefix, like in: i) one, and ii) two
\item a number: to tell the number of items to display
Ex.1: \verb{x = c("Marv", "Nancy"); string_magic("The main characters are \{enum ? x\}.")} leads to
"The main characters are Marv and Nancy.".
Ex.2: \verb{x = c("orange", "milk", "rice"); string_magic("Shopping list: \{enum.i.q ? x\}.")} leads to
"Shopping list: i) 'orange', ii) 'milk', and iii) 'rice'."
}
\item len: gives the length of the vector. Options "letter", "upper", "num".
Option "letter" writes the length in words (up to 100). Option "upper" is the same
as letter but uppercases the first letter.
By default, commas are added to separate thousands. Use uption "num" to preserve
a regular numeric format.
Example: \code{string_magic("Size = {len ? 1:5000}")} leads to "Size = 5,000".
\item swidth: stands for screen width. Formats the string to fit a given width
by cutting at word boundaries and adding newlines appropriately.
Accepts arguments of the form \code{'n'} or \code{'n|s'}, with \code{n} a number and \code{s} a string.
An argument of the form \code{'n|s'} will add \code{s} at the beginning of each line. Further,
by default a trailing white space is added to \code{s}; to remove this
behavior, add an underscore at the end of it.
The argument \code{n} is either an integer giving the target character
width (minimum is 15), or it can be a fraction expressing the
target size as a fraction of the current screen. Finally it can be an expression that
uses the variable \code{.sw} which will capture the value of the current screen width.
Ex.1: \code{string_magic("{15 swidth ! this is a long sentence}")} leads to "this is a long\\nsentence".
Ex.2: \code{string_magic("{15 swidth.#> ! this is a long sentence}")} leads to "#> this is a long\\n#> sentence".
\item difftime: displays a formatted time difference. Option "silent" does not report a warning if the
operation fails. It accepts either objects of class \code{POSIXt} or \code{difftime}.
Example: \verb{x = Sys.time() ; Sys.sleep(0.5) ; string_magic("Time: \{difftime ? x\}")} leads to something
like "Time: 514ms".
}
}

\section{Group-wise operations}{


In \code{string_magic}, the splitting operation \code{s} (or \code{S}) keeps a memory of the strings
that were split. Use the tilde operator, of the form \verb{~(op1, op2)}, to apply operations
group-wise, to each of the split strings.

Better with an example. \code{x = c("Oreste, Hermione", "Hermione, Pyrrhus", "Pyrrhus, Andromaque") ;}
\code{string_magic("Troubles ahead: {S, ~(' loves 'c), C ? x}.")} leads to
"Troubles ahead: Oreste loves Hermione, Hermione loves Pyrrhus and Pyrrhus loves Andromaque.".

Almost all operations can be applied group-wise (although only operations changing the order or
the length of the strings really matter).
}

\section{Conditional operations}{


There are two operators to apply operations conditionally: \code{if} and \code{vif}, the latter
standing for \emph{verbatim if}.

The syntax of \code{if} is \verb{if(cond ; ops_true ; ops_false)} with \code{cond} a
condition (i.e. logical operation) on the value being interpolated, \code{ops_true} a comma-separated
sequence of operations if the condition is \code{TRUE} and \code{ops_false} an \emph{optional} a sequence of
operations if the condition is \code{FALSE}.

Ex.1: Let's take a sentence, delete words of less than 4 characters, and trim
words of 7+ characters.
x = "Songe Cephise a cette nuit cruelle qui fut pour tout un peuple une nuit eternelle"
\code{string_magic("{' 's, if(.nchar<=4 ; nuke ; '7|..'k), c ? x}")}.
Let's break it down. First the sentence is split w.r.t. spaces, leading to a vector
of words. Then we use the special variable \code{.nchar} in \code{if}'s condition to refer
to the number of characters of the current vector (the words). The words with
less than 4 characters are nuked (i.e. removed), and the other words are
trimmed at 7 characters. Finally the modified vector of words is collapsed with
the function \code{c}, leading to the result.

The condition \code{cond} accepts the following special values: \code{.} (the dot), \code{.nchar}, \code{.C}, \code{.len}, \code{.N}.
The dot, \code{.}, refers to the current vector. \code{.nchar} represent the number of characters
of the current vector (equivalent to \code{nchar(.)}). \code{.C} is an alias to \code{.nchar}.
\code{.len} represent the length of the current vector (equivalent to \code{length(.)}).
\code{.N} is an alias to \code{.len}.

If a condition leads to a result of length 1, then the operations are applied to
the full string vector and not element-wise (as was the case in Ex.1). Contrary to element-wise conditions
for which operations modifying the length of the vectors are forbidden (apart from nuking),
such operations are fine in full-string conditions.

Ex.2: \code{x = string_magic("x{1:10}")}; \code{string_magic("y = {if(.N>4 ; 3 first, '...'insert.right), ' + 'c ? x}")}
leads to "y = x1 + x2 + x3 + ...". the same opration applied to \code{x = string_magic("x{1:4}")}
leads to "y = x1 + x2 + x3 + x4".

For \code{vif}, the syntax is \verb{vif(cond ; verb_true ; verb_false)} with \code{verb_true}
a verbatim value with which the vector will be replaced if the condition is \code{TRUE}.
This is similar for \code{verb_false}. The condition works as in \code{if}.

Ex.3: \verb{x = c(1, 25, 12, 6) ; string_magic("Values: \{vif(.<10 ; <10), C ? x\}")} leads to
"Values: <10, 25, 12 and <10". As we can see values lower than 10 are replaced
with "<10" while other values are not modified.

Ex.4: \code{x = string_magic("x{1:10}")}; \code{string_magic("y = {vif(.N>4 ; {S!{x[1]}, ..., {last?x}}), ' + 'c ? x}")}
leads to "y = x1 + ... + x10".
Let's break it down. If the length of the vector is greater than 4 (here it's 10), then
the full string is replaced with \code{"{S!{x[1]}, ..., {last?x}}"}. Interpolation applies to
such string. Hence the split operation \code{S} breaks the string w.r.t.
the commas (default behavior), leading to the vector \code{c("{x[1]}", "...", "{last?x}")}. Since the
string contains curly brackets, interpolation is applied again. This leads to
the vector \code{c("x1", "...", "x10")}. Finally, this vector is collapsed with ' + ' leading
to the final string.
Note that there are many ways to get to the same result. Here is another example:
\code{string_magic("y = {vif(.N>4 ; {x[1]} + ... + {last?x} ; {' + 'c ? x}) ? x}")}.

The \code{vif} condition allows the use of '.' to refer to the current value in
\code{verb_true} and \code{verb_false}, as illustrated by the last example:

Ex.5: \code{string_magic("{4 last, vif(. \%\% 2 ; x{.} ; y{rev?.}), C ? 1:11}")}
leads to "y10, x9, y8 and x11".
}

\section{Special interpolation}{
 if-else:

Using an ampersand ("&") as the first character of an interpolation leads to an \emph{if-else} operation.
Using two ampersands ("&&") leads to a slightly different operation described at the end of this section.

The syntax is as follows: \verb{\{&cond ; verb_true ; verb_false\}} with \code{cond} a
condition (i.e. logical operation) on the value being interpolated, \code{verb_true}
a verbatim value with which the vector will be replaced if the condition is \code{TRUE} and
\code{verb_false} an \emph{optional} verbatim value with which the vector will be replaced if the condition is \code{FALSE}.
If not provided, \code{verb_false} is considered to be the empty string unless the operator is
the double ampersand described at the end of this section.

Note that in \code{cond}, you can use the function \code{len}, an alias to \code{length}.

Ex.1: \code{x = 1:5}; \code{string_magic("x is {&len(x)<10 ; short ; {`log10(.N)-1`times, ''c ! very }long}")}
leads to "x is short". With \code{x = 1:50}, it leads to "x is long", and to "x is very very long"
if \code{x = 1:5000}.

If a condition leads to a result of length 1, the full string is replaced by the verbatim
expression. Further, this expression will be interpolated if requested. This was the case
in Ex.1 where \code{verb_false} was interpolated.

If the condition's length is greater than 1, then each logical values equal to \code{TRUE} is replaced
by \code{verb_true}, and \code{FALSE} or \code{NA} values are replaced with \code{verb_false}. Note,
importantly, that \strong{no interpolation is perfomed in that case}.

Ex.2: \verb{x = 1:3 ; string_magic("x is \{&x == 2 ; two ; not two\}")} leads to the vector
\code{c("x is not two", "x is two", "x is not two")}.

In that example, when x is odd, it is replaced with "odd", and when even it is
replaced with the elements of y.

Using the two ampersand operator ("&&") is like the simple ampersand version but the
default for \code{verb_false} is the variable used in the condition itself. So the syntax is
\verb{\{&&cond ; verb_true\}} and \emph{it does not accept} \code{verb_false}.

Ex.3: \verb{i = 3 ; string_magic("i = \{&&i == 3 ; three\}")} leads to "i = three", and to "i = 5" if \code{i = 5}.

 Pluralization:

There is advanced support for pluralization which greatly facilitates the writing of messages
in natural language.

There are two ways to pluralize: over length or over value. To trigger a "pluralization" interpolation
use as first character:
\itemize{
\item \code{$} to pluralize over the length of a variable (see Ex.2)
\item \verb{#} to pluralize over the value of a variable (see Ex.1)
}

Ex.1: \verb{x = 5; string_magic("I bought \{N?x\} book\{#s\}.")} leads to "I bought five books.".
If \code{x = 1}, this leads to "I bought one book.".

The syntax is \verb{\{#plural_ops ? variable\}} or \verb{\{#plural_ops\}} where \code{plural_ops} are
specific pluralization operations which will be described below.
The pluralization is perfomed \emph{always} with respect to the value of a variable.
You can either add the variable explicitly (\verb{\{#plural_ops ? variable\}}) or refer
to it implicitly (\verb{\{#plural_ops\}}). If implicit, then the algorithm will look at the
previous variable that was interpolated and pluralize over it. This is exaclty what happens in
Ex.1 where \code{x} was interpolated in \code{{N?x}} and plural operation \code{s} in \verb{\{#s\}} then applied to
\code{x}. It was equivalent to have \verb{\{#s ? x\}}. If a variable wasn't interpolated before, then
the next interpolated variable will be used (see Ex.2). If no variable is interpolated
at all, an error is thrown.

Ex.2: \verb{x = c("J.", "M."); string_magic("My BFF\{$s, are\} \{C?x\}!")} leads to "My BFFs are J. and M.!".
If "x = "S.", this leads to "My BFF is S.!".

Pluralizing accepts the following operations:
\itemize{
\item s, es: adds an "s" (or "es") if it is plural (> 1), nothing otherwise. Accepts the option \code{0} or \code{zero} which
treats a 0-length or a 0-value as plural.
\item y or ies: adds an 'y' if singular and 'ies' if plural (>1). Accepts the option \code{0} or \code{zero} which
treats a 0-length or a 0-value as plural.
\item enum: enumerates the elements (see help for the regular operation \code{enum})
\item n, N, len, Len: add the number of elements ("len") or the value ("n") of the variable as a formatted number or
in letters (upper case versions). Accepts the options \code{letter} (to write in letter)
and \code{upper} (to uppercase the first letter).
\item nth, ntimes: writes the value of the variable as an order (nth) or a frequence (ntimes). Accepts the option \code{letter}
to write the numbers in letters (uppercase version of the operator does the same).
\item is, or any verb: conjugates the verb appropriately
}

You can chain operations, in that case a whitespace is automatically added between them.

Ex.3: \verb{x = c(7, 3, 18); string_magic("The winning number\{$s, is, enum ? sort(x)\}.")}
leads to "The winning numbers are 3, 7 and 18.". With \code{x = 7} this leads to
"The winning number is 7.".

On top of the previous operations, there is a special operation allowing to add verbatim text depending on
the situation. The syntax is as follows:
\itemize{
\item \verb{(s1;s2)}: adds verbatim 's1' if singular and 's2' if plural (>1)
\item \verb{(s1;s2;s3)}: adds verbatim 's1' if zero, 's2' if singular (=1) and 's3' if plural
\item \verb{(s1;;s3)}: adds verbatim 's1' if zero, 's3' if singular or plural (i.e. >=1)
}

These case-dependent verbatim values \strong{are interpolated} (if appropriate). In these interpolations
you need not refer explicitly to the variable for pluralization interpolations.

Ex.4: \verb{x = 3; string_magic("\{#(Sorry, nothing found.;;\{#N.upper\} match\{#es, were\} found.)?x\}")} leads to
"Three matches were found.". If "x = 1", this leads to "One match was found." and if "x = 0" this leads
to "Sorry, nothing found.".
}

\section{Escaping and special cases}{


The opening and closing brakets, \code{{}}, are special characters and cannot be used as regular text.
To bypass their special meaning, you need to escape them with a double backslash.

Ex.1: \code{string_magic("open = \\\\\\\\{, close = }")} leads to \code{"open = {, close = }"}.
Ex.2: \code{string_magic("many {5 times.c ! \\\\\\\\}}")} leads to \verb{many \}\}\}\}\}}.

You only need to escape the special delimiters which the algorithm is currently looking for.
As you can see, you don't need to escape the closing bracket in Ex.1 since no box
was open. On the other hand, you need to escape it in Ex.2.

Alternatively, use the argument \code{.delim} to set custom delimiters.

Ex.3: \code{string_magic("I {'can {write} {{what}} I want'}")} leads to \code{"I can {write} {{what}} I want"}.

Since \code{{expr}} evaluates \code{expr}, the stuff inside the \emph{box}, you can pass a
character string and it will stay untouched.

In the few operations expecting a semi-colon (if-else and pluralization), it can also be
escaped with a double backslash.

In interpolations, the exclamation mark (\code{!}) signals a verbatim expression. But what
if you use it to mean the logical operation \emph{not} in an operation-free interpolation?
In that case, you need a hack: use a question mark (\verb{?}) first to indicate to the
algorithm that you want to evaluate the expression.

Ex.4: \code{string_magic("{!TRUE} is {?!TRUE}")} leads to "TRUE is FALSE". The first expression is
taken verbatim while the second is evaluated.
}

\section{Generic regular expression flags}{


All \code{stringmagic} functions support generic flags in regular-expression patterns.
The flags are useful to quickly give extra instructions, similarly to \emph{usual}
\href{https://javascript.info/regexp-introduction}{regular expression flags}.

Here the syntax is "flag1, flag2/pattern". That is: flags are a comma separated list of flag-names
separated from the pattern with a slash (\code{/}). Example: \code{string_which(c("hello...", "world"), "fixed/.")} returns \code{1}.
Here the flag "fixed" removes the regular expression meaning of "." which would have otherwise meant \emph{"any character"}.
The no-flag verion \code{string_which(c("hello...", "world"), ".")} returns \code{1:2}.

Alternatively, and this is recommended, you can collate the initials of the flags instead of using a
comma separated list. For example: "if/dt[" will apply the flags "ignore" and "fixed" to the pattern "dt[".

The four flags always available are: "ignore", "fixed", "word" and "magic".
\itemize{
\item "ignore" instructs to ignore the case. Technically, it adds the perl-flag "(?i)"
at the beginning of the pattern.
\item "fixed" removes the regular expression interpretation, so that the characters ".", "$", "^", "["
(among others) lose their special meaning and are treated for what they are: simple characters.
\item "word" adds word boundaries (\code{"\\\\b"} in regex language) to the pattern. Further, the comma (\code{","})
becomes a word separator. Technically, "word/one, two" is treated as "\\b(one|two)\\b". Example:
\code{string_clean("Am I ambushed?", "wi/am")} leads to " I ambushed?" thanks to the flags "ignore" and "word".
\item "magic" allows to interpolate variables inside the pattern before regex interpretation.
For example if \code{letters = "aiou"} then \code{string_clean("My great goose!", "magic/[{letters}] => e")}
leads to \code{"My greet geese!"}
}
}

\examples{

#
# BASIC USAGE ####
#

x = c("Romeo", "Juliet")

# {x} inserts x
string_magic("Hello {x}!")

# elements in ... are collapsed with "" (default)
string_magic("Hello {x[1]}, ",
    "how is {x[2]} doing?")

# Splitting a comma separated string
# The mechanism is explained later
string_vec("J. Mills, David, Agnes, Dr Strong")

# Nota: this is equivalent to (explained later)
string_magic("{', *'S ! J. Mills, David, Agnes, Dr Strong}")

#
# Applying low level operations to strings
#

# Two main syntax:

# A) expression evaluation
# {operation ? x}
#            | |
#            |  \-> the expression to be evaluated
#             \-> ? means that the expression will be evaluated

# B) verbatim
# {operation ! x}
#            | |
#            |  \-> the expression taken as verbatim (here 'x')
#             \-> ! means that the expression is taken as verbatim

# operation: usually 'arg'op with op an operation code.

# Example: splitting
x = "hello dear"
string_magic("{' 's ? x}")
# x is split by ' '

string_magic("{' 's ! hello dear}")
# 'hello dear' is split by ' '
# had we used ?, there would have been an error


# There are 50+ string operators
# Operators usually have a default value
# Operations can have options
# Operations can be chained by separating them with a comma

# Example: default of 's' is ' ' + chaining with collapse
string_magic("{s, ' my 'c ! hello dear}")

#
# Nesting
#

# {operations ! s1{expr}s2}
#             |   |
#             |    \-> expr will be interpolated then added to the string
#              \-> nesting requires verbatim evaluation: '!'

string_magic("The variables are: {C ! x{1:4}}.")

# This one is ugly but it shows triple nesting
string_magic("The variables are: {ws, C ! {2 times ! x{1:4}}{','s, 4 each !  ,_sq}}.")

#
# Splitting
#

# s: split with fixed pattern, default is ' '
string_magic("{s ! a b c}")
string_magic("{' b 's !a b c}")

# S: same as 's' but default is ',[ \t\n]*'
string_magic("{S !a, b, c}")
string_magic("{'[[:punct:] ]+'S ! a! b; c}")

# add regex flags: e.g. fixed search
string_magic("{'f/.'s ! hi.there}")


#
# Collapsing
#

# c and C do the same, their default is different
# syntax: 's1|s2' with
# - s1 the string used for collapsing
# - s2 (optional) the string used for the last collapse

# c: default is ' '
string_magic("{c ? 1:3}")

# C: default is ', | and '
string_magic("{C ? 1:3}")

string_magic("{', | or 'c ? 1:4}")

#
# Extraction
#

# extract: to extract patterns (option first)
# x: alias to extract.first
# X: alias to extract
# syntax: 'pattern'x
# Default is '[[:alnum:]]+'

x = "This years is... 2020"
string_magic("{x ? x}") # similar to string_magic("{extract.first ? x}")
string_magic("{X ? x}") # similar to string_magic("{extract ? x}")

string_magic("{'\\\\d+'x ? x}")

#
# STRING FORMATTING ####
#

#
# upper, lower, title

# upper case the first letter
string_magic("{upper.first ! julia mills}")

# title case
string_magic("{title ! julia mills}")

# upper all letters
string_magic("{upper ! julia mills}")

# lower case
string_magic("{lower ! JULIA MILLS}")

#
# q, Q, bq: single, double, back quote

string_magic("{S, q, C ! Julia, David, Wilkins}")
string_magic("{S, Q, C ! Julia, David, Wilkins}")
string_magic("{S, bq, C ! Julia, David, Wilkins}")

#
# format, Format: formats the string to fit the same length

# format: the right side is filled with blanks
# Format: the left side is filled with blanks, the string is right aligned

score = c(-10, 2050)
nm = c("Wilkins", "David")
string_magic("Monopoly scores:\n{'\n'c ! - {format ? nm}: {Format ? score} US$}")

# OK that example may have been a bit too complex,
# let's make it simple:

string_magic("Scores: {format ? score}")
string_magic("Names: {Format ? nm}")

#
# ws: white space normalization

# ws: suppresses trimming white spaces + normalizes successive white spaces
# Add the following options in any order to:
# - punct: remove punctuation
# - digit: remove digits
# - isolated: remove isolated characters

string_magic("{ws ! The   white  spaces are now clean.  }")

string_magic("{ws.punct ! I, really -- truly; love punctuation!!!}")

string_magic("{ws.digit ! 1, 2, 12, a microphone check!}")

string_magic("{ws.i ! 1, 2, 12, a microphone check!}")

string_magic("{ws.d.i ! 1, 2, 12, a microphone check!}")

string_magic("{ws.p.d.i ! 1, 2, 12, a microphone check!}")

#
# \%: applies sprintf formatting

 # add the formatting as a regular argument
string_magic("pi = {'.2f'\% ? pi}")
# or right after the \%
string_magic("pi = {\%.2f ? pi}")

#
# paste: appends text on each element
# Accepts the options: right, both, front and back
# It accepts the special values :1:, :i:, :I:, :a:, :A: to create enumerations

# adding '|' on both sides
string_magic("{'|'paste.both, ' + 'c ! x{1:4}}")


# Enumerations
acad = string_vec("you like admin, you enjoy working on weekends, you really love emails")
string_magic("Main reasons to pursue an academic career:\n {':i:) 'paste, C ? acad}.")

# You can also use the enum command
string_magic("Main reasons to pursue an academic career:\n {enum.i ? acad}.")

#
# stopwords: removes basic English stopwords
# the list is from the Snowball project:
#  http://snowball.tartarus.org/algorithms/english/stop.txt

string_magic("{stop, ws ! It is a tale told by an idiot, ",
                         "full of sound and fury, signifying nothing.}")

#
# k: keeps the first n characters
# syntax: nk: keeps the first n characters
#         'n|s'k: same + adds 's' at the end of shortened strings
#         'n||s'k: same but 's' counts in the n characters kept

words = string_vec("short, constitutional")
string_magic("{5k ? words}")

string_magic("{'5|..'k ? words}")

string_magic("{'5||..'k ? words}")

#
# K: keeps the first n elements
# syntax: nK: keeps the first n elements
#         'n|s'K: same + adds the element 's' at the end
#         'n||s'K: same but 's' counts in the n elements kept
#
# Special values :rest: and :REST:, give the number of items dropped

bx = string_vec("Pessac Leognan, Saint Emilion, Marguaux, Saint Julien, Pauillac")
string_magic("Bordeaux wines I like: {3K, ', 'C ? bx}.")

string_magic("Bordeaux wines I like: {'3|etc..'K, ', 'C ? bx}.")

string_magic("Bordeaux wines I like: {'3||etc..'K, ', 'C ? bx}.")

string_magic("Bordeaux wines I like: {'3|and at least :REST: others'K, ', 'C ? bx}.")

#
# Ko, KO: special operator which keeps the first n elements and adds "others"
# syntax: nKo
# KO gives the rest in letters

string_magic("Bordeaux wines I like: {4KO, C ? bx}.")

#
# r, R: string replacement 
# syntax: 's'R: deletes the content in 's' (replaces with the empty string)
#         's1 => s2'R replaces s1 into s2

string_magic("{'e'r, ws ! The letter e is deleted}")

# adding a perl look-behind
string_magic("{'(?<! )e'r !The letter e is deleted}")

string_magic("{'e => a'r !The letter e becomes a}")

string_magic("{'([[:alpha:]]{3})[[:alpha:]]+ => \\\\1.'r ! Trimming the words}")

# Alternative way with simple operations: split, shorten, collapse
string_magic("{s, '3|.'k, c ! Trimming the words}")

#
# times, each
# They accept the option c to collapse with the empty string

string_magic("N{10 times.c ! o}!")

string_magic("{3 times.c ? 1:3}")
string_magic("{3 each.c ? 1:3}")

#
# erase: replaces the items by the empty string
# -> useful in conditions

string_magic("{erase ! I am going to be annihilated}")

#
# ELEMENT MANIPULATION ####
#

#
# rm: removes the elements
# Its (optional) argument is a regular expression giving which element to remove
# Many options: "empty", "blank", "noalpha", "noalnum", "all" 

x = c("Destroy", "All")
string_magic("{'A'rm ? x}")

string_magic("{rm.all ? x}")

x = string_vec("1, 12, 123, 1234, 123456, 1234567")
# we delete elements whose number of characters is lower or equal to 3
# => see later section CONDITIONS
string_magic("{if(.nchar > 3 ; nuke) ? x}")

#
# PLURALIZATION ####
#

# Two ways to enable pluralization:
# {$ command}: means the plural is decuced from the length of the variable
# {# command}: means the plural is decuced from the value of the variable

# Explanatory example
x = c("Eschyle", "Sophocle", "Euripide")
n = 37
string_magic("The author{$s, enum, have ? x} written {#N ? n} play{#s}.")

x = "Laurent Berge"
n = 0
string_magic("The author{$s, enum, have ? x} written {#N ? n} play{#s}.")

# How does it work?
# First is {$s, enum, have ? x}.
# The commands `s`, `enum` and `have` are applied to `x` which must come after a `?`
#    => there the plural (whether an s is added and how to conjugate the verb have) depends
#       on the **length** of the vector `x`
#
# Second comes {#N ? n}.
# The double dollar sign means that the command `N` will be applied to the **value** n.
# The value must come after the `?`
#
# Third is {#s}.
# The object to which `s` should be applied is missing (there is no `? n`).
# The default is to apply the command to the previous object. In this case,
#  this is `n`.

# Another similar example illustrating that we need not express the object several times:
x = c("Eschyle", "Sophocle", "Euripide")
string_magic("The {Len ? x} classic author{$s, are, enum}.")



#
# ARGUMENTS FROM THE ENVIRONMENT ####
#

# Arguments can be evaluated from the calling environment.
# Simply use backticks instead of quotes.

dollar = 6
reason = "glory"
string_magic("Why do you develop packages? For {`dollar`times.c ! $}?",
    "For money? No... for {upper,''s, c ? reason}!", .sep = "\n")

#
# Alias generation
#

# Let's create a formula filler
# - we use .local_ops to create the ad hoc operation "add" which adds variables
# - we transform into a formula ex post

fml = string_magic_alias(.post = as.formula, .local_ops = list(add = "' + 'collapse"))

# example with mtcars
lhs = "mpg"
rhs = c("hp", "drat")
fml("{lhs} ~ {add?rhs} + am")
   

}
\seealso{
Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_clean_alias}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\concept{string_magic functions}
\concept{tools with aliases}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_tools.R
\name{string_is}
\alias{string_is}
\alias{string_any}
\alias{string_all}
\alias{string_which}
\alias{stis}
\alias{stany}
\alias{stall}
\alias{stwhich}
\title{Detects whether a pattern is in a character string}
\usage{
string_is(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame(),
  last = NULL
)

string_any(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame()
)

string_all(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame()
)

string_which(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame()
)

stis(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame(),
  last = NULL
)

stany(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame()
)

stall(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame()
)

stwhich(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  pattern = NULL,
  envir = parent.frame()
)
}
\arguments{
\item{x}{A character vector.}

\item{...}{Character scalars representing the patterns to be found. By default they are (perl) regular-expressions.
Use ' & ' or ' | ' to chain patterns and combine their result logically (ex: \code{'[[:alpha:]] & \\\\d'} gets strings
containing both letters and numbers). You can negate by adding a \code{!} first (ex: \code{"!sepal$"} will
return \code{TRUE} for strings that do not end with \code{"sepal"}).
Add flags with the syntax 'flag1, flag2/pattern'. Available flags are: 'fixed', 'ignore', 'word' and 'magic'.
Ex: "ignore/sepal" would get "Sepal.Length" (wouldn't be the case w/t 'ignore').
Shortcut: use the first letters of the flags. Ex: "if/dt[" would get \code{"DT[i = 5]"} (flags 'ignore' + 'fixed').
For 'word', it adds word boundaries to the pattern. The \code{magic} flag first interpolates
values directly into the pattern with "{}".}

\item{fixed}{Logical scalar, default is \code{FALSE}. Whether to trigger a fixed search instead of a
regular expression search (default).}

\item{ignore.case}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then case insensitive search is triggered.}

\item{word}{Logical scalar, default is \code{FALSE}. If \code{TRUE} then a) word boundaries are added to the pattern,
and b) patterns can be chained by separating them with a comma, they are combined with an OR logical operation.
Example: if \code{word = TRUE}, then pattern = "The, mountain" will select strings containing either the word
'The' or the word 'mountain'.}

\item{or}{Logical, default is \code{FALSE}. In the presence of two or more patterns,
whether to combine them with a logical "or" (the default is to combine them with a logical "and").}

\item{pattern}{(If provided, elements of \code{...} are ignored.) A character vector representing the
patterns to be found. By default a (perl) regular-expression search is triggered.
Use ' & ' or ' | ' to chain patterns and combine their result logically (ex: \code{'[[:alpha:]] & \\\\d'} gets strings
containing both letters and numbers). You can negate by adding a \code{!} first (ex: \code{"!sepal$"} will
return \code{TRUE} for strings that do not end with \code{"sepal"}).
Add flags with the syntax 'flag1, flag2/pattern'. Available flags are: 'fixed', 'ignore', 'word' and 'magic'.
Ex: "ignore/sepal" would get "Sepal.Length" (wouldn't be the case w/t 'ignore').
Shortcut: use the first letters of the flags. Ex: "if/dt[" would get \code{"DT[i = 5]"} (flags 'ignore' + 'fixed').
For 'word', it adds word boundaries to the pattern. The \code{magic} flag first interpolates
values directly into the pattern with "{}".}

\item{envir}{Environment in which to evaluate the interpolations if the flag \code{"magic"} is provided.
Default is \code{parent.frame()}.}

\item{last}{A function or \code{NULL} (default). If a function, it will be applied to the vector
just before returning it.}
}
\value{
It returns a logical vector of the same length as \code{x}.

The function \code{string_which} returns a numeric vector.
}
\description{
Function that detects if one or more patterns are in a string. The patterns can be
chained, by default this is a regex search but special flags be triggered with a
specific syntax, supports negation.
}
\details{
The internal function used to find the patterns is \code{\link[base:grep]{base::grepl()}} with \code{perl = TRUE}.
}
\section{Functions}{
\itemize{
\item \code{string_any()}: Detects if at least one element of a vector matches a regex pattern

\item \code{string_all()}: Detects if all elements of a vector match a regex pattern

\item \code{string_which()}: Returns the indexes of the values in which a pattern is detected

\item \code{stis()}: Alias to \code{string_is}

\item \code{stany()}: Alias to \code{string_any}

\item \code{stall()}: Alias to \code{string_all}

\item \code{stwhich()}: Alias to \code{string_which}

}}
\section{Generic regular expression flags}{


All \code{stringmagic} functions support generic flags in regular-expression patterns.
The flags are useful to quickly give extra instructions, similarly to \emph{usual}
\href{https://javascript.info/regexp-introduction}{regular expression flags}.

Here the syntax is "flag1, flag2/pattern". That is: flags are a comma separated list of flag-names
separated from the pattern with a slash (\code{/}). Example: \code{string_which(c("hello...", "world"), "fixed/.")} returns \code{1}.
Here the flag "fixed" removes the regular expression meaning of "." which would have otherwise meant \emph{"any character"}.
The no-flag verion \code{string_which(c("hello...", "world"), ".")} returns \code{1:2}.

Alternatively, and this is recommended, you can collate the initials of the flags instead of using a
comma separated list. For example: "if/dt[" will apply the flags "ignore" and "fixed" to the pattern "dt[".

The four flags always available are: "ignore", "fixed", "word" and "magic".
\itemize{
\item "ignore" instructs to ignore the case. Technically, it adds the perl-flag "(?i)"
at the beginning of the pattern.
\item "fixed" removes the regular expression interpretation, so that the characters ".", "$", "^", "["
(among others) lose their special meaning and are treated for what they are: simple characters.
\item "word" adds word boundaries (\code{"\\\\b"} in regex language) to the pattern. Further, the comma (\code{","})
becomes a word separator. Technically, "word/one, two" is treated as "\\b(one|two)\\b". Example:
\code{string_clean("Am I ambushed?", "wi/am")} leads to " I ambushed?" thanks to the flags "ignore" and "word".
\item "magic" allows to interpolate variables inside the pattern before regex interpretation.
For example if \code{letters = "aiou"} then \code{string_clean("My great goose!", "magic/[{letters}] => e")}
leads to \code{"My greet geese!"}
}
}

\examples{

# NOTA: using `string_get` instead of `string_is` may lead to a faster understanding 
#       of the examples 

x = string_vec("One, two, one... two, microphone, check")

# default is regular expression search
# => 3 character items
string_is(x, "^...$")

# to trigger fixed search use the flag 'fixed'
string_is(x, "fixed/...")
# you can just use the first letter
string_is(x, "f/...")

# to negate, use '!' as the first element of the pattern
string_is(x, "f/!...")

# you can combine several patterns with " & " or " | "
string_is(x, "one & c")
string_is(x, "one | c")

#
# word: adds word boundaries
#

# compare
string_is(x, "one")
# with
string_is(x, "w/one")

# words can be chained with commas (it is like an OR logical operation)
string_is(x, "w/one, two")
# compare with
string_is(x, "w/one & two")
# remember that you can still negate
string_is(x, "w/one & !two")

# you can combine the flags
# compare
string_is(x, "w/one")
# with
string_is(x, "wi/one")

#
# the `magic` flag
#

p = "one"
string_is(x, "m/{p}")
# Explanation:
# - "p" is interpolated into "one"
# - we get the equivalent: string_is(x, "one")


#
# string_which
#

# it works exactly the same way as string_is
# Which are the items containing an 'e' and an 'o'?
string_which(x, "e", "o")
# equivalently
string_which(x, "e & o")


}
\seealso{
String operations: \code{\link[=string_is]{string_is()}}, \code{\link[=string_get]{string_get()}}, \code{\link[=string_clean]{string_clean()}}, \code{\link[=string_split2df]{string_split2df()}}.
Chain basic operations with \code{\link[=string_ops]{string_ops()}}. Clean character vectors efficiently
with \code{\link[=string_clean]{string_clean()}}.

Use \code{\link[=string_vec]{string_vec()}} to create simple string vectors.

String interpolation combined with operation chaining: \code{\link[=string_magic]{string_magic()}}. You can change \code{string_magic}
default values with \code{\link[=string_magic_alias]{string_magic_alias()}} and add custom operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}}.

Display messages while benefiting from \code{string_magic} interpolation with \code{\link[=cat_magic]{cat_magic()}} and \code{\link[=message_magic]{message_magic()}}.

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent R. Berge
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_tools.R
\name{string_get}
\alias{string_get}
\alias{stget}
\title{Gets elements of a character vector}
\usage{
string_get(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  seq = FALSE,
  seq.unik = FALSE,
  pattern = NULL,
  envir = parent.frame()
)

stget(
  x,
  ...,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  or = FALSE,
  seq = FALSE,
  seq.unik = FALSE,
  pattern = NULL,
  envir = parent.frame()
)
}
\arguments{
\item{x}{A character vector.}

\item{...}{Character scalars representing the patterns to be found. By default they are (perl) regular-expressions.
Use ' & ' or ' | ' to chain patterns and combine their result logically (ex: \code{'[[:alpha:]] & \\\\d'} gets strings
containing both letters and numbers). You can negate by adding a \code{!} first (ex: \code{"!sepal$"} will
return \code{TRUE} for strings that do not end with \code{"sepal"}).
Add flags with the syntax 'flag1, flag2/pattern'. Available flags are: 'fixed', 'ignore', 'word' and 'magic'.
Ex: "ignore/sepal" would get "Sepal.Length" (wouldn't be the case w/t 'ignore').
Shortcut: use the first letters of the flags. Ex: "if/dt[" would get \code{"DT[i = 5]"} (flags 'ignore' + 'fixed').
For 'word', it adds word boundaries to the pattern. The \code{magic} flag first interpolates
values directly into the pattern with "{}".}

\item{fixed}{Logical scalar, default is \code{FALSE}. Whether to trigger a fixed search instead of a
regular expression search (default).}

\item{ignore.case}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then case insensitive search is triggered.}

\item{word}{Logical scalar, default is \code{FALSE}. If \code{TRUE} then a) word boundaries are added to the pattern,
and b) patterns can be chained by separating them with a comma, they are combined with an OR logical operation.
Example: if \code{word = TRUE}, then pattern = "The, mountain" will select strings containing either the word
'The' or the word 'mountain'.}

\item{or}{Logical, default is \code{FALSE}. In the presence of two or more patterns,
whether to combine them with a logical "or" (the default is to combine them with a logical "and").}

\item{seq}{Logical, default is \code{FALSE}. The argument \code{pattern} accepts a vector of
patterns which are combined with an \code{and} by default. If \code{seq = TRUE}, then it is like
if \code{string_get} was called sequentially with its results stacked. See examples.}

\item{seq.unik}{Logical, default is \code{FALSE}. The argument \code{...} (or the argument \code{pattern}) accepts
a vector of patterns which are combined with an \code{and} by default. If \code{seq.unik = TRUE}, then
\code{string_get} is called sequentially with its results stacked, and \code{unique()} is
applied in the end. See examples.}

\item{pattern}{(If provided, elements of \code{...} are ignored.) A character vector representing the
patterns to be found. By default a (perl) regular-expression search is triggered.
Use ' & ' or ' | ' to chain patterns and combine their result logically (ex: \code{'[[:alpha:]] & \\\\d'} gets strings
containing both letters and numbers). You can negate by adding a \code{!} first (ex: \code{"!sepal$"} will
return \code{TRUE} for strings that do not end with \code{"sepal"}).
Add flags with the syntax 'flag1, flag2/pattern'. Available flags are: 'fixed', 'ignore', 'word' and 'magic'.
Ex: "ignore/sepal" would get "Sepal.Length" (wouldn't be the case w/t 'ignore').
Shortcut: use the first letters of the flags. Ex: "if/dt[" would get \code{"DT[i = 5]"} (flags 'ignore' + 'fixed').
For 'word', it adds word boundaries to the pattern. The \code{magic} flag first interpolates
values directly into the pattern with "{}".}

\item{envir}{Environment in which to evaluate the interpolations if the flag \code{"magic"} is provided.
Default is \code{parent.frame()}.}
}
\value{
It always return a character vector.
}
\description{
Convenient way to get elements from a character vector.
}
\details{
This function is a wrapper to \code{\link[=string_is]{string_is()}}.
}
\section{Functions}{
\itemize{
\item \code{stget()}: Alias to \code{string_get}

}}
\section{Caching}{


In an exploratory stage, it can be useful to quicky get values from a vector with the
least hassle as possible. Hence \code{string_get} implements caching, so that users do not need
to repeat the value of the argument \code{x} in successive function calls, and can concentrate
only on the selection patterns.

Caching is a feature only available when the user calls \code{string_get} from the global environment.
If that feature were available in regular code, it would be too dangerous, likely leading to hard to debug bugs.
Hence caching is disabled when used within code (i.e. inside a function or inside an
automated script), and function calls without the main argument will lead to errors in such scripts.
}

\section{Generic regular expression flags}{


All \code{stringmagic} functions support generic flags in regular-expression patterns.
The flags are useful to quickly give extra instructions, similarly to \emph{usual}
\href{https://javascript.info/regexp-introduction}{regular expression flags}.

Here the syntax is "flag1, flag2/pattern". That is: flags are a comma separated list of flag-names
separated from the pattern with a slash (\code{/}). Example: \code{string_which(c("hello...", "world"), "fixed/.")} returns \code{1}.
Here the flag "fixed" removes the regular expression meaning of "." which would have otherwise meant \emph{"any character"}.
The no-flag verion \code{string_which(c("hello...", "world"), ".")} returns \code{1:2}.

Alternatively, and this is recommended, you can collate the initials of the flags instead of using a
comma separated list. For example: "if/dt[" will apply the flags "ignore" and "fixed" to the pattern "dt[".

The four flags always available are: "ignore", "fixed", "word" and "magic".
\itemize{
\item "ignore" instructs to ignore the case. Technically, it adds the perl-flag "(?i)"
at the beginning of the pattern.
\item "fixed" removes the regular expression interpretation, so that the characters ".", "$", "^", "["
(among others) lose their special meaning and are treated for what they are: simple characters.
\item "word" adds word boundaries (\code{"\\\\b"} in regex language) to the pattern. Further, the comma (\code{","})
becomes a word separator. Technically, "word/one, two" is treated as "\\b(one|two)\\b". Example:
\code{string_clean("Am I ambushed?", "wi/am")} leads to " I ambushed?" thanks to the flags "ignore" and "word".
\item "magic" allows to interpolate variables inside the pattern before regex interpretation.
For example if \code{letters = "aiou"} then \code{string_clean("My great goose!", "magic/[{letters}] => e")}
leads to \code{"My greet geese!"}
}
}

\examples{

x = rownames(mtcars)

# find all Mazda cars
string_get(x, "Mazda")
# same with ignore case flag
string_get(x, "i/mazda")

# all cars containing a single digit (we use the 'word' flag)
string_get(x, "w/\\\\d")

# finds car names without numbers AND containing `u`
string_get(x, "!\\\\d", "u")
# equivalently
string_get(x, "!\\\\d & u")

# Stacks all Mazda and Volvo cars. Mazda first
string_get(x, "Mazda", "Volvo", seq = TRUE)

# Stacks all Mazda and Volvo cars. Volvo first
string_get(x, "Volvo", "Mazda", seq = TRUE)

# let's get the first word of each car name
car_first = string_ops(x, "extract.first")
# we select car brands ending with 'a', then ending with 'i'
string_get(car_first, "a$", "i$", seq = TRUE)
# seq.unik is similar to seq but applies unique()
string_get(car_first, "a$", "i$", seq.unik = TRUE)

#
# flags
#

# you can combine the flags
x = string_magic("/One, two, one... Two!, Microphone, check")
# regular
string_get(x, "one")
# ignore case
string_get(x, "i/one")
# + word boundaries
string_get(x, "iw/one")

# you can escape the meaning of ! with backslashes
string_get(x, "\\\\!")

#
# Caching
#

# Caching is enabled when the function is used interactively
# so you don't need to repeat the argument 'x'
# Mostly useful at an exploratory stage

if(interactive() && is.null(sys.calls())){
   
   # first run, the data is cached
   string_get(row.names(mtcars), "i/vol")

   # now you don't need to specify the data
   string_get("i/^m & 4")
}




}
\seealso{
String operations: \code{\link[=string_is]{string_is()}}, \code{\link[=string_get]{string_get()}}, \code{\link[=string_clean]{string_clean()}}, \code{\link[=string_split2df]{string_split2df()}}.
Chain basic operations with \code{\link[=string_ops]{string_ops()}}. Clean character vectors efficiently
with \code{\link[=string_clean]{string_clean()}}.

Use \code{\link[=string_vec]{string_vec()}} to create simple string vectors.

String interpolation combined with operation chaining: \code{\link[=string_magic]{string_magic()}}. You can change \code{string_magic}
default values with \code{\link[=string_magic_alias]{string_magic_alias()}} and add custom operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}}.

Display messages while benefiting from \code{string_magic} interpolation with \code{\link[=cat_magic]{cat_magic()}} and \code{\link[=message_magic]{message_magic()}}.

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent R. Berge
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aliases.R, R/string_tools.R
\name{string_clean_alias}
\alias{string_clean_alias}
\alias{string_clean}
\alias{string_replace}
\alias{stclean}
\alias{streplace}
\title{Cleans a character vector from multiple patterns}
\usage{
string_clean_alias(
  replacement = "",
  pipe = " => ",
  split = ",[ \\n\\t]+",
  ignore.case = FALSE,
  fixed = FALSE,
  word = FALSE,
  total = FALSE,
  single = FALSE,
  namespace = NULL
)

string_clean(
  x,
  ...,
  replacement = "",
  pipe = " => ",
  split = ",[ \\n\\t]+",
  ignore.case = FALSE,
  fixed = FALSE,
  word = FALSE,
  total = FALSE,
  single = FALSE,
  envir = parent.frame(),
  namespace = NULL
)

string_replace(
  x,
  pattern,
  replacement = "",
  pipe = " => ",
  ignore.case = FALSE,
  fixed = FALSE,
  word = FALSE,
  total = FALSE,
  single = FALSE,
  envir = parent.frame()
)

stclean(
  x,
  ...,
  replacement = "",
  pipe = " => ",
  split = ",[ \\n\\t]+",
  ignore.case = FALSE,
  fixed = FALSE,
  word = FALSE,
  total = FALSE,
  single = FALSE,
  envir = parent.frame(),
  namespace = NULL
)

streplace(
  x,
  pattern,
  replacement = "",
  pipe = " => ",
  ignore.case = FALSE,
  fixed = FALSE,
  word = FALSE,
  total = FALSE,
  single = FALSE,
  envir = parent.frame()
)
}
\arguments{
\item{replacement}{Character scalar, default is the empty string. It represents the default
value by which the patterns found in the character strings will be replaced. For example
\code{string_clean(x, "e", replacement = "a")} turn all letters "e" in \code{x} into "a".}

\item{pipe}{Character scalar, default is \code{" => "}. If thevalue of \code{pipe} is found in a pattern,
then the string is split w.r.t. the pipe and anything after the pipe becomes the replacement.

For example in \code{string_clean(x, "e => a")} the default pipe is found in "e => a", so the pattern
"e" will be replaced with "a". In other terms, this is equivalent to \code{string_clean(x, "e", replacement = "a")}.
Example changing the pipe: you can obtain the previous result with \code{string_clean(x, "e|>a", pipe = "|>")}.}

\item{split}{Character scalar, default is \code{",[ \\t\\n]+"} (which means a comma followed with spaces
and/or new lines). By default the patterns to be replaced are comma separated, that is
the pattern is split w.r.t. the argument \code{split} and a replacement is done for each sub-pattern.

Use \code{NULL} or the empty string to disable pattern separation.

For example: let's look at \code{string_clean(x, "w/one, two => three")}. First the flag "word" is extracted from
the pattern (see arg. \code{...}) as well as the replacement (see arg. \code{pipe}), leading to "one, two" the
pattern to be replaced. Then the pattern is split w.r.t. \code{split}, leading
to two patterns "one" and "two". Hence the words (thanks to the flag "w") "one" and "two" from
the string \code{x} will be replaced with "three".}

\item{ignore.case}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then case insensitive search is triggered.}

\item{fixed}{Logical scalar, default is \code{FALSE}. Whether to trigger a fixed search instead of a
regular expression search (default).}

\item{word}{Logical scalar, default is \code{FALSE}. If \code{TRUE} then a) word boundaries are added to the pattern,
and b) patterns can be chained by separating them with a comma, they are combined with an OR logical operation.
Example: if \code{word = TRUE}, then pattern = "The, mountain" will select strings containing either the word
'The' or the word 'mountain'.}

\item{total}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then when a pattern is found
in a string, the full string is replaced (instead of just the pattern). Note, \emph{importantly},
that when \code{total = TRUE} you can use logical operators in the patterns.

Example: \code{string_clean(x, "wi/ & two, three & !four => ", total = TRUE)}}

\item{single}{Logical scalar, default is \code{FALSE}. Whether, in substitutions, to stop at
the first match found. Ex: \code{string_clean("abc", "[[:alpha:]] => _", single = TRUE)} leads
to \code{"_bc"}, while \code{string_clean("abc", "[[:alpha:]] => _")} leads to \code{"___"}.}

\item{namespace}{Character scalar or \code{NULL} (default). \strong{Only useful for package developers.}
As a regular end-user you shouldn't care! If your package uses \code{string_magic}, you should care.
It is useful \strong{only} if your package uses 'custom' \code{string_magic} operations, set with
\code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or \code{\link[=string_magic_register_ops]{string_magic_register_ops()}}.

If so pass the name of your package in this argument so that your function can access
the new \code{string_magic} operations defined within your package.}

\item{x}{A character vector.}

\item{...}{Character scalars representing patterns. A pattern is of the form
"flags/pat1, pat2 => replacement". This means that patterns 'pat1' and 'pat2' will be replaced
with the string 'replacement'. By default patterns are comma separated and the replacement comes
after a ' => ' (see args \code{split} and \code{pipe} to change this). By default the replacement is the empty string
(so "pat1, pat2" \emph{removes} the patterns).

Available regex flags are: 'word' (add word boundaries), 'ignore' (the case), 'fixed' (no regex),
'total', 'single' and 'magic'.
The flag \code{total} leads to a \emph{total replacement} of the string if the pattern is found.
The flag 'magic' allows to interpolate variables within the pattern.  Use flags
with comma separation ("word, total/pat") or use only their initials ("wt/pat").

Starting with an '@' leads to operations in \code{\link[=string_ops]{string_ops()}}. Ex: "@ascii, lower, ws" turns
the string into ASCII, lowers the case and normalizes white spaces (see help of \code{\link[=string_ops]{string_ops()}}).}

\item{envir}{Environment in which to evaluate the interpolations if the flag \code{"magic"} is provided.
Default is \code{parent.frame()}.}

\item{pattern}{A character scalar containing a regular expression pattern to be replaced.
You can write the replacement directly in the string after a pipe: ' => ' (see arg. \code{pipe} to change this).
By default the replacement is the empty string (so "pat1" \emph{removes} the pattern).

Available regex flags are: 'word' (add word boundaries), 'ignore' (the case), 'fixed' (no regex),
'total', 'single' and 'magic'.
The flag \code{total} leads to a \emph{total replacement} of the string if the pattern is found.
The flag 'magic' allows to interpolate variables within the pattern.  Use flags
with comma separation ("word, total/pat") or use only their initials ("wt/pat").}
}
\value{
The main usage returns a character vector of the same length as the vector in input.
Note, however, that since you can apply arbitrary \code{\link[=string_ops]{string_ops()}} operations, the length and type
of the final vector may depend on those (if they are used).
}
\description{
Recursively cleans a character vector from several patterns. Quickly handle the
tedious task of data cleaning by taking advantage of the syntax.
You can also apply all sorts of cleaning operations by summoning \code{\link[=string_ops]{string_ops()}} operations.
}
\section{Functions}{
\itemize{
\item \code{string_clean_alias()}: Create a \code{string_clean} alias with custom defaults

\item \code{string_replace()}: Simplified version of \code{string_clean}

\item \code{stclean()}: Alias to \code{string_clean}

\item \code{streplace()}: Alias to \code{string_replace}

}}
\section{Regular expression flags specific to replacement}{


This function benefits from two specific regex flags: "total" and "single".
\itemize{
\item "total" replaces the \emph{complete string} if the pattern is found (remember that the
default behavior is to replace just the pattern).
\item "single" performs a single substitution for each string element and stops there.
Only the first match of each string is replaced. Technically we use \code{\link[base:grep]{base::sub()}}
internally instead of \code{\link[base:grep]{base::gsub()}}.
}
}

\section{Generic regular expression flags}{


All \code{stringmagic} functions support generic flags in regular-expression patterns.
The flags are useful to quickly give extra instructions, similarly to \emph{usual}
\href{https://javascript.info/regexp-introduction}{regular expression flags}.

Here the syntax is "flag1, flag2/pattern". That is: flags are a comma separated list of flag-names
separated from the pattern with a slash (\code{/}). Example: \code{string_which(c("hello...", "world"), "fixed/.")} returns \code{1}.
Here the flag "fixed" removes the regular expression meaning of "." which would have otherwise meant \emph{"any character"}.
The no-flag verion \code{string_which(c("hello...", "world"), ".")} returns \code{1:2}.

Alternatively, and this is recommended, you can collate the initials of the flags instead of using a
comma separated list. For example: "if/dt[" will apply the flags "ignore" and "fixed" to the pattern "dt[".

The four flags always available are: "ignore", "fixed", "word" and "magic".
\itemize{
\item "ignore" instructs to ignore the case. Technically, it adds the perl-flag "(?i)"
at the beginning of the pattern.
\item "fixed" removes the regular expression interpretation, so that the characters ".", "$", "^", "["
(among others) lose their special meaning and are treated for what they are: simple characters.
\item "word" adds word boundaries (\code{"\\\\b"} in regex language) to the pattern. Further, the comma (\code{","})
becomes a word separator. Technically, "word/one, two" is treated as "\\b(one|two)\\b". Example:
\code{string_clean("Am I ambushed?", "wi/am")} leads to " I ambushed?" thanks to the flags "ignore" and "word".
\item "magic" allows to interpolate variables inside the pattern before regex interpretation.
For example if \code{letters = "aiou"} then \code{string_clean("My great goose!", "magic/[{letters}] => e")}
leads to \code{"My greet geese!"}
}
}

\examples{

x = c("hello world  ", "it's 5 am....")

# we clean the o's and the points (we use 'fixed' to trigger fixed-search)
string_clean(x, "o", "f/.")
# equivalently
string_clean(x, "fixed/o, .")
# equivalently
string_clean(x, "o, .", fixed = TRUE)
# equivalently
string_clean(x, "o", ".", fixed = TRUE)

#
# chaining operations: example using cars
#

cars = row.names(mtcars)
new = string_clean(cars, 
           # replace strings containing "Maz" with Mazda
           "total/Maz => Mazda", 
           # replace the word 'Merc' with Mercedes
           "wi/merc => Mercedes",
           # replace strings containing "Merc" and a digit followed with an 'S'
           "t/Merc & \\\\dS => Mercedes S!",
           # put to lower case, remove isolated characters and normalize white spaces
           "@lower, ws.isolated")

cbind(cars, new)


}
\seealso{
String operations: \code{\link[=string_is]{string_is()}}, \code{\link[=string_get]{string_get()}}, \code{\link[=string_clean]{string_clean()}}, \code{\link[=string_split2df]{string_split2df()}}.
Chain basic operations with \code{\link[=string_ops]{string_ops()}}. Clean character vectors efficiently
with \code{\link[=string_clean]{string_clean()}}.

Use \code{\link[=string_vec]{string_vec()}} to create simple string vectors.

String interpolation combined with operation chaining: \code{\link[=string_magic]{string_magic()}}. You can change \code{string_magic}
default values with \code{\link[=string_magic_alias]{string_magic_alias()}} and add custom operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}}.

Display messages while benefiting from \code{string_magic} interpolation with \code{\link[=cat_magic]{cat_magic()}} and \code{\link[=message_magic]{message_magic()}}.

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent R. Berge
}
\concept{tools with aliases}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_funs.R
\name{get_interpolated_expr}
\alias{get_interpolated_expr}
\alias{get_interpolated_vars}
\title{Lists the expressions used for interpolation in a \code{string_magic} call}
\usage{
get_interpolated_expr(x, parse = FALSE, delim = c("{", "}"))

get_interpolated_vars(x, delim = c("{", "}"))
}
\arguments{
\item{x}{A character scalar for which the variables will be recovered.
For example \code{x = "Hi {person}"} will  return \code{"person"}
(the variable that will be interpolated).}

\item{parse}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, the result is a list of R
expressions. If \code{FALSE} (default), the result is a character vector of expressions.}

\item{delim}{Character vector of length 1 or 2. Default is \code{c("{", "}")}. Defines
the opening and the closing delimiters for interpolation.

If of length 1, it must be of the form: 1) the opening delimiter,
2) a single space, 3) the closing delimiter. Ex: \code{".[ ]"} is equivalent to \code{c(".[", "]")}.
The default value is equivalent to \code{"{ }"}.

[ ]: R:\%20
[", "]: R:\%22,\%20\%22}
}
\value{
If the argument \code{parse = FALSE}, the default, then this function returns a
character vector containing all the expressions that will be interpolated.
This vector can be empty if there is no interpolation.

If the argument \code{parse = TRUE}, then a list is returned, containing the R expressions.

The function \code{get_interpolated_vars} always return a character vector.
}
\description{
Tool intended for development: use \code{get_interpolated_expr} to obtain the list of expressions
which will be interpolated in a \code{\link[=string_magic]{string_magic()}} call.
The function \code{get_interpolated_vars} provides the variables instead.
}
\details{
Note that this function captures even deeply nested interpolations.
}
\section{Functions}{
\itemize{
\item \code{get_interpolated_vars()}: Obtain the variables used in \code{string_magic()} interpolations

}}
\examples{

# let's create a simple interpolation
x = c("Ken", "Barbie")
sma_expr = "{' loves 'c ? x}. But does {' love 'c ? rev(x)}?"
string_magic(sma_expr)

# We recover the two expressions
(char = get_interpolated_expr(sma_expr))

# same with parsing
(expr = get_interpolated_expr(sma_expr, parse = TRUE))

# see the difference
eval(char[[1]])
eval(expr[[1]])

# and only the variables:
get_interpolated_vars(sma_expr)

}
\seealso{
String operations: \code{\link[=string_is]{string_is()}}, \code{\link[=string_get]{string_get()}}, \code{\link[=string_clean]{string_clean()}}, \code{\link[=string_split2df]{string_split2df()}}.
Chain basic operations with \code{\link[=string_ops]{string_ops()}}. Clean character vectors efficiently
with \code{\link[=string_clean]{string_clean()}}.

Use \code{\link[=string_vec]{string_vec()}} to create simple string vectors.

String interpolation combined with operation chaining: \code{\link[=string_magic]{string_magic()}}. You can change \code{string_magic}
default values with \code{\link[=string_magic_alias]{string_magic_alias()}} and add custom operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}}.

Display messages while benefiting from \code{string_magic} interpolation with \code{\link[=cat_magic]{cat_magic()}} and \code{\link[=message_magic]{message_magic()}}.

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_magic_main.R
\name{timer_magic}
\alias{timer_magic}
\title{Sets up a timer that can be used within \verb{_magic} functions}
\usage{
timer_magic()
}
\value{
This function does not return anything and is only intended to be used in
conjunction with future calls of \code{\link[=string_magic]{string_magic()}}.
}
\description{
Sets up a timer which can later be summoned by \code{\link[=string_magic]{string_magic()}} functions via
the \code{.timer}, \code{.timer_lap} and \code{.timer_total} variables. Useful to report
timings within functions with the function \code{\link[=cat_magic]{cat_magic()}} or \code{\link[=message_magic]{message_magic()}}.
}
\details{
This functions sets up a timer with \code{\link[base:Sys.time]{base::Sys.time()}}. This timer can then be tracked
and modified with the \code{.timer}, \code{.timer_lap} and \code{.timer_total} variables within
\code{\link[=cat_magic]{cat_magic()}} or \code{\link[=message_magic]{message_magic()}}.

Note that the timer is precise at +/- 1ms, hence it should \strong{not} be used to time
algorithms with very short execution times.

It works by saving the current system time in R options (\code{stringmagic_timer} and \code{stringmagic_timer_origin}).
Hence, since it uses options, it should not be used in parallel processes.
}
\examples{

# simple example where we time the execution of some elements in a function
# we trigger the message conditionally on the value of the argument `debug`.
rnorm_crossprod = function(n, mean = 0, sd = 1, debug = FALSE){
  # we set the timer
  timer_magic()
  # we compute some stuff
  x = rnorm(n, mean, sd)
  # we can report the time with .timer
  message_magic("{15 align ! Generation}: {.timer}", .trigger = debug)
  
  res = x \%*\% x
  message_magic("{15 align ! Product}: {.timer}",
                "{15 align ! Total}: {.timer_total}", 
                .sep = "\n", .trigger = debug)
  res
}

rnorm_crossprod(1e5, TRUE)


}
\seealso{
Other tools with aliases: 
\code{\link{string_clean_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent Berge
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_tools.R
\name{parse_regex_pattern}
\alias{parse_regex_pattern}
\title{\code{stringmagic}'s regular expression parser}
\usage{
parse_regex_pattern(
  pattern,
  authorized_flags,
  parse_flags = TRUE,
  parse_logical = TRUE,
  envir = parent.frame()
)
}
\arguments{
\item{pattern}{Character scalar, the regular expression pattern to parse.}

\item{authorized_flags}{Character vector representing the flags to be parsed.
Use the empty string if no flags are allowed.}

\item{parse_flags}{Logical scalar, default is \code{TRUE}. Whether to parse the
optional regex flags.}

\item{parse_logical}{Logical scalar, default is \code{TRUE}. Whether to parse logical
regex operations, similarly to \code{\link[=string_get]{string_get()}}.}

\item{envir}{An environment, default is \code{parent.frame()}. Only used if the flag
\code{magic} is present, it is used to find the variables to be interpolated.}
}
\value{
This function always returns a list of 4 elements:
\itemize{
\item \code{flags}: the character vector of flags. If no flags were found, this is the empty string.
\item \code{patterns}: the vector of regex patterns.
\item \code{is_or}: logical vector of the same length as \code{patterns}. Indicates for each
pattern if it should be attached to the previous patterns with a logical OR (\code{FALSE}
means a logical \code{AND}).
\item \code{is_not}: logical vector of the same length as \code{patterns}. Indicates for each pattern
if it should be negated.
}
}
\description{
Parse regular expression with custom flags and obtain the final pattern
to be parsed as well as the vector of flags
}
\details{
This is an internal tool that is exposed in order to facilitate checking what's going on.

There is no error handling.
}
\examples{

parse_regex_pattern("f/hello", c("fixed", "ignore"))

x = "john"
parse_regex_pattern("fm/{x} | Doe", c("fixed", "ignore", "magic"))

}
\seealso{
String operations: \code{\link[=string_is]{string_is()}}, \code{\link[=string_get]{string_get()}}, \code{\link[=string_clean]{string_clean()}}, \code{\link[=string_split2df]{string_split2df()}}.
Chain basic operations with \code{\link[=string_ops]{string_ops()}}. Clean character vectors efficiently
with \code{\link[=string_clean]{string_clean()}}.

Use \code{\link[=string_vec]{string_vec()}} to create simple string vectors.

String interpolation combined with operation chaining: \code{\link[=string_magic]{string_magic()}}. You can change \code{string_magic}
default values with \code{\link[=string_magic_alias]{string_magic_alias()}} and add custom operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}}.

Display messages while benefiting from \code{string_magic} interpolation with \code{\link[=cat_magic]{cat_magic()}} and \code{\link[=message_magic]{message_magic()}}.

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent Berge
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_tools.R
\name{string_fill}
\alias{string_fill}
\title{Fills a character string up to a size}
\usage{
string_fill(
  x = "",
  n = NULL,
  symbol = " ",
  right = FALSE,
  center = FALSE,
  center.right = TRUE,
  na = "NA"
)
}
\arguments{
\item{x}{A character vector.}

\item{n}{Integer scalar, possibly equal to \code{NULL} (default). The size up to which the character
vector will be filled. If \code{NULL} (default), it is set to the largest width in the character vector \code{x}.
To handle how the character is filled, see the arguments \code{symbol}, \code{right} and \code{center}.}

\item{symbol}{Character scalar of length 1, default is a space (" "). It is the symbol with which
the string will be filled.}

\item{right}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then the filling of the string is
done from the left, leading to right-alignment.}

\item{center}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then the filling of the string will
be balanced so as to center the strings.}

\item{center.right}{Logical scalar, default is \code{TRUE}. Only used when \code{center = TRUE},
ignored otherwise. If \code{TRUE}, then when the width is odd and the number of characters of the
string is even (or vice versa), the text is centered with one character on the right.
If \code{FALSE}, this is one character on the left.}

\item{na}{Character scalar or \code{NA}. Default is "NA" (a character string!). What happens to NAs: by default
they are replaced by the character string "NA".}
}
\value{
This functions returns a character vector of the same lenght as the vector in input.
}
\description{
Fills a character string up to a size and handles multibyte encodings
(differently from sprintf).
}
\details{
If you use character filling of the form \code{sprintf("\% 20s", x)} with \verb{x``containing multibyte characters, you may be suprised that all character strings do not end up at the same lenght (the occurrence of this problem depends on many things: encodings are a mess). }string_fill`
uses only base R functions to compensate this. It is slightly slower but, in general, safer.

It also looks a bit like \code{\link[base:format]{base::format()}}, but slightly different (and a bit faster, but more restrictive).
}
\examples{

x = c("apple", "pineapple") 

# simple fill with blank
cat(paste0(string_fill(x), ":", c(3, 7), "€"), sep = "\n")

# center fill
cat(paste0(string_fill(x, center = TRUE), ":", c(3, 7), "€"), sep = "\n")

# changing the length of the fill and the symbol used for filling
cat(paste0(string_fill(x), ":", 
           string_fill(c(3, 7), 3, "0", right = TRUE), "€"), sep = "\n")

# na behavior: default/NA/other
x = c("hello", NA) 
string_fill(x)
string_fill(x, na = NA)
string_fill(x, na = "(missing)")


}
\seealso{
String operations: \code{\link[=string_is]{string_is()}}, \code{\link[=string_get]{string_get()}}, \code{\link[=string_clean]{string_clean()}}, \code{\link[=string_split2df]{string_split2df()}}.
Chain basic operations with \code{\link[=string_ops]{string_ops()}}. Clean character vectors efficiently
with \code{\link[=string_clean]{string_clean()}}.

Use \code{\link[=string_vec]{string_vec()}} to create simple string vectors.

String interpolation combined with operation chaining: \code{\link[=string_magic]{string_magic()}}. You can change \code{string_magic}
default values with \code{\link[=string_magic_alias]{string_magic_alias()}} and add custom operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}}.

Display messages while benefiting from \code{string_magic} interpolation with \code{\link[=cat_magic]{cat_magic()}} and \code{\link[=message_magic]{message_magic()}}.

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent R. Berge
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_tools.R
\name{string_extract}
\alias{string_extract}
\alias{stextract}
\title{Extracts a pattern from a character vector}
\usage{
string_extract(
  x,
  pattern,
  single = FALSE,
  simplify = TRUE,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  unlist = FALSE,
  envir = parent.frame()
)

stextract(
  x,
  pattern,
  single = FALSE,
  simplify = TRUE,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  unlist = FALSE,
  envir = parent.frame()
)
}
\arguments{
\item{x}{A character vector.}

\item{pattern}{A character scalar. It represents the pattern
to be extracted from \code{x}. By default
this is a regular expression. You can use flags in the pattern in
the form \verb{flag1, flag2/pattern}.
Available flags are \code{ignore} (case), \code{fixed} (no regex), word (add word boundaries),
single (select only the first element), and magic (add interpolation with \code{{}}) . Example:
if \code{"ignore/hello"} and \verb{x = "Hello world} extracted text is \code{"Hello"}.
Shortcut: use the first letters of the flags. Ex: "iw/one" will extract the word
"one" (flags 'ignore' + 'word').}

\item{single}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, only the first pattern
that is detected will be returned. Note that in that case, a character vector is returned
of the same length as the vector in input.}

\item{simplify}{Logical scalar, default is \code{TRUE}. If \code{TRUE}, then when the vector input \code{x}
is of length 1, a character vector is returned instead of a list.}

\item{fixed}{Logical scalar, default is \code{FALSE}. Whether to trigger a fixed search instead of a
regular expression search (default).}

\item{ignore.case}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then case insensitive search is triggered.}

\item{word}{Logical scalar, default is \code{FALSE}. If \code{TRUE} then a) word boundaries are added to the pattern,
and b) patterns can be chained by separating them with a comma, they are combined with an OR logical operation.
Example: if \code{word = TRUE}, then pattern = "The, mountain" will select strings containing either the word
'The' or the word 'mountain'.}

\item{unlist}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, the function \code{unlist} is applied
to the resulting list, leading to a character vector in output (instead of a list).}

\item{envir}{Environment in which to evaluate the interpolations if the flag \code{"magic"} is provided.
Default is \code{parent.frame()}.}
}
\value{
The object returned by this functions can be a list or a character vector.

If \code{single = TRUE}, a character vector is returned, containing the value of the first match.
If no match is found, an empty string is returned.

If \code{single = FALSE} (the default) and \code{simplify = TRUE} (default), the object returned is:
\itemize{
\item a character vector if \code{x}, the vector in input, is of length 1: the character vector contains
all the matches and is of length 0 if no match is found.
\item a list of the same length as \code{x}. The ith element of the list is a character vector
of the matches for the ith element of \code{x}.
}

If \code{single = FALSE} (default) and \code{simplify = FALSE}, the object returned is always a list.
}
\description{
Extracts the first, or several, patterns from a character vector.
}
\section{Functions}{
\itemize{
\item \code{stextract()}: Alias to \code{string_extract}

}}
\section{Generic regular expression flags}{


All \code{stringmagic} functions support generic flags in regular-expression patterns.
The flags are useful to quickly give extra instructions, similarly to \emph{usual}
\href{https://javascript.info/regexp-introduction}{regular expression flags}.

Here the syntax is "flag1, flag2/pattern". That is: flags are a comma separated list of flag-names
separated from the pattern with a slash (\code{/}). Example: \code{string_which(c("hello...", "world"), "fixed/.")} returns \code{1}.
Here the flag "fixed" removes the regular expression meaning of "." which would have otherwise meant \emph{"any character"}.
The no-flag verion \code{string_which(c("hello...", "world"), ".")} returns \code{1:2}.

Alternatively, and this is recommended, you can collate the initials of the flags instead of using a
comma separated list. For example: "if/dt[" will apply the flags "ignore" and "fixed" to the pattern "dt[".

The four flags always available are: "ignore", "fixed", "word" and "magic".
\itemize{
\item "ignore" instructs to ignore the case. Technically, it adds the perl-flag "(?i)"
at the beginning of the pattern.
\item "fixed" removes the regular expression interpretation, so that the characters ".", "$", "^", "["
(among others) lose their special meaning and are treated for what they are: simple characters.
\item "word" adds word boundaries (\code{"\\\\b"} in regex language) to the pattern. Further, the comma (\code{","})
becomes a word separator. Technically, "word/one, two" is treated as "\\b(one|two)\\b". Example:
\code{string_clean("Am I ambushed?", "wi/am")} leads to " I ambushed?" thanks to the flags "ignore" and "word".
\item "magic" allows to interpolate variables inside the pattern before regex interpretation.
For example if \code{letters = "aiou"} then \code{string_clean("My great goose!", "magic/[{letters}] => e")}
leads to \code{"My greet geese!"}
}
}

\examples{

cars = head(row.names(mtcars))

# Let's extract the first word:
string_extract(cars, "\\\\w+", single = TRUE)

# same using flags
string_extract(cars, "s/\\\\w+")

# extract all words composed on only letters
# NOTE: we use the flag word (`w/`)
string_extract(cars, "w/[[:alpha:]]+")

# version without flag:
string_extract(cars, "\\\\b[[:alpha:]]+\\\\b")

# If a vector of length 1 => a vector is returned
greet = "Hi Tom, how's Mary doing?"
string_extract(greet, "w/[[:upper:]]\\\\w+")

# version with simplify = FALSE => a list is returned
string_extract(greet, "w/[[:upper:]]\\\\w+", simplify = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string_tools.R
\name{string_split2df}
\alias{string_split2df}
\alias{string_split2dt}
\title{Splits a character vector into a data frame}
\usage{
string_split2df(
  x,
  data = NULL,
  split = NULL,
  id = NULL,
  add.pos = FALSE,
  id_unik = TRUE,
  fixed = FALSE,
  ignore.case = FALSE,
  word = FALSE,
  envir = parent.frame(),
  dt = FALSE,
  ...
)

string_split2dt(
  x,
  data = NULL,
  split = NULL,
  id = NULL,
  add.pos = FALSE,
  id_unik = TRUE,
  fixed = FALSE
)
}
\arguments{
\item{x}{A character vector or a two-sided formula. If a two-sided formula, then the
argument \code{data} must be provided since the variables will be fetched in there.
A formula is of the form \code{char_var ~ id1 + id2} where \code{char_var} on the left is a
character variable and on the right \code{id1} and \code{id2} are identifiers which will be
included in the resulting table. Alternatively, you can provide identifiers via
the argument \code{id}.}

\item{data}{Optional, only used if the argument \code{x} is a formula. It should
contain the variables of the formula.}

\item{split}{A character scalar. Used to split the character vectors. By default
this is a regular expression. You can use flags in the pattern in the form \verb{flag1, flag2/pattern}.
Available flags are \code{ignore} (case), \code{fixed} (no regex), word (add word boundaries),
magic (add interpolation with \code{"{}"}). Example:
if "ignore/hello" and the text contains "Hello", it will be split at "Hello".
Shortcut: use the first letters of the flags. Ex: "iw/one" will split at the word
"one" (flags 'ignore' + 'word').}

\item{id}{Optional. A character vector or a list of vectors. If provided, the
values of \code{id} are considered as identifiers that will be included in the resulting table.}

\item{add.pos}{Logical, default is \code{FALSE}. Whether to include the position of each split element.}

\item{id_unik}{Logical, default is \code{TRUE}. In the case identifiers are provided,
whether to trigger a message if the identifiers are not unique. Indeed, if
the identifiers are not unique, it is not possible to reconstruct the original texts.}

\item{fixed}{Logical, default is \code{FALSE}. Whether to consider the argument \code{split}
as fixed (and not as a regular expression).}

\item{ignore.case}{Logical scalar, default is \code{FALSE}. If \code{TRUE}, then case insensitive search is triggered.}

\item{word}{Logical scalar, default is \code{FALSE}. If \code{TRUE} then a) word boundaries are added to the pattern,
and b) patterns can be chained by separating them with a comma, they are combined with an OR logical operation.
Example: if \code{word = TRUE}, then pattern = "The, mountain" will select strings containing either the word
'The' or the word 'mountain'.}

\item{envir}{Environment in which to evaluate the interpolations if the flag \code{"magic"} is provided.
Default is \code{parent.frame()}.}

\item{dt}{Logical, default is \code{FALSE}. Whether to return a \code{data.table}. See also the function \code{string_split2dt}.}

\item{...}{Not currently used.}
}
\value{
It returns a \code{data.frame} or a \code{data.table} which will contain: i) \code{obs}: the observation index,
ii) \code{pos}: the position of the text element in the initial string (optional, via add.pos),
iii) the text element, iv) the identifier(s) (optional, only if \code{id} was provided).
}
\description{
Splits a character vector and formats the resulting substrings into a data.frame
}
\section{Functions}{
\itemize{
\item \code{string_split2dt()}: Splits a string vector and returns a \code{data.table}

}}
\examples{

x = c("Nor rain, wind, thunder, fire are my daughters.",
      "When my information changes, I alter my conclusions.")

id = c("ws", "jmk")

# we split at each word
string_split2df(x, "[[:punct:] ]+")

# we add the 'id'
string_split2df(x, "[[:punct:] ]+", id = id)

# TO NOTE:
# - the second argument is `data`
# - when it is missing, the argument `split` becomes implicitly the second
# - ex: above we did not use `split = "[[:punct:] ]+"`

#
# using the formula

base = data.frame(text = x, my_id = id)
string_split2df(text ~ my_id, base, "[[:punct:] ]+")

#
# with 2+ identifiers

base = within(mtcars, carname <- rownames(mtcars))

# we have a message because the identifiers are not unique
string_split2df(carname ~ am + gear + carb, base, " +")

# adding the position of the words & removing the message
string_split2df(carname ~ am + gear + carb, base, " +", id_unik = FALSE, add.pos = TRUE)


}
\seealso{
String operations: \code{\link[=string_is]{string_is()}}, \code{\link[=string_get]{string_get()}}, \code{\link[=string_clean]{string_clean()}}, \code{\link[=string_split2df]{string_split2df()}}.
Chain basic operations with \code{\link[=string_ops]{string_ops()}}. Clean character vectors efficiently
with \code{\link[=string_clean]{string_clean()}}.

Use \code{\link[=string_vec]{string_vec()}} to create simple string vectors.

String interpolation combined with operation chaining: \code{\link[=string_magic]{string_magic()}}. You can change \code{string_magic}
default values with \code{\link[=string_magic_alias]{string_magic_alias()}} and add custom operations with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}}.

Display messages while benefiting from \code{string_magic} interpolation with \code{\link[=cat_magic]{cat_magic()}} and \code{\link[=message_magic]{message_magic()}}.

Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aliases.R, R/string_tools.R
\name{string_ops_alias}
\alias{string_ops_alias}
\alias{string_ops}
\alias{stops}
\title{Chains basic operations to character vectors}
\usage{
string_ops_alias(op = NULL, pre_unik = NULL, namespace = NULL)

string_ops(
  x,
  ...,
  op = NULL,
  pre_unik = NULL,
  namespace = NULL,
  envir = parent.frame()
)

stops(
  x,
  ...,
  op = NULL,
  pre_unik = NULL,
  namespace = NULL,
  envir = parent.frame()
)
}
\arguments{
\item{op}{Character \strong{vector} or \code{NULL} (default). Character scalar containing the comma separated values
of operations to perform to the vector. The 50+ operations are detailed in the help
page of \code{\link[=string_magic]{string_magic()}}. Note that if this argument is provided, then the values in
\code{...} are ignored.}

\item{pre_unik}{Logical scalar, default is \code{NULL}. Whether to first unique the vector
before applying the possibly costly string operations, and merging back the result.
For very large vectors with repeated values the time gained can be substantial. By
default, this is \code{TRUE} for vector of length 1M or more.}

\item{namespace}{Character scalar or \code{NULL} (default). \strong{Only useful for package developers.}
As a regular end-user you shouldn't care! If your package uses \code{string_magic}, you should care.
It is useful \strong{only} if your package uses 'custom' \code{string_magic} operations, set with
\code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or \code{\link[=string_magic_register_ops]{string_magic_register_ops()}}.

If so pass the name of your package in this argument so that your function can access
the new \code{string_magic} operations defined within your package.}

\item{x}{A character vector. If not a character vector but atomistic (i.e. not a list),
it will be converted to a character vector.}

\item{...}{Character \strong{scalars}. Character scalar containing the comma separated values
of operations to perform to the vector. The 50+ operations are detailed in the help
page of \code{\link[=string_magic]{string_magic()}}.}

\item{envir}{Environment in which to evaluate the interpolations if the flag \code{"magic"} is provided.
Default is \code{parent.frame()}.}
}
\value{
In general it returns a character vector. It may be of a length different from the original
one, depending on the operations performed.
}
\description{
Simple tool to perform multiple operations to character vectors.
}
\details{
This function is a simple wrapper around string_magic. Formally, \code{string_ops(x, "op1, op2")}
is equivalent to \code{string_magic("{op1, op2 ? x}")}.
}
\section{Functions}{
\itemize{
\item \code{string_ops_alias()}: \code{string_ops} alias with custom defaults

\item \code{stops()}: Alias to \code{string_ops}

}}
\examples{

# data on car models
cars = row.names(mtcars)

# let's get the brands starting with an "m"
string_ops(cars, "'i/^m'get, x, unik")

# Explainer:
# 'i/^m'get: keeps only the elements starting with an m,
#            i/ is the 'regex-flag' "ignore" to ignore the case
#            ^m means "starts with an m" in regex language
# x: extracts the first pattern. The default pattern is "[[:alnum:]]+"
#    which means an alpha-numeric word
# unik: applies unique() to the vector
# => see help in ?string_magic for more details on the operations


# let's get the 3 largest numbers appearing in the car models
string_ops(cars, "'\\\\d+'x, rm, unik, num, dsort, 3 first")

# Explainer:
# '\\d+'x: extracts the first pattern, the pattern meaning "a succession"
#          of digits in regex language
# rm: removes elements equal to the empty string (default behavior)
# unik: applies unique() to the vector
# num: converts to numeric
# dsort: sorts in decreasing order
# 3 first: keeps only the first three elements

# You can use several character vectors as operations:
string_ops(cars, 
           "'\\\\d+'x, rm, unik",
           "num, dsort, 3 first")

}
\seealso{
Other tools with aliases: 
\code{\link{cat_magic_alias}()},
\code{\link{string_clean_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_vec_alias}()}
}
\author{
Laurent R. Berge
}
\concept{tools with aliases}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index.R
\docType{package}
\name{stringmagic-package}
\alias{stringmagic}
\alias{stringmagic-package}
\title{String operations and interpolation: magic edition}
\description{
The stringmagic package introduces a new language tailored to create complex character strings.
Use the function string_magic to interpolate a variable inside a character string and to apply,
on-the-fly, any arbitrary string operation.
}
\details{
It supports over 50 basic operations, seamless customization, nesting,
pluralization, and much more. Writing complex character strings has never been that easy!

To get an overview of the package: see the \href{https://lrberge.github.io/stringmagic/index.html}{Readme}.
To get started with string_magic,
see \href{https://lrberge.github.io/stringmagic/articles/guide_string_magic.html}{the vignette}.

This package also enhances regular string functions (like grep and co) with
specialized functions to:
\itemize{
\item detect combinations of regular expressions
\item chain basic string operations
\item clean character string vectors
}

See more in the \href{https://lrberge.github.io/stringmagic/articles/guide_string_tools.html}{dedicated vignette}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://lrberge.github.io/stringmagic/}
  \item \url{https://github.com/lrberge/stringmagic}
  \item Report bugs at \url{https://github.com/lrberge/stringmagic/issues}
}

}
\author{
Laurent Berge

For a few functions (e.g. string_split2df) I use internally an algorithm to turn vectors
into indexes (i.e. integer vectors of the same length ranging to 1 to the number of groups).
This algorithm was inspired by Sebastian Krantz's \href{https://cran.r-project.org/package=collapse}{collapse}
(if you don't know about it, you should definitely check it out!) who uses an algorithm
from Morgan Jacob's \href{https://cran.r-project.org/package=kit}{kit} package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aliases.R, R/string_magic_main.R
\name{cat_magic_alias}
\alias{cat_magic_alias}
\alias{message_magic_alias}
\alias{cat_magic}
\alias{catma}
\alias{message_magic}
\alias{mema}
\title{Display messages using interpolated strings}
\usage{
cat_magic_alias(
  .sep = "",
  .end = "",
  .width = FALSE,
  .leader = "",
  .delim = c("{", "}"),
  .last = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .namespace = NULL
)

message_magic_alias(
  .sep = "",
  .end = "\\n",
  .width = ~min(100, .sw),
  .leader = "",
  .delim = c("{", "}"),
  .last = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .namespace = NULL
)

cat_magic(
  ...,
  .sep = "",
  .end = "",
  .width = FALSE,
  .leader = "",
  .data = list(),
  .envir = parent.frame(),
  .delim = c("{", "}"),
  .last = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .help = NULL,
  .namespace = NULL
)

catma(
  ...,
  .sep = "",
  .end = "",
  .width = FALSE,
  .leader = "",
  .data = list(),
  .envir = parent.frame(),
  .delim = c("{", "}"),
  .last = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .help = NULL,
  .namespace = NULL
)

message_magic(
  ...,
  .sep = "",
  .end = "\\n",
  .width = NULL,
  .leader = "",
  .envir = parent.frame(),
  .data = list(),
  .delim = c("{", "}"),
  .last = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .help = NULL,
  .namespace = NULL
)

mema(
  ...,
  .sep = "",
  .end = "\\n",
  .width = NULL,
  .leader = "",
  .envir = parent.frame(),
  .data = list(),
  .delim = c("{", "}"),
  .last = NULL,
  .trigger = TRUE,
  .check = TRUE,
  .help = NULL,
  .namespace = NULL
)
}
\arguments{
\item{.sep}{Character scalar, default is the empty string \code{""}. It is used to collapse all
the elements in \code{...} before applying any operation.}

\item{.end}{Character scalar, default is \code{""} (the empty string) for \code{cat_magic},
and \code{"\\n"} (a newline) for \code{message_magic}.
This string will be collated at the end of the message (a common alternative is \code{"\\n"}).}

\item{.width}{Can be 1) a positive integer, 2) a number in (0;1), 3) \code{FALSE} (default
for \code{cat_magic}), 4) \code{NULL} (default for \code{message_magic}), or 5) a one-sided formula. It represents the target
width of the message on the user console. Newlines will be added \emph{between words} to fit the
target width.
\enumerate{
\item positive integer: number of characters
\item number (0;1): fraction of the screen
\item \code{FALSE}: does not add newlines
\item \code{NULL}: the min between 100 characters and 90\% of the screen width
\item one-sided formula: an expression in which you can use the special variable
\code{.sw} which represents the current screen width
}

Note that the value \code{NULL} is equivalent to using \code{~min(100, 0.9*.sw)}.}

\item{.leader}{Character scalar, default is \code{TRUE}. Only used if argument \code{.width} is not \code{FALSE}.
Whether to add a leading character string right after the extra new lines.}

\item{.delim}{Character vector of length 1 or 2. Default is \code{c("{", "}")}. Defines
the opening and the closing delimiters for interpolation.

If of length 1, it must be of the form: 1) the opening delimiter,
2) a single space, 3) the closing delimiter. Ex: \code{".[ ]"} is equivalent to \code{c(".[", "]")}.
The default value is equivalent to \code{"{ }"}.

[ ]: R:\%20
[", "]: R:\%22,\%20\%22}

\item{.last}{Character scalar, a function, or \code{NULL} (default). If provided and character:
it must be an \code{string_magic} chain of operations of the form \code{"'arg1'op1, op2, etc"}. All these operations
are applied just before returning the vector. If a function,
it will be applied to the resulting vector.}

\item{.trigger}{Logical, default is \code{TRUE}. If \code{FALSE}, this function is not run. Can be
useful in debugging situations where we want conditional evaluations.}

\item{.check}{Logical scalar, default is \code{TRUE}. Whether to enable error-handling (i.e.
human readable error messages).
Without error-handling you can save something of the order of 40us.Useful only
in long loops.}

\item{.namespace}{Character scalar or \code{NULL} (default). \strong{Only useful for package developers.}
As a regular end-user you shouldn't care! If your package uses \code{string_magic}, you should care.
It is useful \strong{only} if your package uses 'custom' \code{string_magic} operations, set with
\code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or \code{\link[=string_magic_register_ops]{string_magic_register_ops()}}.}

\item{...}{Character scalars that will be collapsed with the argument \code{sep}. Note that
named arguments are used for substitution.

To interpolate, you can
use \code{"{x}"} within each character string to insert the value of \code{x} in the string.
You can add string operations in each \code{"{}"} instance with the syntax \code{"'arg'op ? x"}
(resp. \code{"'arg'op ! x"}) to apply the operation \code{'op'} with the argument \code{'arg'} to \code{x}
(resp. the verbatim of \code{x}). Otherwise, what to say? Ah, nesting is enabled, and since
there's over 50 operators, it's a bit complicated to sort you out in this small space.

Use the argument \code{.help = "keyword"} (or \code{.help = TRUE}) to obtain a selective help
from the main documentation.

Note that in interpolations you have access to the special variables: \code{.now} and \code{.date}
to get the current time; and the special function \code{.now("format")} to
format the time. Ex: \code{.now('\%Y-\%m \%H:\%m')}.}

\item{.data}{A list used to evaluate the variables in \code{"{}"}. Default is the empty list.
By default the variables are evaluated using the environment from where the
function is called or using the named arguments passed to the function.}

\item{.envir}{An environment used to evaluate the variables in \code{"{}"}. By default the variables are
evaluated using the environment from where the function is called or using the named
arguments passed to the function.}

\item{.help}{Character scalar or \code{TRUE}, default is \code{NULL}. This argument
is used to generate a dynamic help on the console. If \code{TRUE}, the user can select which
topic to read from the main documentation, with the possibility to search for keywords and
navigate the help pages. If a character scalar, then a regex search is perfomed on the main
documentation and any section containining a match is displayed. The user can easily
navigate across matches.}
}
\value{
The functions \code{cat_magic()} and \code{message_magic()} do not return anything, they simply print on the console.

The function \code{cat_magic_alis()} returns a function behaving identically to \code{\link[=cat_magic]{cat_magic()}} but for which the
default values have been altered.
Same for \code{message_magic_alias()}.
}
\description{
Utilities to display messages using \code{string_magic} interpolation and operations to generate the message.
}
\details{
These functions are \code{\link[base:cat]{base::cat()}}/\code{\link[=message]{message()}} wrappers aroung \code{\link[=string_magic]{string_magic()}}. There is one notable difference
with respect to \code{cat}/\code{message}. It's the ability to add newlines after words for
the message to fit a target width. This is controlled with the argument \code{.width}. This is
active by default for \code{message_magic} (default is \code{.width = NULL} which leads to the
minimum betwen 100 characters and 90\% of the screen width).

You can very easily change the default values with the alias generators \code{cat_magic_alias} and
\code{message_magic_alias}.

\verb{[Advanced]} A note for package developers who would use these functions \strong{and}
also use custom \code{string_magic} operations created with \code{\link[=string_magic_register_fun]{string_magic_register_fun()}} or
\code{\link[=string_magic_register_ops]{string_magic_register_ops()}}. To ensure forward compatibility the new operations created
should be defined in the package namespace (see the \emph{ad hoc} section in \code{\link[=string_magic_register_fun]{string_magic_register_fun()}} help).
To access these operators in their specific namespaces, you must use an alias with
\code{cat_magic_alias} or \code{message_magic_alias} with the argument \code{.namespace = "myPackageName"}
(to avoid having to provide the \code{.namespace} argument repeatedly).
}
\section{Functions}{
\itemize{
\item \code{cat_magic_alias()}: Create an alias of \code{cat_magic} with custom defaults

\item \code{message_magic_alias()}: Create an alias of \code{message_magic} with custom defaults

\item \code{catma()}: Alias to \code{cat_magic}

\item \code{message_magic()}: Display messages using interpolated strings

\item \code{mema()}: Alias to \code{message_magic}

}}
\examples{

start = Sys.time()
Sys.sleep(0.05)
message_magic("This example has run in {difftime ? start}.")

cat_magic("Let's write a very long message to illustrate how .width work.", 
          .width = 40)

# Let's add a leader
cat_magic("Let's write a very long message to illustrate how `.width` work.", 
          "And now we add `.leader`.", .width = 40, .leader = "#> ")

# newlines respect the introductory spaces
cat_magic("Here's a list:", 
          "    + short item", 
          "    + this is a very long item that likely overflows", 
         .width = 30, .sep = "\n")


}
\seealso{
Other tools with aliases: 
\code{\link{string_clean_alias}()},
\code{\link{string_magic}()},
\code{\link{string_magic_alias}()},
\code{\link{string_ops_alias}()},
\code{\link{string_vec_alias}()}
}
\concept{tools with aliases}
